/***************************************************************************
                          graphnode.h  -  description
                             -------------------
    begin                : Mon Jan 23 2006
    copyright            : (C) 2006 by r.n.
    email                : rniekamp@sc.cs.tu-bs.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <ctl/ci/graph.ci>

#ifndef _GRAPHNODE_H_
#define _GRAPHNODE_H_

class graphNode
{
public:
	typedef graphNode* pointerT;
private:
	typedef std::list<pointerT>  neighborT;
	neighborT neighborM;
	std::string nameM;
	virtual void printPriv(std::ostream&) const =0;
//	{}
// used to resolve cycles
	mutable bool visitedM;
	void reset() const
	{
		if(!visitedM)
			return;
		visitedM = false;
		for(neighborT::const_iterator nb=neighborM.begin(); nb != neighborM.end(); nb++)
			if(!!*nb)
				(*nb)->reset();
	}
public:
	CTL_AbstractType(graphNode, tupel, (nameM,neighborM), 2);
	graphNode(const std::string &name ="anonymous"): nameM(name), visitedM(false)
	{}
	virtual ~graphNode()
	{}
	void addNeighbor(const pointerT &node)
 	{ neighborM.push_back(node);}
// print graph avoiding infinite recursion
	void print(std::ostream &os, bool first = true) const
	{
		if(visitedM)
			return;
		visitedM = true;
		os<<nameM<<": ";
		printPriv(os);
		neighborT::const_iterator nb;
		os<<"neighbor ( ";
		for(nb=neighborM.begin(); nb != neighborM.end(); nb++)
			if(!!*nb)
				os<<(*nb)->nameM<< " ";
			else
				os<<"nil ";
		os<<")\n\n";
			
		for(nb=neighborM.begin(); nb != neighborM.end(); nb++)
			if(!!*nb)
				(*nb)->print(os, false);
		if(first) // reset to non-visited state
			reset();		
	}
};

class nodeTypeA: public graphNode
{
	double valueM;
	void printPriv(std::ostream& os) const
	{ os << "nodeTypeA { value = " <<valueM<< "}\n"; }
public:
	nodeTypeA(const std::string &name ="anonymous", double value =0): graphNode(name), valueM(value)
	{}
	CTL_Type(nodeTypeA, tupel, ((graphNode&)(*this), valueM), 2);
};

class nodeTypeB: public graphNode
{
	long  iM, jM;
	void printPriv(std::ostream& os) const
	{ os << "nodeTypeB { i = " <<iM<< ", j = "<<jM<<"}\n"; }
public:
	nodeTypeB(const std::string &name ="anonymous", int i =0, int j =0): graphNode(name), iM(i), jM(j)
	{}
	CTL_Type(nodeTypeB, tupel, ((graphNode&)(*this), iM, jM), 3);	
};

#endif
