#include <math.h>
#include <impl.h>
#include <ctl.h>

struct mandelT {};
struct juliaT {};

inline double iter(juliaT, double x0,double y0, double xx0, double yy0, int n, int &N)
{ 
  double x=x0,y=y0,q,s=x+y;
 
  for(int k=0; k<n; k++)
  {
    x*=x ;y*=y; s*=s;
    if((q=x+y)>=4.)
    { N += k; return k+exp(1-q/4); } 
    x+=xx0-y; y=s-q+yy0; s=x+y;
  }
  N += n;
  return 0;
}

inline double iter(mandelT, double x0,double y0, double, double, int n, int &N)
{ 
  double x=x0,y=y0,q,s=x+y;
 
  for(int k=0; k<n; k++)
  {
    x*=x; y*=y; s*=s;
    if((q=x+y)>=4.)
    { N += k; return k+exp(1-q/4); } 
    x+=x0-y; y=s-q+y0; s=x+y;
  }
  N += n;
  return 0;
}

struct pixel
{
	int i, j;
	pixel(int ii=0, int jj=0): i(ii), j(jj)
	{}
};

inline void setCand(ctl::vector<float> &col, ctl::vector<pixel> &pic, int &z, int i, int j, int l, int h)
{
	if(i>=0 && j>=0 && i<l && j<h && col[l*j+i]<0)
	{
		pic[z++] = pixel(i,j);
		col[l*j+i] = 0;
	}
}

template<class Type>
	ctl::vector<float> iter(Type, int l, int h, double x0, double y0, double xx0, double yy0, double dx, int &n)
{
	int size=l*h, i, j, N = n, a=0, o=0, d=1;
	ctl::vector<pixel> cand(size);
	ctl::vector<float> col(size, -1);
	n = 0;
	for(i=0; i<l; i++) 
	{ setCand(col,cand,o,i,0,l,h); setCand(col,cand,o,i,h-1,l,h); }
	for(j=1; j<h-1; j++) 
	{ setCand(col,cand,o,0,j,l,h); setCand(col,cand,o,l-1,j,l,h); }
	while(a<o)
	{
		pixel p=cand[a++];
		if((col[p.j*l+p.i]=iter(Type(),x0+p.i*dx,y0+p.j*dx,xx0,yy0,N,n) ) >0 )
			for(i=p.i-d; i<=p.i+d; i++)	for(j=p.j-d; j<=p.j+d; j++)
				setCand(col,cand,o,i,j,l,h);
	}
	for(i=0; i<size; i++) 
		if(col[i]<0)
			col[i] = 0;
	return col;			
}

ctl::vector<float> mandel(int l, int h, double x0, double y0, double dx, int &n)
{ return iter(mandelT(),l,h,x0,y0,0,0,dx,n); }

ctl::vector<float> julia(int l, int h, double x0, double y0, double X, double Y, double dx, int &n)
{ return iter(juliaT(),l,h,x0,y0,X,Y,dx,n); }

ctl::vector<float> Mandel::mandel(int l, int h, double x0, double y0, double dx, int &n) const
{ return iter(mandelT(),l,h,x0,y0,0,0,dx,n); }

ctl::vector<float> Mandel::julia(int l, int h, double x0, double y0, double X, double Y, double dx, int &n) const
{ return iter(juliaT(),l,h,x0,y0,X,Y,dx,n); }

