/****************************************************************************
** $Id: qt/examples/drawlines/connect.cpp   2.3.0   edited 2001-01-26 $
**
** Copyright (C) 1992-2000 Trolltech AS.  All rights reserved.
**
** This file is part of an example program for Qt.  This example
** program may be used, distributed and modified without limitation.
**
*****************************************************************************/

#ifdef QT4
#include <Qt/qwidget.h>
#include <Qt/qpixmap.h>
#include <Qt/qpainter.h>
#include <Qt/qnamespace.h>
#include <Qt/qimage.h>
#include <Qt/qcursor.h>
#include <Qt/qapplication.h>
#include <Qt/qevent.h>
#include <Qt/qrubberband.h>
#else
#include <qwidget.h>
#include <qpixmap.h>
#include <qpainter.h>
#include <qnamespace.h>
#include <qimage.h>
#include <qcursor.h>
#include <qapplication.h>
#endif

#include <stdlib.h>

#include <ctl.h>

#include <ctl/ci/mandel.ci>

#ifdef QT4
using namespace Qt;
#endif

struct Block
{
ctl::vector<float> iterM;
double x0M, y0M, XM, YM, dxM;
int lM, hM, i0M, j0M, nM;
bool typeM;
float factM;
	char txtM[24];
	Block(): 
			x0M(0), y0M(0), XM(0), YM(0), dxM(0), lM(0), hM(0), i0M(0), j0M(0),nM(0), typeM(false),factM(0)
	{ txtM[0] = 0; }
	Block(double x0, double y0, double dx, int l, int h, int i0, int j0, int n, bool type=false):
			x0M(x0), y0M(y0), XM(0), YM(0), dxM(dx), lM(l), hM(h), i0M(i0), j0M(j0),nM(n), typeM(type), factM(1)
	{ txtM[0] = 0; }
	void set(double x0, double y0, double dx, int n, bool type, double X, double Y)
	{ iterM.clear(); x0M = x0; y0M = y0; dxM = dx; nM = n; typeM=type; XM=X; YM=Y; }
	Block & setBlock(Block &bb, const ctl::vector<float> &col, int p)
	{
			bb.iterM = col;
			int z=0;
 			for(int j=bb.j0M; j<bb.j0M+bb.hM; j++)
				for(int i=bb.i0M; i<bb.i0M+bb.lM; i++)
					iterM[j*lM+i] = col[z++];
			bb.factM = factM;
			sprintf(bb.txtM,"%d", p);
			return bb;
	}
	ctl::result<ctl::vector<float> > eval(const mandelbrot::MandelCI& f, int& count) const
	{
		count = nM;
		if(typeM)
			return f.julia(lM, hM, x0M, y0M, XM, YM, dxM, count);
		return f.mandel(lM, hM, x0M, y0M, dxM, count);
	}
};

class MandelWidget : public QWidget
{
enum { mandel = false, julia = true};
public:
	MandelWidget(int l, int h,const ctl::vector<mandelbrot::MandelCI> &worker, QApplication *app, QWidget *parent=0, const char *name=0):
							QWidget( parent, name, Qt::WNorthWestGravity ), workerM(worker), levelM(0), appM(app), incrM(false), procNoM(false)
	{
		if(l>0 && h>0)
			setFixedSize(l, h);
		else
			setWindowState(Qt::WindowFullScreen);
    setBackgroundColor( black );	
	}
protected:
	void init()
	{
		if(!!blockM)
			return;
		bsizeM = 64;
		const int L=width(), H=height();
		blockM.push_back(Block(-1.5-double(L)/H, -2.0, 4./H, L, H, 0, 0, 200) );
    Block &b=blockM[0];
		for(int j=0; j<H; j+=bsizeM)
			for(int i=0; i<L; i+=bsizeM)
				taskM.append(Block(b.x0M+b.dxM*i,b.y0M+b.dxM*j,b.dxM,
										std::min(bsizeM,L-i),std::min(bsizeM,H-j),i,j,b.nM) );
	}
	void keyPressEvent(QKeyEvent *e)
	{
//printf("%X\n", e->key());
		const int shift = 15;
		int xShift=0, yShift=0; 
    switch(e->key())
    {
			case Key_Down: yShift = -1; break;
			case Key_Up:  yShift = +1; break;
			case Key_Left: xShift = +1; break;
			case Key_Right: xShift = -1; break;
	    case Key_Prior:
			{
				if(levelM==0)
					return;
				levelM--;
				paint();
				break;
			}
	    case Key_Next:
			{
				if(levelM+2>blockM.size())
					return;
				levelM++;
				paint();
				break;
			}
			case Key_Plus:
			{
				int dX=width() * 0.15, dY = height() * 0.15;
				addBlock(dX, width()-dX, dY, height()-dY);
				paint();
				break;
			}
			case Key_Minus:
			{
				int dX=width() * 0.15, dY = height() * 0.15;
				addBlock(-dX, width()+dX, -dY, height()+dY);
				paint();
				break;
			}
	    case Key_C:
			{
				float f=(e->state() & ShiftButton) ? 1.f/1.2f : 1.2f;
				blockM[levelM].factM *= f;
				paint();
				break;
			}
	    case Key_I:
			{
				blockM[levelM].nM += blockM[levelM].nM;
				incrM = true;
				paint();
				incrM = false;
				break;
			}
      case Key_Q:
      {
				hide();
				appM->exit();
      }
      case Key_P:
      {
				procNoM = ! procNoM;
				if(!procNoM)
					paint();
      }

		}
		if(xShift || yShift)
		{
				int count=0;
				Block& b=blockM[levelM];
				Block stripe(b);
				if(xShift>0)
				{	stripe.x0M -= shift*b.dxM; stripe.lM = shift; }
				else if(xShift<0)
				{	stripe.x0M += b.lM*b.dxM; stripe.lM = shift; }
				else if(yShift>0)
				{	stripe.y0M -= shift*b.dxM; stripe.hM = shift; }
				else if(yShift<0)
				{	stripe.y0M += b.hM*b.dxM; stripe.hM = shift; }

				ctl::result<ctl::vector<float> > res=stripe.eval(workerM[0], count);
				float *d=b.iterM.begin();
				if(xShift>0)
				{
					for(int i=0; i<b.hM; i++)	
						for(int j=b.lM-shift-1; j>=0; j--)	
							d[i*b.lM+j+shift] = d[i*b.lM+j];
					b.x0M -= shift*b.dxM;
					ctl::vector<float> s = res;
					for(int i=0; i<b.hM; i++)	
						for(int j=0; j<shift; j++)	
							d[i*b.lM + j ] = s[i*shift + j ];
				}
				else if(xShift<0)
				{
					for(int i=0; i<b.hM; i++)	
						for(int j=0; j<b.lM-shift; j++)	
							d[i*b.lM+j] = d[i*b.lM+j+shift];
					b.x0M += shift*b.dxM;
					ctl::vector<float> s = res;
					for(int i=0; i<b.hM; i++)	
						for(int j=0; j<shift; j++)	
							d[i*b.lM + j + b.lM-shift] = s[i*shift + j ];
				}
				else if(yShift>0)
				{
					for(int i=b.hM-shift-1; i>=0; i--)	
						for(int j=0; j<b.lM; j++)	
							d[(i+shift)*b.lM+j] = d[i*b.lM+j];
					b.y0M -= shift*b.dxM;
					ctl::vector<float> s = res;
					for(int i=0; i<shift; i++)	
						for(int j=0; j<b.lM; j++)	
							d[i*b.lM + j ] = s[i*b.lM + j ];
				}
				else if(yShift<0)
				{
					for(int i=0; i<b.hM-shift; i++)	
						for(int j=0; j<b.lM; j++)	
							d[i*b.lM+j] = d[(i+shift)*b.lM+j];
					b.y0M += shift*b.dxM;
					ctl::vector<float> s = res;
					for(int i=0; i<shift; i++)	
						for(int j=0; j<b.lM; j++)	
							d[(i+b.hM-shift)*b.lM + j ] = s[i*b.lM + j ];
				}
				paint();
		}
	}
	void	paintEvent( QPaintEvent * )
	{ paint(); }
	void  mousePressEvent( QMouseEvent *e)
	{
		downM = true;
		e1M = e0M = e->pos();
	}
	void  mouseReleaseEvent( QMouseEvent *)
	{
		if(!downM)
			return;
		downM = false;
		int xm = std::min(e0M.x(), e1M.x());
		int xM = std::max(e0M.x(), e1M.x());
		int ym = std::min(e0M.y(), e1M.y());
		int yM = std::max(e0M.y(), e1M.y());
		addBlock(std::min(e0M.x(), e1M.x()), std::max(e0M.x(), e1M.x()), 
						std::min(e0M.y(), e1M.y()), std::max(e0M.y(), e1M.y()));
		paint();
	}
	void addBlock(int xm, int xM, int ym, int yM)
	{
		double s = std::min(double(xM-xm)/width(), double(yM-ym)/height());
		double dx = blockM[levelM].dxM;
		double x0 = blockM[levelM].x0M + ((xm + xM) - s*width() )*dx/2;
		double y0 = blockM[levelM].y0M + ((ym + yM) - s*height() )*dx/2;
		Block newBlock=blockM[levelM];
		if(s>0.003)
			newBlock.set(x0,y0,dx*s, newBlock.nM, blockM[levelM].typeM, newBlock.XM, newBlock.YM);
		else
		{
			if(newBlock.typeM == julia)
				return;
			newBlock.set(-2*double(width())/height(),-2.0, 4./height(), blockM[0].nM, julia, x0, y0);
		}
		if(++levelM<blockM.size())
			blockM[levelM] = newBlock;
		else
			blockM.append(newBlock);
	}
	void  mouseMoveEvent( QMouseEvent *e)
	{
		if(!downM)
			return;
		QPainter paint( this );
#ifdef QT4
		QRubberBand rb(QRubberBand::Rectangle, this);
#else
		paint.setRasterOp(XorROP);
		paint.setPen( QColor(255,255,255));
		paint.drawRect(QRect(e0M, e1M));
#endif
		QPoint p=e->pos()-e1M;
		e1M = e->pos();
		paint.drawRect(QRect(e0M, e1M));
	}
private:
	void paint()
	{
		init();
		if(!blockM[levelM].iterM || incrM)
			compute(blockM[levelM]);
		else
			paint(blockM[levelM] );
	}
	void paint(const Block &);
	void compute(Block &);
	bool finished(const Block &b) const
	{
		const float *col = blockM[levelM].iterM.begin();
		if(!incrM || !col)
			return false;
		int L = width();
 		for(int j=b.j0M; j<b.j0M+b.hM; j++) for(int i=b.i0M; i<b.i0M+b.lM; i++)
			if(col[j*L+i]==0.0f)
				return false;
		return true;
	}
	ctl::vector<mandelbrot::MandelCI> workerM;
	ctl::vector<Block> blockM, taskM;
	QPoint e0M,e1M;
	bool downM;
	unsigned int levelM;
	int bsizeM;
	QApplication *appM;
	bool incrM, procNoM;
};

void MandelWidget::compute(Block &b)
{
CTL_Profile("compute");
	ctl::timer().t();
	appM->setOverrideCursor (WaitCursor);
	b.iterM.resize(b.lM*b.hM);
	float *pb=b.iterM.begin();
	int L=width(), H=height(), z=0;
	
	for(int j=0; j<H; j+=bsizeM)
	 	for(int i=0; i<L; i+=bsizeM)
			taskM[z++].set(b.x0M+i*b.dxM, b.y0M+j*b.dxM, b.dxM, b.nM, b.typeM, b.XM, b.YM);
	ctl::vector<ctl::result<ctl::vector<float> > > res(workerM.size());
	unsigned int curr=0,read=0;
	const int idle=-1;
	ctl::vector<int> state(workerM.size(), idle);
	ctl::vector<int> count(workerM.size() );
	ctl::int8 sum = 0;
	while(read<taskM.size())
	{
		for(unsigned int p=0; p<workerM.size(); p++)
		{
			int tid = state[p];			
			if(tid>=0 && !!res[p])
			{
				read++;
				ctl::vector<float> col;
				try
				{ col = res[p]; }
// if any problems fire the worker
				catch(...)
				{
					printf(" lost worker %d while working on task %d\n",p, tid);
					workerM.erase(p);
					res.erase(p);
					state.erase(p);
					count.erase(p);
					continue;
				}
				sum += count[p];
				paint(b.setBlock(taskM[tid], col, p) );
				tid = state[p] = idle;
			}	
			while(curr<taskM.size() && finished(taskM[curr]) )
			{	read++; curr++; }
			if(curr<taskM.size() && tid==idle)
			{
				state[p] = curr;
				const Block &bb=taskM[curr++];
				count[p] = bb.nM;
				try 
				{ res[p] = bb.eval( workerM[p], count[p]); }
				catch(const std::exception &e)
				{ printf("caught by call %s\n", e.what() ); }
			}
		}
	}
	appM->setOverrideCursor (ArrowCursor);
	double t=ctl::timer().dt();
	printf(" #iteration = %e depth = %d time = %e #ter/s =%e\n", double(sum), b.nM, t, sum/t);
}

void MandelWidget::paint(const Block &b)
{
CTL_Profile("paint");
	
	static float f1=2*sqrt(2.0);
	static float f2=sqrt(5.0);
	static float f3=1.5*sqrt(7.0);
	const int size=b.hM*b.lM, M = 511, m=255;
 	QPainter painter( this );
	QPen w(white);
	painter.setPen(white);
	QImage img(b.lM, b.hM, 32);
	QRgb *rgb = (QRgb*)img.bits();
	const float *pc= b.iterM.begin();
	for(int z=0; z<size; z++)
	{
		float c=b.factM * *pc++;
		int I1=int(f1*c) & M;
		int I2=int(f2*c) & M;
		int I3=int(f3*c) & M;
		*rgb++ = qRgb(I1>m?M-I1:I1, I2>m?M-I2:I2, I3>m?M-I3:I3);
	}
{
CTL_Profile("drawImage");
	painter.drawImage(b.i0M, b.j0M,img);
	if(procNoM)
		painter.drawText(b.i0M+b.lM/2, b.j0M+b.hM/2, b.txtM);
}
}

int qtmain(int n, char **arg)
{
	QApplication a(n, arg);
	ctl::vector<ctl::location> locVec=ctl::readLocation("mandel.rsc");
	unsigned int p;
		
	ctl::vector<mandelbrot::MandelCI> worker;
// hire the worker
	for(p=0; p<locVec.size(); p++)
	{	
		ctl::link lnk(locVec[p]);
		if(!!lnk)
		{
			mandelbrot::MandelCI wrk=mandelbrot::MandelCI(lnk);
			if(!!wrk)
				worker.push_back(wrk);	
			if(!!wrk && lnk.getType()!=ctl::lib)
				worker.push_back(wrk);	
		}
	}
	if(!worker)
		return 1;
	int l=0, h=0;
	if(n>2)
	{	
		sscanf(arg[1],"%d", &l);
		sscanf(arg[2],"%d", &h);
	}
	MandelWidget mandel(l, h, worker, &a);
	a.setMainWidget( &mandel );

	mandel.show();
	a.exec();
	return 1;
}
