#ifndef PI_MANAGER_H
#define PI_MANAGER_H

#include <string>
#include <vector>
#include <map>

#include "PiTaskTypes.h"
#include "PiWorker.h"

namespace wire{
	class PiManager{
	public:
		typedef TaskTypes::ParameterT ParameterT;
		typedef TaskTypes::PartialParameterT PartialParameterT;
		typedef TaskTypes::ResultT ResultT;
		typedef TaskTypes::PartialResultT PartialResultT;
		typedef TaskTypes::JobIDT JobIDT;
		typedef TaskTypes::WorkerT WorkerT;
		typedef TaskTypes::StatusT StatusT;

	private:
		ParameterT requestCount_;
		ResultT pi_;
		
		int requestsPerWorker_;
		int totalProcessedRequests_;
		int successfulProcessedRequests_;
		int createdCalls_;
		int neededCalls_;
				
		JobIDT nextSubTaskParameter_;
		std::vector<WorkerT> workers_;
		int nextWorkerID_;

	public:
		PiManager();
		~PiManager();
		PiManager(const std::string &);
		JobIDT getParameter(PartialParameterT* parameter, bool* finished);
		WorkerT& getNextWorker();
		void reportProblem(const WorkerT& worker, JobIDT jobid, StatusT status);
		void takeResult(const JobIDT jobID, const PartialResultT result);
		ResultT work(const ParameterT parameters);
	};
};

#endif
