#ifndef WORLD_H
#define WORLD_H

namespace wire{
	template<class Manager> class World{
		// Import external types
		typedef typename Manager::WorkerT WorkerT;
		typedef typename Manager::PartialParameterT PartialParameterT;
		typedef typename Manager::JobIDT JobIDT;
		typedef typename Manager::PartialResultT PartialResultT;
		typedef typename Manager::StatusT StatusT;
		typedef typename SubTaskTypes::ExceptionT SubTaskExceptionT;
		
		// Local types
		typedef std::map<JobIDT, ctl::result<PartialResultT> > resMapT;
		typedef std::map<JobIDT, WorkerT> workMapT;
		
		// Variables
		Manager* manager_;
		resMapT futureResult;
		workMapT assignedJobs;
		
	public:
		World(Manager* manager) : manager_(manager) {
		}
		
		void spin(){
			timespec sleepTimespec;	/*nanosleep brauch das timespec Structure*/
			sleepTimespec.tv_sec  = 0;		/*was tv_sec und tv_nsec (Sek. und Nanosek.) enthält*/
			sleepTimespec.tv_nsec = 5000;
			bool finished = false;
			PartialParameterT p;

			
			while(true){
				for(typename resMapT::iterator i=futureResult.begin(); i!=futureResult.end(); ){
					try{
						if(!!i->second){
							manager_->takeResult(i->first, i->second);
							assignedJobs.erase(i->first);
							futureResult.erase(i++);
						}
						else{
							i++;
						}
							
					} catch(SubTaskExceptionT& e){
						std::cout << "SubTaskException: " << e.what << std::endl;
						manager_->reportProblem(assignedJobs[i->first], i->first, PermanentJobError);
						assignedJobs.erase(i->first);
						futureResult.erase(i++);
						
					} catch(ctl::exception& e){
						std::cout << "ctl::exception in worker: " <<  " " << e << std::endl;
						manager_->reportProblem(assignedJobs[i->first], i->first, PermanentWorkerError);
						assignedJobs.erase(i->first);
						futureResult.erase(i++);
					}
				}
				
				
				JobIDT id = manager_->getParameter(&p, &finished);
				
				if(finished && futureResult.size() == 0){
					break;
				}
				
				if(finished){
					//nanosleep(&sleepTimespec, NULL);
					sched_yield();
				}
				else{
					WorkerT& w = manager_->getNextWorker();
					//futureResult[id] = ctl::result<PartialResult>(ctl::late);
					futureResult[id] = w.work(p);
					assignedJobs[id] = w;
				}
			}
		}
	};
};

#endif /*WORLD_H_*/
