#include <iostream>
#include <exception>

#include "PiManager.h"
#include "World.h"

namespace wire{
	PiManager::PiManager() : nextSubTaskParameter_(0), nextWorkerID_(0), totalProcessedRequests_(0), successfulProcessedRequests_(0), createdCalls_(0){
		// use resource manager
		//	throw ctl::exception("unallowed constructor"); // besser?
		
		double parm[] = {sqrt(2), sqrt(3), sqrt(11)-2, sqrt(13)-2, sqrt(17)-3,sqrt(19)-3};
		
		for(int i = 0; i<3; i++)
		{ 
			try{
				std::vector<double> parms(2);
				parms[0] = *(parm + 2*i);
				parms[1] = *(parm + 2*i + 1);
				workers_[i] = WorkerT(parms); // uses the resource manager
			}
			catch(ctl::exception &e){
				std::cout << e << std::endl;
			}
		}
	}
	
	PiManager::PiManager(const std::string &resources) : nextSubTaskParameter_(0), nextWorkerID_(0), totalProcessedRequests_(0), successfulProcessedRequests_(0), createdCalls_(0) {
		ctl::vector<ctl::location> locVec = ctl::readLocation(resources.c_str());
		workers_.resize(locVec.size());
		double parm[] = {sqrt(2), sqrt(3), sqrt(11)-2, sqrt(13)-2, sqrt(17)-3,sqrt(19)-3};
		for(int i = 0; i<locVec.size(); i++)
		{ 
			try{
				std::vector<double> parms(2);
				parms[0] = *(parm + 2*i);
				parms[1] = *(parm + 2*i + 1);
				workers_[i] = WorkerT(ctl::link(locVec[i]), parms);
			}
			catch(ctl::exception &e){
				std::cout << e << std::endl;
			}
		}
		
	}
	
	PiManager::JobIDT PiManager::getParameter(PiManager::PartialParameterT* parameter, bool* finished){
		if(createdCalls_ == neededCalls_){
			parameter = NULL;
			(*finished) = true;
			return -1;
		}
		else{
			(*finished) = false;
			(*parameter) = requestsPerWorker_;
			createdCalls_++;
			return nextSubTaskParameter_++;
		}
	}

	PiManager::WorkerT& PiManager::getNextWorker(){
		return workers_[nextWorkerID_++ % workers_.size() ];
	}
	
	void PiManager::reportProblem(const PiManager::WorkerT& worker, JobIDT jobid, StatusT status){
		std::cout << "Problem with job " << jobid << " on worker " << std::endl;
		
		switch (status) {
			case PermanentWorkerError:
				{
					for(std::vector<WorkerT>::iterator i = workers_.begin(); i != workers_.end();i++){
						if(*i == worker){
							workers_.erase(i);
							std::cout << "Permanent worker error: removed worker from scheduling" << std::endl;
							break;
						}
					}				
				}
				break;
			case PermanentJobError:
				{
					std::cout << "Permanent Job error! This should never happen.." << std::endl;
				}
				break;
			default:
				break;
		}
		
		createdCalls_--;
	}
	
	void PiManager::takeResult(const PiManager::JobIDT jobID, const PiManager::PartialResultT result){
		totalProcessedRequests_ += requestsPerWorker_;
		successfulProcessedRequests_ += result;
		pi_ = ((double)successfulProcessedRequests_ / (double)totalProcessedRequests_) * 4.0;
		printf("PI: %.15E\n", pi_);
	}
	
	PiManager::ResultT PiManager::work(const PiManager::ParameterT totalRequestCount){
		neededCalls_ = 10;
		requestsPerWorker_ = totalRequestCount / neededCalls_;
		requestCount_ = totalRequestCount;
		World<PiManager> world(this);
		try{
			world.spin();
		} catch(std::exception &e){
			std::cout << "!!Exception in World loop!! " << e.what() << std::endl;
		}
		return pi_;
	}
	
	PiManager::~PiManager(){
		/*for(std::vector<WorkerT*>::iterator i = workers_.begin(); i != workers_.end();i++){
			delete *i;
			i++;
		}*/
	}
};
