#include <ctl/ci/evalarray.ci>
#include <ctl/ci/paint2d.ci>

#include <string>
#include <map>


int main()                                                                                  
{
	std::map<std::string, std::string> src;

	src["f.h"] = 
"\
double f(double x)\
{\
	int n=int(x), z=0;\
\
	while(n!=1)\
	{\
		if(n&1)\
		{ n+=(n+1)>>1; z+=2; }\
		else\
		{ n>>=1; z++; }\
	}\
	return z/log(x+1);\
};\
";

printf(src["f.h"].c_str());

	ctl::vector<ctl::location> location=ctl::readLocation("./rsc.txt");
	ctl::link builder(location[0]);

	if(!builder)
	{
		printf(" could not open \"builder\"\n");
		return -1;
	}
	std::string f_lib;
	try
	{
		printf("building eval\n");
		ctl::result<std::string> rf_lib(evalBuilder::build(builder, src) );
		f_lib = rf_lib;
	}
	catch(ctl::exception &e)
	{ std::cout<<e<<"\n"; }	

	ctl::vector<double> x(12001), y;
	for(size_t i=0; i<x.size(); i++)
		x[i] = i+1;
std::cout<<f_lib<<"?\n";

	ctl::link worker(f_lib);	
	try {
		printf("invoking eval\n");
		ctl::result<ctl::vector<double> > 	ry(eval::f<double,double>(worker, x) );
		y = ry;
	}
	catch(const ctl::exception &e)
	{ std::cout<<e<<"\n"; exit(-1); }	

	ctl::link viewer(location[1]);
	if(!viewer)
	{
		printf(" could not open 'painter'\n");
		return -1;
	}
	try {
		ctl::vector<float> X(x), Y(y);
		printf("invoking paint\n");
		paint2dCI painter(viewer);
		painter.run();
		painter.setPen(1,2);
		painter.paint(X,Y, true);

		std::cerr<<"please enter <return>\n";
		getchar();
	}
	catch(const ctl::exception &e)
	{ std::cout<<e<<"\n"; }	
	return 0;
}

