#include "types.h"
#include <assert.h>
#include <stdio.h>
#define ndof 3
#define ncpl 1
#define npar 2

struct System
{
double x[ndof];
double b[ndof];
double y[ncpl];
int    i[ncpl];
int    p[ncpl];
double mat[ndof][ndof];
};

struct System sys;

void simu_init_impl(int8 *sfilename, char *filename)
{
	int i;
	printf(" simu.c: '%s'\n",filename);
// set initial values (e.g. by reading filename)
	for(i=0; i<ndof; i++)
		sys.x[i] = sys.b[i] = 0;

	sys.b[0] = 1;
	for(i=0; i<ncpl; i++)
	{
		sys.y[i] = 0;
		sys.i[i] = 2;
	}
}

void simu_new1_impl(int8 *sfilename, char *filename)
{ return simu_init_impl(sfilename, filename); }

void simu_delete_impl()
{}

void simu_new2_impl(int8 *sfilename, char *filename, int *npara, double *para)
{ return simu_init_impl(sfilename, filename); }

void simu_getparam_impl(int8* sp,real8* p) 
{
	int i=0;
	if(*sp==npar)
		for(i=0; i<npar; i++)
			p[i] = sys.p[i]; 
	else
		*sp = npar;
}

void simu_setparam_impl(int8* sp,real8* p) 
{
	int i=0;
	assert(*sp==npar);
	for(i=0; i<npar; i++)
		sys.p[i] = p[i];
}

void simu_getstate_impl(int8* sx,real8* x) 
{
	int i=0;
	if(*sx==ndof)
		for(i=0; i<ndof; i++)
			x[i] = sys.x[i]; 
	else
		*sx = ndof;
}

void simu_setstate_impl(int8* sx,real8* x) 
{
	int i=0;
	assert(*sx==ndof);
	for(i=0; i<ndof; i++)
		sys.x[i] = x[i]; 
}

void simu_setload_impl(int8* sload,real8* load) 
{
	int i=0;
	assert(*sload==ndof);
	for(i=0; i<ndof; i++)
		sys.b[i] = load[i]; 
}

void simu_getcoupling_impl(int8 *si, int4 *I,int8* sy,real8* y)
{
	int i=0;
	if(*si==ncpl && *sy==ncpl)
		for(i=0; i<ncpl; i++)
		{
			I[i] = sys.i[i];
			y[i] = sys.x[I[i]]; 
		}
	else
		*si = *sy = ncpl;
}

void simu_setcoupling_impl(int8* sy,real8* y)
{
	int i=0;
	assert(*sy==ncpl);
	for(i=0; i<ncpl; i++)
		sys.y[i] = y[i]; 
}

void simu_residual_impl(int8* sr,real8* r) 
{
	int i,j;
/*	printf("residual called with %d\n", *sr); */
	if(*sr<ndof)
	{
		*sr = ndof;
		return;
	}
	for(i=0; i<ndof; i++)
		r[i] = -sys.b[i];
	for(i=0; i<ndof-1; i++)
	{
		r[i]   += sys.p[i] * (sys.x[i] - sys.x[i+1]);
		r[i+1] += sys.p[i] * (sys.x[i+1] - sys.x[i]);
	}
/*	printf("residual res = (%e %e %e) \n", r[0], r[1], r[2]); */
}

void simu_solve_impl(real8* eps, int8* sx,real8* x) {}

void simu_timestep_impl(real8* eps, real8* dt, int8* sr, real8* r, int8* sx,real8* x) {}

void simu_precond_impl(int8* sr,real8* r, int8* spr,real8* pr) {}

void simu_getnodes_impl(int8* dim,int *scoor, real8* ccor) {}

void simu_dirderivative_impl(int8* sx0,real8* x0, int8* sdx,real8* dx, int8* sdr,real8* dr) {}

void simu_command_impl(int4 *sc, char *s, int *ret) {}
