
#ifndef _PENALTY_SOLVER_H_
#define _PENALTY_SOLVER_H_

#include "qns.h"

namespace solver
{

class penaltyResiduum
{
  ctl::group groupM;
  mutable simuCI<double> simuM;
  double lambdaM;
  ctl::link partnerM;
  int ndofM,ncplM;

public:
// Implements the parallel scalar product
  real scalMult(const realVec &x, const realVec &y) const
  {
    real s=0;
    for(unsigned int i=0; i<x.size(); i++)
      s += x[i]*y[i];
// build global sum of local scalar products
    s = groupM + s;
    return s;
  }
// Implements the parallel residual function
  penaltyResiduum(const ctl::group &grp,simuCI<double> sim, double lambda):
          groupM(grp), simuM(sim), lambdaM(lambda)
  {
    realVec x,y;
    intVec I;
    x = simuM.getstate();
    ndofM = x.size();

    simuM.getcoupling(I,y);
    ncplM = y.size();

    int partnerID = 1 - grp.logId();
    partnerM = grp[partnerID];
  }
  realVec operator()(const realVec &xc) const
  {
    realVec x=xc, res(ndofM), y0(ncplM), y1;
    intVec I(ncplM);
    if(!x)
      x.resize(ndofM);

    simuM.setstate(x);
// exchange coupling data
    simuM.getcoupling(I,y0);
    partnerM<<y0;
    partnerM>>y1;

    res = simuM.residual();
// add the penalty term
    for(int i=0; i<ncplM; i++)
      res[ I[i] ] += lambdaM*(y0[i] - y1[i]);
    return res;
  }
};
} // solver

#endif
