/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_BINARYTYPE_H_
#define CTL_BINARYTYPE_H_

#include "utl/types.h"
#include "gen/ctl.mcr"
#include "utl/pointer.h"
#include "system.h"
#include "utl/serialize.inl"
#include <complex>

namespace ctl
{

template<class T> struct cstring;
template<class T, int8 D1= -1, int8 D2= 0, int8 D3= 0, int8 D4= 0, int8 D5= 0, int8 D6= 0, int8 D7= 0, int8 D8= 0> struct array;
template<class T> struct reference;
template<class T0, class T1, class T2=ctl::empty, class T3=ctl::empty,
			class T4=ctl::empty, class T5=ctl::empty, class T6=ctl::empty, class T7=ctl::empty,
 			class T8=ctl::empty, class T9=ctl::empty, class T10=ctl::empty, class T11=ctl::empty,
			class T12=ctl::empty, class T13=ctl::empty, class T14=ctl::empty, class T15=ctl::empty> struct tupel;

	char isBinary(const void*);
	struct binarybase
	{};
	int isBinary(const binarybase*);
	template<class T = ctl::empty> struct isBinaryRep
	{
		struct sink { sink(const binarybase&); };
		enum { size = sizeof(isBinary((T*)0) ) };
		enum { value = size>1 };
		typedef ctl::int2type<value> resultT;
		typedef typename ctl::select<value, T, ctl::empty>::resultT sw;
		typedef typename ctl::select<value, sink, T>::resultT ws;
	};
	template<int n> struct property<hasBinaryCheck,char[n]>
	{
		enum { value = true };
		typedef ctl::int2type<value> resultT;
	};
	
	template<> struct isBinaryRep<void>
	{
		enum { value = false };
		typedef ctl::int2type<value> resultT;
		typedef ctl::empty sw;
	};
	char constCheck(void*);
	int constCheck(const void*);

	struct binaryType: public binarybase
	{
		enum { maxSize=16 };
		union { mutable char objBufM[maxSize]; double d;};
		mutable lPointer<ctl::iStream> bufferM;
		mutable void *objM;
		const std::type_info *typeM;
		ctl::oStream & (*writeM)(ctl::oStream &, const void*);
  	ctl::iStream &(*readM)(ctl::iStream &is, void *&);
		typedef void( *delT)(const void *);
 		void *(*cloneM)(const void *obj, void *mem, delT&);
 		mutable delT delM;

		template<typename T> struct genericIOFunctionBase
		{
			static void del(const void* obj)
			{ delete (T*)obj; }
			static void delVec(const void* obj)
			{  delete[] (T*)obj; }
			static void dtr(const void* obj)
			{ ((T*)obj)->~T(); }
			static void* clone(const void *obj, void *mem, delT &d )
			{ 
 				enum { abstract = conversion<abstractT<T>, T>::exist };
				void *p=0;
				T *r=0;
				if(sizeof(T)<=maxSize && mem)
				{	d = &dtr; p = mem; }
				else
					d = &del;
				cloneObject(r, p,*(const T*)obj, int2type<abstract>() );
				return r;
			}
		};
	  template<typename T, class Bin> struct genericIOFunction: genericIOFunctionBase<T>
	  {
			typedef Bin CTL_identT;
			mutable T t;
			operator T&() const
			{ return t; }
			ctl::oStream& write(ctl::oStream &os) const
			{ return Bin::write_(os, t); }
			ctl::iStream& read(ctl::iStream &is)
			{ return Bin::read_(is, t); }
			
   		static ctl::oStream &write(ctl::oStream &os, const void* obj)
	 		{ return Bin::write_(os, *(T*)obj); }
   		static ctl::iStream &read(ctl::iStream &is, void*& obj)
			{ return Bin::read_(is, *(T*)obj); }
	  	 };

		void clone() const
		{
			delT del = 0;
			void *cpy = (*cloneM)(objM, objBufM, del);
			if(delM)
				(*delM)(objM);
			objM = cpy;
			delM = del;
		}
		
		template<typename U, class Bin> void set2(const U &u, bool alloc, falseT, Bin*)
		{
			typeM=&typeid(U);	
			writeM = &genericIOFunction<U, Bin>::write; 
			readM  = &genericIOFunction<U, Bin>::read; 
			cloneM = &genericIOFunction<U, Bin>::clone;
			if(alloc)
			{
				delM = &genericIOFunction<U, Bin>::del;
				objM = (*cloneM)(&u, objBufM, delM);
			}
			else
				delM = 0;
		}
// remove the 'const'
		template<typename U, class Bin> void set2(const U *u, bool alloc, trueT, Bin*)
		{	set2( (U*)u, alloc, falseT(), (Bin*)0); }

		template<typename U, class Bin> void set(const U &u, bool alloc, falseT, Bin*)
		{ objM = (void*)&u; set2((U&)u, alloc, int2type<pointerTraits<U>::is>(), (Bin*)0 ); }

		void assign(const binaryType &b, bool, trueT)
		{ 
			bufferM = b.bufferM; 
			if(b.objM == b.objBufM)
			{
				if(objM && objM==objBufM)
					return;
				memcpy(objBufM, b.objBufM, maxSize);
				objM = objBufM;
			}
			else
				objM = b.objM;
			typeM=b.typeM; writeM=b.writeM; 
			readM=b.readM; cloneM=b.cloneM; 
			delM=b.delM; b.delM = 0; 
		}

		binaryType(): objM(0), typeM(0),writeM(0),readM(0), cloneM(0), delM(0) {}
		binaryType(const binaryType &b): objM(0), typeM(0),writeM(0),readM(0), cloneM(0), delM(0)
		{ assign(b, true, trueT() ); }
		template<typename U, class Bin>  explicit binaryType(const U &u, bool alloc, Bin*): objM((void*)&u), typeM(0),writeM(0),readM(0), cloneM(0), delM(0)
		{ 
			enum { bin=sizeof(isBinary((const U*)0) )>1 };
			set(u, alloc, int2type<bin>(), (Bin*)0);
		}
// U is referenceRef
		template<typename U, class Bin>  explicit binaryType(U &u, trueT, Bin*): objM((void*)&u), typeM(0), writeM(0),readM(0), cloneM(0), delM(0)
		{	set2(u,false,falseT(), (Bin*)0 ); }
		~binaryType()
		{
			if(delM)
				(*delM)(objM);
		}
		binaryType &operator=(const binaryType &b)
		{ assign(b, true, trueT() ); return *this;}

		ctl::oStream &write(ctl::oStream &os) const
		{
			if(writeM)
				return (*writeM)(os, objM);
			return os;
		}
		ctl::iStream &read(ctl::iStream &is)
		{
			if(readM && objM)
			{ return (*readM)(is, objM); }
			bufferM = lPointer<iStream>(new iStream(is) );
			bufferM->commuM = is.commuM;
			bufferM->seek(is.tell());
			return is;
		}
		template<class U> void cast_(U &u) const
		{
			if( objM &&  equal(typeid(U), *typeM) )
				u = *(U*)(const U*)objM;
			else
			{
				if(writeM)
				{
					ctl::oStream os=ctl::serialize(*this);
					ctl::iStream is(os);
	// if(typeof(u) in { [const]char*, [const]char**) owner
					is>>u;
				}
				else
				{
					 if(!bufferM)
						throwStr(" *** returning empty binary ***");
					ctl::iStream& is=(ctl::iStream&)*bufferM; 
					is>>u; 
				}
			}	
		}

		template<class U, class Bin> U cast__(const U*, Bin*) const
		{

			if( objM &&  equal(typeid(U), *typeM) )
				return *(U*)(const U*)objM;
			if(writeM)
			{
				ctl::oStream os=ctl::serialize(*this);
				ctl::iStream is(os);
	// if(typeof(u) in { [const]char*, [const]char**) owner
				U u; 
				Bin::read_(is, u);
				return u;
			}
			else
			{
				 if(!bufferM)
					throwStr(" *** returning empty binary ***");
	// if(typeof(u) in { [const]char*, [const]char**) owner
				U u; 
				ctl::iStream& is=(ctl::iStream&)*bufferM; 
				is>>u; 
				return u;
			}	
		}
		template<class U, class Bin> void recast_(const U &u, Bin*)
		{
			if( equal(typeid(U), *typeM) )
				*(U*)objM = u;
			else
			{
				ctl::oStream os=ctl::serialize(u);
				ctl::iStream is(os);
				read(is);
			}	
		}
	};

	template<class Y> int congruence(const Y*, const void*)
	{
		enum { value = ctl::same<Y,void>::value };
		ctl::Assert<value> CTL_ERROR__no_cast_available;
		(void) CTL_ERROR__no_cast_available;
		return 0;
	}
	
// set binary Type of pointer
	template<class X, class Bin> inline char  congruence(const Bin*, X *const*)
	{ 	return Bin::referenceCheck((X*)0);	}
	template<class charT1, class charT2> inline char  congruence(const cstring<charT1>*, charT2 *const*)
	{ 	return cstring<charT1>::stringCheck((charT2*)0);	}

#ifndef CTL_PartialSpec
	template<class T> class noReference
	{
	public:
		enum { size = sizeof(T) };
	private:
		struct { char d[size]; } dat;
		mutable bool owner;
		template<class X> static void init(const X &x)
		{ new((void*)&x) X();}
		template<class X> static void del(X &x)
		{ x.~X(); }
	public:
		typedef typename ctl::select<ctl::referenceTraits<T>::is, noReference<T>, T>::resultT resultT;
		noReference(): owner(true)
		{ init((T)dat);}
		noReference(const noReference<T> &nr): dat(nr.dat), owner(true)
		{ nr.owner = false; }
		~noReference()
		{ if(owner) del((T)dat);}
		operator T() const
		{ return (T)dat; }
	};
	template<class T, class Bin> inline char  congruence(const Bin*, const noReference<T> *s)
	{ 	return congruence((Bin*)0, &(T)(*s) );	}

#else
	template<class T> struct  noReference
	{ typedef typename unRefConst<T>::resultT resultT; };
#endif


template<class T> inline ctl::oStream &write(ctl::oStream &os, const ctl::noReference<T> &x)
{ return os << (T)x; }

template<class T> inline ctl::iStream &read(ctl::iStream &is, ctl::noReference<T> &x)
{ return is >> ctl::unconst((T)x); }

namespace internal
{
		template<class U, class Bin> struct binSwitch
		{
			enum { same = ctl::same<U,Bin>::value, U_hasCheck= ctl::property<hasBinaryCheck,U>::intrinsic, sw=same ? 2 : U_hasCheck};
			typedef ctl::int2type<sw> resultT;	
		};
		template<class U, class Bin> char verifyCongruence(const U*, const Bin*, ctl::falseT)
		{
			enum { isBin=(sizeof(isBinary((const Bin*)0) )>1) & (sizeof(congruence((const Bin*)0, (const U*)0))==1) };
			Assert<isBin>   CTL_ERROR__non_congruent_types;
			(void) CTL_ERROR__non_congruent_types;
			(void)congruence((const Bin*)0, (const U*)0); return 0;
		}
		template<class U, class Bin> char verifyCongruence(const U *u, const Bin*, ctl::int2type<1>)
		{ (void)u->congruence((const Bin*)0); return 0; }
		template<class U, class Bin> char verifyCongruence(const U*, const Bin*, ctl::int2type<2>)
		{ return 0; }
} // internal

#define CTL_VerifyCongruence(_U,_oT)\
{\
	typedef typename ctl::binArgH<_oT,isBinaryRep<_oT>::value> ::ncT _T;\
	typedef typename internal::binSwitch<_U,_T>::resultT _sw;\
	char (*_x)(const _U*, const _T*, _sw) = &internal::verifyCongruence;\
	(void)(&_x); 	\
}

	template<class T>	struct retString;
	template<class T>	struct cstringRef;
	template<class T>	struct retReference;
	template<class T>	struct referenceRef;
	template<class T, int8 D1= -1, int8 D2= 0, int8 D3= 0, int8 D4= 0, int8 D5= 0, int8 D6= 0, int8 D7= 0, int8 D8= 0> struct arrayRef;
	template<class T, int8 D1= -1, int8 D2= 0, int8 D3= 0, int8 D4= 0, int8 D5= 0, int8 D6= 0, int8 D7= 0, int8 D8= 0> struct retArray;
	template<class T0, class T1, class T2, class T3, class T4, class T5, class T6, class T7, class T8, class T9, class T10, class T11, class T12, class T13, class T14, class T15> struct retTupel;
	template<class T0, class T1, class T2, class T3, class T4, class T5, class T6, class T7, class T8, class T9, class T10, class T11, class T12, class T13, class T14, class T15> struct tupelRef;

	template<class oT, bool bin> struct binArgH
	{
		struct identT
		{ typedef oT CTL_identT;  };
		typedef typename ctl::select<bin, oT, identT >::resultT::CTL_identT ncT;
	};
	template<class oT> struct binArg
	{
		enum { c=sizeof(constCheck((oT*)0 ) )>1 , bin=sizeof(isBinary((const oT*)0) )>1,	binRef = !c & bin};
		typedef typename binArgH<oT, bin>::ncT ncT;
		typedef typename ctl::select<c, const ncT, ncT >::resultT T;
		struct ref
		{ typedef T& refType;  };
		struct ident
		{ typedef T refType;  };
		typedef typename ctl::select<binRef, T, ref >::resultT::refType resultT;
		typedef typename ctl::select<binRef, T, ident >::resultT::refType thisT;
	};

	template<class oY> struct returnType
	{
		enum { isBin = isBinaryRep<oY>::value };
		struct ident
		{
			typedef typename ctl::noVoid<oY>::resultT retType;
			typedef oY CTL_identT;
		};
		typedef typename ctl::select<isBin, oY, ident >::resultT::CTL_identT Y;
		struct ret
		{	typedef Y retType; };

		typedef typename ctl::select<isBin, Y, ident >::resultT::retType resultT;	
		typedef typename ctl::select<isBin, Y, ret>::resultT::retType binType;	
	};


	template<class Bin, class Y, class  Y_, class  isRef> struct binaryCaster
	{
		class temporary
		{
			temporary& operator=(const temporary&);
		public:
			ctl::binaryType &binM;
			Y_  yM;
			temporary(Bin &bin): binM(bin), yM(binM.cast__((const Y_*)0, (Bin*)0) )
			{	CTL_VerifyCongruence(Y_, Bin)	}
			operator Y()
			{ return yM; }
			void recast(trueT)
			{ ((ctl::binaryType &)binM).recast_(yM, (Bin*)0); }
			void recast(falseT)
			{}
			~temporary()
			{ recast(isRef()); }
		};
		struct constTemporary
		{
			Y_  yM;
			constTemporary(const Bin &bin): yM(bin.cast__((const Y_*)0, (Bin*)0) )
			{	CTL_VerifyCongruence(Y_, Bin)		}
			operator Y()
			{ return yM; }
		};
	};

	} //ctl

	// set some standard tyes/templates	
CTL_SetType(std::string, isString, (char), 1);
#ifdef CTL_wchar
CTL_SetType(std::wstring, isString, (wchar_t), 1);
#endif
CTL_SetType(ctl::string, isString, (char), 1);

CTL_SetTemplate(std::vector,(T, alloc),2 ,isArray, (T), 1); 
CTL_SetTemplate(std::list,(T, alloc),2 ,stdArray, (T), 1); 
CTL_SetTemplate(std::set,(T, cmp, alloc),3 ,isArray, (T), 1); 
CTL_SetTemplate(std::queue,(T, seq),2 ,isArray, (T), 1); 
CTL_SetTemplate(ctl::vector,(T),1, isArray, (T), 1); 
CTL_SetTemplate(std::pair,(T0, T1),2 ,tupel, (_.first, _.second), 2); 
CTL_SetTemplate(std::complex,(T),1 ,tupelCtor, (_.real(), _.imag() ), 2); 
CTL_SetTemplate(std::map,(key, val, cmp, alloc),4 ,isArray,((std::pair, (key, val), 2)), 1); 
CTL_SetTemplate(std::auto_ptr,(T),1 ,isReference,(T), 1); 
CTL_SetTemplate(ctl::cPointer,(T),1 ,isReference,(T), 1); 
CTL_SetTemplate(ctl::lPointer,(T),1 ,isReference,(T), 1); 


namespace ctl
{
namespace internal
{
	template<class T> inline bool isZero(const T &t)
	{ return !t; }
	template<> inline bool isZero(const std::string &t)
	{ return t.empty(); }
	template<> inline bool isZero(const ctl::string &t)
	{ return t.empty(); }
 	template<class T> struct strType
	{
		ctl::vector<T> v;
		typedef typename ctl::vector<T>::const_iterator const_iterator;
		static size_t strsize(const T *t)
		{
			if(!t)
				return 1;
			const T* i=t;
			while(!internal::isZero(*i++)) ;
			return i-t;
		}	
		strType() {}
		strType(const T* const t): v(t, t+strsize(t) )
		{}
		oStream &write(oStream &os) const
		{
			const_iterator t=v.begin();
			for(size_t i=0; i<v.size(); i++)
				os<<t[i];
			return os;
		}
		iStream &read(iStream &is)
		{
			v.clear();
			T t;
			do
			{ is>>t; v.push_back(t); 	} while(!internal::isZero(t));
			return is;
		 }
		T* getCopy() const
		{
			T* n=new T[v.size()+1];
			const_iterator t=v.begin();
			for(size_t i=0; i<v.size(); i++)
				n[i] = *t++;
			return n;
		}
	};
}
	template<class T> struct cstring: public ctl::binaryType
	{
	 CTL_TemplateNameS(cstring, (T), 1);
		typedef retString<T>  retType;
		typedef cstringRef<T>  refType;
		typedef cstring<T>  thisT;
		typedef thisT CTL_identT;

		cstring(const T * const &s):  ctl::binaryType(internal::strType<T>((T*)s),true, (CTL_identT*)0) {}
		template<class U> static char stringCheck(const U *)
		{ CTL_VerifyCongruence(U,T); return 0; }
		cstring() {}
		template<class U> static const U* verify(const U *u,ctl::trueT)
		{ return u; }
		template<class U> static const U& verify(const U &u,ctl::falseT)
		{ CTL_VerifyCongruence(U, thisT); return u; }
		template<class U> cstring(const U &u, bool alloc=false): 
						ctl::binaryType(u ,alloc, (CTL_identT*)0)
		{ verify(u,int2type<pointerTraits<U>::is>() ); }
		cstring(const  thisT &a): ctl::binaryType((const ctl::binaryType&)a)
		{}

		template<class U> static ctl::oStream& write__(ctl::oStream &os, U* u, trueT)
		{  
			internal::strType<U> s(u); 
			return os<<s; 
		}
		static ctl::oStream& write__(ctl::oStream &os, char** u, trueT)
		{ return os<<u;	}
		template<class U> static ctl::oStream& write__(ctl::oStream &os,const U& u, falseT)
		{ return os<<u; }
		template<class U> static ctl::oStream& write_(ctl::oStream &os,const U& u)
		{ return write__(os, u, int2type<ctl::pointerTraits<U>::is>() ); }

		template<class U> static ctl::iStream& read_(ctl::iStream &is, U& u)
		{ return is>>u; }
		template<class U> static ctl::iStream& read_(ctl::iStream &is, U*& u)
		{ 
			internal::strType<U> s;
			is>>s;
			u=s.getCopy();
			return is;
	 }
		static ctl::iStream& read_(ctl::iStream &is, char* &u)
		{  return is>>u; }
		static ctl::iStream& read_(ctl::iStream &is, char** &u)
		{  return is>>u; }
	};	

	template<class T> struct reference: public ctl::binaryType
	{
	 CTL_TemplateNameS(reference, (T), 1);
		typedef retReference<T>  retType;
		typedef referenceRef<T>  refType;
		typedef reference<T>  thisT;
		typedef thisT CTL_identT;
		template<class U> reference(const U &u, bool alloc=false): ctl::binaryType(u, alloc, (CTL_identT*)0)
		{ CTL_VerifyCongruence(U, thisT); }
		template<class U> reference(const U &u, trueT ): ctl::binaryType(u, trueT(), (CTL_identT*)0)
		{ CTL_VerifyCongruence(U, thisT); }
		template<class U> static char referenceCheck(const U *)
		{
			enum { isBase = ctl::conversion<const T*, U*>::exist };
			ctl::Assert<isBase> CTL_ERROR_is_no_base;
			(void)CTL_ERROR_is_no_base;
			return 0;
		}
		reference() {}
		reference(const thisT &a): ctl::binaryType((const ctl::binaryType&)a)
		{}

		template<class U> static ctl::oStream& write_(ctl::oStream &os, const U& u)
		{ return os<<u; }
		template<class U> static ctl::iStream& read_(ctl::iStream &is, U& u)
		{ return is>>u; }
	};	

	template<class T, int8 D1, int8 D2, int8 D3, int8 D4, int8 D5, int8 D6, int8 D7, int8 D8> struct array: public ctl::binaryType
	{
	 CTL_TemplateNameS(array, (T, D1, D2, D3, D4, D5, D6, D7, D8), 9);
		typedef arrayRef<T, D1, D2, D3, D4, D5, D6, D7, D8>  refType;
		typedef retArray<T, D1, D2, D3, D4, D5, D6, D7, D8>  retType;
		typedef array<T, D1, D2, D3, D4, D5, D6, D7, D8>  thisT;
		typedef thisT CTL_identT;
		array() {}

		template<class U> static char arrayCheck__(const U*, trueT)
		{ return 1; }
		template<class U> static char arrayCheck__(const U*, falseT)
		{
			CTL_VerifyCongruence(U, T);
			return 1;
		}
		template<class U, class intT> static char arrayCheck_(const U*, trueT, intT)
		{ 
			enum { c=same<const T, const U>::value };
			return arrayCheck__((const U*)0, int2type<c>()); 
		}
		template<class U, class intT> static char arrayCheck_(const U *, falseT, intT)
		{
			typedef array<T, D2, D3, D4, D5, D6, D7, D8> tailT;
			enum { fixedDims = (D2>=0) & (D3>=0) & (D4>=0) & (D5>=0) & (D6>=0) & (D7>=0) & (D8>=0) };
			ctl::Assert<fixedDims> CTL_ERROR__Assuming_fixed_dimensions;
			(void)CTL_ERROR__Assuming_fixed_dimensions;
			CTL_VerifyCongruence(U, tailT);
			return 1;
		}

		template<class U, class intT> static char arrayCheck(const U*, intT*)
		{ 
			enum { c=same<intT, int2type<D1, D2, D3, D4, D5, D6, D7, D8> >::value };
			return arrayCheck_((const U*)0,  int2type<c>(), intT()); 
		}
		template<class U> array(const U &u, bool alloc=false): ctl::binaryType(u, alloc, (CTL_identT*)0)
		{ CTL_VerifyCongruence(U, thisT); }
		array(const array<T, D1, D2, D3, D4, D5, D6, D7, D8> &a): ctl::binaryType((const ctl::binaryType&)a)
		{}

		template<class U> static ctl::oStream& write_(ctl::oStream &os, const U& u)
		{ return os<<u; }
		template<class U> static ctl::iStream& read_(ctl::iStream &is, U& u)
		{ return is>>u; }
	};	



	template<class T0, class T1, class T2, class T3, class T4, class T5, class T6, class T7,
						class T8, class T9, class T10, class T11, class T12, class T13, class T14, class T15> struct tupel: public ctl::binaryType
	{
	 CTL_TemplateNameS(tupel, (T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15), 16);
		typedef retTupel<T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15>  retType;
		typedef tupelRef<T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15>  refType;
		typedef tupel<T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15>  thisT;
		typedef thisT CTL_identT;
		tupel() {}
    	tupel(const thisT &t): ctl::binaryType((const ctl::binaryType&)t)
    	{}
		template<class U> tupel(const U &u, bool alloc=false): ctl::binaryType(u, alloc, (CTL_identT*)0)
		{ CTL_VerifyCongruence(U,thisT); }
#define CTL_TupelCheckH(U, n)  CTL_VerifyCongruence(U##n,T##n);
#define CTL_TupelCheckArg(U, n)  const U##n*
#define CTL_TupelCheck(n)\
		CTL_Tmpl_Dclr2((CTL_List(U,n)),n) static char tupelCheck(CTL_ExprList(CTL_TupelCheckArg, U, Comma, n) )\
		{\
			CTL_ExprList(CTL_TupelCheckH, U, Space, n)\
			Assert<ctl::same<T##n, ctl::empty>::value>   CTL_ERROR__non_congruent_types;\
			(void) CTL_ERROR__non_congruent_types;\
			return 0;\
		}
		CTL_TupelCheck(1)
		CTL_TupelCheck(2)
		CTL_TupelCheck(3)
		CTL_TupelCheck(4)
		CTL_TupelCheck(5)
		CTL_TupelCheck(6)
		CTL_TupelCheck(7)
		CTL_TupelCheck(8)
		CTL_TupelCheck(9)
		CTL_TupelCheck(10)
		CTL_TupelCheck(11)
		CTL_TupelCheck(12)
		CTL_TupelCheck(13)
		CTL_TupelCheck(14)
		CTL_TupelCheck(15)
//		CTL_TupelCheck(16)
#undef  CTL_TupelCheck
#undef  CTL_TupelCheckH
#undef  CTL_TupelCheckArg

		template<class U> static ctl::oStream& write_(ctl::oStream &os, const U& u)
		{ return os<<u; }
		template<class U> static ctl::iStream& read_(ctl::iStream &is, U& u)
		{ return is>>u; }

	};

	template<class T> struct cstringRef: public cstring<T>
	{
		typedef cstring<T> baseT;
		typedef cstringRef<T>  thisT;
		cstringRef() {}
		template<class U> cstringRef(const U &u)
		{
			ctl::Assert<false> CTL_ERROR__discarding_const_requested;
			(void) CTL_ERROR__discarding_const_requested;
		}
		template<class U> cstringRef(U &u, bool alloc=false): baseT(u,alloc)
		{}
#ifdef CTL_TmplBeforeConst
		cstringRef(cstringRef<T> &a): baseT((const baseT&)a)
		{}
#endif
		cstringRef(const thisT &a): baseT((const baseT&)a)
		{}
	};		


	template<class T, int8 D1,int8 D2, int8 D3, int8 D4, int8 D5, int8 D6, int8 D7, int8 D8> struct arrayRef: public array<T,D1, D2, D3, D4, D5, D6, D7, D8>
	{
		typedef array<T, D1, D2, D3, D4, D5, D6, D7, D8> baseT;
		arrayRef() {}
		template<class U> arrayRef(const U &)
		{
			ctl::Assert<false> CTL_ERROR__discarding_const_requested;
			(void) CTL_ERROR__discarding_const_requested;;
		}
		template<class U> arrayRef(U &u, bool /*alloc*/=false): baseT(u)
		{}
#ifdef CTL_TmplBeforeConst
		arrayRef(arrayRef<T, D1, D2, D3, D4, D5, D6, D7, D8> &a): baseT((const baseT&)a)
		{}
#endif
		arrayRef(const arrayRef<T, D1, D2, D3, D4, D5, D6, D7, D8> &a): baseT((const baseT&)a)
		{}
	};
	
	template<class T> struct referenceRef: public reference<T>
	{
		typedef reference<T> baseT;
		typedef referenceRef<T>  thisT;
		referenceRef() {}
		template<class U> referenceRef(const U &u)
		{
			ctl::Assert<false> CTL_ERROR__discarding_const_requested;
			(void) CTL_ERROR__discarding_const_requested;
		}
		template<class U> referenceRef(U &u): baseT(u,trueT())
		{}
#ifdef CTL_TmplBeforeConst
		referenceRef(referenceRef<T> &a): baseT((const baseT&)a)
		{}
#endif
		referenceRef(const referenceRef<T> &a): baseT((const baseT&)a)
		{}
	};		

	template<class T0, class T1, class T2, class T3, class T4, class T5, class T6, class T7,
					class T8, class T9, class T10, class T11, class T12, class T13, class T14, class T15> struct tupelRef:
							public tupel<T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15>
	{
		typedef tupel<T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15>  baseT;
		typedef tupelRef<T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15>  thisT;

		tupelRef() {}
		template<class U> tupelRef(const U &)
		{
			ctl::Assert<false> CTL_ERROR__discarding_const_requested;
			(void) CTL_ERROR__discarding_const_requested;;
		}
		template<class U> tupelRef(U &u, bool alloc=false): baseT(u, alloc)
		{}
#ifdef CTL_TmplBeforeConst
		tupelRef(thisT &a): baseT((const baseT&)a)
		{}
#endif
		tupelRef(const thisT &t): baseT((const baseT&)t)
		{}
	};

	template<class T> struct retReference: public reference<T>
	{
		typedef  reference<T> baseT;
		retReference() {}
		retReference(const retReference &r): baseT((const baseT&)r)
		{}
		template<class U> retReference(const U &u): baseT(u, true)
		{}	

	};

	template<class T> struct retString: public cstring<T>
	{
		typedef  cstring<T> baseT;
		retString() {}
		retString(const retString &r): baseT((const baseT&)r)
		{}
		template<class U> retString(const U &u): baseT(u, true)
		{}	
	};

	template<class T, int8 D1,int8 D2, int8 D3, int8 D4, int8 D5, int8 D6, int8 D7, int8 D8> struct retArray: public array<T, D1, D2, D3, D4, D5, D6, D7, D8>
	{
		typedef array<T, D1, D2, D3, D4, D5, D6, D7, D8>  baseT;
		retArray() {}
		retArray(const retArray<T, D1, D2, D3, D4, D5, D6, D7, D8> &r): baseT((const baseT&)r)
		{}
		retArray(const array<T, D1, D2, D3, D4, D5, D6, D7, D8> &r): baseT((const baseT&)r)
		{}
		template<class U> retArray(const U &u): baseT(u, true)
		{}	
	};

	template<class T0, class T1, class T2, class T3, class T4, class T5, class T6, class T7, class T8, class T9, class T10, class T11, class T12, class T13, class T14, class T15>
		struct retTupel: public tupel<T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15>
	{
		typedef tupel<T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15>  baseT;
		retTupel() {}
		retTupel(const retTupel &r): baseT((const baseT&)r)
		{}
		template<class U> retTupel(const U &u): baseT(u, true)
		{}	
	};	


	template<class Y, class X> inline Y binaryCast_(const X &x,ctl::trueT)
	{
		CTL_VerifyCongruence(Y,X);
		return x.cast__((const Y*)0,(X*)0);
	}
	template<class Y, class X> inline Y binaryCast_(const X &x,ctl::falseT)
	{ return Y(x); }

	template<class Y, class X> inline Y binaryCast(const X &x, const Y* =0)
	{
		typedef typename isBinaryRep<X>::resultT isBin;
		return binaryCast_<Y>(x, isBin());
	}

	template<bool b, class X, class Y> struct binarySelectH
	{ typedef X resultT; };

	template<class X, class Y> struct binarySelectH<true, X, Y>
	{ 
		enum { isConst = ctl::constTraits<X>::is };
		typedef binaryType::genericIOFunction<Y, typename X::CTL_identT> genX;
		typedef typename ctl::select<isConst, const genX, genX>::resultT resultT;
	};

	template<class X, class Y> struct binarySelect
	{
		typedef X Xref;
		enum { value = isBinaryRep<Xref>::value, isConst = ctl::constTraits<X>::is };
		typedef typename binarySelectH<value, X, Y>::resultT resultT;
	};

} // ctl

#endif
