/******************************************************************************

 This file is part of the Component Template Library (CTL)

 Copyright (C) 2006  Rainer Niekamp
 email:              r.niekamp@tu-bs.de

 The CTL is free software; you can redistribute it and/or modify it under the
 terms of the GNU Lesser General Public License as published by the Free Software
 Foundation; either version 2.1 of the License, or (at your option) any later
 version. We only ask that proper credit be given to the author.

 This library is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 for more details.

 You should have received a copy of the GNU Lesser General Public License along
 with this library; if not, write to the Free Software Foundation,
 Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

 ******************************************************************************/

#ifndef CTL_DMN_H_
#define CTL_DMN_H_

#include "ctl/utl/timer.h"
#include "ctl/inl/communicator.inl"

namespace ctl
{

	inline const location &myLocation();

	template<> class CommunicatorImpl<dmn> : public bCommunicator<
	    internal::ioBuffer>
	{
		public:
		void initDaemon(int port, bool verbose);
		protected:
		typedef internal::ioBuffer ioBuffer;
		typedef std::map<int/*port*/, socket> sockMap;
		mutable sockMap extLinkM;
		mutable socket socketM;
		mutable ctl::int4 portM;
		mutable PID ownerIdM;
		ctl::string hostM;
		ctl::string userIdM;
		PID synomIdM;
		mutable bool connectedM;
		double timeoutM;
		ctl::vector<location> componentM;
		socket *getWriter(const PID &pid, bool &ext) const
		{
			ext = false;
			return getSocket(pid, 0, ext);
		}
		socket *getReader(const PID &pid, int usec) const
		{
			bool b;
			return getSocket(pid, usec, b);
		}
		socket *getSocket(const PID &pid, int usec, bool &ext) const;
		bool disconnect(socket &s, int logid = -1)
		{
			return s.writeHd(ioBuffer(0, -1, myPIDM, logid));
		}
		bool accept(bool blocking, double dt = 0.0) const;
		bool
		    readAllCatch(const PID &source, bool blocking, bool first, int tag) const;
		bool readAll(const PID &source, bool blocking, bool first) const;
		public:
		bool addSocket(socket &);
		int nProcs() const
		{
			return extLinkM.size() + 1;
		}
		int connect(const PID &dest, const char *args, ctl::string &error) const;
		ctl::socket *getSocket() const
		{
			return &socketM;
		}
		int write_(oStream&, const PID & dest, int type) const;
		int test_(PID &source, int *type) const;
		int read_(iStream&, PID &source, int *type) const;
		CommunicatorImpl(int n, const char * const *arg);
		CommunicatorImpl(const PID & pid, const PID &logId);
		bool closed() const
		{
			return !isValid();
		}
		bool isValid() const
		{
			return !ownerIdM || extLinkM.find((int) ownerIdM) != extLinkM.end();
		}
		~CommunicatorImpl();
		channelT getType() const
		{
			return dmn;
		}
		void clean();
	};

} // ctl

#endif

