/******************************************************************************

 This file is part of the Component Template Library (CTL)

 Copyright (C) 2006  Rainer Niekamp
 email:              r.niekamp@tu-bs.de

 The CTL is free software; you can redistribute it and/or modify it under the
 terms of the GNU Lesser General Public License as published by the Free Software
 Foundation; either version 2.1 of the License, or (at your option) any later
 version. We only ask that proper credit be given to the author.

 This library is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 for more details.

 You should have received a copy of the GNU Lesser General Public License along
 with this library; if not, write to the Free Software Foundation,
 Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

 ******************************************************************************/

#ifndef CTL_DMN_INL_
#define CTL_DMN_INL_

#include "dmn.h"

namespace ctl
{

	CTL_Inline socket *CommunicatorImpl<dmn>::getSocket(const PID &pid, int usec, bool &ext) const
	{
		if(!!pid)
		{
			ext = true;
			sockMap::iterator link = extLinkM.find((int) pid);
			if(link != extLinkM.end()) return &link->second;
			return 0;
		}
		for(sockMap::iterator link = extLinkM.begin(); link != extLinkM.end(); link++)
			if(link->second.test(usec)) return &link->second;
		return 0;
	}

	CTL_Inline int CommunicatorImpl<dmn>::connect(const PID &dest, const char *args, ctl::string &error) const
	{
		internal::ioBuffer m(0, DAT, PID(myPIDM.port, magic().host), -1);

		log(envM, "#u ++ dmn::connect ->  (%d, %s) (%s)\n", dest.port, dest.host.c_str(), args);
		PID physDest = dest;
		if(!physDest) return -1;

		socket writer;
		if(!writer.connect(physDest.port, physDest.host.c_str(), false)) return -1;
		ctl::log(envM, "#lm ++ dmn <- %d@%s\n", physDest.port, physDest.host.c_str());

		//		ctl::log("#u connect args= %s\n", args);

		if(args) m.setData(args, ::strlen(args) + 1);
		if(!writer.writeHd(m)) return -1;

		internal::ioBuffer m2;
		if(!writer.readHd(m2))
		{
			ctl::log(envM, "#md !! reading version string failed\n");
			return -1;
		}
		char *mess = m2.buffer.buffer();
		ctl::log(envM, "#m ?? read version = '%s'\n", mess);
		char *at = strchr(mess, '#');
		if(!at)
		{
			error = "invalid response";
			return -1;
		}
		*at = 0;
		int destPort = -1;
		(void)sscanf(at + 1, "%d", &destPort);
		if(destPort <= 0)
		{
			error = at + 1;
			return -1;
		}
		if(args && destPort > 0) ownerIdM.port = physDest.port = destPort;

		/*
		 if(version==CTL && incompatible)
		 send invalid header     + close writer return
		 if(  version!=CTL)
		 close writer   return
		 */
		writer.setRPid(physDest);

		extLinkM[destPort] = writer;
		ctl::link p(ownerM->getCommunicator(), physDest, RMI, false);
		envM->setRemoteLink(physDest, p);
		setLastConnect(physDest);

		writer.release();
		return destPort;
	}

	CTL_Inline bool CommunicatorImpl<dmn>::accept(bool blocking, double dt) const
	{
		socket reader;
		PID pid;
		if(blocking && dt > 0.0)
		{
			for(double i = 0; i < dt * 100; i++)
			{
				pid = socketM.accept(reader, false, 1);
				if(!reader)
					sys::usleep(0);
				else
					break;
			}
		}
		else
			pid = socketM.accept(reader, blocking);

		if(!!reader)
		{
			internal::ioBuffer m;
			m.pid.host = magic().host;
			if(!reader.readHd(m, true))
			{
				ctl::log(envM, "#md !! dmn: invalid connection read failed\n");
				return false;
			}
			if(!m.pid.isLogical()) m.pid.host = pid.host;
			reader.setRPid(m.pid);
			ctl::string error;
			if(envM->redirectSocket(pid, reader, m.buffer.buffer(), error))
			{
				return false;
			}

			ctl_version version = ctl_version::get();
			internal::ioBuffer m2(0, DAT, PID(myPIDM.port, magic().host), -1);
			ctl::string mess = version.str;
			mess += '#';
			if(!error)
				mess += portM;
			else
			{
				mess += int(0);
				mess += error;
			}

			//				ctl::log(envM,"#m accept mess= %s\n", mess.c_str());
			m2.setData(mess.c_str(), mess.size());
			if(!reader.writeHd(m2))
			{
				ctl::log(envM, "#md !! dmn:invalid connection: write failed\n");
				return false;
			}
			if(!!error) return false;
			ctl::log(envM, "#lm ++ dmn -> %d@%s\n", m.pid.port, pid.host.c_str());
			setLastConnect(m.pid);
			if(m.logId >= 0)
				throwStr(" invalid logical id in accept<dmn>\n");
			else
			{
				extLinkM[(int) m.pid] = reader;
				envM->setRemoteLink(m.pid, ctl::link(ownerM->getCommunicator(), m.pid, RMI, false));
			}
			reader.release();
			return true;
		}
		return false;
	}

	CTL_Inline bool CommunicatorImpl<dmn>::addSocket(socket &reader)
	{
		PID pid = reader.getRPid();
		if(portM==1) // socketM not set
		{
//			socketM = reader;
			portM = reader.getPort();
			set(0, 1, PID(portM, reader.getIP()));
//			set(0, 1, PID(portM, socketM.getIP()));

			ctl::log(envM, "#m ++ dmn:socket -> %d\n", portM);
		}
//		else

		{
			extLinkM[(int) pid] = reader;
			ownerIdM = pid;
		}
		//		ctl::log(envM,"#m main socket = (%d, %s)\n", int(pid), pid.host.c_str() );
		internal::ioBuffer m2(0, DAT, PID(myPIDM.port, magic().host), -1);
		ctl_version version = ctl_version::get();
		ctl::string mess = version.str;
		mess += '#';
		mess += portM;
//		mess += portM;
		m2.setData(mess.c_str(), mess.size());
		if(!reader.writeHd(m2))
		{
			ctl::log(envM, "#m !! dmn:invalid connection: write failed\n");
			return false;
		}
		envM->setRemoteLink(pid, ctl::link(ownerM->getCommunicator(), pid, RMI, false));
		return true;
	}

	CTL_Inline bool CommunicatorImpl<dmn>::readAllCatch(const PID &source, bool blocking, bool first, int tag) const
	{
		CTL_Try
		{	return readAll(source, blocking, first);}
		CTL_Catch(const ctl::exception &e)
		{
			const char* b[2] =
			{	"non ", ""};
			rethrowStr(e," *** in dmn::read: src=%d tag=%d %sblocking ***", int(source), tag, b[blocking]);
		}
		return false;
	}

	CTL_Inline bool CommunicatorImpl<dmn>::readAll(const PID &source, bool blocking, bool first) const
	{
		bool rd = false;
		int usec = blocking && !first ? 1 : 0;
		for(;;)
		{
			internal::ioBuffer m;
			socket *pReader = getReader(source, usec);
			if(!pReader || !*pReader)
			{
				bool b = extLinkM.empty() && blocking;
				// 				if(!accept(b ? 1 : 0) && b);
				accept(b ? 1 : 0);
				return rd;
			}
			if(!((blocking && first && !!source) || pReader->test(usec)))
			{
				return rd;
			}
			if(!pReader->readHd(m) || (m.size == 0 && m.tag == -1))
			{
				if(m.tag == -2 || (m.tag == -3 && blocking)) throwStr(" *** read failed ***");
				ctl::log(envM, "#lm -- dmn -> %d@%s by tag %d\n", m.pid.port, m.pid.host.c_str(), int(m.tag));
				extLinkM.erase((int) m.pid);
				m.tag = RMI;
				pReader = 0;
			}
			rd = true;
			m.setNo(mCounterM++);
			if(source.port == -1)
			{
				if(!m.pid.isLogical() && pReader)
				{
					m.pid.host = pReader->getRPid().host;
				}
			}
			else
			{
				m.pid = source;
				if(source.port < 0) m.pid.port = -source.port;
			}
			messageM.insert(m);
			if(blocking || m.tag == CONTROL) break;
		}
		return rd;
	}

	CTL_Inline int CommunicatorImpl<dmn>::write_(oStream &os, const PID& dest, int type) const
	{
		int size = os.size();
		if(!size) size = -1;
		internal::ioBuffer m(size, type, getLogId(), int(synomIdM), os);
		// use pointer to avoid socket destruction

		bool ext = false;
		//		PID dest = this->getPhysId(dest_);
		socket *pWriter = getWriter(dest, ext);
		if(!pWriter || !*pWriter)
		{
			//			int me=-1;
			ctl::string err;
			if(connect(dest, 0, err) < 0) return -1;
			pWriter = getWriter(dest, ext);
		}
		m.pid = myPIDM;
		if(!pWriter || !pWriter->writeHd(m))
		{
			ctl::log(envM, "#d !! fatal write error\n");
			return -1;
		}
		ctl::log(envM, "#m << dmn: dest=(%d, %s) size=%d tag=%d \n", dest.port, dest.host.c_str(), size, type);
		return size;
	}

	CTL_Inline int CommunicatorImpl<dmn>::test_(PID &source, int *type) const
	{
		std::set<internal::ioBuffer>::iterator mI = find(source, *type);
		if(mI == messageM.end())
		{
			readAllCatch(source, false, true, *type);
			mI = find(source, *type);
		}
		if(mI != messageM.end())
		{
			source = mI->pid;
			*type = mI->tag;
			if(mI->size < 0)
			{
				extract(source, *type);
				return -1;
			}
			return int(mI->size);
		}
		return -1;
	}

	CTL_Inline int CommunicatorImpl<dmn>::read_(iStream &is, PID &source, int *type) const
	{
		internal::ioBuffer m = extract(source, *type);
		bool first = true;
		while(!m)
		{
			if(readAllCatch(source, true, first, *type))
				m = extract(source, *type);
			else if(!socketM || !isValid())
			{
				ctl::log(envM, "#md !! dmn::read failed\n");
				return -1;
			}
			first = false;
		}
		source = m.pid;
		// otherwise problem with fast return
		if(source.host == myPIDM.host) sys::usleep(0);
		*type = m.tag;
		int size = int(m.size);
		is.set(m.buffer);
		if(m.swap) is.setSwap();
		ctl::log(envM, "#m >> dmn: src=(%d,%s) size=%d tag=%d\n", source.port, source.host.c_str(), size, *type);
		return int(m.size);
	}

	CTL_Inline void CommunicatorImpl<dmn>::initDaemon(int port, bool verbose)
	{
		//		const char *host="0.0.0.0";
		ownerIdM = -1;
		portM = port;
		if(portM==1) // indicating invalid port
		{
			set(0, 1, PID(portM) );
			return;
		}
		if(portM < 0) portM = 0;
		if(!socketM.open(portM, true))
		{
			ctl::log(envM, "#d !! Could not open socket at port %d\n", portM);
			return;
		}
		portM = socketM.getPort();
		set(0, 1, PID(portM, socketM.getIP()));
		if(verbose)
		{
			fprintf(envM->stderrM, "  *** %s ***\n", CTL_VersionStr);
			fprintf(envM->stderrM, "  *** listening %s:%d ***\n", hostM .c_str(), portM);
		}
		ctl::log(envM, "#m ?? listening on:  %d@%s\n", portM, hostM.c_str());
		return;
	}

	CTL_Inline CommunicatorImpl<dmn>::CommunicatorImpl(int n, const char * const *arg) :
		bCommunicator<internal::ioBuffer> (true), connectedM(true), timeoutM(0)
	{
		hostM = myLocation().host;
		location locarg(n, arg);
		timeoutM = locarg.timeout;

		// daemon
		if(locarg.linkage == dmn || locarg.groupSize == 0)
		{
			initDaemon(locarg.portL, locarg.verbose);
			return;
		}
		set(-1, 0);
	}

	CTL_Inline CommunicatorImpl<dmn>::CommunicatorImpl(const PID & pid, const PID &logId) :
		bCommunicator<internal::ioBuffer> (true), ownerIdM(pid), timeoutM(0)
	{
		hostM = myLocation().host;
		if(logId.port < 0)
		{
			while(!socketM.open(0, true))
				;
			portM = socketM.getPort();
			if(pid.host.size() + 80 > CTL_StringBufferSize) ctl::throwStr("string overflow in Communicator<dmn>");

			char mess[CTL_StringBufferSize];
			sprintf(mess, "%d@%s", pid.port, pid.host.c_str());
			set(0, 1, PID(portM, socketM.getIP()), ctl::string("-- dmn -> ") + mess);
		}
		else
			initDaemon(pid.port, false);
	}

	CTL_Inline CommunicatorImpl<dmn>::~CommunicatorImpl()
	{
		ctl::log(envM, "#d -- Communicator<dmn>(%p)\n", this);
		if(!extLinkM.empty()) readAll(-1, false, true);
		ctl::log(envM, "#d ?? in ~Communicator<dmn>: myPID = (%d, %s),\n", myPIDM.port, myPIDM.host.c_str());
		for(sockMap::iterator link = extLinkM.begin(); link != extLinkM.end(); link++)
			link->second.writeHd(internal::ioBuffer(0, -1, myPIDM, -1));
		sys::usleep(10000);
	}

	void CTL_Inline CommunicatorImpl<dmn>::clean()
	{}

} // ctl

#endif

