/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_FILE_INL_
#define CTL_FILE_INL_

#include<stdio.h>
#include <string>
#include <list> 
 
namespace ctl
{

template<> class CommunicatorImpl<file> : public Communicator
{
	mutable FILE *fileM;
	ctl::string fileNameM;
	int write_(oStream &os, const PID &, int ) const
	{ 
		if(!fileM)
			return 0;		 
//		fwrite(os.buffer(),os.size(),1,fileM);
		const vector<dataObject> &data = os.getData();
		for(size_t i=0; i<data.size(); i++)
			fwrite((char*)data[i].data, data[i].size, data[i].n, fileM);
	
		return os.size();
	}
	static int fileSize(FILE *fp)
	{
		if(!fp)
			return 0;
 	  fseek(fp,0,SEEK_END);
   	int size = ftell(fp);
  	rewind(fp);
    return size;
  }
   int test_(PID &source, int *type) const
   {
   		source = PID(); *type = 0; 
		int size=fileSize(fileM);
		return size;
   }
  int read_(iStream& /*is*/, PID &source, int* type) const
	{ 
			if(isM.size()==0)
			{
	   		source = PID(); *type = 0; 
 		   	size_t s = fileSize(fileM);
   		 	if(s>0)
    		{
#ifdef CTL_MSCX
					vector<char> buf(s);
 		 			fread(buf.begin(),1,s,fileM);
					char *end=strchr(buf.begin(), -51);
					if(end)
						buf.resize(end-buf.begin());
					isM.set(buf);
#else
					isM.set(s, 0);
 				 	int fread_err=::fread(isM.buffer(),1,s,fileM);
					(void)fread_err;
#endif
				}
			}
 		 	return 1;
   }
public:
	bool isValid() const
	{ return fileM!=0 ; }
	channelT getType() const
	{ return file; }
   CommunicatorImpl(const char *arg, const char *mod=0): Communicator(false)
   {
   	ctl::log ( envM, "#d ++ Communicator<file>('%s', '%s')\n", arg, mod );
		char name[CTL_StringBufferSize];
		name[CTL_StringBufferSize-1] = 0;
		strncpy(name, arg, CTL_StringBufferSize-1);
		char *space=strstr(name," -");
		if(space)
		{
			space += 2;
			while(*space == ' ' && *space)
				*space++ ='\0';
	   	fileNameM = name;
			fileM = fopen(name, space);								
		}
		else
		{
   		fileNameM = arg;
			fileM = ctl::fopen(arg, mod ? mod : "r+");
		}
		if(!fileM)
		{
			ctl::log("#d !! communicator<file>: Could not open  '%s' (%s) error =%s [%d]\n", arg, mod, strerror(errno), errno);
			set(-1,0);
		}
		else
			set(0,1);
   }
  ~CommunicatorImpl()
	{
  	ctl::log ( envM, "#d -- Communicator<file>\n" );
		if(fileM)
			ctl::fclose(fileM);
	}
};

} // ctl

#endif
