/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_LAMMPI_H_
#define CTL_LAMMPI_H_
// Communicationrinterface for LAMMPI

#include <stdlib.h>
#include <ctl.h>

namespace ctl
{
	typedef const void* mpi_comm;

template<> class CommunicatorImpl<lam> : public Communicator
{
	ctl::vector<char> mpiBufferM;
	mutable mpi_comm interCommM;
	int mpiBufferSizeM;
	void attachBuffer(int size);
	int mpiID(int logId) const
	{ return logId<=0 ? logId : logId-1; 	}
	int logID(int mpiId, bool inter) const
	{ return (inter != ( getLogId()==0) ) ? 0 :  mpiId + 1; }
	mpi_comm mpiCommu(int logId) const;
	int test(PID &source, int &type,  mpi_comm commu) const;
	int read(iStream&, PID &source, int &type, mpi_comm commu) const;
	void  initWithoutParent(const ctl::vector<location> &loc);
public:
   int write_(oStream &, const PID &dest ,int type) const;
   int test_(PID &source, int *type) const ;
	int read_(iStream&, PID &source, int *type) const;
   CommunicatorImpl(int n, const char * const  *arg);
	CommunicatorImpl(const ctl::vector<location> &loc): Communicator(false)
	{
		CommunicatorImpl<mpi>::initMPI();
		initWithoutParent(loc); 	
	}
	~CommunicatorImpl();
	channelT getType() const
	{ return lam; }

};


} // ctl

#endif



