/******************************************************************************

This file is part of the Component Template Library (CTL)

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
for more details.

You should have received a copy of the GNU Lesser General Public License along
with this library; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_SSH_H_
#define CTL_SSH_H_

#include "ctl/inl/communicator.inl"

namespace ctl
{

	inline const location &myLocation();

	template<> class CommunicatorImpl<pipe> : public bCommunicator<internal::ioBuffer>
	{
			typedef internal::ioBuffer ioBuffer;

			File fdIM, fdOM;
			ctl::string tmpfileIM,tmpfileOM;
			mutable ctl::vector<char> bufferM;
			mutable ctl::uint8 posM;
			std::pair<FILE*, int> servOutM;
			bool read ( ctl::int8 size, bool blocking ) const;
			bool  read ( ioBuffer &m, bool blocking ) const;
			void init ( const location &loc );
		protected:
			ctl::string hostM;
			ctl::string userIdM;
			double timeoutM;
			bool readAll ( const PID&, bool blocking, bool first ) const;
		public:
			int write_ ( oStream&, const PID & dest, int type ) const;
			int test_ ( PID &source, int *type ) const;
			int read_ ( iStream&, PID &source, int *type ) const;
			CommunicatorImpl ( int n, const char * const *arg );
			CommunicatorImpl ( const location &loc ) : bCommunicator<internal::ioBuffer> ( true ), posM ( 0 ), servOutM ( (FILE*)0,0 )
			{	init ( loc ); }
			bool isValid() const
			{ return fdIM>=0 && fdOM>=0; }
			~CommunicatorImpl();
			channelT getType() const
			{ return pipe; }
	};

} // ctl



#endif

