/******************************************************************************

This file is part of the Component Template Library (CTL)

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
for more details.

You should have received a copy of the GNU Lesser General Public License along
with this library; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_PIPE_INL_
#define CTL_PIPE_INL_

#include "pipe.h"

namespace ctl
{
	CTL_Inline bool CommunicatorImpl<pipe>::read ( ctl::int8 size, bool /*blocking*/ ) const
	{
//		CTL_Profile("pipe::read0");
		if ( size==0 )
			return true;
		ctl::int8 sbuf = bufferM.size();
		bufferM.resize ( uint4 ( sbuf+size ) );
//		if(blocking)
//		   	return fdIM.read(bufferM.begin()+sbuf, int4(size) )>0;
//		else
		{
			int rd=fdIM.read ( bufferM.begin() +sbuf, int4 ( size ) );
			sbuf += rd;
			if ( rd<size )
			{
				bufferM.resize ( sbuf );
				return false;
			}
		}
		return true;
	}
	CTL_Inline bool  CommunicatorImpl<pipe>::read ( ioBuffer &m, bool blocking ) const
	{
//		CTL_Profile("pipe::read1");
		if ( tmpfileIM.empty() )
		{
			if ( !fdIM.readHd ( m ) )
			{
				ctl::log ( envM, "#m !! fatal read error in pipe::read\n" );
				return false;
			}
			return true;
		}

		enum { shead=MessHeader::fixsize, skey=20, tail=24, key = shead-tail };
		const char *M=magicM.begin();
		int z=0;
		ctl::uint8 sbuf=bufferM.size();
		if ( !posM )
			posM = key;

		for ( ;; )
		{
			if ( posM+skey>sbuf )
			{
				if ( !read ( posM+tail-sbuf, blocking ) )
				{
					if ( fdOM!=1 ) // stdout
						sys::write ( 1, bufferM.begin(), bufferM.size(), false ), fflush ( stdout );
					return false;
				}
				else
					sbuf = bufferM.size();
			}
			for ( z=0; z<skey && bufferM[posM+z]==M[z]; z++ );
			if ( z==skey )
				break;
			posM += 1;
		}
		if ( fdOM!=1 ) // stdout
			sys::write ( 1, bufferM.begin(), posM -key, false ), fflush ( stdout );
		CTL_Assert( posM + tail>=sbuf);
		read ( posM + tail-sbuf, true );
		ctl::iStream is ( shead, bufferM.begin() + ( posM-key ) );
		is>>m;

		posM = 0;
		bufferM.resize ( 0 );

		if ( m.size )
		{
			m.buffer.setBuffer ( m.size, 0 );
			return fdIM.read ( m.buffer.buffer(), m.size ) >0;
		}
		return true;
	}
	CTL_Inline bool CommunicatorImpl<pipe>::readAll ( const PID&, bool blocking, bool first ) const
	{
//		CTL_Profile("pipe::readAll");
		if ( fdOM==1 ) // stdout
			fflush ( stdout );
		bool read1 = false;
		int usec = blocking && !first ? 1 : 0;
		for ( ;; )
		{
			ioBuffer m;
			if ( !fdIM.test ( usec ) || !read ( m, usec>0 ) )
				return read1;
			read1 = true;

			m.setNo ( mCounterM++ );
			m.pid = 1-getLogId();
			messageM.insert ( m );
			if ( blocking )
				break;
			return true;
		}
		return read1;
	}
	CTL_Inline int CommunicatorImpl<pipe>::write_ ( oStream &os, const PID & dest ,int type ) const
	{
//		CTL_Profile("pipe::write");
		int size = os.size();
		if ( fdOM==1 ) // stdout
			fflush ( stdout );
// reader doesn't know about endian ==> write magic in fixed endian
		internal::ioBuffer m ( os.size(),type,magicSwp(),-1,os );
		if ( dest==getLogId() )
		{
			m.setNo ( mCounterM++ );
			messageM.insert ( m );
			return size;
		}
		if ( !fdOM.writeHd ( m ) )
		{
			ctl::log ( envM, "#m !! fatal error in pipe::write\n" );
			return -1;
		}
		if ( fdOM==1 ) // stdout
			fflush ( stdout );
		ctl::log ( envM, "#m << pipe: dest=%d size=%d tag=%d\n",dest.port,size,type );
		return size;
	}
	CTL_Inline int CommunicatorImpl<pipe>::test_ ( PID &source, int *type ) const
	{
//   		CTL_Profile("pipe::test");

		if ( source!=getLogId() )
			readAll ( source, false, true );
		std::set<ioBuffer>::iterator mI=find ( source, *type );
		if ( mI!=messageM.end() )
		{
			source = mI->pid;
			*type = mI->tag;
			return mI->size;
		}
		return -1;
	}
	CTL_Inline int CommunicatorImpl<pipe>::read_ ( iStream &is, PID &source, int *type ) const
	{
		internal::ioBuffer m=extract ( source, *type );
		bool first = true;
		while ( !m )
		{
			if ( readAll ( source, true, first ) )
				m=extract ( source, *type );
//			else
//				sys::usleep(0);
			first = false;
		}
		is.set ( m.buffer );
		if ( m.swap )
			is.setSwap();
		source = m.pid;
// otherwise problem with fast return
		*type = m.tag;
		int size = m.size;
		ctl::log ( envM, "#m >> pipe: src=%d size=%d tag=%d\n",int ( source ),size,*type );

		return size;
	}
	CTL_Inline CommunicatorImpl<pipe>::CommunicatorImpl ( int n, const char * const *arg ) : bCommunicator<internal::ioBuffer> ( true ), posM ( 0 ),servOutM ( (FILE*)0,0 )
	{
		hostM = myLocation().host;

		location locarg ( n, arg );
		int fdin = locarg.portL>0 ? locarg.portL : 0;
		int fdout = locarg.portR>0 ? locarg.portR : 1;
		fdIM = File ( fdin,false ); // stdin
		fdOM = File ( fdout, false ); // stdout
//		fdIM = File(0,false); // stdin
//		fdOM = File(1, false); // stdout

		if ( locarg.linkage!=pipe )
			return;
		ctl::string host=locarg.host;
		ctl::log ( envM, "#d ?? linkage = %s host = %s  pid = 1 nProcs=2\n", "pipe", host.c_str() );
		ctl_version version;
		if ( !fdIM.read ( &version, sizeof ( version ) ) )
		{
			ctl::log ( envM, "#m !! reading version string failed\n" );
			set ( -1,0 );
			return;
		}
/*
		version= ctl_version::get();
		if ( !fdOM.write ( &version, sizeof ( version ) ) )
		{
			ctl::log ( envM, "#m !! writing version string failed\n" );
			set ( -1,0 );
			return;
		}
*/
		ctl::log ( envM, "#d ?? read version = '%s'\n",version.str );

		ctl::log ( envM, "#lm ++ pipe <- %s\n", host.c_str() );
		if(host.size() + 80 > CTL_StringBufferSize)
			ctl::throwStr("string overflow in Communicator<pipe>");

		char mess[CTL_StringBufferSize];
		sprintf ( mess, "-- pipe <- %s", host.c_str() );
		set ( 1, 2, PID(), mess );
	}

	CTL_Inline void CommunicatorImpl<pipe>::init ( const location &loc )
	{
		timeoutM = loc.timeout;
		hostM = myLocation().host;
		ctl::string message = loc.pathExec(); message += "@"; message += loc.host;
		set ( 0, 2, PID ( 0 ), ctl::string ( "-- pipe -> " ) +message );
		PID me ( 0, hostM );
		if ( loc.exec=="*" )
		{
#ifdef CTL_PIPE
			int fdI[2], fdO[2];
			int pipe_err;
			pipe_err= ::pipe ( fdI );
			pipe_err= ::pipe ( fdO );
		

			char tmpI[256] ,tmpO[256];
			sprintf ( tmpI, " -R %d ", fdI[1] );
			sprintf ( tmpO, " -L %d ", fdO[0] );
			if ( !ctl::spawn ( me, loc, 1, 2, "pipe", 0, tmpI, tmpO ) )
			{
				ctl::log ( envM, "#m !! pipe: Could not spawn \"%s\" \n",loc.getSSH().c_str() );
				set ( -1,0 );
				return;
			}
			close ( fdI[1] );
			close ( fdO[0] );
			fdIM = File ( fdI[0], false );
			fdOM = File ( fdO[1], false );
#else
			ctl::log ( envM, "#m !! pipe: Could not create pipe \"%s\" \n",loc.getSSH().c_str() );
			set ( -1,0 );
			return;

#endif
		}
		else
		{
#ifdef CTL_FIFO
			char tmpI[256] ,tmpO[256];
			sprintf ( tmpI,"/tmp/ctlpipeI%d%p", int ( sys::getPid() ),this );
			sprintf ( tmpO,"/tmp/ctlpipeO%d%p", int ( sys::getPid() ),this );

			sys::mkfifo ( tmpI,0700 );
			sys::mkfifo ( tmpO,0700 );

			FILE *fI = fopen ( tmpI, "r+" );
			FILE *fO = fopen ( tmpO, "w+" );

			fdIM = fileno ( fI );
			fdOM = fileno ( fO );

			tmpfileIM = tmpI;
			tmpfileOM = tmpO;
// spawn process
			if ( !ctl::spawn ( me, loc, 1, 2, "pipe", 0, tmpfileIM.c_str(), tmpfileOM.c_str() ) )
			{
				ctl::log ( envM, "#m !! pipe: Could not spawn \"%s\" \n",loc.getSSH().c_str() );
				set ( -1,0 );
				return;
			}
#else
			char tmpI[256];
			sprintf ( tmpI,"tmpctlpipeI%d%p", int ( sys::getPid() ),this );
			tmpfileIM = tmpI;
			servOutM.second = -1;
			if ( !ctl::spawn ( me, loc, 1, 2, "pipe", 0, tmpfileIM.c_str(), 0, &servOutM ) )
			{
				ctl::log ( envM, "#m !! pipe: Could not spawn \"%s\" \n",loc.getSSH().c_str() );
				set ( -1,0 );
				return;
			}
			FILE *fI = 0;
			for(;;)
			{
				if ( ! ( fI=fopen ( tmpI, "r" ) ) )
					sys::usleep ( 1 );
				else
					break;
			}
CTL_Here;
		if(feof(servOutM.first) || (timeoutM > 1 && !fdIM.test ( 1000*timeoutM )) )
		{
			ctl::log ( envM,"#m !! testing pipe failed '%s'\n", strerror ( errno ) );
			set ( -1,0 );
			return;
		}
			fdIM = File ( fileno ( fI ), false );
			fdOM = File ( fileno ( servOutM.first ), false );
#endif
		}
		ctl_version version= ctl_version::get();
		if ( !fdOM.write ( &version, sizeof ( version ) ) )
		{
			ctl::log ( envM,"#m !! write into pipe failed '%s'\n", strerror ( errno ) );
			set ( -1,0 );
			return;
		}

/*
		if ( !fdIM.read ( &version, sizeof ( version ) ) )
		{
			ctl::log ( envM,"#m !! reading from pipe failed '%s'\n", strerror ( errno ) );
			set ( -1,0 );
			return;
		}
*/
		ctl::log ( envM, "#lm ++ pipe -> %s\n", message.c_str() );

	}
	CTL_Inline CommunicatorImpl<pipe>::~CommunicatorImpl()
	{
		ctl::log ( envM, "#d -- Communicator<pipe>\n" );
		if ( fdOM==1 ) // stdout
			fflush ( stdout );

//		readAll(PID(), false, true);
		sys::usleep ( 0 );
		if ( fdIM>1 )
			fdIM.close();
		if ( fdOM>1 )
			fdOM.close();
		while ( !!tmpfileIM && remove ( tmpfileIM.c_str() ) <0 )
			sys::usleep ( 0 );
//			ctl::log("#m !! Could not remove file '%s' error = '%s'\n", tmpfileIM.c_str(), strerror(errno));

		if ( !!tmpfileOM && remove ( tmpfileOM.c_str() ) <0 )
			ctl::log ( "#m !! Could not remove file '%s' error = '%s'\n",tmpfileOM.c_str(), strerror ( errno ) );
	}

} // ctl

#endif

