/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#include "pvm.h"
#include <pvm3.h>

namespace ctl
{

	CTL_Inline int CommunicatorImpl<pvm>::getLogIdS(int n , const char* const *arg, const location &loc)
	{ return pvm_mytid (); }
	CTL_Inline CommunicatorImpl<pvm>::CommunicatorImpl(int , const char * const *): Communicator(false)
	{
		int parent = pvm_parent ();
		if(parent==PvmNoParent)
		{
			ctl::log(envM, "#d !! have no pvm parent\n");
			return;
		}
		initWithParent(parent);
	}

	CTL_Inline void CommunicatorImpl<pvm>::initWithoutParent(const ctl::vector<location> &locV)
	{
		int n = locV.size();
		int nProcs = n+1;
		PID myPID = 0;
		ctl::vector<int> ptid(nProcs);
		ptid[0] = pvm_mytid ();
		int i;
		pvm_catchout( stdout);
		for(i=0; i<nProcs-1; i++)
		{
			int tid;
		 	const location &loc=locV[i];
		 	int flag = loc.host.empty() ? 0 : 1;
			if(!loc.debug.empty())
				flag |= 4;
			string arg0Str = "-l ";
			arg0Str += string("pvm");
			char  *arg1 = loc.args.empty() ? 0 : (char*)loc.args.c_str();
		 	char *args[] = { (char*)arg0Str.c_str() , arg1 , 0};
		 	
			ctl::string cmd="";
			cmd +=  loc.absolutePathExec();
			ctl::string host=loc.host;
			if(host=="localhost")
				host = myLocation().host;
			int spawned = pvm_spawn ((char*)cmd.c_str(), args, flag, (char*)host.c_str(), 1, &tid);
			ptid[i+1] = tid;
			if(spawned!=1)
			{
				ctl::log(envM, "#e !! Could not spawn \"%s\"\n",loc.pathExec().c_str());
				return;
			}
			else
				ctl::log(envM, "#d >$ pvm: %s spawned on %s(%s, %s)\n" ,loc.pathExec().c_str(), loc.host.c_str(), args[0], args[1] );
		}

  // send number of processes to each spawned process ...
		pvm_initsend (PvmDataDefault);
		pvm_pkint (&nProcs, 1, 1);
		pvm_mcast (ptid.p()+1, nProcs-1, 110);
		pvm_initsend (PvmDataDefault);
		pvm_pkint (ptid.p(), nProcs, 1);
		pvm_mcast (ptid.p()+1, nProcs-1, 112);
		for (i = 0; i < nProcs; i++)
		{
      		pid2lid[ptid[i]] = i;
     		lid2pid[i] = ptid[i];
    	}
		set(myPID, nProcs);
	}
	CTL_Inline void CommunicatorImpl<pvm>::initWithParent(int parent)
	{
    // wait for lid <-> pid map  message of the spawner ...
		int nProcs =-1;
		PID myPID = 0;

		pvm_recv (parent, 110);
		pvm_upkint (&nProcs, 1, 1);

		ctl::vector<int>  ptid(nProcs);
		pvm_recv (parent, 112);
		pvm_upkint (ptid.p(), nProcs, 1);

    	int mytid = pvm_mytid ();
		int i = 0;
		for (i = 0; i < nProcs; i++)
		{
			if (mytid == ptid[i])
				myPID = i;
      		pid2lid[ptid[i]] = i;
     		lid2pid[i] = ptid[i];
    	}
    	set(myPID, nProcs);
  }
  CTL_Inline int CommunicatorImpl<pvm>::write_(oStream &os,const PID &dest, int type) const
  {
   		int size = os.size();
    pvm_initsend (PvmDataRaw);
    pvm_pkbyte ((char*) os.buffer(), size, 1);

    int destTid = lid2pid[dest];
    // asynchronous send
    pvm_send (destTid, type);
    ctl::log(envM, "#m << pvm: dest=%d size = %d\n",int(dest),size);
    return size;
  }

  CTL_Inline int CommunicatorImpl<pvm>::read_ (iStream &is, PID &source, int *type) const
  {
    int size = is.size();
	if(!size)
		size = -1;
    // in pvm -1 matches everything
    int PVMsource = !source   ? -1  : lid2pid[source];
    int PVMtag 	  = *type   < 0  ? -1  : *type;

    int bufid = pvm_recv(PVMsource, PVMtag);

    source = pid2lid[PVMsource];

    pvm_bufinfo (bufid, &size, type, &PVMsource);

    if (bufid) {
      source = pid2lid[PVMsource];
    }

    CTL_Assert (size>0);
	is.set(size,0);
    pvm_upkbyte ((char*) is.buffer(), size, 1);
    ctl::log(envM, "#m >> pvm: src=%d size = %d\n",int(source),size);

    return size;
  }

  CTL_Inline int CommunicatorImpl<pvm>::test_(PID &source, int *type) const
  {
//    CTL_Profile("g_ntest");
    int flag=0;
    int PVMsource = !source  ? -1 : lid2pid[source];
    int PVMtag    = *type < 0   ? -1 : *type;

    int bufid = pvm_probe (PVMsource, PVMtag);
    int size  = -1;

    if (bufid) {
      flag = pvm_bufinfo (bufid, &size, type, &PVMsource);
      source = pid2lid[PVMsource];
    }

    return size;
  }

} // ctl
