/******************************************************************************

This file is part of the Component Template Library (CTL)

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
for more details.

You should have received a copy of the GNU Lesser General Public License along
with this library; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_TCPIP_H_
#define CTL_TCPIP_H_

#include "ctl/utl/timer.h"
#include "ctl/inl/communicator.inl"

namespace ctl
{

	inline const location &myLocation();

	template<> class CommunicatorImpl<tcp> : public bCommunicator<internal::ioBuffer>
	{
		public:
			PID getPhysId ( const PID &pid ) const
			{
				if ( !pid.isLogical() )
					return pid;
				return getPID ( pid );
			}
		protected:
			typedef internal::ioBuffer ioBuffer;

//	mutable std::set<ioBuffer> messageM;
			socket socketM, *psocketM;
			mutable ctl::vector<PID> physIdM;
			mutable ctl::vector<socket> linkM;
//	typedef  std::map<PID, socket> sockMap;
			typedef  std::map<int/*port*/, socket> sockMap;
			mutable sockMap extLinkM;
			mutable std::map<int, PID> synonymLinkM;
			ctl::int4 portM;
			PID ownerIdM;
			ctl::string hostM;
			ctl::string userIdM;
			PID synomIdM;
			mutable bool connectedM;
			double timeoutM;
			std::pair<FILE*, int> servOutM;
			ctl::vector<location> componentM;
			socket *getWriter ( const PID &pid, bool &ext ) const
			{ ext = false; return getSocket ( pid,0,ext );	}
			socket *getReader ( const PID &pid,int usec ) const
			{ bool b; return getSocket ( pid,usec,b );	}
			socket *getSocket ( const PID &pid, int usec, bool &ext ) const;
			bool connect ( const PID &dest, int logid=-1, int destid=-1 ) const;
			bool disconnect ( socket &s, int logid=-1 )
			{ return s.writeHd ( ioBuffer ( 0,-1, myPIDM, logid ) ); }
			bool accept ( bool blocking, double dt =0.0 ) const;
			bool readAllCatch ( const PID &source , bool blocking, bool first, int tag ) const;
			bool readAll ( const PID &source , bool blocking, bool first ) const;
//	std::set<ioBuffer>::iterator find(const PID &src, int tag) const;
			ioBuffer extract ( const PID &src, int tag ) const;
			void printServStdout() const;
			void init ( const ctl::vector<location> &loc, const char *typ /* ="tcp" */, bool late );

		public:
			PID getPID ( unsigned int i ) const
			{ return i<physIdM.size() ? physIdM[i] : PID(); }
			int write_ ( oStream&, const PID & dest ,int type ) const;
			int test_ ( PID &source, int *type ) const;
			int read_ ( iStream& , PID &source, int *type ) const;
//   CommunicatorImpl() {}
			CommunicatorImpl ( int n, const char * const *arg );
			CommunicatorImpl ( const ctl::vector<location> &loc, const char *typ /* ="tcp" */, bool late = false ) : bCommunicator<internal::ioBuffer> ( true ), psocketM ( envM->getSocket() ),
					connectedM ( false ), timeoutM ( 0 ), servOutM ( (FILE*)0,0 ), componentM ( loc )
			{ init ( loc, typ, late ); }
			CommunicatorImpl ( const PID & pid, const PID &logId );
			void  getConnections() const;
			bool isValid() const
			{
				if ( connectedM ) CTL_Assert ( !ownerIdM || extLinkM.find ( ( int ) ownerIdM ) !=extLinkM.end() );
				if ( connectedM )
					for ( size_t i=0; i<linkM.size(); i++ )
						if ( i!=logIdM && !linkM[i] )
							return false;
				return !ownerIdM || extLinkM.find ( ( int ) ownerIdM ) !=extLinkM.end();
			}
			~CommunicatorImpl();
			channelT getType() const
			{ return tcp; }
			void clean();
	};

} // ctl

#endif

