/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_TCPIP_INL_
#define CTL_TCPIP_INL_

#include "tcpip.h"

namespace ctl
{
    struct ctl_version
	{
		char str[19];
		static  ctl_version get()
		{
			ctl_version v = { CTL_VersionStr };
			return v;
		}
	};

	CTL_Inline void CommunicatorImpl<tcp>::printServStdout() const
	{
#ifndef CTL_MSC
		if(servOutM.first)
		{
			char line[CTL_StringBufferSize];
			ctl::File f(fileno(servOutM.first) );
			if(f.test(0))
			{	
				int r=f.read(line,-CTL_StringBufferSize);
			 	sys::write(1, line, r);
				fflush(stdout);
			}
		}
#endif
	}		

	CTL_Inline socket *CommunicatorImpl<tcp>::getSocket(const PID &pid, int usec, bool &ext) const
	{
		if(!connectedM)
			getConnections();
		if(!!pid)
		{
			if(pid.isLogical())
			{
				if(!!linkM)
					return  &linkM[pid];
				std::map<int, PID>::iterator s=synonymLinkM.find(int(pid));
				if(s!=synonymLinkM.end())
				{
					ext = true;
	         		sockMap::iterator link=extLinkM.find((int)s->second);
   		      		if(link!=extLinkM.end())
  	       			return &link->second;
				}
			}
     	else
     	{
//				return 0;
		   	for(unsigned int i=0; i<physIdM.size(); i++)
			 		if(physIdM[i]==pid)
			 			return  &linkM[i];
				ext = true;
        sockMap::iterator link=extLinkM.find((int)pid);
        if(link!=extLinkM.end())
        	return &link->second;
      }
			return 0;
    	}

		for(unsigned int src=0; src<linkM.size() ; src++)
			if(src!=getLogId() && linkM[src].test(usec))
				return &linkM[src];
		for(sockMap::iterator link=extLinkM.begin(); link!=extLinkM.end(); link++)
			if(link->second.test(usec))
				return &link->second;
		return 0;	
	}

	CTL_Inline bool CommunicatorImpl<tcp>::connect(const PID &dest, int logid, int destid) const
   	{
		internal::ioBuffer m(0, DAT, PID(((Communicator*)this)->getPhysId().port, magic().host), logid);
		PID physDest = this->getPhysId(dest);
		if(!physDest)
			return false;

//		ctl::log(envM, "#4 ++ tcp::connect -> %s\n",componentM[destid-1].getSSH().c_str() );
//		ctl::log(envM, "#1 connecting to port %d@%s %d\n",  physDest.port, physDest.host.c_str(), destid );
		socket writer;
		if(!writer.connect (physDest.port, physDest.host.c_str(), false ) )
  			return false;
		ctl::log(envM, "#lm ++ tcp <- %d@%s\n",	physDest.port, physDest.host.c_str() );	
		ctl_version version;
		if(writer.read(&version, sizeof(version) ) != sizeof(version) )
		{
			ctl::log(envM, "#md !! reading version string failed: got %s\n", version.str);
			return false;
		}
/*
		if(version==CTL && incompatible)
			send invalid header     + close writer return
		if(  version!=CTL)
			close writer   return
*/
		ctl::log(envM, "#m ?? read version = '%s'\n",version.str);
		if(!writer.writeHd(m))
			return false;
		writer.setRPid(physDest);

		if(destid<0)
			extLinkM[(int)dest] = writer;
		else
			if((unsigned int)destid<linkM.size())
				linkM[destid] = writer;

		writer.release();
		return true;
   	}

	CTL_Inline bool CommunicatorImpl<tcp>::accept(bool blocking, double dt) const
	{
		socket reader;
		PID pid;
		if(blocking && dt>0.0)
		{
	;
				for(double i=0; i<=dt; i+=1)
				{
					pid=psocketM->accept(reader, false, 1);
					if(!reader)
						sys::usleep(1000);
					else
						break;
				}
		}
		else
			pid = psocketM->accept(reader, blocking);

		if(!!reader)
		{
			internal::ioBuffer m;
			ctl_version version= ctl_version::get();
			if(!reader.write(&version, sizeof(version)) )
			{
				ctl::log(envM, "#md !! tcp:invalid connection: write failed\n");
				return false;
			}
				
			m.pid.host = magic().host;
			if(!reader.readHd(m, true))
			{
				ctl::log(envM, "#md !! tcp: invalid connection read failed\n");
				return false;
			}
			ctl::log(envM, "#lm ++ tcp -> %d@%s\n",	m.pid.port, pid.host.c_str() );	
//			ctl::log(envM, "#4 ++ tcp -> %d@%s\n",	m.pid.port, pid.host.c_str() );	
			if(!m.pid.isLogical())
				m.pid.host = pid.host;
			reader.setRPid(m.pid);

			setLastConnect(m.pid);
			if(m.logId>=0)
			{
				synonymLinkM[m.logId] = m.pid;
				if(int(linkM.size() )>m.logId && !linkM[m.logId])
					linkM[m.logId] = reader;	
				if(int(physIdM.size() )>m.logId && !physIdM[m.logId])
					physIdM[m.logId] = m.pid;	
			}
			else
				extLinkM[(int)m.pid] = reader;
			reader.release();
			return true;
		}
		return false;
	}

	CTL_Inline bool CommunicatorImpl<tcp>::readAllCatch(const PID &source , bool blocking, bool first, int tag) const
	{
		CTL_Try
		{ return readAll(source, blocking, first); }
		CTL_Catch(const ctl::exception &e)
		{
			const char* b[2] = { "non ", "" };
			rethrowStr(e," *** in tcp::read: src=%d tag=%d %sblocking ***", int(source), tag, b[blocking]);
		}
		return false;
	}

	CTL_Inline bool CommunicatorImpl<tcp>::readAll(const PID &source , bool blocking, bool first) const
	{
		bool rd = false;
		printServStdout();
		int usec = blocking && !first ? 1 : 0;
 		for(;;)
		{
	     	internal::ioBuffer m;
 			socket *pReader = getReader(source, usec);
 			if(!pReader || !*pReader)
 				return rd;
 			if(!( (blocking && first && !!source) || pReader->test(usec) ) )
  			{ return rd; }

			if(!pReader->readHd(m) || (m.size==0 && m.tag==-1))
 			{
// 				ctl::log(envM, "#1 closing connection (%d, %s)\n",	m.pid.port, m.pid.host.c_str());		
				
				if(m.tag==-2)
				{	throwStr(" *** connection closed? ***");  }
				ctl::log(envM, "#lm -- tcp -> %d@%s\n",	m.pid.port, m.pid.host.c_str() );	

 				extLinkM.erase((int)m.pid);
 				return rd;
 			}
 			rd = true;
 			m.setNo(mCounterM++);
 			if(source.port==-1)
 			{
 				if(!m.pid.isLogical())
 				{ m.pid.host = pReader->getRPid().host; }
 			}
 			else
 			{
 				m.pid = source;
 				if(source.port<0)
 					m.pid.port = -source.port; 	
 			}
//ctl::log("tcp::readAll: (%d %s)[%s]\n", m.pid.port, m.pid.host.c_str(), pReader->getRPid().host.c_str() );
 			messageM.insert(m);
 			if(blocking || m.tag == CONTROL)
 				break;
 		}	
 		return rd;
 	}	
	CTL_Inline internal::ioBuffer CommunicatorImpl<tcp>::extract(const PID &src, int tag) const
 	{
 		std::set<internal::ioBuffer>::iterator mI = find(src, tag);
 		if(mI!=messageM.end() )
 		{
 			internal::ioBuffer m=*mI;
 			messageM.erase(mI);
 			return m;
 		}
 		std::map<int, PID>::iterator s=synonymLinkM.find(int(src));
 		if(s!=synonymLinkM.end())
 		{
 			std::set<internal::ioBuffer>::iterator mI = find(s->second, tag);
 			if(mI!=messageM.end() )
 			{
 				internal::ioBuffer m=*mI;
 				messageM.erase(mI);
 				return m;
 			}
 		}
 		return internal::ioBuffer();	
 	}

   CTL_Inline int CommunicatorImpl<tcp>::write_(oStream &os, const PID & dest ,int type) const
   {
		int size = os.size();
		if(!size)	
			size = -1;
    internal::ioBuffer m(size,type,getLogId(),int(synomIdM), os);
    if((dest==getLogId() || dest==physIdM[getLogId()]) && !synomIdM  )
    {
			m.setNo(mCounterM++);
			messageM.insert(m);
			return size;
    }
// use pointer to avoid socket destruction

		bool ext=false;
    socket *pWriter= getWriter(dest, ext);
		if(!pWriter || !*pWriter)
		{
			int me=dest.isLogical() ? int(getLogId() ) : -1;
			if(!connect(dest, me, dest) )
				return -1;
			pWriter= getWriter(dest, ext);
		}
		if(ext)
     		m.pid=myPIDM;
 	  if(!pWriter->writeHd(m))
		{
			ctl::log(envM, "#d !! fatal write error\n");
			return -1;
		}
 		  ctl::log(envM, "#m << tcp: dest=(%d, %s) size=%d tag=%d \n",dest.port, dest.host.c_str(),size,type);
      return size;
   }

   CTL_Inline int CommunicatorImpl<tcp>::test_(PID &source, int *type) const
   {
   		std::set<internal::ioBuffer>::iterator mI=find(source, *type);
		if(source!=getLogId() && mI==messageM.end() )
		{
   			readAllCatch(source, false, true, *type);
			mI=find(source, *type);
		}
   		if(mI!=messageM.end() )
   		{
   			source = mI->pid;
   			*type = mI->tag;
   			return int(mI->size);
   		}
   		return -1;
   }

   CTL_Inline int CommunicatorImpl<tcp>::read_(iStream &is, PID &source, int *type) const
   {
   		internal::ioBuffer m=extract(source, *type);
   		bool first = true;
		while(!m)
		{
			if(readAllCatch(source, true, first, *type))
				m=extract(source, *type);
			else if(!*psocketM || !isValid())
			{	ctl::log(envM, "#m !! tcp::read failed\n"); return -1; }
			first = false;
		}
		source = m.pid;
// otherwise problem with fast return
		if(source.host == myPIDM.host)
			sys::usleep(0);
		*type = m.tag;
		int size = int(m.size);
		is.set(m.buffer);
		if(m.swap)
			is.setSwap();
    ctl::log(envM, "#m >> tcp: src=(%d,%s) size=%d tag=%d\n",source.port,source.host.c_str(), size, *type);
		return int(m.size);		
   }

   CTL_Inline CommunicatorImpl<tcp>::CommunicatorImpl(int n, const char * const *arg): bCommunicator<internal::ioBuffer>(true), psocketM(envM->getSocket() ),connectedM(true), timeoutM(0),servOutM((FILE*)0,0)
   {
//			PID me=envM->getPhysId();
//    	hostM = me.host;
    	hostM = myLocation().host;
		location locarg(n, arg);
		timeoutM = locarg.timeout;
		if(!psocketM)
		{ 
			psocketM = &socketM;
			if(!psocketM->open() )
   		{
   			ctl::log(envM, "#w !! Could not open port %d\n",portM);
			set(-1,0);
			return;   			
   		}
		}
		
		bool iConnect = locarg.portR==0;
 		portM = psocketM->getPort();
		if(!iConnect)
		{
			fprintf(stdout,"%d\n",portM), fflush(stdout);
			ctl::log(envM, "#d ?? port = %d portR = %d portM = %d\n",portM, locarg.portR, portM);
		}
    	ctl::string tcp=locarg.getChannelString(), host=locarg.host;

    	ctl::uint4 port=locarg.portL, pid=locarg.logId, nProcs=locarg.groupSize;

    	physIdM.resize(nProcs);
    	physIdM[0] = PID(port, host);  	
      	linkM.resize(nProcs);     	

//    	set(pid, nProcs, PID(portM, host) );
			char mess[CTL_StringBufferSize];
			if(host.size() + 80 > CTL_StringBufferSize)
				ctl::throwStr("string overflow in Communicator<tcp>");
			sprintf(mess,"-- tcp -> %d@%s\n", port, host.c_str() );
    	set(pid, nProcs, PID(portM, psocketM->getIP()), mess );
   	  	ctl::log(envM, "#d ?? linkage = %s myPID = (%d, %s) host = %s port = %d pid = %d nProcs=%d\n", tcp.c_str(), myPIDM.port,  myPIDM.host.c_str(),host.c_str(), port, pid, nProcs);
		if(iConnect)
		{
			if(!connect(physIdM[0], pid, 0) )
			{ 	set(-1,0); return; }
		}
		else
			if(!accept(true, timeoutM))
			{ 	set(-1,0); return; }
		PID source = 0;
		int tag = 7;
		if(nProcs>2)
		{
			linkM.resize(nProcs);
			ctl::iStream is;  	
			CommunicatorImpl<ctl::tcp>::read(is,source, tag);
			is>>physIdM;
		}
		ctl::uint4 pidC;
		for(pidC = 1; pidC<getLogId(); pidC++)
			if(!connect (physIdM[pidC], pid, pidC) )
				{ set(0,-1); return; }
		for(pidC = getLogId()+1; pidC<nProcs; pidC++)
			if(!accept(true, timeoutM ))
			{ 	set(-1,0); return; }
	}

   CTL_Inline void CommunicatorImpl<tcp>::init(const ctl::vector<location> &loc, const char *typ, bool late)
   {
    	hostM = myLocation().host;
   		ctl::uint4 i;
		int port = loc.size()==1 ? loc[0].portL : 0;
		int portR = loc.size()==1 ? loc[0].portR : 0;
		timeoutM = loc[0].timeout;
		bool iAccept = portR==0;
		if(!psocketM)
		{ 
			psocketM = &socketM;
			if(!psocketM->open(port, true) )
   		{
   			ctl::log(envM, "#w !! Could not open port %d\n",portM);
			set(-1,0);
			return;   			
   		}
		}
   		portM = psocketM->getPort();
		ctl::string message;
		for(i=0; i<loc.size(); i++)
		{
			if(i)
				message += "           "; 	
			message += loc[i].pathExec(); message += "@"; message += loc[i].host; message += "\n";
		}

    	set(0, int(loc.size())+1, PID(portM, psocketM->getIP()), ctl::string("-- tcp -> ")+message);
			ctl::log(envM, "#d ?? myPID = (%d, %s),\n", myPIDM.port,  myPIDM.host.c_str());
    	physIdM.resize(nProcs());
		linkM.resize(nProcs());
  		
// spawn processes
		PID me(portM, psocketM->getIP());
    	for(i=0; i<loc.size(); i++)
			if(!ctl::spawn(me, loc[i], i+1, nProcs(), typ, 0, 0, 0,iAccept ? 0 : &servOutM))
			{
				ctl::log(envM, "#w !! Could not spawn tcp\"%s\" \n",loc[i].getSSH().c_str());
				set(-1,0);
				return;
			}
//		ctl::log("#4 ++ tcp -> %s\n", message.c_str());		
		
    	physIdM[0] = PID(portM, hostM);

		for(i=1; i<(unsigned int)nProcs(); i++)
		{
			physIdM[i].host = loc[i-1].host;
		}
// get port number of processes
		if(!late)
			getConnections();
	}

	CTL_Inline void  CommunicatorImpl<tcp>::getConnections() const
	{
		connectedM = true;
		if(servOutM.first || servOutM.second)
		{
			int port =0;
			if(servOutM.second)
				port = servOutM.second;
			else
			{
				char line[128];
				if(!feof(servOutM.first) )
				{
					if(fgets(line,128, servOutM.first) )
						port = atoi(line);
					if(!port)
						ctl::log(envM, "#d !! got response '%s' by popened process\n", line);
				}
			}
			if(!port)
				ctl::throwStr(" *** no readable response of popened process ***");
			physIdM[1].port = port;
			if(!connect(physIdM[1], 0, 1))
				((CommunicatorImpl<tcp>*)this)->set  (-1, 0);
			return;
		}
		int i;
    	for(i=0; i<nProcs()-1; i++)
			if(!accept(true, timeoutM))
			{ 	ctl::throwStr(" *** accept timed out ***"); return; }
// tell port numbers of processes
		if(nProcs()>2)
		{
			char buf[CTL_defaultBufferSize];
	 		ctl::oStream os=ctl::serializeB(physIdM,buf, CTL_defaultBufferSize);
    		for(i=1; i<nProcs(); i++)
				CommunicatorImpl<tcp>::write(os, i, 7);
		}
   }

	CTL_Inline CommunicatorImpl<tcp>::~CommunicatorImpl()
	{
		ctl::log(envM, "#d -- Communicator<tcp>(%p)\n", this);
		if(!extLinkM.empty())
			readAll(-1, false,true);
		ctl::log(envM, "#d ?? in ~Communicator<tcp>: myPID = (%d, %s),\n", myPIDM.port,  myPIDM.host.c_str());
//Disconnect external links
		for(sockMap::iterator link=extLinkM.begin(); link!=extLinkM.end(); link++)
				link->second.writeHd(internal::ioBuffer(0,-1, myPIDM,-1));

#ifndef CTL_MSC
		if(servOutM.first)
			pclose(servOutM.first);
#endif
	}
	void CTL_Inline CommunicatorImpl<tcp>::clean()
	{
		printServStdout();
	}

} // ctl

#endif





