/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_CONFIG_H_
#define CTL_CONFIG_H_

//! the ctl config
#define CTL_VER 131
#define CTL_VersionStr "CTL version 1.3.1"


#define CTL_MaxTmplParamList 8
#define CTL_MaxArgList 16
#define CTL_MaxFunction 36
#define CTL_MaxConstructor 36
#define CTL_MaxMethod 36
#define CTL_MaxStaticMethod 36
 

#if defined(DEBUG) || defined(_DEBUG) || defined(__DEBUG) || defined(DEBUG_)
#	define CTL_Debug
#else
#	define CTL_NoLog
#endif

//#define CTL_AllowRoot
//#define CTL_Dmn2Thread
//#define CTL_Dmn0Thread

#ifdef CTL_AllowReConfig
#	undef CTL_Lib
#	define CTL_Inl
#endif

#if !defined  CTL_Inl && !defined  CTL_Lib
#	define CTL_Lib
#endif

#ifdef CTL_Lib
#	define CTL_Inline
#else
#	define CTL_Inline inline
#endif

#define CTL_Static static
//#define CTL_Static
#define CTL_Alias typedef

#ifdef CTL_Tick
#	define CTL_Prof
#endif


// Maximal size copied into bytebuffer
#ifndef CTL_MaxInline
#	define  CTL_MaxInline 65536
#endif
#ifndef CTL_StringBufferSize
#	define  CTL_StringBufferSize 2048
#endif
#ifndef CTL_MaxIncludeDepth
#	define CTL_MaxIncludeDepth 10
#endif

// Initial bytebuffer size 
#ifndef CTL_defaultBufferSize
#	define CTL_defaultBufferSize 65600
#endif

// Let result receive ask all channels
//#define CTL_AsyncMultiReading

// Let environment::run ask all group
#define CTL_ListenAllGroups

#define CTL_EXCEPTION

#ifdef CTL_EXCEPTION
#	define CTL_Try try
#	define CTL_Catch(x) catch(x)
#	define CTL_CatchAll catch(...)
#	define CTL_Throw  throw
#else
#	define CTL_Try
#	define CTL_Catch(x) x; if(false)
#	define CTL_CatchAll if(false)
#	define CTL_Throw
#endif

#define CTL_PartialSpec
#define CTL_TmplBeforeConst
#define CTL_TmplCastOper
#define CTL_PIPE
#define CTL_FIFO
#define CTL_VNstdio

#ifndef CTL_FortranPostFix
#	define CTL_FortranPostFix _
#endif
#ifndef CTL_FortranPostFix_
#	define CTL_FortranPostFix_ __
#endif

#ifdef _MSC_VER
#	define CTL_MSC_HEADER
#	ifndef __ICL
#		define CTL_VC
#	endif

#	define CTL_MSC
#	define CTL_wchar
#	ifndef WIN32
#		define WIN32
#	endif
#	if   _MSC_VER<1400
#		undef CTL_TmplBeforeConst
#	endif
# pragma warning(disable : 4127)
# pragma warning(disable : 4181)
# pragma warning(disable : 4244)
# pragma warning(disable : 4675)
# pragma warning(disable : 4521)
# pragma warning(disable : 4706)
# pragma warning(disable : 4996)

#	define CTL_WinPath
#endif

#ifdef __CYGWIN32__
#	define CTL_CygWin
#	define CTL_SelfDelete
#	undef CTL_FIFO
#	define CTL_WinPath
#	define CTL_XtermAvailable
#endif

#ifdef WIN32
# ifdef CTL_MinGW32
#		define CTL_MSC_HEADER
#		define CTL_PUTENV
#		define CTL_WinPath
#	endif
#	define CTL_Win32
#	define CTL_Export  __declspec(dllexport)
#	define CTL_SelfDelete
#	undef CTL_FIFO
#	undef CTL_PIPE

#else
#	define CTL_Export  
#endif

#if defined __unix || defined  __unix__ || defined unix 
#	define CTL_Unix
#	ifdef _XOPEN_SOURCE
#		define CTL_XtermAvailable
#	endif
#endif 

#if defined(__hpux) || defined (__hpux__)
#	define CTL_HPUX
#endif

#if defined(_SGIAPI) || defined(sgi)
#	define CTL_SGI
#endif

#ifdef  __APPLE__
#	define CTL_OSX
#endif

#ifdef __sparc
#	define CTL_Sparc
#endif

#ifdef _AIX
#	define CTL_Aix
#endif

#ifdef __alpha
#	define CTL_Alpha
#endif

#ifdef __GNUC__
#	if __GNUC__ ==2 && __GNUC_MINOR__ <=95
#		define  CTL_TmplNotToSub
#		undef CTL_VNstdio
#	endif
#	define CTL_GNUC
#endif

#	define CTL_NO_RTLD_DEEPBIND
#if defined CTL_Unix && !defined CTL_CygWin
#	undef CTL_NO_RTLD_DEEPBIND
#endif  

#if defined(CTL_GNUC) || defined (CTL_Aix)
#	define CTL_TemplateDclr template
#else
#	define CTL_TemplateDclr
#endif

#if defined(CTL_GNUC)
#	define CTL_PrettyFunction __PRETTY_FUNCTION__
#endif

#if defined(CTL_SGI) || defined(CTL_MSC) || defined(CTL_Alpha)
#	define CTL_PUTENV
#endif

#if defined(CTL_VC) // || defined(__ICC)
#	undef CTL_TmplCastOper
#	if   _MSC_VER<1400
#		undef CTL_PartialSpec
#	endif
#	undef CTL_VNstdio
#endif

#define CTL_Monitor
//#define CTL_CatchSignals

#ifndef CTL_Debug
#	define CTL_CONT
#endif

#ifndef  CTL_LocatorCI
#	ifdef CTL_LOCATE
#		define CTL_Locate  CTL_LOCATE
#	endif
#	ifdef CTL_Locate
#		define CTL_LocatorCI <ctl/ci/locator.ci>
#	endif
#else
#	define CTL_Locate
#endif

#ifdef CTL_Locate
#	define CTL_UseLocator
#	if CTL_Locate==1
#		undef   CTL_Locate
#		define  CTL_Locate  CTL_Locate
#	endif
#endif

#ifndef CTL_SGI
#	define CTL_Main main
#else
#	define CTL_Main ctl_main
#endif

#ifndef CTL_Shell
#	define CTL_Shell ssh
#	define CTL_ShellPort 22
#else
#	if CTL_Shell==rsh  && !defined CTL_ShellPort
#		define CTL_ShellPort 414
#	endif
#endif

#ifndef CTL_DBG
#	ifdef CTL_SGI
#		define  CTL_DBG dbx
#	else
#		define  CTL_DBG gdb
#	endif
#endif

#ifndef CTL_Terminal
#	define CTL_Terminal xterm
#endif

#endif
