/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_CONNECT_H_
#define CTL_CONNECT_H_

#include "binaryType.h"
#include "remote.h"
#include "utl/functiontraits.h"

namespace ctl
{

template<class CI> struct CI_Alias
{
	typedef typename ctl::constTraits<CI>::resultT CI_;
	mutable CI *ci;
public:	
	mutable void (*delM)(const void*);
	static void deleteS(const void *p)
	{ delete (CI*)p; }
	typedef CI thisT;
	operator ctl::type2type<CI_Alias<CI> >&() const;
	CI_Alias(): ci(0), delM(0)
	{}
	CI_Alias(const CI_Alias<CI> &m): ci(m.ci), delM(m.delM)
	{ m.ci = 0; m.delM = 0; }
	explicit CI_Alias(const CI &m): ci(new CI(m)), delM(deleteS)
	{}
	template<class Impl> CI_Alias(const Impl& impl): ci(new CI(impl) ), delM(deleteS)
	{}
	~CI_Alias()
	{ if(delM) (*delM)(ci); }
	CI_Alias& operator=(const CI_Alias<CI> &m)
	{ ci = m.ci; delM = m.delM, m.ci = 0; m.delM = 0; return *this;  }
	operator CI_ &()
	{ return *(CI_*)ci; }
	operator const CI &() const
	{ return *ci; }
	ctl::oStream &write(ctl::oStream &os) const
	{ return ci->write(os); }
	ctl::iStream &read(ctl::iStream &is)
	{ 
		if(!ci)
		{ ci=new CI; delM = deleteS; }
		return unconst(*ci).read(is);
	}
};
template<> struct CI_Alias<void> {};

template<int typ> struct aliasSwitch
{ template<class T> struct R { typedef T resultT; }; };

template<> struct aliasSwitch<2>
{ template<class T> struct R { typedef CI_Alias<const T> resultT; }; };

template<> struct aliasSwitch<3>
{ template<class T> struct R { typedef const CI_Alias<T> resultT; }; };

template<> struct aliasSwitch<4>
{ template<class T> struct R { typedef typename T::CTL_identT resultT; }; };

template<> struct aliasSwitch<5>
{ template<class T> struct R { typedef const typename T::CTL_identT resultT; }; };
template<class CTL_Y> struct alias
{
	enum { isCI=sizeof(CTL_isCI((CTL_Y*)0,(ctl::empty*)0) )>1 };
	enum { isConst=sizeof(ctl::constCheck((CTL_Y*)0) )>1 };
	enum { isBin=sizeof(ctl::isBinary((CTL_Y*)0) )> 1 };
	enum { flag = isConst + 2*isCI + 4*isBin };
	typedef typename ctl::aliasSwitch<flag>:: CTL_TemplateDclr R<CTL_Y>::resultT resultT;
};
template<class CTL_Y> struct argAlias
{
	typedef typename alias<CTL_Y>::resultT Y;
	typedef typename ctl::binArg<Y >::resultT resultT;
};
template<class CTL_Y> struct resAlias
{
	typedef typename alias<CTL_Y>::resultT Y;
	typedef typename resultType<Y >::resultT resultT;
};

}  // ctl
char CTL_isCI(const void*, void*);
namespace ctl {

template<class RMI> struct rFunction  : public Remote
{
	using Remote::linkM;
	typedef rPointer<rFunction<RMI> > RP;
	typedef typename RMI::origY Y;
	typedef typename RMI::retY vY;
	typedef typename RMI::optArgs args;	
	typedef typename RMI::inArgs inArgs;	
	typedef typename RMI::resY resY;
	virtual resY operator()(const args &) const
	{ CTL_Assert(false); return resY(); }
};


template<class CI> struct CI_Interface
{
	typedef typename CI::CTL_Interface interf;
	typedef typename CI:: CTL_TemplateDclr CTL_RecvDef<interf> def;
	typedef typename def::CTL_IBaseT interface;
};

template<class CI> struct Interface : public Remote,  public  CI_Interface<CI>::interface
{
	typedef Interface<CI> thisT;
	typedef typename CI_Interface<CI>::interface baseT;
	using Remote::envM;
	using Remote::linkM;
	ctl::string nameM;
public:
	ctl::status receive(ctl::iStream &is)
	{
		ctl::MID mid; is>>mid;
		ctl::status ok= ((baseT*)this)->receiveMID(is, mid);
		if(ok==ctl::failed)
			ctl::throwStr(" *** method %d not implemented in class %s ***", int(mid), (CI::CTL_typeNameS()).c_str() );
		return ok;
	}
	Interface(bool guarded=true, const ctl::link &p=ctl::link(), bool getenv = true): Remote(guarded, p, getenv)
	{}
	void setName(const ctl::string &name)
	{ nameM = name; }
	typeId fullTypeName() const
	{ return CI::CTL_fullTypeNameS(); }
	// reimplement virtual ctl::Remote::methods
	void putCredit(uint4 v) const
	{ static_cast<const ctl::Remote*>(this)->putCredit(v); }
	void getCredit() const
	{ static_cast<const ctl::Remote*>(this)->getCredit(); }
	const ctl::CommunicatorSP &getCommunicator() const
	{ return  static_cast<const ctl::Remote*>(this)->getCommunicator(); }
	ctl::link &link() const
	{ return static_cast<const ctl::Remote*>(this)->linkM; }
	bool residence(PID &rootId, PID &threadId) const
	{ rootId = getRootId(Remote::envM); threadId = getThreadId(Remote::envM); return false; }
	bool isLocal() const
	{ return true;}

	ctl::Remote *getRemoteP() const
	{ return (thisT*)this; }
	ctl::shared *getSharedP() const
	{ return static_cast<const Remote*>(this)->getSharedP(); }
	ctl::result<void> destruct_()
	{
		if(!nameM)
			delete this;
		return ctl::result<void>();
	}
	void CTL_delete(const ctl::link &p)
	{
		if(p.getType()==ctl::lib)
			destruct_();
		if(p.getType()==ctl::thread)
		{
			typedef ctl::RMICall<void, ctl::field0<> , ctl::defaultCall> RMI;
			RMI::threadCall(ctl::functor(*this,&thisT::destruct_), ctl::field0<>(),p, __FILE__ );
		}
	}
};

namespace internal {

#ifdef CTL_MSC
	CTL_Forward2(CTL_ImplHandle)
#endif

template<class CI, class Impl_, class detail, class Args =ctl::field0<>, bool holdRef_=false> struct createRP
{
	typedef typename ctl::pointerTraits<Impl_>::resultT Impl;
	enum { single = ctl::conversion<Impl&, detail>::exist  };
	enum { holdRef = holdRef_ || single || ctl::pointerTraits<Impl_>::is };
	enum { exist = ctl::conversion<ctl::Interface<CI>*,Impl*>::exist && !holdRef};
#ifdef CTL_MSC
	typedef typename CTL_Get_CTL_ImplHandle<CI,Impl,detail>::resultT CI_Impl;
#else
	typedef typename CI::  CTL_TemplateDclr CTL_ImplHandle<Impl, detail, holdRef>  CI_Impl;
#endif
	typedef typename CI::CTL_RP resultT;
	typedef Args args;
	typedef typename ctl::select<exist, Impl, CI_Impl>::resultT C;
	static C* alloc(falseT)
	{ return new C(); }
	static C* alloc(trueT)
	{ return new C((Impl&)detail() ); }
	static inline resultT store(C *ci)
	{ return ctl::cPointer<typename CI::CTL_Interface>(ci); }
	C *getCIP() const
	{ return alloc(int2type<single>() ); }
	resultT operator()() const
	{ return store(getCIP() ); }
#define CTL_CreateOper(n)\
	template<CTL_ExprList(CTL_ClassList, T, Comma,n) >\
		resultT operator()(CTL_PairList(T,n)) const\
	{ return store(new C(CTL_List(x,n) ) ); }
	 CTL_CreateOper(1)
	 CTL_CreateOper(2)
	 CTL_CreateOper(3)
	 CTL_CreateOper(4)
	 CTL_CreateOper(5)
	 CTL_CreateOper(6)
	 CTL_CreateOper(7)
	 CTL_CreateOper(8)
	 CTL_CreateOper(9)
	 CTL_CreateOper(10)
	 CTL_CreateOper(11)
	 CTL_CreateOper(12)
	 CTL_CreateOper(13)
	 CTL_CreateOper(14)
	 CTL_CreateOper(15)
	 CTL_CreateOper(16)
	 CTL_CreateOper(17)
#undef  CTL_CreateOper
};

} // internal

template<class CI, class Impl, class detail, class Args >  inline ctl::Functor<typename CI::CTL_RP, const ctl::internal::createRP<CI, Impl, detail, Args>, Args>
		functor(const ctl::internal::createRP<CI, Impl, detail, Args> &create)
{ return  ctl::Functor<typename CI::CTL_RP, const ctl::internal::createRP<CI, Impl, detail, Args>, Args>(create); }


template<class RMI, class Functor> struct ImplFunc  : public rFunction<RMI>
{	
	typedef  rFunction<RMI>  baseT;
	using baseT::envM;
	typedef  ImplFunc<RMI, Functor> thisT;
	typedef typename baseT:: resY resY;
	typedef typename baseT:: vY vY;
	typedef typename baseT:: Y Y;
	typedef typename baseT:: args ciArgs;
	typedef typename baseT:: inArgs inArgs;
	typedef typename Functor::args args;
    enum { ok = (int)args::size==(int)ciArgs::size };
	ctl::Assert<ok> CTL_Error_No_of_Args_Mismatch;
	typedef typename Functor::Y resultT;

	struct ref
	{ typedef const ciArgs& carg; };
	struct cast
	{ 
		typedef typename Functor::args args;
		typedef typename Functor::Y resultT;
		typedef const typename ctl::fieldMap2<args,inArgs, internal::binCast>::resultT carg;
	};
	typedef typename ctl::select<RMI::hasBinRep, cast, ref>::resultT::carg carg;

	typedef  typename RMI::origCall caller;
	struct CallerRes
	{
		typedef typename Functor::args args;
		typedef typename Functor::Y resultT;
		Functor fM;
		const char *prettyM;
		resY operator()(const args &arg) const
		{
			cPointer<excptHandle> excpt;
			vY y (caller().call(fM, arg, (resultT*)0,  excpt, prettyM) );
			resY  r(y, excpt, prettyM);
			return r;
		}
		CallerRes(Functor f, const char *pretty): fM(f), prettyM(pretty) {}		
	};
	struct Caller
	{
		Functor fM;
		const char *prettyM;
		resultT operator()(const ciArgs &arg, cPointer<excptHandle> &excpt) const
		{ return caller().call(fM, carg(arg), (resultT*)0, excpt, prettyM); }
		Caller(Functor f, const char *pretty ): fM(f), prettyM(pretty) {}		
	};
	Functor funcM;
	const char *prettyM;
	ImplFunc(Functor func, const char *pretty): funcM(func), prettyM(pretty)
	{}
	~ImplFunc()
	{}
	virtual resY operator()(const ciArgs &arg) const
	{
		cPointer<excptHandle> excpt;
		Caller cll(funcM,prettyM);
		vY y (cll(arg, excpt) );
		resY r (y, excpt, prettyM);
		return r;
	}
	status receive(ctl::iStream &is)
	{ return RMI::call(int2type<RMI::hasBinRep>(), CallerRes(funcM,prettyM), is); }
};

template<class RMI, class Functor> inline typename ImplFunc<RMI,Functor>::CallerRes callerRes(Functor f, RMI* =0, const char *pretty=0)
{
	typedef  typename ImplFunc<RMI,Functor>::CallerRes callR;
	return callR(f, pretty);
}

template<class RMI, class C, class Meth> struct  ImplClassMeth : public rFunction<RMI>
{
	typedef typename rFunction<RMI>:: vY vY;
	typedef typename rFunction<RMI>:: Y Y;
	typedef typename rFunction<RMI>:: resY resY;
	typedef typename rFunction<RMI>:: args args;

	C &cM;
	Meth methM;
	ImplClassMeth(C &c, Meth meth): cM(c), methM(meth)
	{}
	virtual  resY operator()(const args &arg) const
	{ return ctl::functor(cM,methM)(arg); }
	status receive(ctl::iStream &is)
	{ return RMI::call(ctl::falseT(), ctl::functor(cM,methM), is); }
};


template<class Int> struct hasConstrType {};
template<class Int> struct hasStaticType {};
template<class Int> struct hasMethodType {};


template<class RMI, typename Func> inline ImplFunc<RMI,Func> *createImplFunc(Func f, const char *pretty)
{ return new ImplFunc<RMI,Func>(f, pretty); }

template<class RFICall, class fid, class FI>
	inline ctl::rFunction<RFICall> *connectFunction(FI fI, typename RFICall::origArgs*, const char *pretty)
{
	ctl::rFunction<RFICall> *fobj=0;
	fobj = createImplFunc<RFICall>(functor(fI), pretty);	
//	ctl::log("#c ** connecting %s **\n", pretty ); 
	addFunction(pretty,fobj);
	return  fobj;
}

template<class CI, class Impl, class detail, int holdRef> struct CTL_createH
{
	typedef ctl::field0<> args;
	typedef typename CI::CTL_resRP Y;
	const Impl &implM;
	CTL_createH(const Impl &impl): implM(impl) {}
	CTL_createH(Impl *pimpl): implM(*pimpl) {}
	Y operator()(const args&) const
	{ return ctl::internal::createRP<CI, Impl, detail, ctl::field0<>, holdRef>()((Impl&)implM); }  
};

template<class CI, class Impl_, class detail> inline CI create(const Impl_ &impl, channelT type=ctl::serial, const char *loc=0, CI* =0, detail* =0)
{
	typedef typename ctl::pointerTraits<Impl_>::resultT Impl;
	enum { holdRef = ctl::pointerTraits<Impl_>::is };

	CI ci;
	if(type==ctl::serial || type==ctl::local)
	{	ci.CTL_createDetail(impl, (detail*)0); return ci; }
	ctl::string Pretty =CI::CTL_typeNameS()+ "()";
	const char *pretty = Pretty.c_str();
	CTL_createH<CI, Impl, detail, holdRef> creator(impl);
	typedef ctl::RMICall<typename CI::CTL_RP, ctl::field0<>, ctl::tryCallCatch<> > RMI;

	ctl::string s(loc);
	if(type!=ctl::undefined)
		s += " -l "; s += ctl::location::channel2string(type);

	setTmpFunction(pretty, ctl::createImplFunc<RMI>(creator, pretty)); 
	ci.CTL_setLink(ctl::link(s) );
	ci.CTL_setCIPv(true);
	setTmpFunction(0,0); 
	return ci;
}

template<class CI, class Impl> inline CI create(const Impl &impl, channelT typ, const char *loc=0, CI* =0)
{ return create<CI, Impl, ctl::empty>(impl,typ,loc); }

template<class T> inline bool connect(T* =0)
{ return !ctl::connectType2((T*)0,ctl::typeName<T>().string()).empty(); }

template<class C, class Impl, class detail> inline void connect(C* =0, Impl* =0, detail* =0)
{ C::CTL_connect(falseT(), (Impl*)0,(detail*)0); }
template<class C, class Impl, class detail> inline void connectR(C* =0, Impl* =0, detail* =0)
{ C::CTL_connect(trueT(), (Impl*)0,(detail*)0); }

template<class C, class Impl> inline void connect(C* =0, Impl* =0)
{ C::CTL_connect(falseT(), (Impl*)0,(empty*)0); }
template<class C, class Impl> inline void connectR(C* =0, Impl* =0)
{ C::CTL_connect(trueT(), (Impl*)0,(empty*)0); }

template<class C, class Impl, class detail> inline void connect(ctl::link &p, C* =0, Impl* =0, detail* =0)
{ C::CTL_connect(falseT(), p, (Impl*)0,(detail*)0); }
template<class C, class Impl, class detail> inline void connectR(ctl::link &p, C* =0, Impl* =0, detail* =0)
{ C::CTL_connect(trueT(), p, (Impl*)0,(detail*)0); }

template<class C, class Impl> inline void connect(ctl::link &p, C* =0, Impl* =0)
{ C::CTL_connect(falseT(), p, (Impl*)0,(empty*)0); }
template<class C, class Impl> inline void connectR(ctl::link &p, C* =0, Impl* =0)
{ C::CTL_connect(trueT(), p, (Impl*)0,(empty*)0); }

inline bool isInBaseList(const ctl::typeId &D,  const ctl::typeId &B)
{
	const char *d=D.c_str(),*b=B.c_str() ;
	const char *base = strstr(d, b);
	if(!base)
		return false;
	if(base!=d && *(base-1)!=',' && *(base-1)!='(')
		return false;
	size_t len=strlen(b);
	base = base +len;
	if(*base && *base!=',' &&*base!=')' && *base!='(' )
		return false;
	return true;
}    
 
template<class C> inline C &addStaticParent(C* c)
{ ctl::addStatic(std::pair<void*, void(*)(const void*)>(c, &ctl::binaryType::genericIOFunctionBase<C>::del), true); return *c; }
template<class C> inline C &addStatic(C* c)
{ ctl::addStatic(std::pair<void*, void(*)(const void*)>(c, &ctl::binaryType::genericIOFunctionBase<C>::del)); return *c; }                               	
template<class C> inline C &addStaticVec(C* c)
{ ctl::addStatic(std::pair<void*, void(*)(const void*)>(c, &ctl::binaryType::genericIOFunctionBase<C>::delVec)); return *c; }                               	
template<class C> inline C& createStatic()
{ return addStatic(new C); }
template<class C, class X0>	inline C& createStatic(const X0 &x0)
{ return addStatic(new C(x0) ); }
template<class C, class X0>	inline C& createStaticParent(const X0 &x0)
{ return addStaticParent(new C(x0) ); }
template<class C, class X0, class X1>	inline C& createStatic(const X0 &x0,const  X1 &x1)
{ return addStatic(new C(x0, x1) ); }
template<class C, class X0, class X1, class X2>	inline C& createStatic(const X0 &x0, const X1 &x1,const X2 &x2)
{ return addStatic(new C(x0, x1, x2) ); }
template<class C, class X0, class X1, class X2, class X3> inline C& createStatic(const X0 &x0,const X1 &x1, const X2 &x2, const X3 &x3)
{ return addStatic(new C(x0, x1, x2, x3) ); }


} //  ctl


#endif


