/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#include "ctl.h"

#ifdef CTL_inClass
#	error nested class definitions,  probably missing CTL_ClassEnd
#endif

#ifdef CTL_LibName
#	define CTL_Prefix CTL_LibName::CTL_typeNameS()+"::"
#else
#	define CTL_Prefix ctl::typeId()
#endif

#undef CTL_Typename

#ifdef CTL_Extends
#	define CTL_BaseList : CTL_CI_BaseDclr((CTL_Extends))
#	define CTL_IF_BaseList :CTL_IF_BaseDclr((CTL_Extends))
#	define CTL_RB_BaseList  CTL_RB_BaseDclr((CTL_Extends))
#	define CTL_RB_Base CTL_Cat_2(CTL_RBaseT, CTL_Get2_2((CTL_Extends)) )
#	define CTL_IB_Base CTL_Cat_2(CTL_IBaseT, CTL_Get2_2((CTL_Extends)) )
#	define CTL_IB_BaseList  CTL_IB_BaseDclr((CTL_Extends))
#	define CTL_IR_BaseList  CTL_IR_BaseDclr((CTL_Extends))
#	define CTL_base_setCICall CTL_setCICall((CTL_Extends))
#	define CTL_base_connectStatic CTL_ConnectStatic((CTL_Extends))
#	define CTL_base_Use CTL_base_use((CTL_Extends))
#	define CTL_base_Init : CTL_base_init((CTL_Extends))
#	define CTL_BaseName  "(" +  CTL_base_name((CTL_Extends)) + ")"
#else
#	define CTL_BaseList
#	define CTL_IF_BaseList
#	define CTL_RB_BaseList
#	define CTL_RB_Base  CTL_RBaseT0
#	define CTL_IB_Base  CTL_IBaseT0
#	define CTL_IB_BaseList
#	define CTL_IR_BaseList
#	define CTL_base_setCICall
#	define CTL_base_connectStatic
#	define CTL_base_Use
#	define CTL_base_Init
#	define CTL_BaseName ""
#endif

#ifdef CTL_ClassTempl
# define CTL_ClassTmpl CTL_ClassTempl
#endif

#ifdef CTL_ClassTmpl
CTL_TemplateFwd(CTL_ClassTmpl)


#	define CTL_Typename typename
#	define CTL_ClassName CTL_Tmpl_NameH((CTL_ClassTmpl))
#	define CTL_Class   CTL_Get3_1 ((CTL_ClassTmpl))

#	define CTL_ClassDclr template<CTL_GetClassList (CTL_Get3_2 ((CTL_ClassTmpl)), CTL_Get3_3 ((CTL_ClassTmpl)) )>

#	define CTL_TypeID CTL_Prefix + CTL_TmplNameStr1 (CTL_ClassTmpl)

#elif defined CTL_Class
// forward
CTL_ClassFwd(CTL_Class)
#	define CTL_Typename
#	define CTL_ClassName CTL_Class
#	define CTL_ClassMethods CTL_Cat2(CTL_Class, Methods)
#	define CTL_ClassDclr 
#	define CTL_TypeID CTL_Prefix + CTL_Str(CTL_Class)
#else 
#	error CTL_Error: No classname specified
#endif

#	define CTL_ClassMethods CTL_Cat2(CTL_Class, Methods)


#define CTL_inClass

CTL_ClassDclr class CTL_Class CTL_BaseList
{
public:
	typedef  CTL_ClassName  ciT;
	typedef ciT thisT;

	CTL_LocalBinaries

//	enum { _ = -1 };
	typedef cstring<char> string;	
	typedef cstring<ctl::wchar> wstring;	
	typedef unsigned char uchar;
	typedef ctl::int8 int8;
	typedef ctl::uint8 uint8;
	typedef ctl::int4 int4;
	typedef ctl::uint4 uint4;
	typedef ctl::int2 int2;
	typedef ctl::uint2 uint2;
	typedef ctl::int1 int1;
	typedef ctl::uint1 uint1;
	typedef ctl::real4 real4;
	typedef ctl::real8 real8;
	typedef ctl::any any;
#ifndef size_t
#	define size_t CTL_Error_size_t_is_Arch_Dependend
#endif

	static ctl::typeId CTL_typeNameS()
	{  return CTL_TypeID; }
	virtual ctl::typeId CTL_typeName() const
	{ return CTL_typeNameS(); }
	static ctl::typeId CTL_fullTypeNameS()
	{ return CTL_typeNameS() + CTL_BaseName; }
//! 
	ctl::typeId CTL_fullTypeName() const
	{
		const CTL_Interface *cip= CTL_getCIP();
		return cip ? cip->fullTypeName() : ctl::typeId();
	 }
	virtual const ctl::link &CTL_getLink() const
	{ return CTL_attrM.linkM; }
	virtual void CTL_setLink(const ctl::link &p) const
	{
		if(!!p && (!CTL_attrM.linkM || CTL_attrM.linkM.getType()==ctl::serial || p.getCommunicator().isHolding() ) )
		{	CTL_attrM.linkM= p; }
	}
	virtual const ctl::string &CTL_getName() const
	{ return CTL_attrM.nameM; }
	virtual ctl::Environment *CTL_getEnv() const
	{ return CTL_attrM.envM; }
	const char *error() const
	{ return 0; }
	operator ctl::property<ctl::hasTypeName> &() const;
	operator ctl::type2type<thisT>&() const;
	typedef thisT CTL_Scope;
	struct CTL_Interface;

	typedef ctl::rPointer<thisT> CTL_RP;
	typedef CTL_RP CTL_identT;
	typedef ctl::result<CTL_RP> CTL_resRP;

	struct CTL_Attr
	{
		mutable ctl::link linkM;
		mutable CTL_resRP rpM;
		ctl::string nameM;
		ctl::Environment *envM;
		mutable CTL_Interface *ciPM;
		CTL_Attr(ctl::Environment *e =0): envM(e ? e : &ctl::env()), ciPM(0)
		{}
		CTL_Attr(const ctl::link &p, const ctl::string &name=ctl::string() ): linkM(p),  nameM(name), envM(&ctl::env()), ciPM(0)
		{}
		CTL_Attr &operator=(const CTL_Attr &a)
		{
			linkM = envM==a.envM ? a.linkM : ctl::getClient();
			rpM = a.rpM;
			nameM = a.nameM;
			ciPM = 0;
			return *this;
		}
  } CTL_attrM;

	typedef ctl::CI_Alias<const thisT> CTL_alias;
	CTL_alias CTL_aliasM;
	operator CTL_alias &()
	{ CTL_aliasM.ci = this; return CTL_aliasM;  }
	operator const CTL_alias &() const
	{ CTL_aliasM.ci = (thisT*)this;  return CTL_aliasM;  }
