/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_inClass
# error not in class scope, probably missing CTL_ClassBegin
#endif

#ifndef CTL_NoEmptyCtor
#	define CTL_DefEmptyCtor
#endif
public:

#define CTL_StaticMethod0 CTL_RP, CTL_getNamed, (const string, const string), 2

#define CTL_ExpandMethod CTL_DefRMITypes
#include CTL_ExpandAll

#undef CTL_StaticMethod0
// construction
	CTL_Class(const thisT &ci) CTL_base_Init
	{ CTL_create(ci, ctl::int2type<2>());  }
	template<class Y> CTL_Class(const ctl::result<Y> &res)
	{ CTL_create(res.getValue()); }
#ifdef CTL_DefEmptyCtor
	explicit CTL_Class (const ctl::link &p): CTL_attrM(p)
	{ CTL_create(p, ctl::trueT() ); }	
  explicit CTL_Class(ctl::Flag flag = ctl::late): CTL_attrM(ctl::link())
	{ CTL_create(flag, ctl::trueT() ); }
#else
	CTL_Class(): CTL_attrM(ctl::link()) 
	{ CTL_init(ctl::link()); }
#endif
	CTL_Class (const ctl::link &p, ctl::namedFlag flag, const ctl::cstring<char>& name1="#", const ctl::cstring<char>& name2=""):CTL_attrM(p)
	{ CTL_create(flag, name1, name2); }	
	explicit CTL_Class(ctl::namedFlag flag, const ctl::cstring<char>& name1="", const ctl::cstring<char>& name2=""): CTL_attrM(ctl::link())
	{ CTL_create(flag, name1, name2); }
	template<class X>  CTL_Class (const X &x): CTL_attrM(ctl::link())
	{ CTL_create(x); CTL_Assert(!!CTL_getLink()); }
//	template<class X> explicit CTL_Class (const ctl::link &p, const X &x): CTL_attrM(p)
//	{ CTL_create(x); CTL_Assert(!!CTL_getLink()); }
	template<class Impl> CTL_Class (const Impl &impl, ctl::channelT type, const char *loc=0): CTL_attrM(ctl::link())
	{ *this = ctl::create<thisT, Impl>(impl, type, loc);	}
	template<class Impl, class detail> CTL_Class (Impl &impl, detail d, ctl::channelT type, const char *loc=0): CTL_attrM(ctl::link())
	{ *this = ctl::create<thisT, Impl>(impl, type, loc, this, &d);	}
	thisT &operator =(const thisT &ci)
	{
		const ctl::link &p=ci.CTL_getLink();
		CTL_attrM.linkM = p;
		CTL_attrM.nameM = ci.CTL_getName();
		const CTL_Interface *cip=ci.CTL_getCIPThrows(false);
		if(cip)
			CTL_attrM.rpM=CTL_RP(cip);	
		CTL_setCIP((CTL_Interface*)cip);	
		return *this;
	}
	template<class X> thisT &operator =(const X &x)
	{ CTL_create(x); return *this; }
	template<class Y> thisT &operator =(const ctl::result<Y> &res)
	{ CTL_create(res.getValue());return *this; }
private:

	template<class CI> void CTL_assign(const CI &ci)
	{
		if(!ci)
			ctl::throwStr(" *** cast %s --> %s not reasonable (%s not yet instantiated) ***",
						ci.CTL_typeName().c_str(), CTL_typeNameS().c_str(), CI::CTL_typeNameS().c_str() );
		if(!ctl::isInBaseList(ci.CTL_fullTypeName().c_str(), CTL_typeNameS().c_str() ) )
			ctl::throwStr(" *** cast %s(%s) --> %s (not derived) ***", ci.CTL_typeName().c_str(), ci.CTL_fullTypeName().c_str(), CTL_typeNameS().c_str() );
		const CTL_Interface *cip =0;
		const ctl::link &p=ci.CTL_getLink();
		if(p.getType()==ctl::lib || p.getType()==ctl::thread || p.getType()==ctl::serial)
			cip=static_cast<const CTL_Interface*>(ci.CTL_getCIP()->CTL_extends(CTL_typeNameS() ) );
		else
		{
			ctl::cPointer<ctl::internal::creditT> credit;
			CTL_Try
			{ credit = ci.CTL_getCIP()->getCreditP(); }
			CTL_Catch(const ctl::exception &e)
			{ ctl::rethrowStr(e, " *** cast %s --> %s failed ***", ci.CTL_typeName().c_str(), CTL_typeNameS().c_str()  ); }
			if(!!credit)
				cip = new ctl::internal::rReference<thisT>(credit);			
		}
		if(cip)
			CTL_attrM.rpM=CTL_RP(cip);	
		else
			ctl::throwStr(" *** cast %s --> %s failed ***", ci.CTL_typeName().c_str(), CTL_typeNameS().c_str() );
		CTL_attrM.linkM = p;
		CTL_attrM.nameM = ci.CTL_getName();
		CTL_setCIP((CTL_Interface*)cip);	

	}
public:
	static char CTL_constrExist(float, ctl::empty);
	static int CTL_constrExist(int, ctl::link);
	static int CTL_constrExist(int, ctl::Flag);
	template<class X> void CTL_create(const X &x)
	{
		enum { isCI = ctl::conversion<ctl::type2type<X>, X>::exist };
		CTL_createH(x, ctl::int2type<isCI>() );
	}
	template<class X, class dum> void CTL_createHH(const X &x, ctl::falseT/*Ctor(X) does'nt exist*/, dum*)
	{
		enum { hasBinCheck = ctl::property<ctl::hasBinaryCheck, X>::value }; 
		enum { constrBinEx =  sizeof(CTL_constrExist(0, *(dum*)0 ) )>1 }; // ctor accepting binary
		enum { takeConstr = (constrBinEx && hasBinCheck) };
		CTL_create(x, ctl::int2type<takeConstr>() );
	}
	template<class X, class dum> void CTL_createHH(const X &x, ctl::trueT/*Ctor(X) exists**/, dum*)
	{	CTL_create(x, ctl::trueT() );	}
	template<class X> void CTL_createH(const X &x, ctl::falseT/*noCI*/)
	{ 
		enum { constrEx =  sizeof(CTL_constrExist(0, x ) )>1 }; // ctor accepting X
		CTL_createHH(x, ctl::int2type<constrEx>(), (ctl::binarybase*)0); 
	}
	template<class CI> void CTL_createH(const CI &ci, ctl::trueT )
	{
		enum { derived = ctl::conversion<const thisT*, typename CI::thisT*>::exist };
		enum { derivedCI = derived ? 2 : 3 };
		CTL_create(ci, ctl::int2type<derivedCI>() );
	}
public:

	template<class Impl, class detail> CTL_Interface * CTL_createDetailH(const Impl &impl, detail*, ctl::int2type<0>)
	{	
		CTL_ImplHandle<Impl,detail, false> *implH = new CTL_ImplHandle<Impl, detail, false>((Impl&)impl); 
		implH->linkM = CTL_attrM.linkM;
		return implH;
	}
	template<class Impl, class detail> CTL_Interface * CTL_createDetailH(const Impl &impl, detail*, ctl::int2type<1>)
	{
		typedef typename ctl::pointerTraits<Impl>::resultT ImplT;
		CTL_ImplHandle<ImplT ,detail, true> *implH = new CTL_ImplHandle<ImplT,detail, true>(*impl);
		implH->linkM = CTL_attrM.linkM;
		return implH;
	}
	template<class Impl, class detail> void CTL_createDetail(const Impl &impl, detail* d)
	{
		ctl::typeId mod=ctl::getLocationModifier(CTL_typeNameS() );
		if(!!mod)
		{
			*this = ctl::create(impl, ctl::undefined, mod.c_str(), (thisT*)0, d);
			return;
		}
		else
			CTL_attrM.linkM = ctl::link(ctl::createCommunicator(0, 0, ctl::serial) );
		CTL_Interface *cip = CTL_createDetailH<Impl, detail>(impl,	d, ctl::int2type<ctl::pointerTraits<Impl>::is>());
		CTL_setCIP(cip);
		CTL_attrM.rpM = CTL_RP(cip);
	}
private:
	template<class Impl> void CTL_create(const Impl &impl, ctl::falseT)
	{ CTL_createDetail(impl, &impl); }
//	{ CTL_createDetail<Impl, const Impl&>(impl, &impl); }

	template<class X> void CTL_create(const X &x, ctl::trueT)
	{
		if(CTL_init(ctl::link()) )
			CTL_attrM.rpM = CTL_new(CTL_attrM.linkM, x);
		else
			ctl::throwStr(" *** in %s::CTL_create<X>: no link found ***", CTL_typeNameS().c_str() );

	}
	template<class CI> void CTL_create(const CI &ci, ctl::int2type<2>)
	{	*this = (const thisT&)ci; 	}
	template<class CI> void CTL_create(const CI &ci, ctl::int2type<3>)
	{	CTL_assign((const typename CI::thisT&)ci); 	}
	void CTL_create(const thisT &ci, ctl::int2type<2>)
	{	*this = ci; 	}
#ifdef CTL_DefEmptyCtor
	void CTL_create(const ctl::link &p, ctl::trueT)
	{
		if(CTL_init(p) )
			CTL_attrM.rpM = CTL_new(CTL_attrM.linkM);
	}
	void CTL_create(ctl::Flag flag, ctl::trueT)
	{
		if(flag==ctl::now)
		{
			if(CTL_init(ctl::link()) )
				CTL_attrM.rpM = CTL_new(CTL_attrM.linkM);
		}
		if(!CTL_attrM.linkM && (flag==ctl::remote || flag==ctl::local2))
			CTL_attrM.linkM = ctl::getClient();		
	}
#endif
	void CTL_create(ctl::namedFlag flag, const ctl::cstring<char>& name1, const ctl::cstring<char>& name2)
	{
		bool swp=false;
		if(flag==ctl::named)
			CTL_attrM.nameM=ctl::binaryCast<ctl::string>(name1);
		else
		{
			ctl::string typId = ctl::binaryCast<ctl::string>(name2);
			if(!typId)
			{
				swp = true;
				typId = ctl::binaryCast<ctl::string>(name1);
				if(!typId || typId[0] == '#')
					ctl::throwStr(" *** in %s(named): no typeId given  ***", CTL_typeNameS().c_str() );	
			}
			if(!CTL_attrM.linkM)
				CTL_attrM.linkM = ctl::getLink(ctl::binaryCast<ctl::string>(typId) );
			if(!CTL_attrM.linkM)
				ctl::throwStr(" *** in %s(named): no link for %s found ***", CTL_typeNameS().c_str(), typId.c_str() );
		}
		if(CTL_init(CTL_attrM.linkM) )
			CTL_attrM.rpM = CTL_getNamed(CTL_attrM.linkM, swp?name2:name1, swp?name1:name2);
	}
public:
	// destruction 
	virtual ~CTL_Class()
	{		
// avoid exception in destr
		CTL_attrM.rpM.clearExcpt();

		if(CTL_attrM.rpM.isEmpty() || CTL_attrM.rpM.getRefCount()>1 )
			return;
		const CTL_RP &rp=CTL_attrM.rpM.yM;
		int refLimit = CTL_attrM.linkM.getType()==ctl::thread ? 2 : 1;
		if(!rp || rp.getRefCount()!=refLimit || !!CTL_attrM.nameM)
			return;
		if(rp.getRefCount()<=1 && (!rp->getCreditP() || rp->getCreditP()->getRefCount()<=1)) // 0 ==>  local linkage
			ctl::log("#f <- void %s::CTL_delete()\n", CTL_typeNameS().c_str() );
	}
// uses
	static void use()
	{ use(ctl::link() );  }
	static void use(ctl::link &p)
	{
		if(!p)
			p = ctl::getLink(CTL_typeNameS(), ctl::any(), ctl::sys::getLocatorName()) ;
		if(!p)
			ctl::throwStr(" *** no link found in %s::use ***", CTL_typeNameS().c_str() );
		use((const ctl::link&)p);
	}

	static void use(const char* s)
	{ use(ctl::link(s) ); }

	static void use(const ctl::string& s)
	{ use(s.c_str() ); }

	static void use(const ctl::link &p)
	{
		if(p.getType()==ctl::invalid)
			ctl::throwStr(" *** invalid link in %s::use ***", CTL_typeNameS().c_str() );
		ctl::log("#f -> %s::use\n", CTL_typeNameS().c_str() );
		if(p.getType()==ctl::modifier)
		{ 
			ctl::setLocationModifier(CTL_typeNameS().c_str(), p.getModifier());
			return; 
		}
		ctl::setLink(CTL_typeNameS(), p);
/// let all basis use same link
CTL_base_Use
	}

	template<class Impl, class detail> static void use(Impl* =0, detail* =0)
	{
		ctl::link p(ctl::serial);
    CTL_connect(ctl::trueT(), p, (Impl*)0, (detail*)0);
		use(p);
	}
	template<class Impl> static void use(Impl* =0)
	{ use((Impl*)0, (ctl::empty*)0); }


/// init & accessors
	CTL_Interface &CTL_getCIRef(bool create=true) const
	{
		if(!CTL_attrM.ciPM)
			CTL_setCIPv(create);		
		return *CTL_attrM.ciPM;
	}
	const CTL_Interface *CTL_getCIPThrows(bool create=true) const
	{
		if(!CTL_attrM.ciPM)
			CTL_setCIPv(create);	
		return CTL_attrM.ciPM;
	}
	CTL_Interface *CTL_getCIPThrows(bool create=true)
	{
		if(!CTL_attrM.ciPM)
			CTL_setCIPv(create);	
		return CTL_attrM.ciPM;
	}
	const CTL_Interface *CTL_getCIP(bool create=true) const
	{
		CTL_Try
		{ return CTL_getCIPThrows(create); }		
		CTL_Catch(const ctl::exception &e)
		{ ctl::log("#x !! in CTL_getCIP() const ***** caught(%s) *****\n", e.what() ); return 0; }
	}
	CTL_Interface *CTL_getCIP(bool create=true)
	{
		CTL_Try
		{ return CTL_getCIPThrows(create); }		
		CTL_Catch(const ctl::exception &e)
		{ ctl::log("#x !! in CTL_getCIP() ***** caught(%s) *****\n", e.what() );return 0; }
	}
private:
	bool CTL_init(const ctl::link &p) const
	{
		if(p.getType()==ctl::invalid)
			ctl::throwStr(" *** in %s::CTL_init(const ctl::link&): link is invalid (%s)***", CTL_typeNameS().c_str(), p.error() );
		if(!CTL_attrM.linkM)
			CTL_attrM.linkM = !p ? ctl::getLink(CTL_typeNameS(), ctl::any(), ctl::sys::getLocatorName()) : p;
		return  !!CTL_attrM.linkM;
	}
	virtual void CTL_initRP() const
	{
		if(CTL_init(CTL_attrM.linkM) )
		{	
#ifdef CTL_DefEmptyCtor
			if(CTL_attrM.rpM.isEmpty() )
				CTL_attrM.rpM = CTL_new(CTL_attrM.linkM);
#endif
		}
		else
			ctl::throwStr(" *** in %s::CTL_initRP(): no link found ***", CTL_typeNameS().c_str() );
	}
	const CTL_RP &CTL_getCI() const
	{
		CTL_initRP();
		const CTL_RP &rp=CTL_attrM.rpM.getValue();
		return rp;
	}

protected:
	void CTL_setCIP(CTL_Interface *ci) const
	{
		CTL_attrM.ciPM = ci;
// call CTL_setCIP of basis
		CTL_base_setCICall
	}
public:
	virtual void CTL_setCIPv(bool create) const
	{
		if(create || !CTL_attrM.rpM.isEmpty() )
		{
			const CTL_RP &rp=CTL_getCI();
			if(!rp)
				ctl::throwStr(" *** in %s::empty object called ***", CTL_typeNameS().c_str() );
			rp->link() = CTL_getLink();
			rp->link().release();
			CTL_setCIP((CTL_Interface*)rp.get());
		}		
	}
	bool operator==(const thisT &ci) const
	{ return CTL_attrM.rpM == ci.CTL_attrM.rpM; }
	bool empty() const
	{ return !CTL_attrM.linkM || !CTL_getCIP(true); }

// create if needed	

	bool operator!() const
	{ 
		if(!CTL_attrM.linkM)
				CTL_attrM.linkM = ctl::getLink(CTL_typeNameS() );
		return !CTL_attrM.linkM || !CTL_getCIP(true);
	}
	void CTL_touch() const
	{
		if(!CTL_attrM.linkM)
			CTL_attrM.linkM = ctl::getLink(CTL_typeNameS() );
		if(!CTL_attrM.linkM)
			ctl::throwStr(" *** link for '%s' is invalid(%s)***", CTL_typeNameS().c_str(), CTL_attrM.linkM.error() );
		else
			(void)CTL_getCIPThrows(true);
	}
	ctl::oStream &write(ctl::oStream &os) const
	{
		if(os.counting() && !CTL_attrM.rpM.isEmpty())
		{ 
			CTL_Try 
			{ CTL_getCIPThrows(false); }
			CTL_Catch(const ctl::exception &e)
			{ ctl::rethrowStr(e," *** can't write empty object %s ***",  CTL_typeNameS().c_str() ); }
		}
		return os<<CTL_RP(CTL_attrM.ciPM);
	}
	ctl::iStream &read(ctl::iStream &is)
	{
		CTL_RP rp;
		is>>rp;
		if(!!rp)
			CTL_attrM.rpM = rp;
		CTL_setCIP((CTL_Interface*)rp.get());
		if(!!rp)
			CTL_attrM.linkM = rp->link();
		return is;
	}

/// connect
	template<class Impl, class detail> struct CTL_Creator: public ctl::Remote
	{ 
		const void* extends(const ctl::typeId &base) const
		{ 
			if(ctl::isInBaseList(CTL_fullTypeNameS(), base) )
				return ctl::internal::createRP<thisT, Impl, detail, ctl::field0<>, false>().getCIP()->CTL_extends(base);
			return 0;
		}
	};

	template<class Impl, class detail> static void CTL_connectStaticH(Impl* =0, detail* =0)
	{
#		define CTL_ExpandConstructor CTL_DefConnect
#		include CTL_ExpandAll
#ifdef CTL_DefEmptyCtor
		CTL_resRP (*ff)() = &thisT::CTL_new;
		thisT::connectFunctionC(ctl::falseT(), ff, ctl::internal::createRP<thisT, Impl, detail>());
#endif
		connectFunctionS(ctl::int2type<0>(), &thisT::CTL_getNamed , &CTL_ImplHandle<Impl, detail>::CTL_getNamed);
// see CTL_getNamed
#ifdef CTL_DefEmptyCtor
		ctl::addFunction(CTL_typeNameS().c_str(), new CTL_Creator<Impl,detail>() );
#endif
#		define CTL_ExpandStaticMethod CTL_DefConnectStatic
#		include CTL_ExpandAll
	}		
	template<class Impl, class detail> static void CTL_connectStatic(ctl::falseT, Impl* =0, detail* =0)
	{	
		CTL_connectStaticH((Impl*)0, (detail*)0); 
	}		
	template<class Impl, class detail> static void CTL_connectStatic(ctl::trueT rec, Impl* =0, detail* =0)
	{	
		(void)rec;
		CTL_connectStaticH((Impl*)0, (detail*)0);
// basis::connectStatic
		CTL_base_connectStatic
	}

	template<class Rec, class Impl, class detail> static void CTL_connect(Rec rec, Impl* =0, detail* =0)
	{
		ctl::log("#c ** connecting %s **\n", CTL_typeNameS().c_str());
		CTL_connectStatic(rec, (Impl*)0, (detail*)0);
		ctl::log("#c \n");
	}	
	template<class Rec, class Impl, class detail> static void CTL_connect(Rec rec, ctl::link &p, Impl* =0, detail* =0)
	{
		ctl::setLink(ctl::typeId(), p);
		CTL_connect(rec, (Impl*)0, (detail*)0);
		ctl::setLink(ctl::typeId(), ctl::link() );
	}
	template<class Rec, class Impl> static void CTL_connect(Rec rec, ctl::link &p, Impl* =0)
	{ CTL_connect(rec, p, (Impl*)0, (ctl::empty*)0);  }
	template<class Rec, class Impl> static void CTL_connect(Rec rec, Impl* =0)
	{ CTL_connect(rec, (Impl*)0, (ctl::empty*)0); }



#	if defined CTL_Constructor0Throws
#		define CTL_Throws ( CTL_Constructor0Throws )
#	else
#		define CTL_Throws (  (), 0 )
#	endif
#ifdef CTL_DefEmptyCtor
#if defined CTL_LogOff || defined CTL_NoLog
#	define CTL_Pretty (const char*)0
#else
#	define CTL_Pretty pretty
#endif
	
	static const char*CTL_printC0(ctl::type2type<>* =0) 
	{ CTL_Static const char *pretty= ctl::prettyNameS (0, CTL_typeNameS().c_str(), 0, 0, "()", 0) ; return pretty; };
	typedef CTL_Tmpl_Name2S(ctl::tryCallCatch, CTL_Throws)  CTL_Cat3(CTL_Catch,C,0);
	CTL_DefFunction8(0, CTL_RP, CTL_new, (), 0, static, "C", (<>), 1, C )
#endif
#	undef  CTL_Throws


///////////  abstract interface used by extending classes //////////////
// inherits baseList::CTL_Interface
struct CTL_Interface CTL_IF_BaseList
{
  	typedef CTL_ClassName ciT;
#if defined CTL_TmplNotToSub
CTL_LocalBinaries
#endif

	typedef CTL_Interface thisT;
	CTL_Interface() {}
//  template<class Impl> CTL_Interface(Impl &) {}

	virtual ctl::result<void> destruct_()
	{ delete this;  return ctl::result<void>(); }
	void CTL_delete(const ctl::link &p)
	{
		if(p.getType()==ctl::thread)
		{
			typedef ctl::RMICall<void, ctl::field0<> , ctl::defaultCall> RMI;
			RMI::threadCall(ctl::functor(*this,&thisT::destruct_), ctl::field0<>(), p, __FILE__ );
		}
		else
			destruct_();
	}

	virtual ~CTL_Interface() {}
	virtual void setImpl(const void*) {}
	virtual void setName(const ctl::string &)=0;
	virtual const ctl::CommunicatorSP &getCommunicator() const =0;
	virtual ctl::link &link() const=0;
	virtual bool isLocal() const=0;
	virtual ctl::Remote *getRemoteP() const =0;
	virtual ctl::shared *getSharedP() const =0;
	virtual ctl::typeId fullTypeName() const=0;
	virtual const ctl::internal::creditT *getCreditP() const
	{ return 0; }
	virtual bool residence(ctl::PID &rootId, ctl::PID &threadId) const =0;
 	ctl::status receiveMID(ctl::iStream&, const ctl::MID&)
	{ return ctl::failed; }
	virtual const void* CTL_extends(const ctl::typeId &) const
	{ return this; }
#define CTL_ExpandMethod CTL_DclrPureMeth
#include CTL_ExpandAll
};



///////////  the CI interface //////////////
#define CTL_StaticMethod0 CTL_RP, CTL_getNamed, (const string, const string), 2

#define CTL_ExpandMethod CTL_MethCallImpl
#define CTL_ExpandConstructor CTL_DefConstr
#define CTL_ExpandStaticMethod CTL_DefStaticMeth
#include CTL_ExpandAll

#undef CTL_StaticMethod0



///////////  remote interface used by extending classes //////////////

template<class base = CTL_Interface> struct CTL_RemoteInstanceBase : public base
{
	typedef CTL_ClassName ciT;
#if defined CTL_TmplNotToSub
CTL_LocalBinaries
#endif
   virtual const ctl::link &checkRP(const ctl::MID &mid, ctl::objectID &trgt, ctl::PID &pid) const=0;
	typedef  CTL_RemoteInstanceBase thisT;
	typedef  base baseT;

#define CTL_ExpandMethod CTL_DefMethSend
#include CTL_ExpandAll
};

template<class base = CTL_Interface> struct CTL_RemoteInstanceBaseExt : public base
{
#if defined CTL_TmplNotToSub
CTL_LocalBinaries
#endif

	typedef base baseT;
	typedef baseT CTL_RBaseT0;
// construct stack of CTL_RemoteInstanceBase into CTL_RB_Base
	CTL_RB_BaseList

	typedef typename CTL_ClassName:: CTL_TemplateDclr CTL_RemoteInstanceBase<CTL_RB_Base> CTL_RBaseT;
};

///////////  implementation classes //////////////
struct  CTL_RemoteBase: public ctl::Remote, public CTL_RemoteInstanceBaseExt<>::CTL_RBaseT
{
	using ctl::Remote::guardedM;
	void CTL_delete(const ctl::link &)
	{}
};

///////////  implementation by direct derivation  //////////////
template<class base=CTL_Interface> struct CTL_Receiver: public base
{
typedef CTL_ClassName ciT;
#if defined CTL_TmplNotToSub
CTL_LocalBinaries
#endif
	typedef base baseT;
	typedef CTL_Receiver<base> thisT;
	using baseT::getRemoteP;
	ctl::status receiveMID(ctl::iStream &is, const ctl::MID &midStr)
	{
		switch(midStr)
		{
			case 0:
			{
				ctl::uint4 refVal;
				is>>refVal;
				is.clear();
				getRemoteP()->putCredit(refVal);
				return ctl::success; 
			}

#define CTL_ExpandMethod CTL_DefCallDerivedSwitch
#include CTL_ExpandAll

			default:
				return  baseT::receiveMID(is, midStr);
		}
	}
};

///////////  implementation interface //////////////
template<class base =CTL_Interface > struct CTL_RecvDef
{
	typedef  base baseT;
	typedef baseT CTL_IBaseT0;
	CTL_IR_BaseList
	typedef typename ciT:: CTL_TemplateDclr CTL_Receiver<CTL_IB_Base>  CTL_IBaseT;
};


template<class Impl, class detail = ctl::empty, class base=CTL_Interface, bool holdRef =false> struct CTL_ImplBase: public base
{
typedef CTL_ClassName ciT;
#if defined CTL_TmplNotToSub
CTL_LocalBinaries
#endif
	typedef CTL_ImplBase<Impl, detail, base, holdRef> thisT;
	enum { single = holdRef  };
//	enum { single = ctl::conversion<const Impl&, detail>::exist  };
	typedef typename ctl::select<holdRef, Impl&, Impl>::resultT implT;
	typedef base baseT;
	using baseT::getRemoteP;
	using baseT::link;
	Impl *pimplM;
  CTL_ImplBase():pimplM(0) {}
	void setImpl(Impl *p)
	{ pimplM = p; ((base*)this)->setImpl(p); } 
	const void* CTL_extends(const ctl::typeId &id) const
	{
		if(id==CTL_ClassName::CTL_typeNameS() )
			return CTL_ClassName::CTL_Interface::CTL_extends(id);
		int atTop = ctl::same<baseT, typename CTL_ClassName::CTL_Interface>::value;
		return atTop ? 0 : baseT::CTL_extends(id);
	}
	ctl::status receiveMID(ctl::iStream &is, const ctl::MID &midStr)
	{
		switch(midStr)
		{
			case 0:
			{
				ctl::uint4 refVal;
				is>>refVal;
				is.clear();
				getRemoteP()->putCredit(refVal);
				return ctl::success;
			}

#define CTL_ExpandMethod CTL_DefCallSwitch
#include CTL_ExpandAll

			default:
				return  baseT::receiveMID(is, midStr);
		}
	}

#define CTL_ExpandMethod CTL_DefTrampoline
#define CTL_ExpandStaticMethod CTL_DefStaticTrampoline
#include CTL_ExpandAll


}; // CTL_ImplBase

template<class Impl, class detail, class base =CTL_Interface, bool holdRef=false> struct CTL_ImplDef
{
	typedef  base baseT;
	typedef baseT CTL_IBaseT0;
	CTL_IB_BaseList
	typedef typename ciT:: CTL_TemplateDclr CTL_ImplBase<Impl, detail, CTL_IB_Base, holdRef>  CTL_IBaseT;
};

///////////  implementation wrapper //////////////

template<class Impl, class detail = ctl::empty, bool holdRef=false> struct CTL_ImplHandle: public ctl::Remote, public CTL_ImplDef<Impl, detail, typename ciT::CTL_Interface, holdRef>::CTL_IBaseT
{
	typedef CTL_ClassName ciT;
	typedef typename ciT::CTL_Interface CTL_Interface;
#if defined CTL_TmplNotToSub
CTL_LocalBinaries
#endif

	typedef CTL_ImplHandle<Impl, detail> thisT;
	typedef typename CTL_ClassName:: CTL_TemplateDclr CTL_ImplDef<Impl, detail, typename ciT::CTL_Interface, holdRef >::CTL_IBaseT baseT;
	enum { single = baseT::single  };
	typedef typename baseT::implT implT;
	using ctl::Remote::envM;
	using ctl::Remote::linkM;
	ctl::string CTL_nameM;
	implT implM;
public:
	ctl::status receive(ctl::iStream &is)
	{
		ctl::MID mid; is>>mid;
		ctl::status ok= ((baseT*)this)->receiveMID(is, mid);
		if(ok==ctl::failed)
			ctl::throwStr(" *** method %d '%s' not implemented in class %s ***", int(mid), mid.strM.c_str(),(CTL_TypeID).c_str() );
		return ok;
	}
  const void *extends(const ctl::typeId &id)  const
	{ return baseT::CTL_extends(id); }
	bool isLocal() const
	{ return true;}
	const void *getCIP() const
	{ return static_cast<const CTL_Interface*>(this); }
	virtual ctl::typeId fullTypeName() const
	{ return CTL_fullTypeNameS(); }
	static CTL_RP CTL_getNamed(const ctl::string &objId, const ctl::typeId &ciname)
	{
		ctl::string type=!ciname ? CTL_typeNameS() : ciname;
		std::pair<ctl::typeId, ctl::string> key(type, objId);
		if(!!objId)
		{
			ctl::cPointer<ctl::Remote> sp=ctl::getNamedObject(key);
			if(!!sp)
				return sp;
		}
		ctl::RemoteSP creator=ctl::getFunction(type.c_str());
		const CTL_Interface *cip=0;
		if(!!creator)
			cip = static_cast<const CTL_Interface*>(creator->extends(type) );
		else
			ctl::throwStr(" *** in %s(named): %s not connected ***", CTL_typeNameS().c_str(), type.c_str() );
		if(!cip)
			ctl::throwStr(" *** in %s(named): %s not derived ***", CTL_typeNameS().c_str(), type.c_str() );

		CTL_RP  rp=ctl::cPointer<CTL_Interface>(cip) ;
		if(!!objId)
		{
			rp->setName(objId);
			ctl::setNamedObject(key,ctl::RemoteSP(rp->getRemoteP()) );
		}
		return rp;
	}

	void setName(const ctl::string &name)
	{ CTL_nameM = name; }
	// reimplement virtual ctl::Remote::methods
	const ctl::CommunicatorSP &getCommunicator() const
	{ return  static_cast<const ctl::Remote*>(this)->getCommunicator(); }
	ctl::link &link() const
	{ return static_cast<const ctl::Remote*>(this)->linkM; }
	bool residence(ctl::PID &rootId, ctl::PID &threadId) const
	{ rootId = ctl::getRootId(ctl::Remote::envM); threadId = ctl::getThreadId(ctl::Remote::envM); return false; }	ctl::Remote *getRemoteP() const
	{ return (thisT*)this; }
	ctl::shared *getSharedP() const
	{ return static_cast<const ctl::Remote*>(this)->getSharedP(); }
	const ctl::link &checkRP(const ctl::MID &mid, ctl::objectID &trgt, ctl::PID &pid) const
	{ return static_cast<const ctl::Remote*>(this)->checkRP(mid, trgt, pid); }

	ctl::result<void> destruct_()
	{
		if(!CTL_nameM)
		{	delete this; }
		return ctl::result<void>();
	}
	void CTL_init()
	{ ((baseT*)this)->setImpl(&implM); }
	~CTL_ImplHandle()
	{ ctl::log("#f -> void %s::CTL_delete(%p)\n",CTL_typeNameS().c_str(), this); }
	CTL_ImplHandle(Impl &impl_): ctl::Remote(true, ctl::getClient()), implM(impl_) 
	{ CTL_init(); }
	CTL_ImplHandle(ctl::Flag=ctl::remote ): ctl::Remote(true, ctl::getClient()) 
	{ CTL_init(); }
	CTL_ImplHandle(ctl::falseT): ctl::Remote(true, ctl::getClient()) 
	{ CTL_init(); }
	CTL_ImplHandle(ctl::trueT): ctl::Remote(true, ctl::getClient()), implM(detail())
	{ CTL_init(); }

#define CTL_MaxCtrArgs 0

#define CTL_ExpandConstructor CTL_DefImplConstr
#include CTL_ExpandAll

#define CTL_ImplConstr(n)\
	CTL_Tmpl_Dclr2((CTL_List(X,n)),n) CTL_ImplHandle(CTL_PairList(X,n)): ctl::Remote(true, ctl::getClient()), implM(CTL_List(x,n) ) { CTL_init(); }

#if CTL_MaxList > 0 
	CTL_ImplConstr(1)
# if CTL_MaxList > 1 
	CTL_ImplConstr(2)
#  if CTL_MaxList > 2 
	CTL_ImplConstr(3)
#   if CTL_MaxList > 3 
	CTL_ImplConstr(4)
#    if CTL_MaxList > 4 
	CTL_ImplConstr(5)
#     if CTL_MaxList > 5 
	CTL_ImplConstr(6)
#      if CTL_MaxList > 6 
	CTL_ImplConstr(7)
#       if CTL_MaxList > 7 
	CTL_ImplConstr(8)
#        if CTL_MaxList > 8 
	CTL_ImplConstr(9)
#         if CTL_MaxList > 9
	CTL_ImplConstr(10)
#          if CTL_MaxList > 10
	CTL_ImplConstr(11)
#           if CTL_MaxList > 11 
	CTL_ImplConstr(12)
#            if CTL_MaxList > 12
	CTL_ImplConstr(13)
#             if CTL_MaxList > 13
	CTL_ImplConstr(14)
#              if CTL_MaxList > 14
	CTL_ImplConstr(15)
#               if CTL_MaxList > 15
	CTL_ImplConstr(16)
#                if CTL_MaxList > 16 
	CTL_ImplConstr(17)
#                endif
#               endif
#              endif
#             endif
#            endif
#           endif
#          endif
#       	endif
#     	 endif
#       endif
#      endif
#     endif
#    endif
#   endif
#  endif
# endif
#endif

#undef CTL_ImplConstr
#undef CTL_MaxCtrArgs

}; // CTL_ImplHandle


}; // CTL_Class



// C/Extern specific
#ifdef CTL_Extern


#ifdef CTL_ClassPrefix
#	define CTL_ClassPrefix_  CTL_ClassPrefix
#else
 #	define CTL_ClassPrefix_  CTL_Class
#endif

#undef  CTL_Typename
#define  CTL_Typename
// Expand global fortran procedures here
  namespace ctl
  {
    namespace Extern
    {

extern "C"
{
#define const
void CTL_FortranName2(CTL_ClassPrefix_, new) (char* location, ctl::int8 *handle)
{
	ctl::Extern::create<CTL_ClassName>(location, handle);
/*
	ctl::link p(location);
	CTL_ClassName ci(p);
	if(!!ci)
		ctl::Extern::objMapper<CTL_ClassName>(*handle, ci);
	else
		*handle = 0;
*/
}

void CTL_FortranName2(CTL_ClassPrefix_, delete_impl) ();

void CTL_FortranName2(CTL_ClassPrefix_, rel)(ctl::int8 *handle)
{ *handle = -*handle; ctl::Extern::objMapper<CTL_ClassName>(*handle); }


void CTL_FortranName2(CTL_ClassPrefix_, recv)(ctl::int8 *res)
{ ctl::Extern::recv(res); }

#define CTL_ExpandMethod CTL_FortranDefClass
#define CTL_ExpandConstructor CTL_FortranDefConstr
#include CTL_ExpandAll
#undef const
} // extern "C"

#if defined CTL_Connect | defined CTL_Connecting

struct CTL_Class
{
	struct detail
	{

#	define CTL_ExpandConstructor CTL_FortranConstrDetail
#	include CTL_ExpandAll
	};
	CTL_Class() {}
	~CTL_Class()
	{ CTL_FortranName2(CTL_ClassPrefix_, delete_impl)();	}
#define CTL_ExpandMethod CTL_FortranImplClass
#define CTL_ExpandConstructor CTL_FortranImplConstr
#include CTL_ExpandAll
};

#endif
 	 } // Extern
 } // ctl
#	if defined(CTL_Connect)

#	endif
#undef CTL_ClassPrefix
#undef CTL_ClassPrefix_
#endif
// fortran
#define CTL_UndefMethods
#include CTL_ExpandAll

#ifdef CTL_Extends
#	undef CTL_Extends
#endif

#ifdef CTL_NoEmptyCtor
#	undef CTL_NoEmptyCtor
#endif
#ifdef CTL_DefEmptyCtor
#	undef CTL_DefEmptyCtor
#endif

#ifdef size_t
#	undef size_t
#endif

#undef CTL_base_Init
#undef CTL_BaseList
#undef CTL_ClassName
#undef CTL_UndefMethods
#undef CTL_ClassTmpl
#undef CTL_ClassTempl
#undef CTL_inClass
#undef CTL_Class
#undef CTL_Typename
#undef CTL_Prefix
#undef CTL_TypeID
#undef CTL_IF_BaseList
#undef CTL_RB_BaseList
#undef CTL_RB_Base
#undef CTL_IB_Base
#undef CTL_IB_BaseList
#undef CTL_IR_BaseList
#undef CTL_base_setCICall
#undef CTL_base_connectStatic
#undef CTL_base_Use
#undef CTL_BaseName
#undef CTL_ClassDclr

