/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_MCR_
#define CTL_MCR_

#include "utl.mcr"

#define CTL_none0()
#define CTL_NullPList(n,x) ctl::nullPointer(x)
#define CTL_NullTList(n,CTL_Bin) (const CTL_Trans1(CTL_Bin)*)0

// macros for intrinsic ctl type declarations in class scope
#define CTL_named(list, n)

#define CTL_isString(list, n) \
	template<class CTL_Bin> inline char congruence(const CTL_Bin* =0) const\
	{ return CTL_Bin::stringCheck(CTL_MacroApply(CTL_NullTList, Comma, list, n) ); }

#define CTL_isReference(list, n) \
	template<class CTL_Bin> inline char congruence(const CTL_Bin* =0) const\
	{ return CTL_Bin::referenceCheck(CTL_MacroApply(CTL_NullTList, Comma, list, n) ); }

#define CTL_isArray(list, n) \
	template<class CTL_Bin> inline char congruence(const CTL_Bin* t=0) const\
		CTL_CheckisArray(CTL_Bin, list, n, tList, nt, _)

#define CTL_isTupel(list, n) \
	template<class CTL_Bin> inline char congruence(const CTL_Bin* =0) const\
	{ return CTL_Bin::tupelCheck(CTL_MacroApply(CTL_NullTList, Comma, list, n) ); }

#define CTL_tupel(list, n) \
	template<class CTL_Bin> inline char congruence(const CTL_Bin* =0) const\
	{ return CTL_Bin::tupelCheck(CTL_MacroApply(CTL_NullPList, Comma, list, n) ); }\
	ctl::oStream &write(ctl::oStream &os) const\
	{  return os<<CTL_MacroApply(CTL_Second, LShift, list, n); }\
	ctl::iStream &read(ctl::iStream &is)\
	{ return is>>CTL_MacroApply(CTL_Second, RShift, list, n); }

#define CTL_stdArray(list, n) \
	CTL_isArray(list, n)\
	ctl::oStream &write(ctl::oStream &os) const\
	{  return ctl::internal::sendStdContainer(os, *this); }\
	ctl::iStream &read(ctl::iStream &is)\
	{ return ctl::internal::readStdContainer(is, *this); }

#define CTL_Type(T, binType, list, n) \
	CTL_TypeName(T);\
	CTL_##binType(list, n)\
	operator ctl::property<ctl::hasBinaryCheck>() const

#define CTL_AbstractType(T, binType, list, n) \
	virtual ctl::typeId CTL_typeName() const =0;\
	static ctl::typeId CTL_typeNameS()\
	{ return #T; }\
	operator ctl::property<ctl::hasTypeName>() const;\
	CTL_##binType(list, n)\
	operator ctl::property<ctl::hasBinaryCheck>() const;\
	operator ctl::abstractT<T>() const

#define CTL_Template(T, Tlist, nt, binType, list, n) \
  CTL_TemplateName (T,Tlist,nt);\
  CTL_##binType(list,n)\
  operator ctl::property<ctl::hasBinaryCheck>() const

// macros for extrinsic ctl type declarations in global scope
#define CTL_CheckisString(T, list, n, Tlist, nt, decl) \
 	{	(void)t; return CTL_Bin::stringCheck(CTL_MacroApply(CTL_NullTList, Comma, list, n) ); }\

#define CTL_CheckisReference(T, list, n, Tlist, nt,  decl) \
 	{	(void)t; return CTL_Bin::referenceCheck(CTL_MacroApply(CTL_NullTList, Comma, list, n) ); 	}\

#define CTL_CheckisArray1(list, n) \
{	(void)t; return CTL_Bin::arrayCheck(CTL_MacroApply(CTL_NullTList, Comma, list, 1), (ctl::int2type<-1>*)0 ); 	}
#define CTL_CheckisArrayH(list, n) \
{	(void)t; return CTL_Bin::arrayCheck( (const CTL_First(list,n)*)0, (ctl::int2type<CTL_Tail(list,n)>*)0  ); 	}

#define CTL_CheckisArray2(list, n) CTL_CheckisArrayH(list, n)
//{	(void)t; return CTL_Bin::arrayCheck( (const CTL_Get2_1(list)*)0, (ctl::int2type<CTL_Get2_2(list)>*)0  ); 	}
#define CTL_CheckisArray3(list, n) CTL_CheckisArrayH(list, n)
#define CTL_CheckisArray4(list, n) CTL_CheckisArrayH(list, n)
#define CTL_CheckisArray5(list, n) CTL_CheckisArrayH(list, n)
#define CTL_CheckisArray6(list, n) CTL_CheckisArrayH(list, n)
#define CTL_CheckisArray7(list, n) CTL_CheckisArrayH(list, n)
#define CTL_CheckisArray8(list, n) CTL_CheckisArrayH(list, n)

#define CTL_CheckisArray(T, list, n, Tlist, nt, decl) CTL_CheckisArray##n (list, n)

#define CTL_CheckstdArray(T, list, n, tList, nt, decl) \
	CTL_CheckisArray##n (list,n)\
	decl inline ctl::oStream &operator<<(ctl::oStream &os, const T& _)\
	{ return ctl::internal::sendStdContainer(os, _); }\
	decl inline ctl::iStream &operator>>(ctl::iStream &is, T& _)\
	{ return ctl::internal::readStdContainer(is, _); }

#define CTL_CheckisTupel(T, list, n, tList, nt, decl) \
	{	(void)t; return CTL_Bin::tupelCheck(CTL_MacroApply(CTL_NullTList, Comma, list, n) ); }

#define CTL_Checktupel(T, list, n, tList, nt, decl) \
 	{\
 		const T &_ = *t;\
 		return CTL_Bin::tupelCheck(CTL_MacroApply(CTL_NullPList, Comma, list, n) );\
	}\
	decl inline ctl::oStream &operator<<(ctl::oStream &os, const T &_)\
	{  return os<<CTL_MacroApply(CTL_Second, LShift, list, n); }\
	decl inline ctl::iStream &operator>>(ctl::iStream &is, T &_)\
	{ return is>>CTL_MacroApply(CTL_Second, RShift, list, n); }

#define CTL_ChecktupelCtor(T, list, n, tList, nt, decl) \
 	{\
 		const T &_ = *t;\
 		return CTL_Bin::tupelCheck(CTL_MacroApply(CTL_NullPList, Comma, list, n) );\
	}\
	decl inline ctl::oStream &operator<<(ctl::oStream &os, const T &_)\
	{  return os<<CTL_MacroApply(CTL_Second, LShift, list, n); }\
	decl inline ctl::iStream &operator>>(ctl::iStream &is, T &_)\
	{ return readByCtor(is, _, CTL_GetWC(list, n) ); }

#define CTL_SetType(T, binType, list, n) \
  CTL_SetTypeName(T);\
namespace ctl	 {\
	int setProperty( ctl::property<ctl::hasBinaryCheck>*,const T*);  \
  template<class CTL_Bin> inline char  congruence(const CTL_Bin* =0, const T* t=0)\
  CTL_Check##binType(T, list, n, (), 0, template<> )\
}

#define CTL_SetTemplate(T, Tlist, nt, binType, list, n) \
  CTL_SetTemplateName(T, Tlist, nt);\
namespace ctl	 {\
 template<CTL_GetClassList(Tlist,nt) > int setProperty( ctl::property<ctl::hasBinaryCheck>*,const CTL_Tmpl_Name(T,Tlist,nt)*);\
 template<class CTL_Bin,  CTL_GetClassList(Tlist,nt) > inline char congruence(const CTL_Bin* =0, const CTL_Tmpl_Name(T,Tlist,nt)* t =0)\
 CTL_Check##binType(CTL_Tmpl_Name(T,Tlist,nt), list, n, Tlist, nt, CTL_Tmpl_Dclr2(Tlist, nt))\
}
 
// forward declaration

#define CTL_ClassFwd(CI)\
class CI;\
int CTL_isCI(const CI*, ctl::empty*);

#define CTL_TemplateFwd(templ)\
template<CTL_GetClassList (CTL_Get3_2 ((templ)), CTL_Get3_3 ((templ)) )> class CTL_Get3_1 ((templ));\
template<CTL_GetClassList (CTL_Get3_2 ((templ)), CTL_Get3_3 ((templ)) )> int CTL_isCI(const CTL_Tmpl_NameH((templ)) *, ctl::empty*);


#define CTL_LocalBinaries\
	static const int _=-1;\
	template<class CTL_T> struct cstring: public ctl::binarybase \
	{ typedef ctl::cstring<CTL_T> CTL_identT; }; \
	template<class CTL_T, ctl::int8 CTL_D1=_, ctl::int8 CTL_D2=0, ctl::int8 CTL_D3=0, ctl::int8 CTL_D4=0, ctl::int8 CTL_D5=0, ctl::int8 CTL_D6=0, ctl::int8 CTL_D7=0, ctl::int8 CTL_D8=0> struct array: public ctl::binarybase  \
	{ typedef ctl::array<CTL_T,CTL_D1,CTL_D2,CTL_D3,CTL_D4,CTL_D5,CTL_D6,CTL_D7,CTL_D8> CTL_identT; }; \
	template<class CTL_T> struct reference: public ctl::binarybase \
	{ typedef ctl::reference<CTL_T> CTL_identT; }; \
	template<class CTL_T0, class CTL_T1, class CTL_T2=ctl::empty, class CTL_T3=ctl::empty, \
			class CTL_T4=ctl::empty, class CTL_T5=ctl::empty, class CTL_T6=ctl::empty, class CTL_T7=ctl::empty, \
 			class CTL_T8=ctl::empty, class CTL_T9=ctl::empty, class CTL_T10=ctl::empty, class CTL_T11=ctl::empty, \
			class CTL_T12=ctl::empty, class CTL_T13=ctl::empty, class CTL_T14=ctl::empty, class CTL_T15=ctl::empty> \
				struct tupel: public ctl::binarybase\
	{ typedef ctl::tupel<CTL_T0,CTL_T1,CTL_T2,CTL_T3,CTL_T4,CTL_T5,CTL_T6,CTL_T7,CTL_T8,CTL_T9,CTL_T10,CTL_T11,CTL_T12,CTL_T13,CTL_T14,CTL_T15> CTL_identT; };


#define CTL_Constructor(mid, list, n)\
	typedef CTL_Field_Type (list, n) CTL_constrType##mid;\
	operator  ctl::hasConstrType<ctl::int2type<mid> >() const

#define CTL_StaticMethod(mid, y, f, list, n)\
	static y (*getFunc(ctl::int2type<mid>)) (CTL_Get(list,n) ) \
	{ y (*func) ( CTL_Get(list,n) ) =f; return func; }\
	operator  ctl::hasStaticType<ctl::int2type<mid> >() const

#define CTL_Method(mid, y, f, list, n)\
	template<class Impl> static y (Impl::*getMeth(const Impl*, ctl::int2type<mid>) ) list\
	{ return &f; }\
	operator  ctl::hasMethodType<ctl::int2type<mid> >() const

#	if defined CTL_NoLog
#		define CTL_Log0(e, f) ;
#		define CTL_Log1(e, f, a0) ;
#		define CTL_Log2(e, f, a0, a1) ;
#		define CTL_Log3(e, f, a0, a1, a2) ;

#	else
#		define CTL_Log0(e, f) ctl::log(e, f);
#		define CTL_Log1(e, f, a0) ctl::log(e, f, a0);
#		define CTL_Log2(e, f, a0, a1) ctl::log(e, f, a0, a1); 
#		define CTL_Log3(e, f, a0, a1, a2) ctl::log(e, f, a0, a1, a2); 
#	endif
	

#endif

