/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef _FORTRAN_H_
#define _FORTRAN_H_

#	ifndef CTL_Extern
#	 define  CTL_Extern
#	endif

#include "ctl/system.h"
#include "ctl/gen/fortran.mcr"

#define CTL_FortranName(class) CTL_Cat2(class,CTL_FortranPostFix)
#define CTL_FortranName2(class, method) CTL_Cat4(class,_,method,CTL_FortranPostFix_)

extern "C"
{
	void CTL_FortranName2(ctl,cwd)(char *path, size_t *len)
	{ 
		const char *cwd=ctl::getcwd();
		size_t s=strlen(cwd);
		if(*len<s )
			*len = -s;
		else
		{
			strncpy(path, cwd, *len); 
			*len = s;
		} 
//		std::cerr<<"ctl_cwd'"<<path<<"'\n";
	}
	void CTL_FortranName2(ctl,getfd)(int *fd)
	{
		FILE *f=fopen("/dev/null", "a");
		if(!f)
			*fd=-1;
		else
		{
			*fd = fileno(f);
			fclose(f);
			fprintf(stderr," ctl_getfd %d\n", (*fd));
		}
	}
}
//gnu pascal
#if defined CTL_ConnectP && defined CTL_GPC
extern "C"
{
extern void _p_initialize (int argc, char **argv, char **envp, int options);
extern void _p_finalize ();
}
inline void CTL_Initialize()
{ _p_initialize(0, 0, 0, 0); }
inline void CTL_Finalize()
{ _p_finalize(); }
#elif defined CTL_ConnectF && defined CTL_F77_Init
extern "C"
{
extern void CTL_FortranName2(f,init)();
extern void CTL_FortranName2(f,exit)();
}
inline void CTL_Initialize()
{ CTL_FortranName2(f,init)(); }
inline void CTL_Finalize()
{ CTL_FortranName2(f,exit)(); }
#else
inline void CTL_Initialize() {}
inline void CTL_Finalize() {}
#endif

namespace ctl
{
template<class C, class Impl> void connectC(C* =0, Impl* =0)
{ C::CTL_connect(ctl::falseT(), (Impl*)0,(typename Impl::detail*)0); }
template<class C, class Impl> void connectF(C* =0, Impl* =0)
{ C::CTL_connect(ctl::falseT(), (Impl*)0,(typename Impl::detail*)0); }
template<class C, class Impl> void connectP(C* =0, Impl* =0)
{ C::CTL_connect(ctl::falseT(), (Impl*)0,(typename Impl::detail*)0); }

namespace Extern
{
	typedef ctl::int2 int2;
	typedef ctl::int4 int4;
	typedef ctl::int8 int8;
	typedef ctl::real4 real4;
	typedef ctl::real8 real8;

	template<class T, class valType = ctl::array<T>, class refType=ctl::arrayRef<T> > struct array
	{
		typedef array<T, valType, refType> thisT;
		typedef T value;
		mutable bool ownerM;
		int8 ssM,osM, *sM;
		T *tM;

		thisT &operator=(const thisT &f)
		{
//ctl::log("fortran = array: %e ssM=%Ld %p\n", tM ? *tM : -1.0, ssM, sM);
			CTL_Assert(!ownerM);
			if(!sM) //called via ctl::array::_cast()
			{
				CTL_Assert(ssM>=f.ssM);
				ssM = f.ssM;
				for(int8 i=0; i<ssM; i++)
					tM[i] = f.tM[i];
			}
			else
			{
				ownerM = f.ownerM;
				ssM = f.ssM;
				osM = ssM;
				sM = f.sM;
				tM = f.tM;
				f.ownerM = false;
			}
			return *this;
		}
		array(int8 *s,T *t):ownerM(false), ssM(*s), osM(ssM),sM(s),tM(t)
		{
			if(t && sizeof(T)==1) // is string
				t[*s-1] = T();
		}
		array(const thisT &f): ownerM(f.ownerM), ssM(f.ssM), osM(ssM), sM(f.sM),tM(f.tM) 
		{ 
//ctl::log("fortran ctor (array): %e ssM=%Ld %p %Ld\n", tM ? *tM : -1.0, ssM, sM, *sM);
			sM = &ssM;
			f.ownerM = false;
		}
		array(): ownerM(true),ssM(0), osM(ssM),sM(&ssM),tM(0) 
		{
//			ctl::log("fortran ctor (): %e ssM=%Ld %p %Ld\n", tM ? *tM : -1.0, ssM, sM, *sM);
		}
		~array()
		{
//			ctl::log("fortran dtor (): %e ssM=%Ld %p %Ld\n", tM ? *tM : -1.0, ssM, sM, *sM);
			if(ownerM)
			{ delete[] tM;  }
		}
		thisT &operator =(const valType &a)
		{  
			sM = 0; 
			thisT f = a.cast__((const thisT*)0, (thisT*)0 ); 
			ssM = f.ssM;
			for(int8 i=0; i<ssM; i++)
				tM[i] = f.tM[i];
//ctl::log("fortran = ctl::array: %e ssM=%Ld %p %Ld\n", tM ? *tM : -1.0, ssM, sM, *sM);
			return *this;
		}
		mutable refType arrayRefTmpM;
		operator refType &()
		{ return arrayRefTmpM=refType(*this, true); }
		operator refType &()  const
		{ return arrayRefTmpM=refType(*this, true); }
		operator const valType &() const
		{ return arrayRefTmpM=refType (*this, true); }
		bool resize()
		{
////fprintf(stderr,"array::resize %e %d %d %d %p %p %p\n", tM ? *tM : -1.0, int(ssM), int(osM), sM?int(*sM):-1, sM, &ssM, tM);
			if(*sM>osM)
			{
				if(ownerM)
					delete [] tM;
				else
					ownerM = true;
				tM = new T[*sM];
//ctl::log("array::resize!! %d %d %d %p %p %p\n", int(ssM), int(osM), sM?int(*sM):-1, sM, &ssM, tM);
				return true;
			}
			ssM = osM = *sM;
			return false;
		}
		template<class U> static ctl::oStream& write_(ctl::oStream &os, const U& u)
		{ return os<<u; }
		template<class U> static ctl::iStream& read_(ctl::iStream &is, U& u)
		{ return is>>u; }
			
	};
// does fortran have char** ?
//	template<typename T> class cstring: public array<T, ctl::cstring<T>, ctl::cstringRef<T> >;

	typedef array<char, ctl::cstring<char>, ctl::cstringRef<char> > string;
} //ctl
	template<class T> inline oStream &operator<<(oStream &os,const  Extern::array<T> &a)
		{
//if(os.writing()) fprintf(stderr,"array::write %e %d %d %d %p %p %p\n", tM ? *tM : -1.0, int(ssM), int(osM), sM?int(*sM):-1, sM, &ssM, tM);
		 	os<<(ctl::uint8)(a.ssM);
			if(a.tM)
				os.write(a.tM, a.ssM);
			else
				os.skip(a.ssM * sizeof(T) );
			return os;
		}
		template<class T> inline iStream &operator>>(iStream &is, Extern::array<T> &a)
		{
//ctl::log("array::read %d %d %d %p %p %p\n", tM ? *tM : -1.0, int(ssM), int(osM), sM?int(*sM):-1, sM, &ssM, tM);
			if(!a.sM)    //called via ctl::array::_cast()
				a.sM = &a.ssM;
			ctl::uint8 size; is>>size;
			if(!a.tM)
			{
				*a.sM = size;
				a.tM = new T[size];
			}
			ctl::uint8 min = (ctl::uint8)*a.sM < size ? *a.sM : size;
			is.read(a.tM,min);
//			std::cout<<"read::t[0] = "<< tM[0]<<"\n";
			is.skip(sizeof(T)*(size-min) );
//assert(sM == &ssM);
			*a.sM = size;
			a.ssM = a.osM = *a.sM;
//ctl::log("array::read!! %d %d %d %p %p %p\n", int(ssM), int(osM), sM?int(*sM):-1, sM, &ssM, tM);
			return is;
		}

	template<> inline oStream &operator<<(oStream &os, const Extern::string &a)
	{ 

			if(a.tM)
				os.write(a.tM, a.ssM-1);
			else
				os.skip((a.ssM-1) * sizeof(char), '?' );
			os << '\0';
			return os;
	}

	template<> inline iStream &operator>>(iStream &is, Extern::string &a)
	{
		if(!a.sM)    //called via ctl::array::_cast()
			a.sM = &a.ssM;
		size_t size;
		bool end = is.strTerm(size);
		if(!end)
			size++;
		if(!a.tM)
		{
			*a.sM = size;
			a.tM = new char[size];
		}
		ctl::uint8 min = (ctl::uint8)*a.sM < size ? *a.sM : size;
		is.read(a.tM,min);
		if(end)
			a.tM[size] = '\0';
		is.skip(sizeof(char)*(size-min) );
		*a.sM = size;
		a.ssM = a.osM = *a.sM;
		return is;		
	}


namespace Extern
{
	template<class T> struct type
	{
		typedef T resultT;
		typedef type<T> value;
		typedef T* pointer;
		T *tM;
		type(T *t=0, bool =false): tM(t) {}
		type(const value &v , bool =false): tM(v.tM) {}
		type<T> &operator=(const T &t)
		{ CTL_Assert(tM); *tM = t; return *this; }
		operator T&() { return *tM; }
		operator const T&() const { return *tM; }
	};
	template<> struct type<const string>
	{
		typedef char resultT;
		typedef char* pointer;
		typedef const string value;
	};
	template<> struct type<string>
	{
		typedef char resultT;
		typedef char* pointer;
		typedef string value;
	};
#ifdef CTL_PartialSpec

	template<class T> struct type<const array<T> >
	{
		typedef T resultT;
		typedef T* pointer;
		typedef const array<T>  value;
	};
	template<class T> struct type<array<T> >
	{
		typedef T resultT;
		typedef T* pointer;
		typedef array<T> value;
	};
#else
#define CTL_FortranSpec(T)\
	template<> struct type<T&>\
	{\
		typedef T resultT;\
		typedef T* pointer;\
		typedef type<T> value;\
	};\
	template<> struct type<array<T> >\
	{\
		typedef T resultT;\
		typedef T* pointer;\
		typedef array<T>  value;\
	};\
	template<> struct type<array<T> &>\
	{\
		typedef T resultT;\
		typedef T* pointer;\
		typedef array<T> value;\
	};

CTL_FortranSpec(char)
CTL_FortranSpec(int2)
CTL_FortranSpec(int4)
CTL_FortranSpec(int8)
CTL_FortranSpec(real4)
CTL_FortranSpec(real8)

#endif

	struct recvWrapBase: public ctl::shared
	{
		virtual ~recvWrapBase()
		{}
		virtual bool recv(ctl::int8)=0;
	};

	template<class Res, class Y> class recvWrap : public recvWrapBase
	{
		typedef Y value;
		Res resM;
		value yM;
	public:
		recvWrap(const Res &res, const value& y):resM(res), yM(y)
		{}
		bool recv(ctl::int8 flag)
		{
//			ctl::clog<<"receiver"<<flag<<"\n";
			while(!resM && (flag<0) );
			if(!!resM)
			{
				yM = resM.getValue();
				return true;
			}
			return false;
		}

	};
	typedef ctl::cPointer<recvWrapBase> resultEntry;


/*
 id = 0 ==> insert: ret=ref in map,				id=newId
 id > 0 ==> get    : ret=ref in map ? none,  id=id ? 0
 id < 0 ==> erase: ret= none                 		,id=-id ? 0
*/
template<class T> inline T &objMapper(ctl::int8 &id, const T &obj = T())
{
static std::map<ctl::int8, T> objMapS;
static ctl::int8 currentIdS = 3;
static T noneS;

	if(id==0)
		return (objMapS[id=currentIdS++] = obj);
	typename std::map<ctl::int8, T>::iterator h=objMapS.find(id);
	bool del=id<0;
	if(del)
		id = -id;	
	if(h!=objMapS.end() && !!h->second)
	{
		if(!del)
			return h->second;
		objMapS.erase(h);
	}
	else		
		id = 0;
	return noneS;
}

template<class CI> inline void create(const char* location, ctl::int8 *handle)
{
	ctl::log("#d ++ by fortran: %s\n", location);
	*handle = 0; ctl::link p;
	if(location && *location)
	{ p = ctl::link(location); }
	else
	{ p = ctl::getLink(CI::CTL_typeNameS()); }
	if(!p)
		ctl::log("#l !! by fortran: connecting %s failed\n", location);
	else
	{
		CI ci(p);
		if( !!ci )
		{
			ctl::Extern::objMapper<CI>(*handle, ci);
			return;
		}
		else
			ctl::log("#c !! by fortran: creation of %s failed\n", CI::CTL_typeNameS().c_str());
	}
	handle = 0;
}

/*
*id < 0 ==> wait for result id = -1
*id >=0 ==> if available id = -1 else id = new id
*/

template<class Y, class Res> inline void setResult(ctl::int8 *id, const Res &rRes, const Y &t)
{
//	ctl::clog<<"setResult?"<<*id<<"\n";
	if(*id<0)
	{
		while(!rRes);
		*id = -1;
	}
	if(*id==-1 || rRes.ready(false))
	{  *id = -1; recvWrap<Res,Y>(rRes, t).recv(1);  return;  }
//	ctl::clog<<"setResult??"<<*id<<"\n";
	objMapper<resultEntry>(*id,  new recvWrap<Res,Y>(rRes, t) );
//	ctl::clog<<"setResult"<<*id<<"\n";
}

inline void setResult(ctl::int8 *res, const ctl::resultType<void>::resultT &rRes)
{ if(!rRes.isEmpty()) ctl::Extern::setResult(res, rRes, ctl::empty() ); }

/*
*id < -2 ==> wait for result id = -1
*id >  2 ==> if available id = -1 else id = old id
*/
inline void recv(ctl::int8 *id)
{
	ctl::int8 resHandle = *id<0 ? - *id : *id;
	if(resHandle < 3)
			return;
	resultEntry r=objMapper<resultEntry>(resHandle);
//	ctl::clog<<"recv "<<resHandle<<" "<<*id<<"\n";
	if(!resHandle)
	{
		*id = -2;
		return;
	}
	if(r->recv(*id) )
	{
		*id = -1;
		ctl::int8 negId = -resHandle;
		objMapper<resultEntry>(negId);
	}
	else
		*id = resHandle;
}

} // Extern

	template<class T> inline bool sizeChanged(const Extern::array<T> &x)
	{ return ((Extern::array<T>&)x).resize(); }


// set binary Type of array
	template<class T, class Bin> inline char  congruence(const Bin*, const Extern::array<T>*)
	{ 	return Bin::arrayCheck((T*)0, (ctl::int2type<-1>*)0);	}

	template<class Bin> inline char  congruence(const Bin*, const Extern::string*)
	{ 	return Bin::stringCheck((char*)0);	}

	template<class T> inline oStream &write(oStream &os, const Extern::type<T> &s)
	{ return os<<*s.tM; }

	template<class T> inline iStream &read(iStream &is, Extern::type<T> &s)
	{ return is>>*s.tM; }
	template<class T> inline T &Fortran_conv(const Extern::type<T>& t)
	{ return (T&)(Extern::type<T>&)t; }
	template<class T> inline T &Fortran_conv(const Extern::type<const T>& t)
	{ return (T&)(const T&)t; }
	template<class T> inline ctl::array<T> Fortran_conv(const Extern::array<T> & t)
	{ return ctl::array<T>(t, true); }
	template<class T> inline ctl::arrayRef<T> Fortran_conv(Extern::array<T> & t)
	{ return ctl::arrayRef<T>(t, true); }
	inline ctl::cstring<char> Fortran_conv(const Extern::string & t)
	{ return ctl::cstring<char>(t, true); }
	inline ctl::cstringRef<char> Fortran_conv(Extern::string & t)
	{ return ctl::cstringRef<char>(t, true); }
} // ctl

#endif
