/******************************************************************************

 This file is part of the Component Template Library (CTL)

 Copyright (C) 2006  Rainer Niekamp
 email:              r.niekamp@tu-bs.de

 The CTL is free software; you can redistribute it and/or modify it under the
 terms of the GNU Lesser General Public License as published by the Free Software
 Foundation; either version 2.1 of the License, or (at your option) any later
 version. We only ask that proper credit be given to the author.

 This library is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 for more details.

 You should have received a copy of the GNU Lesser General Public License along
 with this library; if not, write to the Free Software Foundation,
 Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

 ******************************************************************************/

#ifndef CTL_Group_H_
#define CTL_Group_H_

#include "utl/vector_alg.h"
#include "remote.h"
#include "link.h"
#include "connect.h"

namespace ctl
{

	struct groupCI: public Remote
	{
		enum	{ extends = true };
		typedef groupCI thisT;
		typedef thisT CTL_Interface;
		typedef thisT CTL_RemoteBase;
		using Remote::getSharedP;
		typedef resultType<status>::resultT resultType;
		ioTag tagM;
		bool while1M;
		mutable int lockedM;
		static ctl::typeId CTL_typeNameS()
		{	return "groupCI";	}
		ctl::typeId fullTypeName() const
		{	return CTL_typeNameS();	}
		groupCI(bool while1 = false, bool env = true) :
			Remote(false/*unguarded*/, ctl::link(), env), tagM(RMI), while1M(while1), lockedM(0)
		{}
		typedef RMICall<status, ctl::field2<const ctl::vector<int4>, const int4>,
		    defaultCall> RMI0;
		virtual resultType getCtrl(const ctl::vector<int4> &c, int4) const;
		typedef RMICall<status, ctl::field1<const int4>, defaultCall> RMI1;
		virtual resultType terminate(int4) const;
		typedef RMICall<bool, ctl::field2<const PID, const ctl::string>,
		    defaultCall> RMI3;
		virtual result<bool> recvTermination(const PID &source,
		    const ctl::string &error);
		typedef RMICall<void, ctl::field1<const ctl::string>, defaultCall> RMI4;
		static result<void> recvException(const ctl::link &,
		    const ctl::string &error);
		virtual status receive(ctl::iStream &is);
	};

	typedef ctl::rPointer<groupCI> groupRP;

	struct processInfo
	{
		typedef lexicalAssoc<char, ctl::string> varT;
		PID pid, threadId;
		objectID instance;
		bool while1;
		varT variables;
		oStream &write(oStream &os) const
		{
			return os << pid << threadId << instance << while1 << variables;
		}
		iStream &read(iStream &is)
		{
			//		swap = is.getSwap();
			return is >> pid >> threadId >> instance >> while1 >> variables;
		}
		//	CTL_Type(ctl::processInfo, tupel, (pid, instance), 2);
		processInfo(const PID &p = PID(), const PID &tid = PID(),
		    const objectID &obj = 0, bool wh = true, const varT &var = varT()) :
			pid(p), threadId(tid), instance(obj), while1(wh), variables(var)
		{
		}
	};

	class groupImpl: public groupCI
	{
		friend class Environment;
		friend class link;
		friend class Communicator;

		ctl::string nameM;
		mutable ctl::vector<int4> ControlM, LocalM;

		mutable ctl::vector<processInfo> partnerM;
		CommunicatorSP communicatorM;
		mutable CommunicatorSP localCommM;
		mutable bool terminatedM, CtrlM, inTerminationM, destroyM, activM;
		bool checkTerminationM;
		typedef std::list<ctl::iStream> queueT;
		mutable queueT queueM;
		mutable ctl::string errorM;
		const link *parentM;
		channelT typeM;
		mutable int callCounterM;
		mutable messageForward* forwardM;

		struct tempInstance
		{
			mutable internal::rReference<groupCI> obj;
			mutable groupCI *grp;
			tempInstance(const processInfo &info, const CommunicatorSP &commu,
			    const PID &pid) :
				obj(info.instance, link(commu, pid), 0, "ctl::groupCI", PID(), PID()),
				    grp(0)
			{
			}
			tempInstance(const groupCI *g) :
				obj(0, ctl::link(), 0, "ctl::groupCI", PID(), PID()), grp((groupCI*) g)
			{
			}
			groupCI *operator->()
			{
				if(grp) return grp;
				return &obj;
			}
		};

		groupImpl(const groupImpl&);
		groupImpl &operator=(const groupImpl&);
		groupImpl() :
			terminatedM(0), CtrlM(false), forwardM(0)
		{
		}
		groupImpl(PID &, const PID &logId = PID(), const char *args = 0);
		groupImpl(const location &loc, channelT typ, const ctl::string &name);

		bool operator()(ctl::iStream &) const;
		resultType getCtrl(const ctl::vector<int4> &vec, int4 count) const
		{
			CtrlM = true;
			ControlM = vec;
			callCounterM = count;
			LocalM[myPID()]++;
			//		std::cout<<" getControl"<<int(myPID())<<"  "<< ControlM<<"  "<< LocalM<<"\n";
			return success;
		}
		resultType terminate(int4 count) const
		{
			ctl::log("#d ?? %d received termination (%d calls)", int(logId()), count);
			callCounterM = count;
			terminatedM = true;
			return success;
		}
		result<bool> recvTermination(const PID &source, const ctl::string &error)
		{
			ctl::log("#d ?? link %d of group \"%s \" terminated by \"%s\"\n",
			    ctl::int4(source), nameM.c_str(), error.c_str());
			checkTerminationM = false;
			//		communicatorM.destruct();
			terminatedM = true;
			destroyM = true;
			return true;
		}
		static void recvException(const ctl::string &error)
		{
			ctl::exception e(error);
			CTL_Throw(groupException(e.whatM));
		}
		void init(const location &loc, channelT type);
		void init(const ctl::vector<location> &loc, channelT type);
		bool checkParent() const;
		bool callNext() const;
		void queueMessage(int len, PID &src, int *type) const
		{
			iStream is(0, 0, src, communicatorM->getMyId());
			if(len > 0) is.setBuffer(len, 0);
			is.commuM = communicatorM;
			int rlen = communicatorM->read(is, src, *type);
			if(rlen < 0) return;
			CTL_Assert(!!src && (len<0 || len==rlen));
			//		assert(src.port<1000*1000 || !src.host.empty() );
			is.sourceM = src;
			//		ctl::iStream is(rlen,(char*)buf, src, communicatorM->getMyId());
			queueM.push_back(is);
		}
		void readMessage(bool blocking = false, const PID &src = PID(), ioTag tag =
		    RMI) const;
		void init();
		void init(int n, const char * const *arg, bool enableDmn);
		const ctl::vector<processInfo> &getPartner() const;

		status untilCompletion(bool = false) const;
		void check() const
		{
			getPartner();
		}
		//    static ioTag getDestTag(iStream &);
		public:
		void setForward(messageForward* forward) const
		{
			forwardM = forward;
		}
		void setError(const string &error) const
		{
			errorM = error;
		}
		using groupCI::getSharedP;
		static void connect();
		//! create a groupImpl by the sample of location 's and the channelT type
		groupImpl(const ctl::vector<location> &loc, channelT typ = tcp,
		    const ctl::string &name = "group");
		//! create a groupImpl by the command line arguments,
		CTL_Inline
		groupImpl(int n, const char * const *arg, bool enableDmn = true,
		    bool while1 = true, Environment *env = 0);

		status writeMessage(const ctl::oStream &os, const PID &pid, ioTag tag = RMI) const
		{
			const ctl::vector<processInfo> &partner = getPartner();
			if(!partner) return failed;

			if(checkTerminationM)
			{
				if(pid.isLogical())
					LocalM[pid]++;
				else
				{
					for(unsigned int i = 0; i < partner.size(); i++)
						if(partner[i].pid == pid)
						{
							LocalM[i]++;
							break;
						}
				}
			}
			activM = true;
			return communicatorM->write((oStream&) os, pid, tag) < 0 ? failed
			    : success;
		}
		groupImpl(const CommunicatorSP&);CTL_Inline
		~groupImpl();
		const ctl::string &getName() const
		{
			return nameM;
		}

		tempInstance getInstance(const PID &p) const;
		bool isValid() const
		{
			//		if(destroyM || !communicatorM)
			if(!communicatorM) return false;
			if(!communicatorM->isValid())
			{

				((CommunicatorSP&) communicatorM).destruct();
				deleteGroup(this);
				return false;
			}
			CTL_Try
			{
				getPartner();
			}
			CTL_CatchAll
			{
				return false;
			}
			return true;
		}
		bool closed() const
		{
			return destroyM || !isValid() || communicatorM->closed();
		}
		//! check whether the groupImpl is still valid
		bool operator!() const
		{
			return !isValid();
		}
		bool terminated() const
		{
			return terminatedM;
		}
		channelT getLinkage() const
		{
			return typeM;
		}
		PID getPhysId(const PID &logId) const
		{
			//		if(!logId.isLogical())
			if(!logId.isLogical() || typeM == ctl::ext) return logId;
			return getPartner()[logId].pid;
		}

		const CommunicatorSP &getCommunicator() const
		{
			return communicatorM;
		}
		//! run a distributed application until termination is determined
		CTL_Inline
		bool run() const;CTL_Inline
		bool run(int &read) const
		{
			callCounterM = 0;
			bool ok = run();
			read = callCounterM;
			return ok;
		}
		//
		bool runJob(int &n, bool blocking) const
		{
			if(!checkParent()) return true;
			readMessage(blocking);
			n = queueM.empty() ? 0 : 1;
			callNext();
			return closed();
		}
		bool empty() const
		{
			return queueM.empty();
		}
		void readRMIs(const PID & srcp = PID(), ioTag tag = RMI) const
		{
			readMessage(false, srcp, tag);
			while(!callNext())
				;
		}
		void broadCastError(const ctl::string &error) const;
		link getProcess(const PID &pid = -1, ioTag = RMI, bool owner = true) const;
		//! return the specified link of this process groupImpl
		link operator[](int p) const
		{
			return getProcess(p);
		}
		link operator[](const PID &pid) const;
		link getParent() const;
		int4 nProcs() const
		{
			return !communicatorM ? 0 : communicatorM->nProcs();
		}
		//! return the size of this groupImpl
		uint4 size() const
		{
			return nProcs();
		}
		PID myPID() const
		{
			return !communicatorM ? PID() : communicatorM->getLogId();
		}
		//! returns the logical Id of this instance of the link groupImpl
		int rank() const
		{
			return myPID();
		}
		int logId() const
		{
			return myPID();
		}
		// global operations
		template<class X> struct addAssign
		{
			void operator()(X &y, const X &x) const
			{
				y += x;
			}
		};
		template<class X> struct minAssign
		{
			void operator()(X &y, const X &x) const
			{
				if(x < y) y = x;
			}
		};
		template<class X> struct maxAssign
		{
			void operator()(X &y, const X &x) const
			{
				if(y < x) y = x;
			}
		};
		template<class X> struct mulAssign
		{
			void operator()(X &y, const X &x) const
			{
				y *= x;
			}
		};
		template<class X> struct AndAssign
		{
			void operator()(X &y, const X &x) const
			{
				y &= x;
			}
		};
		template<class X> struct OrAssign
		{
			void operator()(X &y, const X &x) const
			{
				y |= x;
			}
		};
		template<class X> struct XOrAssign
		{
			void operator()(X &y, const X &x) const
			{
				y ^= x;
			}
		};
		// broadcast verteilt Daten vom Proz. 0 auf alle anderen mittels Hypercube-topologie
		template<class X> inline X broadcast(const X &x) const
		{
			char rw = 0;
			int step = 0, partner;
			X y = x;
			while((partner = broad_partner(step, rw)) >= 0)
			{
				if(rw == 'r') getProcess(partner, OPER, false) >> y;
				if(rw == 'w') getProcess(partner, OPER, false) << y;
				step++;
			}
			return y;
		}

		//! blocking: compute the sum of t over all link in this groupImpl
		template<class X, class F> X merge(const X &xx, F f) const
		{
			char rw = 0;
			int step = 0, partner;
			X y = xx, x;
			while((partner = merge_partner(step, rw)) >= 0)
			{
				if(rw == 'r')
				{
					getProcess(partner, OPER, false) >> x;
					f(y, x);
				}
				if(rw == 'w') getProcess(partner, OPER, false) << y;
				step++;
			}
			y = broadcast(y);
			return y;
		}
		template<class X> X Sum(const X& x) const
		{
			return merge(x, addAssign<X> ());
		}
		template<class X> ctl::vector<X> sample(const X &x) const;
		int merge_partner(int step, char &ReadOrWrite) const;
		int broad_partner(int step, char &ReadOrWrite) const;
		static inline int ld2(int n)
		{
			int b = 1, ld = 0;
			while(b < n)
			{
				b <<= 1;
				ld++;
			}
			return ld;
		}

	};

	inline ctl::PID Communicator::getPhysId(const PID & pid) const
	{
		return getOwner()->getPhysId(pid);
	}

	inline void Communicator::check() const
	{
		if(!checkedM && !!getOwner())
		{
			checkedM = true;
			getOwner()->check();
		}
	}

	template<class X> inline ctl::vector<X> groupImpl::sample(const X &x) const
	{
		PID myPID = communicatorM->getLogId();
		uint4 nProcs = communicatorM->nProcs();
		ctl::vector<X> smple(nProcs);
		smple[myPID] = x;
		for(uint4 i = 0; i < nProcs; i++)
			if(i != myPID)
			{
				link p = getProcess(i, OPER, false);
				p << x;
				p >> smple[i];
			}
		return smple;
	}

	//! Implementation of a link group
	class group
	{
		friend class Environment;
		cPointer<groupImpl> grpM;
		public:
		void setError(const string &error) const
		{
			if(!!grpM) grpM->setError(error);
		}
		group(const cPointer<groupImpl> grp) :
			grpM(grp)
		{
		}
		group()
		{
		}
		template<class LocVec>
		group(const LocVec &locT, channelT typ)
		{
			ctl::vector<location> loc(locT.size());
			int z = 0;
			for(typename LocVec::const_iterator l = locT.begin(); l != locT.end(); ++l)
				loc[z++] = *l;
			grpM = new groupImpl(loc, typ);
			if(!grpM->isValid()) grpM.clear();
		}
		group(const ctl::vector<location> &loc, channelT typ = tcp,
		    const ctl::string &name = "group")
		{
			grpM = new groupImpl(loc, typ, name);
			if(!grpM->isValid()) grpM.clear();
		}
		//! create a groupImpl by the command line arguments
		group(int n, const char * const *arg, bool /*envInit*/= true, bool while1 =
		    false)
		//	group(int n, const char *const *arg)
		{
			grpM = new groupImpl(n, arg, true, while1);
			if(!grpM->isValid()) grpM.clear();
		}
		const ctl::string &getName() const
		{
			return grpM->getName();
		}
		//! check whether the group is still valid
		bool closed() const
		{
			return !grpM || grpM->closed();
		}
		bool operator!() const
		{
			return !grpM || !grpM->isValid();
		}
		//! run a distributed application until termination is determined
		bool run() const
		{
			return grpM->run();
		}
		bool run(int &read) const
		{
			return grpM->run(read);
		}
		//! return the specified link of this group
		link operator[](int p) const
		{
			return (*grpM)[p];
		}
		//! return the size of this group
		uint4 size() const
		{
			return !grpM ? 0 : grpM->nProcs();
		}
		//  ! returns the logical Id of this instance of the group
		int rank() const
		{
			return grpM->myPID();
		}
		int logId() const
		{
			return grpM->myPID();
		}
		// global operations
		// broadcast verteilt Daten vom Proz. 0 auf alle anderen mittels Hypercube-topologie
		template<class X> inline X broadcast(const X &x) const
		{
			return grpM->broadcast(x);
		}

		//! blocking: compute the sum of t over all links in this group
		template<class X, class F> X merge(const X &x, F f) const
		{
			return grpM->merge(x, f);
		}
		template<class X> X Sum(const X& x) const
		{
			return merge(x, groupImpl::addAssign<X>());
		}
		template<class X> X operator+(const X &x) const
		{
			return Sum(x);
		}
		template<class X> X Product(const X &t) const
		{
			return merge(t, groupImpl::mulAssign<X>());
		}
		template<class X> X operator*(const X &x) const
		{
			return Product(x);
		}
		template<class X> X Min(const X &t) const
		{
			return merge(t, groupImpl::minAssign<X>());
		}
		template<class X> X operator<(const X &x) const
		{
			return Min(x);
		}
		template<class X> X Max(const X &t) const
		{
			return merge(t, groupImpl::maxAssign<X>());
		}
		template<class X> X operator>(const X &x) const
		{
			return Max(x);
		}
		template<class X> X Or(const X &t) const
		{
			return merge(t, groupImpl::OrAssign<X>());
		}
		template<class X> X operator|(const X &x) const
		{
			return Or(x);
		}
		template<class X> X XOr(const X &t) const
		{
			return merge(t, groupImpl::XOrAssign<X>());
		}
		template<class X> X operator^(const X &x) const
		{
			return XOr(x);
		}
		template<class X> X And(const X &t) const
		{
			return merge(t, groupImpl::AndAssign<X>());
		}
		template<class X> X operator&(const X &x) const
		{
			return And(x);
		}
		template<class X> ctl::vector<X> sample(const X &x) const
		{
			return grpM->sample(x);
		}
	};

	template<class T> inline shared *getSharedP(const T *t)
	{
		return t->getSharedP();
	}
} // ctl


#endif

