/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_COMM_CONN_INL_
#define CTL_COMM_CONN_INL_

#include "comm_conn.h"
#include "ctl/ci/connection.ci"

namespace ctl
{

  	CTL_Inline bool CommunicatorSingle::readAll(const PID&, bool blocking, bool first) const
	{
		bool read1 = false;
//		int usec = blocking && !first ? 1 : 0;
		for(;;)
		{
   			internal::ioBuffer m;
			if(!connectM->test(0) || !connectM->readHd(m))
				return read1;
			read1 = true;		

			m.setNo(mCounterM++);
			m.pid = 1-getLogId();
			messageM.insert(m);
			if(!blocking)
				break;
			return true;
		}	
		return read1;
	}	

   CTL_Inline int CommunicatorSingle::write_(oStream &os, const PID & dest ,int type) const
   {
		int size = os.size();
		if(!size)	
		size = -1;
     internal::ioBuffer m(size,type,getLogId(),getLogId(), os);
     if(dest==getLogId() )
     {
			m.setNo(mCounterM++);
			messageM.insert(m);
			return size;
     }

   	if(!connectM->writeHd(m))
		{
			ctl::log(envM, "#d !! fatal write error\n");
			return -1;
		}
 		ctl::log(envM, "#d << cnct dest=(%d) size=%d tag=%d \n",dest.port,size,type);
    return size;
   }

   CTL_Inline int CommunicatorSingle::test_(PID &source, int *type) const
   {
   		std::set<internal::ioBuffer>::iterator mI=find(source, *type);
		if(source!=getLogId() && mI==messageM.end() )
			{
   				readAll(source, false, true);
				mI=find(source, *type);
			}
   		if(mI!=messageM.end() )
   		{
   			source = mI->pid;
   			*type = mI->tag;
   			return int(mI->size);
   		}
   		return -1;
   }

   CTL_Inline int CommunicatorSingle::read_(iStream &is, PID &source, int *type) const
   {
   		internal::ioBuffer m=extract(source, *type);
   		bool first = true;
		while(!m)
		{
			if(readAll(source, true, first))
				m=extract(source, *type);
			else
 				if(!isValid())
 				{	ctl::log(envM, "#d !! tcp::read failed\n"); return -1; }
         first = false;
		}
		source = m.pid;
// otherwise problem with fast return
		*type = m.tag;
		int size = int(m.size);
		is.set(m.buffer);
		if(m.swap)
			is.setSwap();
	   ctl::log(envM, "#d >> cnct: src=%d size=%d tag=%d\n",int(source),size,*type);
		return int(m.size);		
   }

class rioBase: public ioBase
{
	 connectionCI connectM;
public:                 	
	rioBase(const   connectionCI &c):  connectM(c) {}
	bool write(void *buf, int size) const
	{ return true; }
	size_t read(void *buf, int size) const
	{ return 0; }
	bool write(int8 size, const vector<sys::iovec> &p) const
	{ return 0; }	
	bool operator!() const
	{ return false; }
	bool close() const
	{ return true; }
	bool test(int usec) const
	{ return false; }
};

   CTL_Inline CommunicatorImpl<cnct>::CommunicatorImpl(int n, const char * const *arg):
			CommunicatorSingle(new rioBase(connectionCI(location(n, arg).getSSH() , 1, 2) ), false)
   { set(1, 2); 	}

   CTL_Inline CommunicatorImpl<cnct>::CommunicatorImpl(const location &loc):
		CommunicatorSingle(new rioBase(connectionCI(loc.getSSH() ) ), false)						
	{ set(0, 2); }

	CTL_Inline CommunicatorImpl<cnct>::~CommunicatorImpl()
	{
	}

} // ctl

#endif





