/******************************************************************************

This file is part of the Component Template Library (CTL)

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
for more details.

You should have received a copy of the GNU Lesser General Public License along
with this library; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_COMMU_EXT_H_
#define CTL_COMMU_EXT_H_
#include <stdlib.h>
#include "ctl/utl/timer.h"
#include "ctl/ci/commu.ci"
#include "ctl/inl/communicator.inl"
#include "ctl/com/dmn.h"

namespace ctl
{

class CommunCI_Impl: public messageForward
{
	link pM, clientM;
	groupSP grpM;
public:
	~CommunCI_Impl()
	{ 
		ctl::log("#m ~CommunCI_Impl\n");
		pM.setForward();
	}
	void forward() const
	{
		int tag = -1;
		PID src=1;
//ctl::log("forwarding\n");
		if(pM.getCommunicator()->test_(src, &tag)>=0)
		{
			iStream is;
			pM.getCommunicator()->read_(is, src, &tag);
			ctl::log("#m commCI << size = %d tag = %d\n", (int)is.size(), tag);
			oStream os(is.size(), is.buffer() );
			clientM.writeMessage(os, tag); 
		}
	}
  void write(const ctl::vector<char>& b, const PID& /*pid*/, int type) const
	{ 
		oStream os(b.size(), (void*)b.begin()) ;
		pM.getCommunicator()->write_(os, pM.getPid(), type); 
	}
	CommunCI_Impl (const ctl::vector<ctl::string>& locS=ctl::vector<ctl::string>() )
	{ init(locS[0]); }
	CommunCI_Impl (const ctl::string& locS)
	{ init(locS); }
	void init(const ctl::string& locS)
	{
		clientM = ctl::getClient();
		location loc (locS);
		pM = link(loc);
		if(loc.linkage==thread)
				grpM = new groupImpl(pM.getCommunicator());

		if(!pM)
			ctl::throwStr("Could not create communicator<%s> \n",   location::channel2string(loc.linkage));
		pM.setForward(this);
	}
};

	template<> class CommunicatorImpl<ext> : public Communicator
	{
			commuCI commu;
		public:
	 		PID getConnection(bool /*blocking*/ = true) const
			{ 
				if(commu.CTL_getLink().getCommunicator()->getType()==dmn)
					return commu.CTL_getLink().getCommunicator()->getLastConnect(); 
				return 1; 
			}
			int write_ ( oStream &os,const PID &dest ,int type ) const
			{
//				ctl::log("#m CommunicatorImpl<ext>::write_");
				vector<char> buf ( os.size(), os.buffer() );
				commu.write ( buf, dest, type );
				return os.size();
			}
			int test_ ( PID &source, int *type ) const
			{	return commu.CTL_getLink().getCommunicator()->test_(source, type ); }
			int read_ ( iStream &is, PID &source, int *type ) const
			{
//				ctl::log("#m CommunicatorImpl<ext>::read_");
				return commu.CTL_getLink().getCommunicator()->read_(is, source, type ); 
			}
			channelT getType() const
			{ return ext; }
			CommunicatorImpl (const link& p, const ctl::vector<ctl::string>& locS ) :Communicator ( false )
			{	init (p, location(locS[0]) ); }
			CommunicatorImpl (const link& p, const location &loc) : Communicator ( false )
			{ init (p, loc); }
			void init(const link& p, const location &loc )
			{
				ctl::log("#d ++ CommunicatorImpl<ext>('%s')\n", loc.c_str());
				if(!p)
				{	set(-1,0); return; }
				commu = commuCI (p, ctl::cstring<char>(loc.c_str()) );
				if(!commu)
				{	set(-1,0); return; }
				ctl::string message = loc.pathExec(); message += "@"; message += loc.host;
				set ( 0, 2, PID ( 0 ), ctl::string ( "-- ext -> " ) +message );
			}
			~CommunicatorImpl()
			{ ctl::log ( envM, "#d -- Communicator<ext>\n" ); }
	};

	CTL_Inline CommunicatorSP Communicator::newExternal(const location &loc)
	{ 
		CommunicatorSP commu;
//		ctl::log("#d newExternalCommunicator:loc=%s + '%s'\n", loc.c_str(), loc.pipeTail.c_str() );
//		if ( loc.pipeTail.empty() )
//			return CommunicatorSP();
		location cpy(loc);
		link* p=0;
		channelT typ=ctl::undefined;
		ctl::string tailStr=loc.pipeTail;
		
		if(loc.linkage==dmn && !loc.pipeTail.empty() )
		{
			location tail(tailStr);
			tailStr=tail.pipeTail;
			tail.pipeTail.clear();
			tail.sshString.clear();
			cpy.pipeTail = tail.str();
			cpy.sshString.clear();
			typ = keepAsIs;
		}
		else
		{
			cpy.pipeTail.clear();
		}
		cpy.pipeType= 0;
		p= new link(cpy, typ);
		if(!*p)
		{ delete p; 	return CommunicatorSP(); }

		if(loc.pipeType=='<')
		{
// get connect from end of '<'
			groupSP dGrp=env().getDmnGroup();
//			ctl::rBinResult<ctl::cstring<char> > id=ctlCI::getLANid(*p, dGrp->getCommunicator()->getPhysId().toString(), tailStr);
			ctl::result<ctl::string> id=ctlCI::getLANid(*p, dGrp->getCommunicator()->getPhysId().toString(), tailStr);
			iStream is;
			int tag=2;
			PID pid;
			dGrp->getCommunicator()->read(is, pid, tag);
			ctl::log("#d %s => ctlCI::getLANid(%s, %s)'\n", ((ctl::string)id).c_str(), tailStr.c_str(), dGrp->myPID().toString().c_str());
			ctl::string _tail=id;
			ctl::log("tail='%s' '%s'\n",tailStr.c_str(), _tail.c_str() );
			if(!tailStr)
				commu = dGrp->getCommunicator();
			else
			{
				link p2=link(dGrp->getCommunicator(), pid, RMI);
				commu =  CommunicatorSP ( new CommunicatorImpl<ext> (p2, location(tailStr)) );
			}
		}
		else if(loc.pipeType=='>')
		{
			groupSP dGrp=env().getDmnGroup();
// create dmn to end of '>'
			ctl::string id;
			ctlCI::getLANid(*p, ctl::string(), tailStr).assign(id);
			location idLoc(id);
			PID pid(idLoc.portL, idLoc.host);
			ctl::string error;
			dGrp->getCommunicator()->setLastConnect(pid);
			if(!tailStr)
			{	commu = dGrp->getCommunicator(); commu->ownerLinkM = p; return commu; }
			location tailLoc(tailStr);
//			if(tailLoc.linkage==lib || tailLoc.linkage==thread)
/*
			if(false)// && tailLoc.linkage==lib || tailLoc.linkage==thread)
			{			
				const CommunicatorImpl<dmn>* DCom = dynamic_cast<const CommunicatorImpl<dmn>*>(dGrp->getCommunicator().get());
				int port = DCom->connect(pid, tailStr.c_str(), error);
				if(!!error || port<0)
				{
					ctl::log("#m !! %s\n", error.c_str()+1 );
					delete p;
					return commu;
				}
				pid.port = port;
				DCom->setLastConnect(pid);
				commu = dGrp->getCommunicator();
			}
			else
*/
			{
				link p2=link(dGrp->getCommunicator(), pid, RMI);
				commu = CommunicatorSP ( new CommunicatorImpl<ext> (p2, location(tailStr)) );
			}
		}
		else
		{ 
//ctl::log("#d creating CommunicatorImpl<ext>('%s')\n", tailStr.c_str());
			commu = CommunicatorSP ( new CommunicatorImpl<ext> (*p, location(tailStr) ) );
//			commu = CommunicatorSP ( new CommunicatorImpl<ext> (*p,"| " + tailStr) );
		}
		commu->ownerLinkM = p;
		return commu;
	 }

}


#endif
