/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_ENVIRONMENT_INL_
#define CTL_ENVIRONMENT_INL_

#include <time.h>
#include <ctl.h>
#include "environment.h"
#include "communicator.inl"
#include "link.inl"
#include "group.inl"
#include "remote.inl"
#include "utl.inl"

#include <signal.h>

namespace ctl
{


	CTL_Inline const char * Environment::sigName(int sig)
	{
		switch(sig)
		{
			case SIGINT: return "SIGINT";
			case SIGILL:  return "SIGILL";
			case SIGABRT:  return "SIGABRT";
			case SIGFPE:  return "SIGFPE";
			case SIGTERM:  return "SIGTERM";
#ifndef CTL_Win32
			case SIGCONT: return "SIGCONT";
			case SIGPIPE:  return "SIGPIPE";
			case SIGSYS:  return "SIGSYS";
#endif
			case SIGSEGV:  return "SIGSEGV";
			default:  return "SIGUNKNOWN";
		}
	}
//	CTL_Inline void conthandle(int sig)
//	{ fprintf(stderr, " got signal SIGCONT %d\n",sys::getsTid() ); }
	CTL_Inline void Environment::setSignals()
	{
#ifdef CTL_CatchSignals
#	ifndef CTL_Win32
			sigHandlesM.push_back(signalEntryT(SIGPIPE, ::signal (SIGPIPE, SIG_IGN) ) );
			sigHandlesM.push_back(signalEntryT(SIGCONT, ::signal (SIGCONT, SIG_IGN) ) );
			sigHandlesM.push_back(signalEntryT(SIGABRT, ::signal (SIGABRT, sighandle) ) );
			sigHandlesM.push_back(signalEntryT(SIGSYS, ::signal (SIGSYS, sighandle) ) );
			sigHandlesM.push_back(signalEntryT(SIGINT, ::signal (SIGINT, sighandle) ) );
//			sigHandlesM.push_back(signalEntryT(SIGCHLD, ::signal (SIGCHLD, SIG_IGN ) ) );
//			::signal (SIGCONT, (sigHandle)&conthandle);
#	endif
			sigHandlesM.push_back(signalEntryT(SIGINT, ::signal (SIGINT, sighandle) ) );
			sigHandlesM.push_back(signalEntryT(SIGILL, ::signal (SIGILL, sighandle) ) );
			sigHandlesM.push_back(signalEntryT(SIGFPE, ::signal (SIGFPE, sighandle) ) );
			sigHandlesM.push_back(signalEntryT(SIGSEGV, ::signal (SIGSEGV, sighandle) ) );
			sigHandlesM.push_back(signalEntryT(SIGTERM, ::signal (SIGTERM, sighandle) ) );
#else
// catch at least SIGINT
//	sigHandlesM.push_back(signalEntryT(SIGINT, ::signal (SIGINT, sighandle) ) );
#endif
	}
	void Environment::resetSignals()
	{
		for(size_t i=0; i<sigHandlesM.size(); i++)
			::signal(sigHandlesM[i].first, SIG_DFL);
//			::signal(sigHandlesM[i].first, sigHandlesM[i].second);
	}

	CTL_Inline void Environment::sighandle(int sig)
	{
		fprintf(stderr,"  *** got signal %s(%d) pid = %d ***\n", sigName(sig), sig, getpid() ); fflush(stderr);
		static bool first=true;
		if(first)
		{
			first = false;
			char envname[24];
//			threadIdT tid = 0;
			const char *envStr = global::getName(envname);
			if(envStr)
			{
				Environment &e=env();
				e.setError(ctl::string("  *** got signal ")  + sigName(sig) + " ***");
				fprintf(e.stderrM,"  *** in %s: got signal %s(%d) ***\n", e.cmdArgM[0].c_str(), sigName(sig), sig ); fflush(e.stderrM);
			}
			else
				fprintf(stderr,"  *** got signal %s(%d) ***\n", sigName(sig), sig ); fflush(stderr);
		}
		exit(1);
	}

	inline typeId extractScope(const typeId &fname)
	{
		const char *s=fname.c_str();//, *d=strrchr(s, ':');
		const char *bra = strchr(s,'('), *blank = strchr(s,' '), *dot = 0;
		if(!bra)
			return fname;
// ***** constructor *****
		if(!blank || blank>bra) 
			return fname.substr(0, bra-s); 
// ***** scoped function *****
		typeId scope = fname.substr(blank-s, bra-s);
		dot = strrchr(scope.c_str(),':');
	//	CTL_Assert(bra);
		return scope.substr(0,dot-scope.c_str()-1 );
	}

		CTL_Inline RemoteSP Environment::getFunction(const char *fname)
		{
//log(this,"#d getFunction '%s'\n", fname );
			RemoteSP f;
// ask parent
			if(parentEnvM && !!(f=parentEnvM->getTmpFunction(fname)))
			{
//log(this,"#d getFunction '%s' %p\n", fname, f.get() );
				f->envM = this;
				return f;
			}
// look in own symbol table 
			symbolT const *fp = symbolTableM.find(fname);
			if(fp)
				return *fp;
// 	look for link with scope name
			typeId scope = extractScope(fname);
//log(this,"#d getFunction scope = '%s'\n", scope.c_str() );
			const ctl::link *p=linkMapM.find(scope.c_str());
			ctl::link P;
			if(!p)
			{
//   try to get link from some where
				P = getLink(scope, ctl::any(), 0, false);
//		!P ==> give up
				if(!P)
					return RemoteSP();
				log(this,"#d ?? link added for %s\n",fname);
				linkMapM[scope.c_str()] = P;
				p = &P;
			}
			f = p->getType()!=serial ? p->getFunction(fname) : RemoteSP();
			if(!!f)
				return f;

			return RemoteSP();
		}

	CTL_Inline void Environment::setLink(const typeId &scope, const link &p)
	{
//log(this,"#d setLink '%s'\n", scope.c_str() );
		if(!scope)
			tmpLinkM = !p ? 0 : (link*)&p;
		else
			linkMapM[scope.c_str()] = p;
	}

	CTL_Inline ctl::string Environment::getLocationModifier(const typeId &scope) const
	{
		ctl::string s=scope;
		typeId found, search = "CTL::";
 // look for all containing namespaces 
		for(;;)
		{
			const ctl::string *str=locationModifierMapM.find(s.c_str());
			if(str)
				return *str;
			if(!found)
				found = getVariable((search+s).c_str() );
			int pos= s.rfind(':');
			if(pos<0)
				break;
			s = s.substr(0, pos-1);
		}		
		return ctl::string();		
	}

	CTL_Inline const link* Environment::getLocalLink(const typeId &scope)
	{
		ctl::string s=scope;
		typeId search = "CTL::";
		for(;;)
		{
			const ctl::link *p=linkMapM.find(s.c_str());
			if(p)
				return p;
			int pos= s.rfind(':');
			if(pos<0)
				break;
			s = s.substr(0, pos-1);
		}		
		return 0;		
	}
	CTL_Inline typeId getLocation(const typeId& scope)
	{ return env().getLocation(scope); }
	
	CTL_Inline typeId Environment::getLocation(const typeId &scope, const variableSpaceT& space)
	{
		ctl::string s=scope;
		typeId found;
		typeId search = "CTL::";
		for(;;)
		{
			if(!!(found = space.getVariable((search+s).c_str() ) ) )
				return found;
			int pos= s.rfind(':');
			if(pos<0)
				break;
			s = s.substr(0, pos-1);
		}		
		return found;		
	}

	CTL_Inline link Environment::getLink(const typeId &scope, const ctl::any &property, const char *locator, bool askParent)
	{
		link p;
// scope is needed
		if(scope.empty())
			return p;	
		typeId found;
		if(!property)
		{ 
			const link *pp=getLocalLink(scope);
			if(pp)
				return *pp;
			found = getLocation(scope, envSpaceM); 
		}
		if(!found && locator && *locator==0)
		{ /*printf("not found\n");*/ 	return ctl::link(); }		
		if(!found && !locationSelectorM)
		{
			if(parentEnvM && askParent)
				return parentEnvM->getLink(scope, property);
			if(!locator || !*locator)
				locator = sys::getLocatorName();
			if(setSelector(locator))
				setLocator();
//			if(!setSelector(locator) || !setLocator() )
//				return p;

		}
    if(!found && !!locationSelectorM)
      found = (*locationSelectorM)(scope, property);

		if(!found)
			found = getLocation(scope, _envSpaceM);
		if(!!found)
		{
			ctl::string modifier = getLocationModifier(scope);
			found += modifier;
			location loc(found);
			if(!!loc)
				return link(loc);
		}
		return p;
	}

	CTL_Inline const link &Environment::getParent(const PID &/*logId*/) const
	{
		if(!parentM)
		{
			if(parentStrM=="client" || parentStrM=="ctl_main")
				return  parentM;
			if(parentStrM.empty())
				parentStrM = "dmn";

			location loc;
			loc.setArgs(parentStrM);
      log("#d ?? parent string = %s\n",  parentStrM.c_str());
	   	if(loc.linkage == dmn || loc.groupSize<0 )
			{
				log(this,"#lm ?? connect to parent @ %d %s\n",loc.portL,loc.host.c_str());
   				parentM = link(PID(loc.portL, loc.host),loc.logId);
			}
			else
			{
				const ctl::vector<ctl::string> &cmdLine = getCommandLine();
				const char *const arg[2] = { cmdLine[0].c_str(),  parentStrM.c_str() };

				log(this,"#l ?? creating link %s %s\n", arg[0], arg[1]);
				parentM = link(2, arg);
			}
		}
		return parentM;
	}

	CTL_Inline bool Environment::initSocket()
	{
#ifdef CTL_Win32
		static bool first=true;
		if(first)
		{
			first = false;
			WSADATA info;
			return !WSAStartup(MAKEWORD(1,1), &info);
		}
#endif
		return true;
	}

	CTL_Inline Environment::Environment(): inDestrM(false),locationM("undef","undef","undef"), 
																					parentEnvM(0), childEnvM(0),rootEnvM(0), threadEnvM(0), 
						threadM(false),pclientM(&clientM), lockM(0),connectorM(0),tmpLinkM(0), verboseM(false),
						ioDeviceM(0), fdinM(0), fdoutM(1), fderrM(2), fdlogM(-1),
						stdinM(0), stdoutM(0), stderrM(0),
						loggerM(0),tidM(0),logThemesM(0),monitorThemesM(0),mpiCountM(0), runModeM(0),pSocketTmpM(0),pVarsTmpM(0), ownedByParentEnvM(false),
						inRunJobM(false)
	{

		rootEnvM = threadEnvM = this;
		const char *arg=sys::getEnv("CTL_Parent");
		if(arg)
			setParentStr(arg);
		else
			setParentStr("client");
		initSocket();
		tidM = sys::getTid();
	}
	CTL_Inline ctl::vector<ctl::string> getSymbols()
	{ return env().getSymbols(); }

CTL_Inline status evaluate(iStream& is, link& client);
	CTL_Inline void evaluateMessage(const ctl::vector<char>& buf)
	{
//		ctl::log("#d evaluate(%d) tag = %d", buf.size(), tag );
		iStream is(buf);
		
		ctl::link p = getClient(); 
//CTL_Assert(!!p.getPid());
		is.commuM = p.getCommunicator();
		is.sourceM = 0;
		evaluate(is, p ); 
	}
	struct commConnectDetail
	{
		CTL_Constructor(1, (const ctl::vector<ctl::string>), 1 ); 
		CTL_Constructor(2, (const ctl::string), 1 ); 
		CTL_StaticMethod(1, void, ctl::evaluateMessage, (const ctl::vector<char>&), 1);
//		CTL_Method(1, bool, CommunCI_Impl::write, (const ctl::vector<char>&, const PID &,int) const, 3);
//		CTL_Method(3, bool, CommunCI_Impl::read, (ctl::vector<char>&, PID &,int) const, 3);
	};

	CTL_Inline ctl::location Environment::init()
	{
		static const char*  arg[4] = {"ctl_main","-l serial", "-f .", 0};
		location loc= init(0, 3, arg);
#ifdef CTL_MPI
// called by mpirun?
		if(sys::getEnv("LAMWORLD"))
		{
			ctl::log("#d ++ MPI (<== $LAMWORLD set)\n");
			static const char*  arg[4] = {"ctl_main","-l mpi", "-f .", 0};
			mainGroupM = new groupImpl(3, arg);
			if(!(*mainGroupM))
				mainGroupM.clear();
		}
#endif
		return loc;
	}
	// ****** Environment::init: initialize new Environment ******
	CTL_Inline ctl::location Environment::init(Environment* parentEnv, int n, const char * const *arg)
	{
		cmdArgM.resize(n);
		for(int ii=0; ii<n; ii++)
		{
			cmdArgM[ii] = arg[ii];
//			std::cerr<<arg[ii]<<"??\n";
		}

		
		ctl::location locArgs(n, arg, ctl::keepAsIs);

		if(locArgs.linkage == ctl::undefined)
		{ locArgs.linkage = dmn; }
//		locArgs.print(std::cerr);
		int logId =  Communicator::getLogIdS(n, arg, locArgs);
		(void)logId;
		if(!!locArgs.parent)
			setParentStr(locArgs.parent);
		verboseM = locArgs.verbose;

		if(locArgs.priority)
			sys::setPriority(locArgs.priority);

// ****** get working directory ******
		char localPath[CTL_StringBufferSize];
		sys::getcwd(localPath, CTL_StringBufferSize-1);
		if(!locArgs.path)
			locArgs.path = localPath;
		ctl::string path=sys::absolutePath(localPath, locArgs.path);
		if(!locArgs.wd.empty())
			path = sys::absolutePath(path, locArgs.wd);
		bool err = false;

// ****** locArgs.linkage later changed ****** 
		if(locArgs.linkage!=lib && locArgs.linkage!=thread)
		{
			setSignals(); 	
			if(!path.empty())
				err = ::chdir(path.c_str()) != 0;
		}
		else
			setThread();
		locArgs.path = pathM = ( (err || path.empty())  ? ctl::string(localPath) : path );

		ctl::string home = sys::getEnv("HOME");
		ctl::string cwd1="ctl.env";
		ctl::string cwd2=".ctl.env";
		bool homeErr=false, cwdErr=false;
		ctl::string envFile=locArgs.envFile, readErr;
#ifdef CTL_WinPath
		ctl::string pathS = pathM + "\\";		
#else
		ctl::string pathS = pathM + "/";		
#endif

		if(!home.empty() )
			home += "/.ctl.env";
		else
			homeErr = true;

//		ctl::cout<<locArgs<<"\n";
//		ctl::cout<<locArgs.term<<" <-- term\n";
		cwdErr=readVariables(cwd1.c_str(), readErr, CTL_MaxIncludeDepth, false);

		if(!cwdErr && !envFile)
			envFile = pathS + cwd1;
		if(cwdErr)
			cwdErr = readVariables(cwd2.c_str(), readErr, CTL_MaxIncludeDepth, false);
		if(!cwdErr && !envFile)
			envFile = pathS + cwd2;
		if(cwdErr && !!home)
			homeErr=readVariables(home.c_str(), readErr, CTL_MaxIncludeDepth, false);
		if(!homeErr && !envFile && !!home)
			envFile = home;

//***** overwrite variables given by parent/location *****
		setParentEnv(parentEnv);
		if(parentEnvM)
		{
			if(parentEnvM->pVarsTmpM)
			{
				ctl::string err;
				for(const char *c=parentEnvM->pVarsTmpM; c; c=readVariable(c, "parent environ", err,  0) )
					std::cerr<<c<<"???????\n";
				exportSpaceM.variables = envSpaceM.variables;
				parentEnvM->pVarsTmpM = 0;
			}
			else
				setExportVariables(parentEnvM->getExportVariables());
		}
		if(!!locArgs.varStr)
			for(const char *c=locArgs.varStr.c_str(); c; c=readVariable(c, "command args", readErr, 0) );
		if(!!locArgs.parent)
		{
			location parent(locArgs.parent);
			if(!!parent.varStr)
				for(const char *c=parent.varStr.c_str(); c; c=readVariable(c, "command args", readErr, 0) );
			if(!locArgs.log)
				 locArgs.log = parent.log;
		}
		log("#x ++ variables from ctl.env '%s'\n", envSpaceM.getVariableList().c_str() );
		log("#x ++ variables exported ctl.env '%s'\n", exportSpaceM.getVariableList().c_str() );
		log("#x ++ variables from .ctl.env '%s'\n", _envSpaceM.getVariableList().c_str() );

//***** set locationM *****
		string name=sys::getHostIP();
		locationM = location(locArgs.exec, pathM, name);
		locationM.linkage = locArgs.linkage;
		locArgs.host = name;

//***** invoke connected main *****
		runModeM = locArgs.runMode;
		if(runModeM=='e')
			argStringM = locArgs.argString;

//***** ioDevices *****
		char devType = !!locArgs.term  ? 'x' : 'n';

//		char devType = !!locArgs.term && locArgs.notByTerminal() ? 'x' : 'n';
//locArgs.print(std::cout);
		ioDeviceM = sys::createDev(devType, locArgs, fdinM, fdoutM);
//		fderrM = fdoutM==1 ? 2 : fdoutM;
		stdinM = fdinM==0 ? stdin : ::fdopen(fdinM,"r");
		stdoutM = fdoutM==1 ? stdout : ::fdopen(fdoutM,"w");
		stderrM = fdoutM==1 ? stderr : stdoutM;

#ifndef CTL_NoLog
// ***** logging *****
		int logThemes = 0;
		const char *theme=0;
		if(	(theme = getVariable("CTL_LogThemes")) )
			(void)themes::getThemes(theme, logThemes);
		if(!!locArgs.log)
		{
			if(locArgs.log==".")
			{
				ctl::string logger = getVariable("CTL_Logger");
				locArgs.log = !logger ? ctl::string("ctl.log") : logger;
			}
			char logFile[CTL_StringBufferSize];         		
			logFile[CTL_StringBufferSize-1] = 0;
			ctl::string logPath=sys::absolutePath(pathM, locArgs.log);
			strncpy(logFile, logPath.c_str(), CTL_StringBufferSize-1);

			char *themeLoc = strrchr(logFile,':');			
			if(themeLoc && themeLoc>logFile+1) // Check for C:<PATH> in DOS
			{	*themeLoc++ = 0; (void)themes::getThemes(themeLoc, logThemesM); }
			else if(logThemes)
				logThemesM = logThemes;
			else
				(void)themes::getThemes("adx", logThemesM); 
			
			if(strncmp(locArgs.log.c_str(),"stderr", 6)==0)
				{	loggerM = stderrM; fdlogM = ::fileno(loggerM); }
			else if(strncmp(locArgs.log.c_str(),"stdout", 6)==0)
				{	loggerM = stdoutM; fdlogM = ::fileno(loggerM); }
			else if(!loggerM)
			{
				char *dot=strchr(logFile, '%');
				char out[CTL_StringBufferSize+10];
				if(logId >=0 && dot)
				{
					*dot = 0;
					sprintf(out,"%s%d%s", logFile, logId, dot+1);
				}
				else
					sprintf(out,"%s", logFile);
				loggerM = fopen(out,"w");
				if(loggerM)
					fdlogM = ::fileno(loggerM);
			}
//			fprintf(stderrM, "openLog2:out[%d] (lev=%d)='%s'\n",fdlogM, logThemesM, locArgs.log.c_str());
		}

//***** get time *****
		time_t t = ::time(0);
		char *timeStr= ::ctime(&t);
		char* eol=strchr(timeStr,'\n');
		if(eol)
			*eol='\0';


		ctl::string str;	
		char buf[CTL_StringBufferSize];
		if(locArgs.exec.size() + 80 > CTL_StringBufferSize ||
						locationM.path.size() + 80 > CTL_StringBufferSize )
			ctl::throwStr("string overflow in Communicator<dmn>");

		sprintf(buf,"#a********  ctl-log file: written by %s (PID %d)  ******** \n", locArgs.exec.c_str(), sys::getPid());
		str += buf;
		sprintf(buf,"********  at %s  ********\n", timeStr);
		str += buf;
		str += "********  called with: ";
		str += locArgs.c_str();
//locArgs.print(std::cout);
// 		for(size_t z=1; z<cmdArgM.size(); z++)
//			str += cmdArgM[z].c_str(), str+= " ";
		str += " ********\n"; 
		sprintf(buf,"********  working directory: %s  ********\n", locationM.path.c_str() );
		str += buf;
		sprintf(buf,"********  last compiled: %s %s  ********\n\n", __DATE__, __TIME__);
		str += buf;
		log(this, str.c_str());

		log(this,"#dm ++ Environment(%p) pid = %d tid = %u\n\n", this, sys::getPid(), sys::getsTid() );

		//***** Monitoring *****
#	ifdef CTL_Monitor
		// instantiate Monitor

				ctl::string monitorStr=getVariable("CTL_Monitor");
				if(!!monitorStr && monitorStr!="none")
				{
		// get silent monitor
					setVariable("CTL_Monitor", 0, true);
					log(this, "#d ++ monitored by  '%s'\n", monitorStr.c_str());
					monitorM = link(monitorStr+ " -X CTL_Monitor=none ");
		//			if(monitorM.getType() == thread)
					if(monitorM.getType() != dmn)
					{
						ctl::string id;
						ctlCI::getLANid(monitorM).assign(id);
		//				fprintf(stderr,"got getLANid %s\n", id.c_str());
						setVariable("CTL_Monitor", id.c_str(), true);
						monitorThemesM = 0;
					}
					if(logThemes>0)
						monitorThemesM = logThemes;
					else
						(void)themes::getThemes("f", monitorThemesM);
					log(this, "\n\n");
				}
				logThemes = logThemesM;
				logThemesM = 0;
// monitor the header
				log(this, str.c_str());
				logThemesM = logThemes;
#	endif

#endif

		if(err)
			log(this,"#w ?? cwd -> %s failed\n", path.c_str());	

//***** Late readVar error logging *****
		if(!!envFile)
			ctl::log(this, "#v ?? variables read from '%s'\n", envFile.c_str());
		if(!!readErr)
			log("#v %s\n", readErr.c_str());	

// 		ctl::clog<<"envSpace:\n"<<envSpaceM.variables<<"\n";
// 		ctl::clog<<"exportSpace:\n"<<exportSpaceM.variables<<"\n";
// 		ctl::clog<<"_envSpace:\n"<<_envSpaceM.variables<<"\n";



		ctl::log(this,"#c ** connecting intrinsics **\n");

		groupImpl::connect();
		connect<ctl::exception>();
		log(this,"\n");
		ctl::connect<commuCI, CommunCI_Impl, commConnectDetail >();

		ctl::log(this,"#c ** connecting ctlCI **\n");
		ctlCI::connect<1>(&ctl::getSymbols);
		ctlCI::connect<3>(&ctl::getPID);
		ctlCI::connect<4>(&ctl::getVersion);
		ctlCI::connect<5>(&ctl::getLANid);
		ctlCI::connect<6>(&ctl::getLANidS);
		ctlCI::connect<12>(&ctl::setUsing);

#ifdef CTL_Monitor
		ctlCI::connect<13>(&ctl::monitoring);
#endif 		
		ctl::log(this,"#c\n");
		intrinsicSymbolTableM = symbolTableM; 
		return locArgs;
	}

	CTL_Inline Environment::~Environment()
	{
// already in destruction  ?
		if(inDestrM)
			return;
		ctl::log("#d ?? mainGroup = %p in env %p\n", mainGroupM.get(), this);
		ctl::log("#d ?? dmnGroup = %p in env %p\n", dmnGroupM.get(), this);
		if(!!errorM)
					log(this,"#wl\n !! error = %s", errorM.c_str());
		log(this,"#d\n -- Environment(%p) pid = %d\n\n", this, sys::getPid());

		if(!!dmnGroupM && dmnGroupM!=monitorM.getGroup())
		{
			const char *monStr = getVariable("CTL_Monitor") ;
			if(monStr && strcmp(monStr,"none")==0 )
			for(int z=0; z<100 && dmnGroupM->size()>1; z++)
			{ 
				dmnGroupM->run(); 
				ctl::log("#m %p monitor waiting for dmnGroup %d\n", this, dmnGroupM->size() );
				sys::usleep(1e4);
			}	
		}

		locationSelectorM.clear();

#ifndef CTL_Win32
		signal( SIGPIPE,  SIG_IGN);
#endif
		symbolTableM.clear();		
		namedObjMapM.clear();
		clientM = ctl::link();
		objectListM.clear();
		ctl::uint4 z=0;
		ctl::vector<const Remote*> refV(referenceListM.size());
		for(std::set<const Remote*>::iterator ref=referenceListM.begin(); ref!=referenceListM.end(); ref++)
			refV[z++] = *ref;
		for(z=0; z<refV.size(); z++)
			((Remote*)refV[z])->~Remote();

		linkMapM.clear();

		inDestrM = true;
		groupListT::iterator groupI;
		groupListM.erase(mainGroupM.get());
		groupListM.erase(dmnGroupM.get());
		for(groupI = groupListM.begin(); groupI !=groupListM.end(); groupI++)
		{	
//			if((*groupI)!=monitorM.getGroup() && mainGroupM != *groupI )
			if((*groupI)!=monitorM.getGroup())
			{
				(*groupI)->setError(errorM);
				delete *groupI;
			}
		}
		if(!!dmnGroupM && dmnGroupM!=monitorM.getGroup()  && dmnGroupM!=mainGroupM)
		{
			dmnGroupM->errorM = errorM;
//			dmnGroupM->~groupImpl();
			dmnGroupM.destruct();
		 }
		ownedLinksM.clear();
		mainGroupM.clear();
//		mainGroupM.destruct();

		for(size_t i=0; i<tMarker.size(); i++)
			delete tMarker[i];

		time_t t = ::time(0);
		char *timeStr= ::ctime(&t);
		char* eol=strchr(timeStr,'\n');
		if(eol)
			*eol='\0';

{
	char buf[CTL_StringBufferSize];
	if(cmdArgM[0].size() + 80 > CTL_StringBufferSize)
		ctl::throwStr("string overflow in ~Environment");

	ctl::string str;
	sprintf(buf, "#a \n********  closing ctl-log file of %s (PID %d) ******** \n", cmdArgM[0].c_str(),  sys::getPid());	
	str += buf;
	sprintf(buf,"********  at %s  ********\n", timeStr);
	str += buf;
	ctl::log(this, str.c_str());

}
		logThemesM |= themes::silent;
		monitorThemesM |= themes::silent;
		if(!!monitorM)
			monitorM = ctl::link();
		destroystaticObjects();
		if(!!loggerM && loggerM != stdout && loggerM != stdoutM && loggerM != stderr && loggerM != stderrM)
		{	::fclose(loggerM); loggerM = 0; }
		delete ioDeviceM;
#ifdef CTL_Win32
		if(!parentEnvM)
			WSACleanup();
#endif
	resetSignals();
	}

	
/**/// non blocking read and call of RMI's

CTL_Inline bool Environment::run(int &read, bool blocking, const groupImpl *caller)
{
	if(inRunJobM)
		return true;
	inRunJobM = true;
//log("Environment::run %d\n", threadM);
	read = 1;
	bool closed = !runJob(read, blocking );

	int r=0;
/*
	closed &= dmnGroupM->runJob(r, blocking);
	if(r)
		return closed;
*/
//	bool closed = !runJob(read, blocking || !dmnGroupM );
	if(closed && threadM) // owned by thread
	{ inRunJobM = false; 	return false; }
	if(!read)
		updateOwnedLinks();
	if(ownedByParentEnvM && !!dmnGroupM && dmnGroupM->closed())
	{ inRunJobM = false; 	return false; }
	if(!dmnGroupM || dmnGroupM==caller)
	{ inRunJobM = false; 	return true; }
//	int r=0;
	closed &= dmnGroupM->runJob(r, blocking);
	read += r;
	if(!!mainGroupM && mainGroupM != dmnGroupM)
		closed &= mainGroupM->runJob(r, blocking);
	read += r;
#ifdef CTL_ListenAllGroups
//	if(parentEnvM && parentEnvM!=this)
//		parentEnvM->run(read, false);
	groupListT::iterator groupI;
	for(groupI = groupListM.begin(); groupI !=groupListM.end(); groupI++)
		if(*groupI != mainGroupM.get() && *groupI != dmnGroupM.get()) 
		{	
			(*groupI)->runJob(r, blocking);
			read += r;
		}
#endif
	inRunJobM = false;
	return !closed;
}

/**/// non blocking read and call of RMI's
CTL_Inline bool run(int /*n=0*/, bool blocking/*=false*/)
{ int read=1; env().inRunJobM = false; return env().getThreadEnv()->run(read, blocking); }

#ifdef CTL_Monitor
	inline bool monitoring(const ctl::string &m, bool b)
	{ 
		Environment &e=ctl::env();
		device *dev = e.getDevice();
		static std::map<PID, int> &color = createStatic<std::map<PID, int> >();
		static int z=0;
		PID client = e.getClient().getPid();
		std::map<PID, int>::iterator col=color.find(client);
		if(col==color.end())
			color[client] = (color.size()+7)%8;
		char buf[CTL_StringBufferSize];
		if(client.host.size()+m.size()+80>CTL_StringBufferSize)
			ctl::throwStr("string overflow in monitoring");

		sprintf(buf, "\033[%im%.5d: \033[%im%s\033[0m",30 + color[client], ++z,30 + color[client], m.c_str());
//		sprintf(buf, "(%s:%d): \033[%im%s\033[0m", client.host.c_str(), client.port, 30 + color[client], m.c_str());
		dev->write(buf, strlen(buf)+1);
		if(b)
			dev->read(buf, 1);
		return true; 
	}
#endif 		

//! print link info to file

CTL_Inline void messageInfo(FILE *file, const char *mess)
{
	Environment *pEnv=&env();
	ctl::location loc=pEnv->getMyLocation();
	
	if(!pEnv->verboseM)
		return;
	if(mess)
	{
		fprintf(file, "%s",  mess);	
		fflush(file);
		return;

	}
	int myID = !!pEnv->getDmnGroup() ? int(pEnv->getDmnGroup()->myPID()) : 0;
	int nProc = !!pEnv->getDmnGroup() ? int(pEnv->getDmnGroup()->size()) : 0;
#ifdef CTL_Tick
	if(nProc>2)
	 	fprintf(file,"  *** running %s as process %d on %s (%d MHz) ***\n  *** working dir is %s ***\n  *** last compiled %s %s  ***\n",
    		loc.exec.c_str(),myID, loc.host.c_str(),int(1e-6/ctl::timer().secPerTick()), loc.path.c_str(), __DATE__, __TIME__);
	else
	 	fprintf(file,"  *** running %s on %s (%d MHz) ***\n  *** working dir is %s ***\n  *** last compiled %s %s  ***\n",
    		loc.exec.c_str(), loc.host.c_str(),int(1e-6/ctl::timer().secPerTick()), loc.path.c_str(), __DATE__, __TIME__);

#else

	if(nProc>2)
 		fprintf(file,"  *** running %s as process %d on %s  ***\n  *** working dir is %s ***\n  *** last compiled %s %s  ***\n",
    		loc.exec.c_str(),myID, loc.host.c_str(), loc.path.c_str(), __DATE__, __TIME__);
	else
	 	fprintf(file,"  *** running %s on %s  ***\n  *** working dir is %s ***\n  *** last compiled %s %s  ***\n",
    		loc.exec.c_str(), loc.host.c_str(), loc.path.c_str(), __DATE__, __TIME__);
#endif
		fflush(file);
}

CTL_Inline const PID &Environment::getPhysId(bool initDmn) const
{
	if(!pysIDM)
	{
		static const char *arg[3]={"-l dmn", 0, 0};
		int nArgs = 1;
		if(!initDmn && parentEnvM && parentEnvM->pSocketTmpM)
		{ nArgs = 2; arg[1] = "-L 1"; } // don't init socket
//		ctl::log(this, "#d ++ mainGroup(%p) in env %p (<-%p)\n", mainGroupM.get(), this, parentEnvM);
		dmnGroupM = new groupImpl(nArgs, arg, false, false, (Environment*)this);
		if(!dmnGroupM->isValid())
			throwStr("#creation of dmnGroup failed");
		ctl::log(this, "#d ++ dmnGroup(%p) in env %p (<-%p)\n", dmnGroupM.get(), this, parentEnvM);
		pysIDM = dmnGroupM->getCommunicator()->getPhysId();
		if(parentEnvM && parentEnvM->pSocketTmpM)
		{
			((CommunicatorImpl<dmn>*)dmnGroupM->getCommunicator().get())->addSocket(*parentEnvM->pSocketTmpM);
			pysIDM = dmnGroupM->getCommunicator()->getPhysId();
			parentEnvM->pSocketTmpM = 0;
			ownedByParentEnvM = true;
		}
	}
	else if(initDmn && pysIDM.isLogical() && pysIDM.port==0)
		((CommunicatorImpl<dmn>*)dmnGroupM->getCommunicator().get())->initDaemon(-1, true);

	return pysIDM; 
}
#	if defined(CTL_UseLocator)

} // ctl

#	include CTL_LocatorCI

namespace ctl {

CTL_Inline bool ctl::Environment::setLocator()
{ 		
	ctl::link p(selectStrM);	
	if(!p)		
	{
		selectStrM = "#";
		return false;
	}
	CTL_Locator locate(p,named,"implicit");	
	bool ok = !!locate;
	if(ok)
		setLocator(ctl::lPointer<locationSelectBase>(new locationSelector<CTL_Locator>(locate) ) );	
	return ok;
}

#else

CTL_Inline bool Environment::setLocator()
{ return false; }


#endif

} // ctl

#endif

