/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_Group_INL_
#define CTL_Group_INL_

#include <memory>
#include "ctl/utl/factory.h"
#include "environment.h"
#include <ctl/group.h>
#include <ctl/remote.h>

namespace ctl
{

CTL_Inline groupCI::resultType groupCI::getCtrl(const ctl::vector<int> &c, int4 count) const
{ 
	RMI0::send(*this, MID(0), RMI0::optArgs(c, count), RMI);
	return success; 
}
CTL_Inline groupCI::resultType groupCI::terminate(int4 count) const
{ 
	RMI1::send(*this, MID(1), RMI1::optArgs(count), RMI);
	return success; 
}

CTL_Inline result<bool> groupCI::recvTermination(const PID & source, const ctl::string &error)
{
	typedef RMI3::optArgs args;
	args X(source,error);
	return RMI3::send(*this, MID(3), X, RMI);
//	return true;
}

CTL_Inline result<void> groupCI::recvException(const ctl::link &p, const ctl::string &error)
{
	typedef RMI4::optArgs args;
	args X(error);
//	typeId name="void ctl::groupCI::recvException(const string)";
	RMI4::send(p, "void ctl::groupCI::recvException(const string)", X);	
	return  result<void>();
}

CTL_Inline status groupCI::receive(ctl::iStream &is)
{
	status r = ctl::unknown;
	ctl::MID mid; is>>mid;
	ctl::falseT nocast;
	switch(mid)
	{
		case 0: { r =  RMI0::call(nocast, ctl::functor(*this,&groupCI::getCtrl) ,is); break; }
		case 1: { r = RMI1::call(nocast, ctl::functor(*this,&groupCI::terminate), is); break; }
		case 3: { r = RMI3::call(nocast, ctl::functor(*this,&groupCI::recvTermination), is); break; }
	}
	return r;
}

CTL_Inline groupImpl::tempInstance groupImpl::getInstance(const PID &p) const
	{
		int P=-1;
		if(!p.isLogical())
		{
			for(uint4 i=0; i<size(); i++)
				if(i!=myPID() && partnerM[i].pid==p)
				{ P = i; break; }
			return 0;
		}
		else P = p;
		if(P==(int)myPID() )
			return this;
		return tempInstance(getPartner()[P], 	communicatorM, P);
}

CTL_Inline void groupImpl::broadCastError(const ctl::string &error) const
	{
	ctl::log(envM, "#m !! broadcast termination\n"); 
		if(!isValid())
			return;
		ctl::vector<result<bool> > R(size() );
		for(uint4 i=0; i<size(); i++)
			if(i!=myPID() && i<partnerM.size() && partnerM[i].while1)
				R[i] = getInstance(i)->recvTermination(myPID(), error);
	}

CTL_Inline void groupImpl::readMessage(bool blocking, const PID &src, ioTag tag) const
{
	if(destroyM || !communicatorM || (communicatorM->isEmpty() && (!isValid() || terminatedM) ) )
	{ terminatedM = true;	return;  }
//	checkPartner(src);
	if(!queueM.empty() || !!parentM)
		blocking = false;
  int len=-1, type = tag;
	PID currSrc=src;
	if(!blocking)
	{
		len=communicatorM->test(currSrc, type);
		if(len<0)
		{	
			if(forwardM)
				forwardM->forward();
			return;
		}
	}
	CTL_Assert(blocking || !!currSrc);
	queueMessage(len,  currSrc, &type);
	blocking = false;
}

CTL_Inline void groupImpl::init()
{
	forwardM = 0;
	if(!communicatorM)
	{
		ctl::log("#m !! Could not create communicator<%s> \n",   location::channel2string(typeM));
		return;
	}

	communicatorM->setOwner(this);
	if(communicatorM->nProcs()<2)
		checkTerminationM=false;
  	
 	if(checkTerminationM)
 	{
		LocalM.resize(communicatorM->nProcs());
		ControlM.resize(communicatorM->nProcs());
		ControlM.fill(0); LocalM.fill(0);	
	}
	envM->addGroup(this);
}

CTL_Inline groupImpl::groupImpl(const CommunicatorSP& commu): groupCI(false, true), nameM("main"), terminatedM(false),CtrlM(false) ,
                      inTerminationM(false), destroyM(false), activM(false),  checkTerminationM(false), parentM(0)
{	
	ctl::log(envM, "#d ++ group(%p) in env=%p\n", this, envM);
	communicatorM = commu; 
	init();
}

CTL_Inline void groupImpl::init(int n, const char *const *arg, bool /*byMain*/)
{	communicatorM = createCommunicator(n,arg,typeM); }

CTL_Inline  const ctl::vector<processInfo> &groupImpl::getPartner() const
{
	if(!partnerM)
	{
// beware of recursive call via communicator::check()
		partnerM.resize(1);
		if(getLinkage()==dmn || getLinkage()==ext|| size()<2)
			return partnerM;
		processInfo info(communicatorM->getPhysId(), envM->getThreadId(), this, while1M, envM->getExportVariables());
		CTL_Try
		{	
			partnerM=sample(info);
			for(unsigned int p=0; p<size(); p++)
				if(p!=myPID())
					envM->setRemoteLink(partnerM[p].threadId, getProcess(p, RMI, false) );
			if(int(myPID() )!=0)
				envM->setExportVariables(partnerM[0].variables);
		}
		CTL_Catch(const ctl::exception &)
		{
			partnerM.clear();
			((CommunicatorSP&)communicatorM).destruct();
			deleteGroup(this);
			CTL_Throw;
		}
	}	
	return partnerM;	
}


CTL_Inline groupImpl::groupImpl(int n, const char *const *arg, bool byMain, bool while1, Environment *envP):groupCI(while1, false), nameM("main"), terminatedM(false),CtrlM(false) ,
                      inTerminationM(false), destroyM(false),activM(false),  checkTerminationM(true), parentM(0)
{
	envM = !envP ? &ctl::env(n, arg, false) : envP;
	ctl::log(envM, "#d ++ group(%p) in env=%p\n", this, envM);
	location loc(n,arg);

	typeM = loc.linkage;
	if(typeM == thread || typeM == lib) 
		typeM = serial;
	init(n, arg, byMain);
	init();
	if(!!communicatorM)
	{
		if(byMain && myPID()==0 && !!envM->getParent(myPID()))
			parentM = &envM->getParent(myPID());
		if(!byMain && !parentM && (typeM==mpi) )
			   while1M = false;
	}
	if(!isValid())
		deleteGroup(this);
}

CTL_Inline groupImpl::groupImpl(PID & pid, const PID &logId, const char *args):nameM("main"), terminatedM(false),CtrlM(false),
      inTerminationM(false), destroyM(false), activM(false), checkTerminationM(false), parentM(0), typeM(dmn)
{
	ref(); // beware of destruction in communicatorImpl<dmn>::connect due to temporary link using this
// corresponding unref in link::link
	CommunicatorImpl<dmn> *DCom=new CommunicatorImpl<dmn>(pid,logId);
	communicatorM = CommunicatorSP(DCom);
	init();
	ctl::string error;
	int destPort=DCom->connect(pid, args, error);
	partnerM.resize(2);
	partnerM[1].pid = pid;
//	ctl::log("#m groupImpl(pid, args= %s[%d]\n", args, destPort); 
	if(!!error)
	{
		ctl::log("#m !! %s\n", error.c_str()+1 );
		errorM += error.c_str()+1;
	}
	if(destPort<0 || !!error)
		communicatorM.destruct();
//	ctl::log("#m group::connect (%d -> %d)\n", int(pid), destPort);

	pid.port = destPort;
}


CTL_Inline void groupImpl::init(const ctl::vector<location> &loC, channelT type)
{
	typeM = type;
	ctl::vector<location> loc(loC.size());
	for(size_t i=0; i<loc.size(); i++)
		loc[i] = envM->addExportedVariables(loC[i]);
	communicatorM = createCommunicator(loc,type);
	init();
}

CTL_Inline groupImpl::groupImpl(const ctl::vector<location> &loc, channelT type, const ctl::string &name): nameM(name),
						terminatedM(false), CtrlM(false) ,inTerminationM(false), destroyM(false), activM(false), checkTerminationM(true), parentM(0)
{
	while1M = !!loc && loc[0].replyDstr;
	 init(loc,type); }


CTL_Inline void groupImpl::init(const location &loc, channelT type)
{
	typeM = type;
	while1M = loc.replyDstr;
	if(!location::checkSupport(typeM))
		return;
	bool late= (typeM==tcp);
	ctl::vector<location> locV(1, loc);
	communicatorM = createCommunicator(locV, type, late);
	checkTerminationM = true;
	init();
}

CTL_Inline groupImpl::groupImpl(const location &loc, channelT type, const ctl::string &name): nameM(name),
						terminatedM(false), CtrlM(false) , inTerminationM(false), destroyM(false),activM(false),  checkTerminationM(false), parentM(0)
{
	init(loc,type);
	if(type==mpi || type ==ext)
		checkTerminationM = false;		
}


CTL_Inline groupImpl::~groupImpl()
{
	shared::ref();
//	CTL_Assert(queueM.empty());
	ctl::log("#d -- group(%p) in env %p\n", this, envM);
	if(errorM.empty())
		errorM = "~group";
	if(partnerM.size()<size())
		destroyM = true;
	else
		for(unsigned int p=0; p<size(); p++)
			if(p!=myPID())
				envM->rmRemoteLink(partnerM[p].threadId );

	if(!!communicatorM)
		communicatorM->clean();
	if(!destroyM)
	{
		if(checkTerminationM && !errorM.empty())
		{
			CTL_Try
			{ broadCastError(errorM); }
			CTL_Catch(const ctl::exception &e)
			{ ctl::log(envM, "#m !! in ~group caught %s\n", e.what() ); }
		}
	}
	else
	{ sys::usleep(10); }

	CTL_Try
	{
		communicatorM.destruct();
	}
	CTL_Catch(const ctl::exception& e)
	{
		ctl::log("#m !! caught exception '%s'in destruction of group (%p)\n", e.what(), this);
	}
	CTL_CatchAll
	{
		ctl::log("#m !! caught exception in destruction of group (%p)\n", this);
	}
	deleteGroup(this);
	shared::deref();
// needed ?! by sel -l tcp
//	errorM.clear();
}


CTL_Inline link groupImpl::getProcess(const PID &pid, ioTag tag, bool owner) const
{ 
	if(typeM==dmn)
		return link();
	if(pid!=myPID())
		return link(communicatorM, pid, tag, owner);
	if(!localCommM)
		localCommM = CommunicatorSP(new CommunicatorImpl<serial>() );
	return link(localCommM);
}

CTL_Inline link groupImpl::operator[](const PID &pid) const
{ return getProcess(pid); }

CTL_Inline link groupImpl::getParent() const
{ return envM->getParent(myPID() ); }

CTL_Inline bool groupImpl::checkParent() const
{
	if(destroyM)
		return false; 
	if(parentM && (parentM->grpM->destroyM || !parentM->callNext(true)) )
	{
			ctl::log("#d ?? sending closed by parent\n");
			for(int i=1; i<nProcs(); i++)
				getInstance(i)->recvTermination(PID(0), "closed by parent");
			destroyM = true;
			return false;
	}
	return true;
}

CTL_Inline bool groupImpl::run() const
{
	CTL_Try
	{
		if(!checkParent())
			return false;
		callCounterM =0;
		ctl::status ok =untilCompletion(myPID()==0 && checkTerminationM);
		return !destroyM && ok==success && isValid() ;
	}
	CTL_Catch(const ctl::groupException &e)
	{ CTL_Throw(e); 	}

	CTL_CatchAll
	{
		ctl::log("#mx !! caught exception in group::run\n");
		errorM = "caught exception in group::run\n";
		return size()<2;
	}
}

CTL_Inline status groupImpl::untilCompletion( bool checkTermination) const
{
	if(inTerminationM)
		return failed;
	else
		inTerminationM = true;

	CtrlM =  checkTermination;
//ctl::log(" untilCompletion control = %d\n",int(CtrlM));
	terminatedM = false;
	readMessage();
	ControlM = LocalM;
	LocalM.fill(0);
	Wait W;
	while(!terminatedM)
	{
		bool queueEmpty = callNext();
		int read = 1;
		if(queueEmpty)	
		{	
			if(!envM->run(read, false, this)) // closed by client
			{ inTerminationM = false;	return failed; }
 			W(read==0, false);
			if(!CtrlM && !checkParent() )
				return success;
		}
		else
			W(false);

		if(queueEmpty && CtrlM && ControlM[myPID()] + LocalM[myPID()]<=0)
		{
			int i=0;
			ControlM += LocalM;
			LocalM.fill(0);
			while(i<nProcs() && ControlM[i]<=0) i++;
			if(i==nProcs()) // send stop signal
			{
				ctl::log("#m ?? sending terminate\n");
				for(i=0; i<nProcs(); i++)
					getInstance(i)->terminate(callCounterM);
			}
			else
			{
				CtrlM = false;
				--LocalM[i];
				ctl::log("#m ?? sending getCtrl\n");

				getInstance(i)->getCtrl(ControlM,callCounterM);
			}

		}
		if(!terminatedM && queueM.empty())
			readMessage(false);
	}
	LocalM.fill(0);
	while(!callNext());
	inTerminationM = false;
	terminatedM = false;
	activM = false;
	return success;
}

CTL_Inline ioTag getDestTag(iStream &is)
{
	int8 currPos=is.currentSize(), pos = is.size() - sizeof(ioTag);
	is.seek(pos);
	uint4 tag;  	is>>tag;
	is.seek(currPos);
	return tag;
}

CTL_Inline void groupImpl::connect()
{
	ctl::connectFunction<groupCI::RMI4, int2type<4> >(&groupImpl::recvException,(groupCI::RMI4::origArgs*)0, "void ctl::groupCI::recvException(const string)" ) ;
}

CTL_Inline status evaluate(iStream& is, link& client);
//returning true if no message available
CTL_Inline bool groupImpl::callNext() const
{
// dangerous with multiple thread
//CTL_Profile("group::callNext");
   queueT::iterator mess=queueM.begin();
	if(queueM.begin()!=queueM.end())
	{
		callCounterM++;
		ctl::iStream is = *mess;
		if(is.size()==0) // close connection + continue
		{
			envM->rmRemoteLink(is.sourceM);
			queueM.erase(mess);
			if( (mess=queueM.begin() ) == queueM.end() )
				return true;
			is = *mess;
		}
		queueM.erase(mess);
		PID clientPID;
		clientPID = communicatorM->getPhysId(is.sourceM);	
//		ctl::log("#m callNext <- (%d %s)\n", clientPID.port, clientPID.host.c_str() );
		ctl::link client(communicatorM, clientPID);
		status stat = evaluate(is, client);
		switch(stat)
		{
			case ctl::retry:
			{
				queueM.push_back(is);
				if(queueM.empty())
					sys::usleep(0);
				break;
			}
			case ctl::failed: case ctl::success:
			{
				if(!!communicatorM && checkTerminationM)
						LocalM[myPID()]--;
				break;
			}
   		default: CTL_Assert(false);
		}
		return false;
	}
	return true;	
}

CTL_Inline status evaluate(iStream& is, link& client)
{
		Environment *myEnv = &env();
		objectID trgt;
		is>>trgt;
		Remote *target=trgt;
		status stat=failed;

		PID clientPID = client.getPid();
		ioTag tag = getDestTag(is);
		client.setTag(tag);
		link current = myEnv->getClient();

		if(!!client.getCommunicator())
			env().setClient(client);
		const ctl::link *curr=myEnv->getLink(clientPID);
		bool received=true;
		if(!target) // no target ==> function ?
		{
			link::envSet tmp(curr);
			MID mid;  is>>mid;
			target = (Remote*)myEnv->getFunction(mid.strM.c_str()).get();
			if(!target )
			{
				ctl::string err = ctl::string(" *** function ") + mid.strM + " not connected *** ";
				ctl::log("#f -> %s\n", mid.strM.c_str()); 
				ctl::log("#f <= %s throws ctl::exception('%s')\n", mid.strM.c_str(), err.c_str() ); 
				groupCI::recvException(client, err);
				received=false;
				stat = failed;
			}
		}
		if(target)
		{
		CTL_Assert(!!is.sourceM);
			bool changeEnv = target->envM && (target->envM != myEnv) && target->envM != myEnv->getParentEnv();
			Environment *childEnv = myEnv->getChildEnv();
			if(changeEnv)
				myEnv->setChildEnv(target->envM->getChildEnv());
			CTL_Try
			{	stat =  target->receive(is); }
			CTL_Catch(const ctl::exception &e)
			{ 
				stat = failed;
				received = false; 
				groupCI::recvException(client, e.what() ); 
			}
			if(changeEnv)
				myEnv->setChildEnv(childEnv);
		}
		myEnv->setClient(current);
		if(received && is.currentSize()>0 && is.currentSize() != is.size())
			log("#m ?? warning only %d of %d read\n", int(is.currentSize()), int(is.size()) );
		return stat;
}

CTL_Inline int  groupImpl::broad_partner(int step,char &ReadOrWrite) const
{
	PID myPID = communicatorM->getLogId();
	int nProcs = communicatorM->nProcs();

	ReadOrWrite = 'x';
	const int dim=ld2(nProcs);
	if(step<0 || step>=dim)
		return -1;
	int i = dim-step-1,msk=(1<<i)-1;
	int partner=myPID^(1<<i);

	if(partner>=nProcs)
		return nProcs;
	if(!(msk & myPID))
	ReadOrWrite = (myPID&(1<<i) ? 'r' : 'w');
	return partner;
}

CTL_Inline int groupImpl::merge_partner(int step,char &ReadOrWrite) const

{
	PID myPID = communicatorM->getLogId();
	int nProcs = communicatorM->nProcs();

	ReadOrWrite = 'x';

	const int dim=ld2(nProcs);
	if(step<0 || step>=dim)
		return -1;
	int partner=-1,msk=-1;
	{
		partner=myPID^(1<<step); msk = (1<<step)-1;
		if(partner>=nProcs)
			return nProcs;
		if(!(msk & myPID))
		ReadOrWrite = (myPID&(1<<step) ? 'w' : 'r');
	}
	return partner;
}


} // e.o ctl



#endif
