/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_IO_INL_
#define CTL_IO_INL_

#include <ctl/utl/timer.h>
#ifdef CTL_MemMap
#include <sys/mman.h>
#endif
#include <ctl/inl/system.inl>

namespace ctl
{
	CTL_Inline PID::IPvX::IPvX(const PID &pid): port(pid.port)
	{
		ip[0] = ip[1] = ip[2] = ip[3] = ip[4] = ip[5] = ip[6] = ip[7] = 0;
		if(pid.isLogical())
			return;
		char hip[64],  *h=hip,*dot;
		strncpy(hip, pid.host.c_str(), 63);
		unsigned int r,n;
		if(strchr(h, '.'))
		{ n = 4;}
		else if(strchr(h, ':'))
		{ n = 8;}
		else
			return;
		if(n==4)
		{
			int c[4] = { 0 };
			for(int i=0; i<4; i++)
			{
				c[i] =*h++ - '0';
				CTL_Assert(c[i]<10 && c[i]>=0);
				while(isdigit(*h))
				{ c[i] *= 10; c[i] += *h++ - '0'; }
				h++;
			}
			ip[6] = c[1] + (c[0]<<8);
			ip[7] = c[3] + (c[2]<<8);
		  return;
		}
		else
			for(unsigned int i=0; h && i<n; i++)
			{
/*
				r = 0;
				while(*h!=':' && *h)
				{ 
					r *= 16; 
					if(isdigit(*h))
						 r += *h++ - '0';
					else
						if(*h>= 'A' && *h<='F') 
						 r += *h++ - 'A';
					else
						if(*h>= 'a' && *h<='f') 
						 r += *h++ - 'a';
				}
				h++;
				ip[i] = r;
*/
			(void)sscanf(h,"%X", &r);
			ip[i] = r;
			h = (dot=strchr(h, ':')) ? dot+1 : 0;

			}			
	}
	CTL_Inline PID::IPvX::operator PID() const
	{
//		CTL_Profile("PID->IPvX");
		PID pid;
		pid.port = port;
		int len=0;
		while(!ip[len] && len<8) len++;
		char host[64] = {'\0'};
		switch(len)
		{
			case 6:  case 7:
			{
				int c0=ip[6]>>8,c1 = ip[6]&0xff, c2=ip[7] >> 8, c3 = ip[7]&0xff;
				sprintf(host,"%d.%d.%d.%d",c0,c1,c2,c3);	
				break;
			}
			case 0: case 1: case 2: case 3:
			{
				sprintf(host,"%x:%x:%x:%x:%x:%x:%x:%x",ip[0],ip[1],ip[2],ip[3],ip[4],ip[5],ip[6],ip[7]);	
			}
		}
		pid.host = host;

		return pid;
	}

	CTL_Inline int PID::cmp(const PID &p) const
	{
		if(port != p.port)
			return port - p.port;
		const char *c0=host.empty()? "\0" : host.c_str(), *c1=p.host.empty()? "\0" : p.host.c_str();
		while(*c0 && *c1)
		{
			if(*c0 != *c1)
				return *c0 - *c1;
			c0++; c1++;
		}
		return *c0 - *c1;
	}

	CTL_Inline bool ioBase::writeHd(const internal::ioBuffer &m) const
	{	
		const int headSize=MessHeader::fixsize;
 	   char buf[256];	
		ctl::oStream os=ctl::serializeB(m, buf, headSize);
		const  vector<dataObject> &data = m.buffer.getData();
		vector<sys::iovec> p(data.size()+1);
		p[0] = sys::ioEntry(os.buffer(), headSize);
		int sum=0;
		for(unsigned int i=0; i<data.size(); i++)
		{
			sum +=  data[i].size*data[i].n;
			p[i+1] = sys::ioEntry((char*)data[i].data, data[i].size*data[i].n);
		}
		CTL_Assert(sum==int(m.size) );
		int8 size = headSize + m.size;
		return write(size, p);
	}

	CTL_Inline bool ioBase::readHd(internal::ioBuffer &m, bool check) const
	{	
//	CTL_Profile("readHeaded");
		 if(!(*this))
			return false;
		const int headSize=MessHeader::fixsize;
		char buf[headSize];
		if(!read(buf, headSize) )
		{
			if(errno)
				ctl::log("#m !! read error = '%s'\n", strerror(errno));
			m.tag = errno ? -2 : -3;
			close();
		  return false;
		}
		ctl::iStream is(headSize, buf);
		PID pid=m.pid;
		is>>m;
	  if(check && m.pid.host!=pid.host)
		{
			ctl::log("#m !! invalid key %s %s\n", m.pid.host.c_str(),  pid.host.c_str());
			return false;
		}
	
//		CTL_Profile("reading");
		if(m.size)
			m.buffer.setBuffer(m.size, memmap(m.size) );
		if(m.size>0 && !read(m.buffer.buffer(), uint4(m.size) ))
		{
			close();
			m.tag = -2;
			ctl::log("#m !! read error = '%s'\n", strerror(errno));
			return false;
		}
   	return true;
	}

	CTL_Inline bool File::close() const
	{
		int ok =0;
		if(fdM>=0)
			ok =sys::close(fdM, isSocketM);
		fdM = -1;
		return ok==0;
	}
	CTL_Inline bool File::write(int8 size, const vector<sys::iovec> &p) const
	{
		while(size)
		{
			int wc=sys::writev(fdM, p.begin(), int(p.size()), isSocketM);
			if(wc<=0)
			{
				ctl::log("#m !! write error = %s\n", strerror(errno));
				return false;
			}
			size -=wc;
		}
		return true;
	}
	CTL_Inline bool File::write(void *buf, int size) const
	{
		if(fdM==-1 || !buf)
		   return false;
		char *wBuf = (char*) buf;
   		while(size)
   		{
			int wc=sys::write(fdM, wBuf, size, isSocketM);
	   		if(wc<0)
	   			return false;
	   		wBuf += wc; size -=wc;
	   	}
	   	return true;
	}

	CTL_Inline size_t File::read(void *buf, int ssize) const
	{
		size_t size = ssize<0 ? -ssize : ssize;
		if(fdM==-1)
		   return 0;
		char *rBuf = (char*) buf;
		
   		while(size)
   		{
	   		int rc=sys::read(fdM, rBuf, size, isSocketM);
			if(rc < 0 && (errno == EINTR || errno == EAGAIN))
				continue;
			if(ssize<0)
				return rc;
	   		if(rc<=0)
	   			return rBuf - (char*)buf;
	   		rBuf += rc; size -=rc;
	   	}
	  return rBuf - (char*)buf;
   }

   CTL_Inline bool File::test(int usec) const
   {
#ifdef CTL_MSC
	   if(!isSocketM)
		   return !::eof(fdM);
#endif

   		if(fdM<0)
   			return false;
   		timeval t,  *pt = usec<0 ? 0 : &t;
   		t.tv_sec = 0; t.tv_usec = usec;
		fd_set  rd;
   		FD_ZERO(&rd);
   		FD_SET((unsigned)fdM, &rd);
		int ok=::select(fdM+1, &rd, 0, 0, pt );
		if(ok<0)
			ctl::log("#w !! select  %d error = '%s'\n", (int)fdM, strerror(errno));
		return  ok>=0 && (FD_ISSET(fdM,&rd)!=0);
   }

	CTL_Inline bool socket::getLocalAddr(sockaddr_in &sa, unsigned short port, const char *host)
	{
		memset(&sa, 0, sizeof(sockaddr_in));		
 		char  *hostname=(char*)( host ? host : myLocation().host.c_str() );
// 		char  *hostname=(char*)( host ? host : "localhost" );
// 		char  *hostname=(char*)( host ? host : "0.0.0.0");
		struct hostent *hp=0;
		hp = gethostbyname(hostname);		
		if(hp)
		{  		
   		memcpy((char *)&sa.sin_addr,hp->h_addr,hp->h_length);
			sa.sin_family= hp->h_addrtype;
		}
		sa.sin_port  = htons(port);
		return true;
	}
	CTL_Inline void socket::setPort()
	{
		if(!!fdM)
		{
			::sockaddr_in sat;
	 		sys::sockl_t ssize=sizeof(sat);
	 		
			getsockname(fdM,(::sockaddr*)&sat, &ssize);

			pidM.port = htons(sat.sin_port);
			pidM.host=  inet_ntoa(*(in_addr*)&sat.sin_addr.s_addr);
//			ctl::log("#d ?? setPort:fd = %d port = %d %s\n", int(fdM), pidM.port, pidM.host.c_str());
	 	}
	}

#ifdef CTL_MemMap
	CTL_Inline char *socket::memmap(size_t size) const
	{ 
		void * buf= mmap(0, size, PROT_READ | PROT_WRITE, MAP_SHARED, fdM, 0); 
		if(buf==MAP_FAILED)
		{
			printf(" !! mmmap failed: error = %s\n",strerror(errno));
			return 0;
		}
		return (char*)buf;
	}
#endif
	CTL_Inline bool socket::open(unsigned short port, bool blocking, const char * host)
	{
		if(fdM>0)
			return true;
		if ((fdM= ::socket(AF_INET, SOCK_STREAM, 0)) < 0)
		{	errorM = openE; return false; }
		if(!blocking)
			sys::fcntl(fdM, sys::f_setfl, sys::o_nonblock);

		yesM=1;
		if( setsockopt(fdM, SOL_SOCKET,  SO_REUSEADDR,  &yesM,  sizeof(yesM) ) < 0)
		{	errorM = openE; return false; }
		struct sockaddr_in sa;
		if(!getLocalAddr(sa, port, "0.0.0.0"))
//		if(!getLocalAddr(sa, port, host))
		{	errorM = hostE; return false; }
		if (::bind(fdM, (sockaddr*)&sa, sizeof(sockaddr_in)) < 0)
		{
	 		close();
	 		errorM = bindE;
	 		return false;
	 	}
		if(!getLocalAddr(sa, port, host) )
		{	errorM = hostE; return false; }
	 	setPort();
		pidM.host=  inet_ntoa(*(in_addr*)&sa.sin_addr.s_addr);

   	int backlog = 10;            // queue limit for pending connections
   	if(::listen(fdM, backlog ) < 0 )
   	{
		close();
	 	errorM = listenE;
	 	return false;
		}
		ctl::log("#m ++ opened (%d %d)\n", (int)fdM, (int)pidM.port); 

		return true;
	}

	CTL_Inline bool socket::connect(unsigned short port, const char *host, bool fwd)
	{
		close();
		ctl::log("#m ++ connecting -> (%d %s)\n", (int)port,host); 
  
		::sockaddr_in sa;		
		if(!getLocalAddr(sa, port, host) )
		{	errorM = hostE; return false; }
   		if ((fdM = ::socket(AF_INET, SOCK_STREAM, 0)) < 0)
		{	errorM = openE; return false; }
   		if (::connect(fdM, (sockaddr*)&sa, sizeof(sockaddr_in)) < 0)
   		{   // try ssh forwarding
			if(!fwd)
			 { ctl::log("#d !! connect to host=%s(port=%d)error = %s\n", host, port, strerror(errno)); close(); return false; }
			char cmd[512];
			sprintf(cmd, "ssh -2 -f -N -L %d:%s:%d %s ", port, "localhost", port, host);
			ctl::log("#d ?? forward command = '%s'\n", cmd);
			ctl::log("#d $> '%s'\n", cmd);
			int sys_err=::system(cmd);
			(void)sys_err;
			if(!getLocalAddr(sa, port,"localhost") )
				return false;
			 if (::connect(fdM, (sockaddr*)&sa, sizeof(sockaddr_in)) < 0)
			 { ctl::log("#m !! connect error = %s\n", strerror(errno)); close(); return false; }
			}
   		setPort();
		sys::sockBool on = 1;

		::setsockopt(fdM,IPPROTO_TCP,TCP_NODELAY,&on,sizeof(on));
			ctl::log("#m ++ connected (%d %d)\n", (int)fdM, (int)pidM.port); 
   		return true;
	}
	CTL_Inline bool socket::close() const
	{
		bool ok = true;
		if(!!fdM)
		{
			ctl::log("#m -- closed (%d %d)\n", (int)fdM, (int)pidM.port); 
//			shutdown(fdM, sys::shut_rd);  	
			shutdown(fdM, sys::shut_rdwr);  	
			ok=(sys::close(fdM)==0);
			fdM = -1;
		}
		return ok;
	}
	CTL_Inline PID socket::accept(socket &acc, bool blocking, int usec) const
	{
		sockaddr_in sa;
		sys::addr_t len = sizeof(sa);
//		if(!blocking)
//			usec=1;

		if(!blocking && !test(usec) )
			return PID();
		int ok = ( acc.fdM =  ::accept(fdM,(sockaddr *)&sa, &len)  );
		if(acc.fdM==0)
			acc.fdM = -1;
		PID pid;
				
		if (ok>=0 && acc.fdM>=0)
		{
			pid.port = htons(sa.sin_port);
			pid.host = inet_ntoa(*(in_addr*)&sa.sin_addr.s_addr);	
			acc.pidM = pid;

		ctl::log("#m ++ accepted (%d %d)\n", (int)acc.fdM, (int)pid.port); 
		}
		return pid;
	}

CTL_Inline bool spawn(const PID &local, const location &dest, int logid, int nproc, const char *typ, const char *before, const char *in=0, const char *out=0, std::pair<FILE *, int> *pServOut=0)
	{
//	dest.print(std::cerr);
		ctl::string cmd = " ", host = local.host;
		bool isLocal  = !dest.user && (dest.exec=="*" || location::isLocalHost(dest.host) ||  dest.host==host);
		bool quote = (before && !isLocal);
		ctl::string end=quote ? " ' " : " ";
		bool useSSH = dest.shell!="none";
#ifdef CTL_MSC
		const char sep = '#';
#else
		const char sep = ' ';
#endif
		if(!isLocal && useSSH)
		{
			ctl::string shell = !dest.shell ? ctl::string(CTL_Str(CTL_Shell) ) :  dest.shell;

			shell += sep;
			cmd = cmd + shell;
			int sshPort=dest.sPort == 0 ? CTL_ShellPort : dest.sPort;
			int pos = (int)dest.host.find(':');
			ctl::string destL,destR;
			if(pos>0)
			{
				destL = dest.host.substr(0, pos);
				if(!dest.user.empty())
					cmd += dest.user+"@";
				cmd +=  destL + sep + shell;
				destR = dest.host.substr(pos+1);
			}
			else
				destL = destR = dest.host;
			socket sshCheck;			
			if(!sshCheck.connect (sshPort, destL.c_str(), false ) )
  			{
				ctl::log("#m !! Can't connect sshd@%s at port %d\n", destL.c_str(), sshPort);	
				if(strcmp(typ,"tcp") )
					return false;
				useSSH = false;
				host = local.host;
			}
			else
			{
				host = sshCheck.getIP();
				sshCheck.close();

				const char *pssh = strrchr(shell.c_str(),'/');
#ifdef CTL_WinPath
				if(!pssh)
					pssh = strrchr(shell.c_str(),'\\');
#endif
				pssh = pssh ? pssh+1 : shell.c_str();
				if(strncmp(pssh,"ssh",3)==0)
				{
					if(dest.term.empty() || dest.term == "none")
//					if(!dest.hasTerminal())
						cmd += "-x ";
					else
						cmd += "-X ";
					if(!dest.user.empty())		
       					cmd += dest.user+"@";
				}
				else if(!dest.user.empty() && strncmp(pssh,"rsh",3)==0)
					cmd += "-l ", cmd += dest.user+sep;
				else
					ctl::log("#m !! unknown remote shell %s\n ", pssh);
				cmd += destR + sep;
			}
		}
		ctl::string arg = "-l "; arg += typ;
		if(dest.linkage!=pipe)
		{ 
			arg += " -L "; arg += local.port; 
			if(!isLocal)
			{ arg += " -h "; arg += host; }
//			if(!location::isLocalHost(host) )
//			{ arg += " -h "; arg += host; }
		}
		if(logid != 1)
		{ arg += " -p "; arg += int(logid); }
		if(nproc != 2)
		{ arg += " -S "; arg += int(nproc); }
		if(!dest.args.empty() && (dest.linkage==tcp || dest.linkage==pipe || dest.linkage==ext) )
		{	arg += sep; arg += dest.args; }
		if(in)
		{	
			if(*in!=' ') 
				arg +=  " > ";  
			arg += in; 
		}
		if(out)
		{	
			if(*out!=' ') 
				arg +=  " < ";  
			arg += out; 
		}
		if(dest.hasTerminal())		
{
		if(dest.exec=="*")
			cmd = ctl::string("-x ") + cmd;
		else
			cmd = dest.term + cmd ;
}
		if(quote)
			cmd += " ' " ;
		ctl::string path=dest.linkage == mpi ? dest.pathExec() : dest.absolutePathExec();
//		if(!path)
//			return -1;
		if(!before)
		{
			if(!dest.debug.empty())
				cmd += dest.debug + path + "  \" ' " + arg + " ' \"" ;				
			else
				cmd += path + sep + arg ;
		}
		else
			cmd += ctl::string(before) + arg + " \" ;  " + path;
		cmd += end;
//printf("%s\n", cmd.c_str());
		ctl::log("#d $> '%s' \n",cmd.c_str());
//std::cout<<cmd<<"exec="<<dest.exec<<pServOut<<"\n";
		int ret = sys::spawn(cmd.c_str (), pServOut, dest.exec=="*" );
		return ret>0;
	}

} // ctl

#endif
