/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_LINK_INL_
#define CTL_LINK_INL_

#include <ctl/link.h>
#include <ctl/group.h>

namespace ctl
{

	struct soHandle : public bCommunicator<internal::ioBuffer>
	{
		sys::LibHandle so;
		ctl::string libCopy;
		ctl::string library, logFile, errorM;
		typedef void (*openLibPointer)(void **,void *,void *, const char *, const char *(*)(), void (*)() );
		typedef void (*closeLibPointer)(void *, bool);
		typedef const char * (*locateLibPointer)();
		openLibPointer openLib;	
		closeLibPointer closeLib;	
		locateLibPointer locateLib;	
		bool threadM, closeM ;
		Environment *childEnvM, *parentEnvM;
		mutable Environment *tmpEnvM;
		link clientM;
		sys::threadIdT parentIDM, childIDM;

		static CTL_runDclr threadRun(void *D)
		{
			soHandle *handle = (soHandle*)D;
			handle->childIDM = sys::getsTid();
			(*handle->openLib)((void**)&handle->childEnvM,(void*)handle->parentEnvM, (void*)&handle->clientM, handle->logFile.c_str(),
														 handle->locateLib, handle->parentEnvM->getConnector()) ;
			(*handle->closeLib)((void*)handle->childEnvM, true);
			handle->childEnvM = 0;
			return 0;
		}
		bool open(const link &p)
		{
			if(!!library && !so)
				return false;
			threadM = p.getIType()==thread;
			if(!openLib || !closeLib)
			{
				ctl::string dlerr = sys::dlerror();
				ctl::log("#l !! %s' is not a CTL component %s\n", library.c_str(), dlerr.c_str());
				errorM +=  library +"  is not a CTL component (" + dlerr + ")";
				if(so)
					sys::dlclose(so);
				so = 0;
				return false;
			}
			clientM = p;
			clientM.release();
			if(threadM)
			{
				parentIDM = sys::getsTid();
				
				sys::threadIdT childID = sys::startThread(&threadRun, this);
				if(!childID)
					return false;
				while(!childEnvM)
					sys::usleep(0);
			}
			else
				(*openLib)((void**)&childEnvM,(void*)parentEnvM, (void*)&clientM, logFile.c_str(), locateLib, parentEnvM->getConnector() ) ;
			
			return true;
		}
		void setEnv(bool b) const
		{ 
//			ctl::log("#d setEnv (%d) parent = %p child = %p\n", int(b), parentEnvM, childEnvM);
			if(b)
				tmpEnvM = parentEnvM->setChildEnv(childEnvM); 
			else
				tmpEnvM->clearChildEnv();
		}
			
		soHandle(const location &lib): so(0), openLib(0), closeLib(0), locateLib(0), childEnvM(0), parentEnvM(0), clientM(0), parentIDM(0), childIDM(0)
		{
			parentEnvM =&env();
			if(lib.exec!="*")				
				library=lib.absolutePathExec();
			if(lib.linkage==thread)
			{
				logFile = library + " -l thread " + lib.args;
				if(!!lib.term)
					logFile += + " -x [" + lib.term+ "]";
			}
			else
				logFile = library + " -l lib " + lib.args;
			closeM = lib.portL>=0;
//			ctl::log("#f portL = %d\n", lib.portL);
			threadM = lib.linkage==thread;
			if(!!library)
			{
#ifndef CTL_WinPath
				if(library.size()>2 && library[0] == '~' && library[1] == '/')
				{
					const char *home = sys::getEnv("HOME");
					if(home)
						library = ctl::string(home) + "/" + library.substr(2);
				} 
#endif
				so=sys::dlopen(library.c_str(), /*local binding*/true);
#ifdef CTL_Unix
				if(so && sys::dlsym(so, "CTL_unsaveMT") && envM->addLibHandle(so))
				{
					sys::dlclose(so);
// copy dll to /tmp
					libCopy=sys::copyFile(library);
					if(libCopy.empty())
						ctl::log(envM, "#lw !! Could not open component '%s', error = copy failed\n", library.c_str());
//open copy
					ctl::log(envM, "#lw ++ multi-thread unsave component copied to %s\n", libCopy.c_str());
					so=sys::dlopen(libCopy.c_str(), /*local binding*/true);
				}
#endif
				if(!so)
				{
					string dlerr=sys::dlerror();
					ctl::log(envM, "#lw !! Could not open component '%s', dlerror = %s\n", library.c_str(), dlerr.c_str() );
					errorM += dlerr;
					return;
				}
				openLib = (openLibPointer)sys::dlsym(so, "CTL_openLibrary");
				closeLib = (closeLibPointer)sys::dlsym(so, "CTL_closeLibrary");
				locateLib = (locateLibPointer)sys::dlsym(so, "CTL_getLocatorName");
			}
			else
			{
				openLib = &ctl::openLibrary;
				closeLib = &ctl::closeLibrary;
				locateLib = &sys::getLocatorName;
			}
			ctl::log(envM, "#l ++ %s -> %s\n", location::channel2string(lib.linkage), logFile.c_str());
		}
		~soHandle()
		{
//			CTL_Assert(sys::getTid()==parentEnvM->getTid() );
			channelT type = threadM ? thread : lib;
			if(!library || so)
			{
				if(threadM && childEnvM)
				{
					addJob(0,0); // the terminating job
					while(childEnvM) // wait for thread finished
						sys::usleep(0);
				}
				if(!threadM)
				{
					setEnv(true);
					childEnvM->tidM =0;
					(*closeLib)((void*)childEnvM, false);
					setEnv(false);	
				}
// if in so an atexit-function is registered dlclose causes segfault 
				if(so && closeM)
					sys::dlclose(so);
				if(!libCopy.empty() && !sys::remove(libCopy.c_str()) )
					log(envM, "#d !! removing tempory copy '%s' failed\n", libCopy.c_str());

				ctl::log("#l -- %s -> %s \n", location::channel2string(type), logFile.c_str());
//				ctl::log("#l -- %s -> %s (%d)\n", location::channel2string(type), logFile.c_str(), int(so && closeM));
			}
		}
		bool closed() const
		{ return childEnvM ==0; }
		channelT getType() const
		{ return threadM ? thread : lib; }
		bool operator!() const
		{ return !isValid(); }
		bool isValid() const
		{ return(!library || !!so) && childEnvM; }
		RemoteSP operator()(const char *pretty) const
		{
			if(pretty && childEnvM)
				return childEnvM->getFunction(pretty);
			return RemoteSP(); 	
		}
		bool addJob(void(*fnc)(void*), void *data) const
		{
			if(isValid())
			{
				childEnvM->addJob(fnc, data, parentIDM);
				if(childIDM)
					sys::wakeUp(childIDM);
			}
			return true;
		}
		typedef std::pair<int, ctl::vector<char> > streamedT;
		mutable jobQueue< streamedT > ioStream[2];
		int write_(oStream& os, const PID& dest, int tag) const
		{ 
//CTL_Profile("thread::write_");
			bool clientThread = (parentIDM == sys::getsTid());
			PID pid = !dest ? PID(clientThread ? 1 : 0) : dest;
			ctl::log("#m << thread: dest=%d size=%d tag=%d\n",int(pid.port), (int)os.size() ,tag);
			if(clientThread && tag == RMI)
			{
				ctl::vector<char > buf(os.size(), os.buffer()); 
				link p(clientM);
				p.childTidM=childEnvM->getTid();
				commuCI::evaluate(p, buf);
			}
			else 
			{
				ctl::vector<char > buf(os.size(), os.buffer()); 
				ioStream[!clientThread].put(streamedT(tag, buf));
//ctl::log(" written -> %d this=%p\n", int(!clientThread), this );
			}
			return 0; 
		}
		int test_(PID &src, int *type) const
		{ 
//ctl::log("thread::test_ src = %d tag = %d\n",(int)src, *type);
			bool clientThread = (parentIDM == sys::getsTid() );
		streamedT *view=ioStream[clientThread].view();
//ctl::log(" testing -> %d(%p) this=%p\n", int(clientThread), view, this );
//if(view)
//	ctl::log("FOUND tag=%d\n", view->first);
	//		bool empty = ioStream[clientThread].empty();
			if(view && (*type<0 || view->first==*type) )
			{
				src = clientThread ? 1 : 0;
				*type = view->first;
//ctl::log("thread::test_ success src = %d tag = %d\n",(int)src, *type);

				return 0;
			}
			return -1; 
		}
		int read_ (iStream &is, PID &source, int* tag) const
		{ 
//CTL_Profile("thread::read_");
//			ctl::log("#m >> thread XX: src=%d  tag=%d\n", int(source.port), *tag);
			bool clientThread = (parentIDM == sys::getsTid() );
			streamedT buf(ioStream[clientThread].get());
			is.set(buf.second);
			source = PID(int(clientThread) );
			*tag = buf.first;
			ctl::log("#m >> thread: src=%d size=%d tag=%d\n", int(source.port), is.size(), *tag);
			return 1; 
		}

	};


	CTL_Inline void link::setForward(messageForward* forward)
	{ 
		const groupImpl *grp = getGroup();
		if(grp)
			grp->setForward(forward);
	}
	CTL_Inline void link::init(bool owner)
	{
		if(!!communicatorM)
		{
			if(owner)
				grpM = communicatorM->getOwner();
			else
				grpM = groupSP(communicatorM->getOwner(), false);
				
			linkageM = communicatorM->getType();
			if(linkageM==thread || linkageM==lib)
				childTidM = sys::getTid();
		}
	}
 	CTL_Inline void link::addFunction(const char *pretty, const RemoteSP &f)
	{	symbolTableM[pretty] = f;	}

	CTL_Inline bool link::addJob(void(*fnc)(void*), void *recv) const
	{
		if(getType() != thread || !communicatorM)
			return false;
		return communicatorM->addJob(fnc, recv);
	}

	CTL_Inline void link::initStr(const char *loc, channelT typ, const char *mod)
	{	
		if(loc && mod==0 && typ == undefined && *loc == '+')
		{ 
			while(isspace(*loc) )
				loc++;
			errorM = ++loc;
			linkageM = modifier;
			return;
		}
		if(typ==ctl::file)
		{
			communicatorM = CommunicatorSP(new ctl::CommunicatorImpl<ctl::file>(loc, mod) );
			if(!!communicatorM && !communicatorM->getLogId())
				communicatorM.destruct();
		}
		else
			init(location(loc), typ);
	}
	CTL_Inline void link::setChildEnv(const Environment *e) const
	{
		if(getIType()!=thread)
			return;
		soHandle *so=getHandle();
		if(so)
			so->childEnvM = (Environment*)e;
	}

	CTL_Inline void link::setEnv(bool b) const
	{
		if(getType()!=lib)
			return;
		soHandle *so=getHandleRaw();
		if(so)
			so->setEnv(b);
	}

	CTL_Inline RemoteSP link::getFunction(const char *pretty) const
	{	
		RemoteSP f;
		if(!pretty)
			return f;
		if(getType()==serial)
		{
			symbolT const *fp = symbolTableM.find(pretty);
			if(fp)
				return *fp;
			return  env().getFunction(pretty );
    }
		soHandle *so=getHandle();
		if(so)
			return  (*so)(pretty );
		return f;	
	}

	CTL_Inline link::link(channelT  linkage) : pidM(-1),tagM(UNDEF), linkageM(linkage), childTidM(0)
 	{
		if(linkageM == serial)
		{
			communicatorM = createCommunicator(0,0,serial);
			init();
		}
	}

	CTL_Inline link::link(int n, const char * const *argv): linkageM(undefined), childTidM(0)
	{
		grpM = new groupImpl(n, argv, false);
		tagM = RMI;
		communicatorM = grpM->getCommunicator();
		if(!!communicatorM)
			linkageM = communicatorM->getType();
	}

	CTL_Inline link::link(const CommunicatorSP& commu, const PID & target ,ioTag type, bool owner):
						communicatorM(commu),pidM(target),tagM(type), childTidM(0)
	{ init(owner); }

	CTL_Inline link::link(const PID & pidC, const PID &logId, const char *args): linkageM(dmn), childTidM(0)
	{
		PID pid=pidC;
		grpM = new groupImpl(pid, logId, args);
		grpM->unref(); // ref in groupImpl::groupImpl
		if(!grpM->isValid() )
		{
			errorM += grpM->errorM;
			grpM.destruct();
			return;
		}
		if(logId)
			pidM = pid;
		tagM = RMI;
		communicatorM = grpM->getCommunicator();
	 }


	CTL_Inline link::link(const ctl::string &loc,  channelT typ, const char *mod): pidM(-1),tagM(UNDEF), linkageM(typ), childTidM(0)
	{ initStr(loc.c_str(), typ, mod); }

	CTL_Inline link::link(const std::string &loc,  channelT typ, const char *mod): pidM(-1),tagM(UNDEF), linkageM(typ), childTidM(0)
	{ initStr(loc.c_str(), typ, mod); }

	CTL_Inline link::link(const ctl::any &property): propertyM(property), childTidM(0) {}

	CTL_Inline link::link(const char *loc, channelT typ, const char *mod): pidM(-1),tagM(UNDEF), linkageM(typ), childTidM(0)
	{ initStr(loc, typ, mod); }

	CTL_Inline link::link(const location &loc, channelT typ): childTidM(0)
	{ init(loc, typ); }

	CTL_Inline link::~link()
	{
//    May be ~ctl::environment already executed
//		ctl::log(0,"~link(%p,%d)\n",communicatorM.get(), !!communicatorM ? communicatorM.getRefCount() : -1 );
		if(!communicatorM)
			grpM.release();
	}

//create dmn link -> p
	CTL_Inline link::link(const link &p, const location &locC)
	{
		if(p.linkageM==lib)
		{
			ctl::log("#l !! lib can't be converted to dmn\n");
			return;
		}
		ctl::string id;
		ctlCI::getLANid(p).assign(id);

		location dLoc(id), loc = locC;
		loc.pipeTail.clear();
		loc.pipeType = 0;
		loc.errorStr.clear();
		loc.portL = dLoc.portL;
		loc.host = dLoc.host;
		loc.linkage = ctl::dmn;
		init(loc, ctl::dmn);
	}

	CTL_Inline void link::init(const location &loC, channelT typ)
	{
		if(!loC)
			return;
		location loc = env().addExportedVariables(loC);
		if(!!loc.pipeTail && typ!=ctl::keepAsIs)
		{
			if(loc.linkage==dmn)
			{
				location tail(loc.pipeTail);
				if(!!tail.pipeTail)
					typ = ext;		
			}
			else
				typ =  ctl::ext;
		}
		if(typ==undefined || typ==ctl::keepAsIs)
			typ = loc.linkage;

		linkageM = typ;
		CTL_Try
		{
			switch(linkageM)
			{
				case ctl::logid:
				{
					groupSP grp=env().getMainGroup();
					if(!!grp)
						*this = (*grp)[loc.portL];
					break;
				}
				case ctl::thread: case ctl::lib:
				{
					soHandle* handle = new soHandle(loc);
					communicatorM = CommunicatorSP(handle);
					if(!handle->open(*this))
					{
						errorM += handle->errorM;
						communicatorM.destruct();
						linkageM = ctl::invalid;
						break;
					}
					childTidM = handle->childEnvM->getTid();
					break;
				}
				case ctl::dmn:
				{
					int logId = loc.portL ? -1 : 0;
					ctl::int4 port=logId<0 ? loc.portL : loc.portR;
					if(port)
					{
						ctl::string locStr;
						if(!!loc.pipeTail)
						{ locStr = loc.pipeTail.c_str(); locStr += "#"; locStr += env().getExpVariableList(); }

						*this = link(pidM = PID(port, loc.host), logId, locStr.c_str() );
					}
					else
						linkageM = ctl::invalid;
					break;
				}
				case ctl::undefined:
				{
					if(!!loc.ciName)
						*this = ctl::getLink(loc.ciName, ctl::any(), sys::getLocatorName() );
					break;
				}			
				case ctl::serial:
					childTidM = sys::getTid();
				default:
				{
					grpM = new groupImpl(loc, typ, "link");
					
					if(!grpM->isValid())
					{
						grpM.destruct();
						linkageM = ctl::invalid;
						return ;
					}
					pidM = 1;
					tagM = RMI;
					communicatorM = grpM->getCommunicator();
					ownedLinkM = lPointer<link>(communicatorM->getOwnerLink());
					if(typ == mpi || typ == ext)
					{	
           	pidM = communicatorM->getConnection();
//						if(!pidM)
//							pidM = 1;
//           ctl::log("#d getConnection::got %s\n", pidM.toString().c_str()); 
					}
             
				}
			}
		}
		CTL_Catch(const ctl::exception &e)
		{
			grpM.destruct();
			ctl::log("#x !! %s\n", e.what() );
		}
	}
// return valid
	CTL_Inline bool link::callNext(bool blocking) const
	{
		const groupImpl *groupL = 0;
		if(!communicatorM || !communicatorM->isValid() || !(groupL=communicatorM->getOwner()) )
			return false;
		groupL->readMessage(groupL->queueM.empty() && blocking);
		groupL->callNext();
		return !!communicatorM && communicatorM->isValid() ;
	}
	CTL_Inline ctl::vector<ctl::string> link::getSymbols() const
	{
		ctl::vector<ctl::string> symName;
		if(getType()==serial)
		{
			symName.resize(symbolTableM.size() );
			for(size_t i=0; i<symbolTableM.size(); i++)
				symName[i] = symbolTableM.getKey(int(i)); 
		}
		else
			symName<<ctlCI::getSymbols(*this);
		return symName;
	}
	CTL_Inline bool link::operator!() const
	{
		if(!!grpM)
			return !grpM->isValid();
		return !communicatorM; 	
	}
	CTL_Inline bool link::closed() const
	{ return !communicatorM || communicatorM->closed() || (!!grpM && grpM->closed() ); }

	CTL_Inline bool link::run(int n/*=0*/, bool blocking/*=false*/) const
	{
		soHandle *so=getHandle();
		if(!so && (linkageM == thread || linkageM ==lib) )
			return false;
		if(so && so->childEnvM)
		{
			if(so->parentIDM == sys::getsTid())
				return !closed();
			bool closed=!so->childEnvM->runJob(n, blocking);
//			bool closed=!soHandle::runJob(so->threadQueue,n, false);
			if(closed)
				((CommunicatorSP&)communicatorM).clear();
			return !closed; 	
		}
		const groupImpl *groupL = 0;
		if(!communicatorM || !(groupL=communicatorM->getOwner()) )
			return linkageM==thread || linkageM==lib;
		for(;;)
		{
			CTL_Try
			{
				if(blocking && n<=0)	
					break;
				groupL->readMessage(blocking,pidM,tagM);
				bool queueEmpty=groupL->callNext();
				if(queueEmpty && n<=0)
					break;
				else
					n--;
			}
			CTL_Catch(const ctl::groupException &e)
			{ CTL_Throw(e); 	}
			CTL_CatchAll
			{ return false; }
		}

		return !closed();
//		return !!communicatorM && communicatorM->isValid() && groupL->isValid();
   }

	CTL_Inline status link::writeMessage(const ctl::oStream &os, int tag)  const
	{
		if(tag==-1)
			tag = tagM;
		if(!grpM)
		{	
			if(!!communicatorM)
			{ communicatorM->write((ctl::oStream&)os, pidM, tag); return success; }
			return failed;
		}
		status ok= grpM->writeMessage(os, pidM, tag);
		if(ok==failed)
			((CommunicatorSP&)communicatorM).destruct();
			
		return  ok;
	}

} // ctl

#endif



