/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef	CTL_REMOTE_INL_
#define	CTL_REMOTE_INL_

#include <ctl/remote.h>

namespace ctl
{
 	CTL_Inline bool location::checkSupport(channelT type)
	{
#ifndef CTL_MPI
		if(type==mpi || type==lam)
			return false;
#endif
#ifndef CTL_PVM
		if(type==pvm)
			return false;
#endif
#ifndef CTL_PIPE
		if(type==pipe)
			return false;
#endif
		return true;
	}

	CTL_Inline bool location::isLocalHostName(const ctl::string &host)
	{ return host.empty() || !strcmp(host.c_str(),"localhost") || 
					!strncmp(host.c_str(),"127.",4) || !strncmp(host.c_str(),"10.",3);
	}
	CTL_Inline bool location::isLocalAreaName(const ctl::string &host)
	{ return isLocalHostName(host) ||  
					!strncmp(host.c_str(),"192.168.",8) || !strncmp(host.c_str(),"10.",3) ||
					!strncmp(host.c_str(),"172.16",7); /*-31*/
	}
	CTL_Inline bool location::isLocalHost(const ctl::string &host)
	{ return isLocalHostName(host) || host == localHost(); }
	CTL_Inline ctl::string location::localHost()
	{	return myLocation().host;	}

	CTL_Inline ctl::string location::absolutePath() const
	{ return path=="." ? localPath() :sys::absolutePath(localPath(),path); }

	CTL_Inline ctl::string location::absolutePathExec() const
	{ 
		if(exec=="*")
			return ctl::string(); 
		return !path  ? exec : absolutePath() + "/" + exec; 
	}
	CTL_Inline char *location::seperate(char *str, char *&left, char *&right, char c, bool back)
	{
		char *pos=back ? strrchr(str,c) : strchr(str,c);
		if(pos)
		{ left = str; right = pos+1;	*pos = '\0'; }
		else
		{ left = 0; right = str; }
		while(left && *left && isspace(*left) )
			left++;
		return left;		
	}
//! Constructor accepting a string in ssh syntax 'user@host: path/exec <args>'
	CTL_Inline void location::init(const ctl::string &locStr)
	{
		const char *locC = locStr.c_str();
		if(!locC)
			return;
		char	loc[CTL_StringBufferSize], *s=loc,*left=0, *t=0, *b=0, *arg;
		loc[CTL_StringBufferSize-1] = 0;
		strncpy(loc, locC, CTL_StringBufferSize-1);
		t=strchr(s,'#');
		if(t)
			*t = '\0';
		t=strchr(s,'\n');
		if(t)
			*t = '\0';
		for(; isspace(*s); s++);
//  seperate tail of pipe location
		pipeType = 0;	
		if(*s=='|' || *s=='<' || *s=='>')
			pipeType = *s++;		
		for(t=s; *t && *t!='|' && *t!='<' && *t!='>'; t++) ;
		if(t && *(t-1)!='-')
		{
			pipeTail = t;
			if(!pipeType)
				pipeType = *t;
			*t = '\0';
		}
		for(; isspace(*s); s++);
// unbraced arrow
		if( (t=strstr(s,"->") ) && (!(b=strchr(s,brasep) ) || b>t) )
		{
			char *c=t;			
			*c = 0;
			while(isspace(*--c) )
				*c = 0;
			
			ciName = s;
			for(s=t+2; isspace(*s); s++);		
		}
		if(*s=='-' && s[1] !='>')
			arg = s;
		else
			for(arg = s+1; *arg && (!(isspace(*(arg-1)) || *(arg-1)==':') || *arg!='-');  arg++);
		if(*arg)
		{
			(void)localHost();
			setArgs(arg);
			if(!!errorStr)
				return;
			if(arg!=s)
				*(arg-1) = 0;
			else
			{
				host = localHost();
				exec = "*"; // only args given ==> exec == self
				return;
			}
		}
		else
			arg = 0;		
// typ and/or args
		if(!*s)
			return;
		substSlash(s);
		for(t = s+strlen(s)-1; isspace(*t); t--);
		t[1] = '\0';	t = s;
		if(seperate(t, left, s, '@', false) )
			user = left;
		bool sIsHost = false;
		if(seperate(s, left, s, ':',false) )
		{
			sIsHost = true;
#	ifdef CTL_Win32
			if(!left[1])
			{ s = left; s[1] = ':'; left = ""; }
#	endif
			host = left;
#	ifndef CTL_Win32
			if(seperate(s, left, s, ':',false) )
			{	host +=":"; host += left; }
#	endif
		}
		else
		{
			if(linkage==dmn)
			{ sIsHost = true;	host = s ; }
			if(!!user)
			{ errorStr += "daemon + user!?"; 	return; }
		}
		bool isLocal = host.empty()  || host=="localhost" || strncmp(host.c_str(),"127.",4)==0;
		if(isLocal)
			host = localHost();

		if(seperate(s, left, s, '/',true) )
		{ sIsHost = false; 	path = left; }
		if(isLocal && path.empty()  )
			path = localPath();

		while(path.size()>2 && path[path.size()-3]=='/' && path[path.size()-2]=='.')
		{	path.resize(path.size()-2); }

		if(!s || !*s)
		{
			char pstr[24] = {0};
			if(portR)
			{			
				sprintf(pstr,"%d", portR);
				exec = pstr;
				linkage	= dmn;
			}
			return;
		}
		for(t=s; *t && !isspace(*t); t++);
		if(*t)
		{
			*t++ = 0;
			while(*t && isspace(*t)) t++;
			if(isdigit(*t))
			{
				(void)sscanf(t, "%d", &portL);
				linkage = dmn;
			}
			if(linkage==undefined)
				linkage = getChannel(t);
		}
		if(!sIsHost)
			exec = s;
		if(isdigit(*s))
		{
			linkage	= dmn;
			portL = atoi(s);
		}
		if(linkage==undefined)
		{
			char *dot = strrchr(s,'.');
			if(!dot || strcmp(dot+1,"exe")==0 )
				linkage = tcp;
			else  if(isLocal && (strcmp(dot+1,"so")==0  || strcmp(dot+1,"lib")==0 || strcmp(dot+1,"dll")==0) )
				linkage = thread;
		}
		else 
			if((linkage==thread || linkage==lib) && !!exec && exec!="*" && !strchr(exec.c_str(), '.'))
#ifdef CTL_MSC
				exec += ".dll";
#elif defined CTL_OSX
				exec += ".dylib";
#else
				exec += ".so";
#endif
		
	}

	CTL_Inline void location::setup(channelT ch)
	{
//		if(!exec && (portL || portR))
		if(linkage==ctl::keepAsIs)
			linkage=undefined;
		if(!exec && portR)
		 errorStr="invalid combination of local and port";
		if(!!errorStr)
		{ clear(); return; }
		if(linkage == tcp || linkage == pipe)
		{ 
			if(logId==-1)
				logId = 1;
			if(groupSize==0)
				groupSize = 2;
		}
		if(!!varStr)
		{
			args += "-X ";
			if(strchr(varStr.c_str(), comma) )
			{	args += (char)brasep; args += varStr; args += (char)ketsep; }
			else
				args += varStr;
		}
		if(linkage==pipe)
			debug.clear();
		if(!!debug)
		{	
			if(term.empty() || term.size()<2)
				setTerminal();
		}
		if(!!term)
		{
			if( noTerminal() && ch!=keepAsIs )
			{	args += " -x ["; args += term; args += "]";  }
	//		{	args += " -x "; term = " "; }
			else 
			{ localArgs += " -x "; localArgs += term;  }
		}
		if(linkage==dmn)
		{
//			if(portL<1)
//			{ errorStr += "daemon w/o port number"; return; }
			if(!exec)
			{
				wd.clear();
				exec = "daemon";
			}
			else
			{
				const char *c= strrchr(host.c_str(),'/');
				if(c && exec == c+1)
					host = localHost();
			}
		}
		if(!errorStr && !exec)
			exec="*";
		if(!host || host=="undef")
			host = localHost();
	}
	CTL_Inline void location::set()
	{
		pipeType = 0;
		if(host.empty())
			host = localHost();
		if(path.empty())
			path = localPath();
	}

	CTL_Inline const char *readString(const char *c,char *str, bool readIP)
	{
		int z=0;
		if(!*c)
			return c;
		while(isspace(*c))
			c++;
		if(!readIP && isdigit(*c) )
			return c;
		str[z] = 0;
		int quote = 0;
		while(*c && (( z && !isspace(str[z-1])) || *c!='-' || isdigit(c[1]) || quote) )
		{
			if(*c==location::brasep)
				quote++;
			if(quote && *c==location::ketsep)
				quote--;
// z==0 ==> !isspace(*c)
			if(isspace(*c) )
			{
				if(str[z-1]!=' ' )
					str[z++] = ' ';
				c++;
			}
			else
				str[z++] = *c++;
		}
		while(isspace(str[z-1]) )
			z--;
		str[z] = 0;
		while(isspace(*c))
			c++;
		return c;
	}

	CTL_Inline const char *readNum(const char *c,int &num, char *str =0)
	{
		while(isspace(*c))
			c++;
		num = 0;
		bool neg=false;
		if(*c=='-' && isdigit(c[1]))
		{	neg = true; c++; }
		while(isdigit(*c))
		{
			if(str)
				*str++ = *c;
			num = 10 *num + *c++ -'0';
		}
		if(neg)
			num = -num;
		if(str)
				*str = 0;
		while(isspace(*c))
			c++;
		return c;
	}
	CTL_Inline bool isCommand(const char *start, const char *c)
	{
		if(!*c)
			return false;
		if(c!=start && !isspace(*(c-1)) )
			return false;
		return *c == '-' && !isdigit(c[1]);
	}

	CTL_Inline ctl::string location::getBraced(char *x, bool b)
	{
		char * bra=0, *ket=0;
		ctl::string braced;
		if(b) braced=x;
   	if( (bra=strchr(x,brasep)) )
		{
			ket = strrchr(x,ketsep);
			if(ket==bra || !ket)
				return braced;
			*ket = *bra = 0;
			while(* --ket==' ')
				*ket = 0;			
			while(isspace(*++bra) ) ;
			braced = bra;
		}
		return braced;
	}

	CTL_Inline void location::setArgs(const ctl::string &argStr)
	{
		typedef enum {cypher_, wd_, run_, log_, dbg_, host_, help_, type_, priority_, logicalId_,
									parent_, shell_, timeout_, verbose_, env_, terminal_, var_, port_} commandT;
		const char *c=argStr.c_str(), *start=c;
		bool readIP=false;
		for(unsigned int i=0; i<argStr.size(); i++)
		{
			readIP = false;
			char cmd=0, x[CTL_StringBufferSize];
			x[0] = 0;
			int num=0;
			while(isspace(*c))
				c++;
			if(isdigit(*c))
			{
				linkage = dmn;
				cmd = 'L';
			}
			else 	
			{
				if(getChannel(c)!=undefined)
				 	 	 	cmd = 'l';
				else
				{
					while(*c && !isCommand(start,c))
						c++;
					if(!*c++)
						break;
					cmd = *c++;
					if(cmd=='-')
					{
						c = readString(c, x, readIP);
//						std::cout<<"got x="<<x<<"\n";
						if(strcmp(x,"help")==0)
							cmd = 'H';
					}
					readIP = (cmd == 'h') || (cmd == 'e');
					while(isspace(*c))
						c++;
				}
			}
			if(!isCommand(start,c))
				c = readString(c, x, readIP);
			switch(cmd)
			{
				case 'c':
					cypher=true;
					break;
				case 'd':
					wd = (*x ? ctl::string(x) : ctl::string("."));
					args += "-d "; args += wd; args += " ";
					break;
				case 'D':
					fprintf(stderr, " Debug mode not yet implemented\n");
					break;
				case 'e':
					args += "-e "; runMode = 'e';
					if(isdigit(x[0]) )
						runMode = x[0];
					argString = getBraced(x, true);
					args += ctl::string("[") + argString+ "] ";
					break;
				case 'f':
					log = *x ? x : "ctl.log";
					args += "-f "; args += log; args += " ";
					break;
				case 'g':
					setDebug(*x ? ctl::string(x) : ctl::string());
					localArgs += "-g ";
					localArgs += debug;
					break;
				case 'h':
					host = *x ? x : "localhost";
					break;
				case 'H':
					runMode = 'H';
					break;
				case 'I':
					runMode = 'I';
					if(*x==':' && *(x+1)=='a')
						runMode = 'A';
					break;
				case 'l':
					if(*x)
					{
						char *s=0;
						linkage=getChannel(x);
						localArgs += "-l ";
						s = x;
						while(*s && !isspace(*s) ) s++;
						localArgs += x;
						localArgs += " ";
						if(linkage==dmn || linkage==tcp || linkage==logid || linkage == lib || linkage == thread)
						{
							readNum(s, num);
							if(num)
								portL = num;
						}
					}
					break;
				case 'n':
					c = readNum(c, num, x);
					args += "-n ";
					if(num)
					{
						priority=num;
						args += x; args += " ";
					}
					else
						priority = 1;											
					break;
				case 'p':
					c = readNum(c, num, x);
					if(num)
					{
						args += "-p "; args += x; args += " ";
						logId = num;
					}
					break;
				case 'P':
					if(*x)
						parent = getBraced(x, true);
					break;
				case 's':
					if(*x)
					{
						char *px = x+strlen(x);
						while(isspace(*--px) );
						if(isdigit(*px) )
						{
							while(isdigit(*--px));
							readNum(px+1, num);
							while(isspace(*px)) px--;
							*++px =0;
						}
						shell = getBraced(x, true);
					}
					else
						c = readNum(c, num);
					if(num>0)
						sPort = num;
					break;

				case 'S':
					c = readNum(c, num, x);
					if(num)
					{
						args += "-S "; args += x; args += " ";
						groupSize = num;
					}
					break;
				case 'T':
					timeout = 1.0; // millisec
					c = readNum(c, num, x);
					if(num)
					{
						args += "-T "; args += x; args += " ";
						timeout = num;
					}
					break;
				case 'v':
					args += "-v ";
					verbose = true;
					break;
				case 'V':
					args += "-V ";
					envFile = x;
					args += envFile + " ";
					break;
				case 'x':
					setTerminal(ctl::string(getBraced(x, true)) );
					break;
				case 'X':
					if(!!varStr)
						varStr += char(comma);
					varStr += getBraced(x, true);
					break;
				case 'L': case 'R':
				{
					int &port=cmd=='L' ? portL : portR;
					args += cmd =='L' ? "-L " : "-R ";
					c = readNum(c, num, x);
					if(num)
					{
						port=num;
						if(linkage == dmn)
							exec = x;
						else
						{ args += x; args += " "; }
					}
					else
						port = -1;
					break;
				}
			}
		}
		
//		ctl::log("'%s'\n", args.c_str());
	}

	CTL_Inline void location::print(std::ostream &os) const
	{
		const char* yes[] = {"no", "yes" };
		if(!!ciName)
			os<<"ci_name= '"<<ciName<<"'\n";
		os<<"exec= '"<<exec<<"'\n";
		os<<"path= '"<<path<<"'\n";
		os<<"user= '"<<user<<"'\n";
		os<<"host= '"<<host<<"'\n";
		os<<"shell= '"<<shell<<"'("<<sPort<<")\n";
		os<<"term= '"<<term<<"'\n";
		os<<"debug= '"<<debug<<"'\n";
		os<<"log= '"<<log<<"'\n";
		os<<"wd= '"<<wd<<"'\n";
		os<<"port= ("<<portL<<", "<<portR<<")\n";
		os<<"linkage= "<<ctl::location::channel2string(linkage) <<"\n";
		if(!!pipeTail) 
			os<<"piping'"<<pipeType<<"' '"<<pipeTail<<"'\n";
		os<<"cypher= " <<yes[cypher]<<"\n";
		os<<"verbose= "<<yes[verbose]<<"\n";
		os<<"logId= "<<logId<<"("<<groupSize<<")\n";
		os<<"priority= "<<priority<<"\n";
		os<<"error= '"<<errorStr<<"'\n";
		os<<"timeout= "<<timeout<<"\n";
		os<<"parent= '"<<parent<<"'\n";
		os<<"variables= '"<<varStr<<"'\n";
		os<<"args= '"<<args<<"'\n";
		os<<"localArgs= '"<<localArgs<<"'\n";
		os<<"string= '"<<c_str()<<" "<<"'\n";
	}	
		
	CTL_Inline ctl::vector<ctl::location> readLocation(const char *file)
	{
		ctl::vector<ctl::location> loc;
		FILE *fp = fopen(file,"r");
		if(!fp)
		{
			log("#w !! in readLocation: Could not open '%s' \n", file);
			return loc;
		}
		char line[CTL_StringBufferSize], *c;
		while( (c=readLine(line,CTL_StringBufferSize,fp)) )
		{	
			if(*c) loc.push_back(ctl::location(c)); 
//			std::cerr<<"'"<<c<<"'";
		}
		ctl::fclose(fp);
		return loc;
	}


	CTL_Inline ctl::string location::getSSH() const
	{ 
		ctl::string str;
		if(pipeType)
		{	str += pipeType; str +=' '; }
				
		str += user;
		if(!str.empty())
			str += "@";
		if(linkage !=thread && linkage !=lib)
		{
			if(host.empty())
				str += "localhost";
			else
				str += host;
			str += ":";
		}
		if(path.empty())
			str += "./";
		else
			str += path+ "/";
		str += exec;
	
		str += " ";
		str += localArgs;
		str += " ";
		str += args;
		if(!!pipeTail)
			str+= " " + pipeTail;
//		str += channel2string(linkage);
		return str;
	}

	CTL_Inline bool refCounter::put(uint4 v)
	{
		const int lg=5, base=(1<<lg);
//		ctl::log("this=%p,put: v = %d\n",this, v);
		if(!v--)
			return --creditM==0;
		uint4 p=v>>lg, b = (1<<(base-1-(v & (base-1))));
		if(p+1>valueM.size())
			valueM.append(0);
		if(valueM[p]<=~b)
		{
			valueM[p] += b;
			return false;
		}
		valueM[p] -= ~b + 1;
		while(p && !~valueM[--p])
			valueM[p] = 0;
		if(p==0)
			return --creditM==0;
		valueM[p]++;
		return false;

	}

	CTL_Inline void Remote::putCredit(uint4 v) const
	{
		CTL_Assert(guardedM);
		if(creditM.put(v))
			releaseObject(RemoteSP(this) );
	}

inline char *deleteSpace(char *dest, char *end, const char *src, bool ket=false)
{
	if(!src)
		return dest;
	while(*src)
	{
		if(dest+10>end)
			throwStr("");
		while(isspace(*src))
			src++;
		if(*src == '<' || *src == '(' || *src == ',')
		{
			*dest++ = *src++;
			while(isspace(*src))
					src++;
			if(src[0] == 'c' && src[1] == 'o' && src[2] == 'n' && src[3] =='s' && src[4] == 't' && isspace(src[5]) )
			{ *dest++ = 'c'; *dest++ = 'o'; *dest++ = 'n'; *dest++ = 's'; *dest++ = 't'; *dest++ = ' ';  src += 6; }
		}
		if(!ket || *src!=')')
			*dest++ = *src;
		src++;
	}
	return dest;
}

CTL_Inline typeId prettyName(const char *y, const char *scope, const char *f, const char *param, const char *args, const char *args_)
{
//CTL_Profile("prettyName");
	char pretty[CTL_StringBufferSize];
	char *c=pretty, *end = pretty + CTL_StringBufferSize;
	if(y && y[0]=='(')
		y++;
	if(y && (y[0] != 'C' || y[1] != 'T' || y[2] != 'L'  || y[3] != '_' || y[4] != 'R' || y[5] != 'P') )
	{	c = deleteSpace(c, end, y, true);	*c++ = ' '; }
	c = deleteSpace(c, end, scope);
	if(f)
	{
		*c++ = ':';	*c++ = ':';
		while(*f)
			*c++ = *f++;	
	}
	if(param)
		c=deleteSpace(c, end, param);
	c = deleteSpace(c, end, args);

	if(args_)
	{
		const char *cn= args_+ strlen(args_) - 5;
		if(cn[0] == 'c' && cn[1] == 'o' && cn[2] == 'n' && cn[3] =='s' && cn[4] == 't' )
		{ *c++ = ' '; *c++ = 'c'; *c++ = 'o'; *c++ = 'n'; *c++ = 's'; *c++ = 't'; }
	}

	*c++=0;
	return typeId(pretty);
	/*
	char *name = new char[c-pretty];
	memcpy(name, pretty,c-pretty);

	return name;
	*/
}

CTL_Inline const char *prettyNameS(const char *y, const char *scope, const char *f, const char *param, const char *args, const char *args_)
{
	typeId pretty=prettyName(y,scope,f,param,args,args_);
	// locate prettyName c-string on heap
	char *name=new char[pretty.size()];
	memcpy(name, pretty.c_str(), pretty.size());
	return name;
}
//{ return createStaticParent<typeId>(prettyName(y,scope,f,param,args,args_)).c_str(); }
//{ return createStatic<typeId>(prettyName(y,scope,f,param,args,args_)).c_str(); }


} // ctl

#endif
