/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_SYSTEM_INL_
#define CTL_SYSTEM_INL_

#include <fstream>
#ifndef CTL_MSC_HEADER
#include <dlfcn.h>
#include <unistd.h>
#include <sys/socket.h>    
#include <netinet/tcp.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pthread.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <signal.h>
#include <pwd.h>
#ifdef CTL_OSX
	#include <sys/wait.h>
#else
	#include <wait.h>
#endif
#include <sys/resource.h>
#ifdef _POSIX_PRIORITY_SCHEDULING
#	define CTL_SchedYield
#endif
#else

#	include <Winsock2.h>
#	include <direct.h>
#	include <process.h>
#	include <io.h>

#endif

#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>


namespace ctl
{
	namespace sys
	{
		CTL_Inline const char *getLocatorName()
		{
#ifdef CTL_Locate
			return CTL_Str(CTL_Locate);
#else
			return 0;
#endif
		}

 	CTL_Inline int getPid()
	{ return ::getpid(); }
	CTL_Inline const char *getEnv(const char *name)
	{
		const char *var = getenv(name);
		return var && *var ? var : 0;;
	}
	CTL_Inline ctl::string getHostIP()
	{
		char buf[255];
		::gethostname(buf, 254);
//		printf("%s\n", buf);
		struct hostent *hp = gethostbyname(buf), *hp2;
		if(!hp)
			return ctl::string();
#ifndef CTL_SGI
		hp2 = gethostbyname(hp->h_name);
		if(hp2)
			hp = hp2;
#endif
		ctl::string ip0;
		char **addrList=hp->h_addr_list;
		if(!addrList)
			return ip0; 
//		printf("%s\n", hp->h_name);
		if(addrList[0])
			ip0 = inet_ntoa(*(struct in_addr *) *addrList);
		while (*addrList)
		{
			ctl::string ip(inet_ntoa(*(struct in_addr *) * addrList++) );
//			printf("%s\n", ip.c_str());
			if(!ctl::location::isLocalHostName(ip) )
				return ip;
		}
		return ip0;
	}
#ifdef CTL_PUTENV
	CTL_Inline void setEnv(const char *name, const char *var)
	{
		char *str= new char[strlen(var)+strlen(name)+2];
		sprintf(str,"%s=%s", name,var);
		(void)putenv(str);
	}
	CTL_Inline void unsetEnv(const char *name)
    {
			if(getEnv(name))
				setEnv(name,""); 	
	}
#else
		CTL_Inline void setEnv(const char *name, const char *var)
		{ ::setenv(name,var,1); }
		CTL_Inline void unsetEnv(const char *name)
    { ::unsetenv(name); }
#endif
	CTL_Inline ctl::string absolutePath(ctl::string mypath,const ctl::string &path)
	{
		bool rel = !path || path[0] != '~';
#ifdef CTL_WinPath
		if(rel && path.size()>1 && path[1]!=':' && path[0]!='/')
#else
		if(rel && !!path && path[0]!='/')
#endif
		{
			int up=0, lst= (int)path.size()-1;
			for(int i=0; i<lst; i+=3)
			{
				if(path[i]!='.' || path[i+1]!='.' )
					break;
				up = i+2;
				if(up<=lst && path[up]=='/')
					up++;
				size_t size=mypath.size();
				if(size>1)
				{
					int pos = int(size)-2;
					if(mypath[pos]=='.')
						pos--;
					while(pos>=0 && mypath[pos]=='/')
						pos--;
					mypath=mypath.substr(0,pos+1);
				}
				int p=(int)mypath.rfind('/');
				while(p>=2 && mypath[p-1]=='.' && mypath[p-2]=='/')
					p -= 2;
				mypath = mypath.substr(0,p);
			}			
			return mypath+"/"+path.substr(up);
		}
		return path;
	}


	CTL_Inline const char* getcwd(char *path,int len)
	{
		const char* cwd_err = ::getcwd(path, len);
		(void)cwd_err;
		location::substSlash(path);
		return path; 
	}
	CTL_Inline bool _copyFile(const ctl::string& src, const ctl::string& dest)
	{
		std::ifstream is(src.c_str(), std::ios::binary);
		if(!is.good())
		{
			ctl::log("#d !! _copyFile: could not open source file '%s'\n", src.c_str());
			return false;
		}
		std::ofstream os(dest.c_str(), std::ios::binary);
		if(!os.good())
		{
			ctl::log("#d !! _copyFile: could not open destination file '%s'\n", dest.c_str());
			return false;
		}
		os << is.rdbuf();
		return os.good();
	}



#ifdef CTL_Win32

	CTL_Inline ctl::string copyFile(const ctl::string& src, const ctl::string& dest)
	{
		if(_copyFile(src, dest))
			return dest;
		return ctl::string();
	}
	CTL_Inline void setPriority(int prior)
	{
// IDLE_PRIORITY_CLASS NORMAL_PRIORITY_CLASS HIGH_PRIORITY_CLASS REALTIME_PRIORITY_CLASS
		HANDLE handle=GetCurrentProcess();
		if(prior>0)
			SetPriorityClass(handle,IDLE_PRIORITY_CLASS);
	}
	  CTL_Inline threadIdT getsTid()
		{ return GetCurrentThreadId(); }
	  CTL_Inline threadIdT getTid()
		{ return getsTid()^(threadIdT)(getPid()<<7); }
		CTL_Inline void wakeUp(threadIdT /*pid*/)
		{ ::Sleep(0); }
		CTL_Inline int fcntl(int,int,int)
		{ return -1; }

		CTL_Inline int close(int s, bool isSock =true)
		{ 
			if(isSock)
				return closesocket(s);
			else
				return ::close(s);
		}
		inline iovec ioEntry(char *dat, unsigned long b)
		{
			iovec io= { b, dat };
			return io;
		}
		CTL_Inline int write(int s, const void *buf, size_t n, bool isSock=true)
		{ 
			if(isSock)
				return ::send(s, (const char*)buf, int(n), 0); 
			else
				return ::write(s,buf,n);
		}

		CTL_Inline int writev(int s, const iovec *v, int n, bool isSock=true)
		{ 
			int w=0;
			if(isSock)
				WSASend (s, (iovec*)v, n,  (LPDWORD) &w, 0, NULL, NULL);
			else
				for(int i=0; i<n; i++)
					w+=::write(s, v[i].buf, v[i].len);
			return w;
		}

		CTL_Inline int read(int s, void *buf, size_t n, bool isSock=true)
		{ 
			if(isSock)
				return ::recv(s, (char*)buf, int(n), 0); 
			else
				return ::read(s,buf,n);
		}

		CTL_Inline HMODULE dlopen(const char *lib, bool)
		{
			HMODULE so= LoadLibrary(lib);
			return so;
//			return (UINT)so < (UINT)HINSTANCE_ERROR  ? 0 : so;
		}
		CTL_Inline void *dlsym(HMODULE handle, const char *fname)
		{ return (void*)GetProcAddress(handle, LPCSTR(fname)); }
		CTL_Inline void dlclose(HMODULE handle)
		{ FreeLibrary(handle); }
		CTL_Inline const char* dlerror()
		{ static char err[8]; itoa(int(GetLastError()),err, 10); return err; }
		CTL_Inline ctl::string getDOScwd(const ctl::string &path)
		{ 
			char buf[CTL_StringBufferSize];
			buf[CTL_StringBufferSize-1] = 0;
			strncpy(buf, path.c_str(),CTL_StringBufferSize-1);
			char *s=buf,*p=0;
			ctl::string dosPath;
			while(p=strchr(s,'/') )
			{
				*p=0;
				dosPath += ctl::string(s) + "\\";
				s = p+1;
			}
			return dosPath += s;
		}

		CTL_Inline int system(const char *cmd, const ctl::string &path)
		{ 
			ctl::string cmdStr = ctl::string("cd ") + getDOScwd(path) + " & " + cmd;
			return ::system(cmdStr.c_str()); 
		}
		CTL_Inline int spawn(const char *cmd, std::pair<FILE*, int> *pServOut=0, bool=false)
		{
			if(pServOut)
			{
				pServOut->first = pServOut->second <0 ? _popen(cmd,"wt") : _popen(cmd,"rt");
				return pServOut->first!=0;
			}
			char prg[CTL_StringBufferSize], *p = prg;
			prg[CTL_StringBufferSize-1] = 0;
//			memcpy(prg, cmd, strlen(cmd)+1);
			strncpy(prg, cmd, CTL_StringBufferSize-1);
			while(isspace(*p))
				p++;
			char *sp=strchr(p,'#');
			if(sp)
				*sp = '\0';
			const char *args[3]={ p,sp+1,0 };
			while( sp=strchr(sp+1,'#') )
				*sp = 0;
			int ret = (int)_spawnvp(_P_NOWAIT,p,args);
			return ret<0 ? 0 : ret;
		} 
		typedef unsigned int threadType;

		CTL_Inline void usleep(double usec)
		{::Sleep(int(usec*1e-6) ); }
		CTL_Inline int mkfifo(const char * /*name*/, int /*mode*/ =0)
		{ return 0; }
// 		threadType CTL_Inline startThread(void (__cdecl *run) (void *), void *arg)
//		{ threadType h= _beginthread(run, 0, arg); return GetThreadId((HANDLE)h); }
#define CTL_runDclr		sys::threadResult __stdcall
		threadIdT CTL_Inline startThread(threadType (__stdcall *run) (void *), void *arg)
		{ 
			unsigned int thr=0; 
			int ok =_beginthreadex(0,0,run, arg, 0, &thr);

			return ok<0 ? 0 : threadIdT(thr);
		}
		CTL_Inline int getCommandArgs(char** /*arg*/, int /*maxArg*/)
		{ return -1; }


// not win32
#else
#	ifdef CTL_XtermAvailable
// xterm based device
class Xterm: public fileNoDevice
{
  int client, serv;
	pid_t child_pid;
	char error[CTL_StringBufferSize];
	void setError(const char *m)
	{ 
		sprintf(error,"%s: %s\n", m, ::strerror(errno) );
		fprintf(stderr,"%s: %s\n", m, ::strerror(errno) );
	}
public:
	const char* getError() const
	{ return error[0] ? error : 0; }
	Xterm(const char *title, const char* term):child_pid(0)
	{ open(title, term); }
	static char* eatSpaces(char* s)
	{
		char *o = s+strlen(s);
		while(*s && isspace(*s))
			s++;
		while(s<=o && isspace(*--o))
			*o = 0;
		return s;
	}
	void open(const char *title, const char* term)
	{
		const char *xterm="xterm";
	  char *servName=0;
		error[0] = 0;
		if((client = ::open("/dev/ptmx", O_RDWR)) < 0) 
			return;
		if(::grantpt(client) || ::unlockpt(client) || !(servName = (char *)::ptsname(client)) ) 
		{
 		   setError("failed to open client device");
 		   ::close(client);
   		 return;
  	}
 		if ((serv = ::open(servName, O_RDWR)) == -1) 
		{ setError("failed to open serv"); return; }
  	child_pid = fork();
  	if (child_pid == -1) 
		{ setError("failed to fork"); return; }
  	if (child_pid == 0) 
		{
   		::close(serv);
     	char arg1[80];
     	sprintf(arg1, "-Sz/%d", client);
			vector<const char*> args(2);
			vector<vector<char> > xargs, xargs1;
			args[0] = title;
			args[1] = arg1;
//#define CTL_Terminal xterm -geometry 40x40 -fg red
			::printf("LINE = %d",__LINE__);
			if(strncmp(CTL_Str(CTL_Terminal),"xterm",5)==0)
			{	
				xargs = split(string(CTL_Str(CTL_Terminal)), ' ', '[',  ']');
				int nargs= xargs.size() ;
				for(int i=0; i<nargs; i++ )
					args.push_back(eatSpaces(xargs[i].begin() ) );
			}
			if(strncmp(term, "xterm", 5)==0 && strlen(term)>6)
			{
				xargs1 = split(string(term+6), ' ', '[',  ']');
				int nargs= xargs1.size();//keep out the -e option at the end
				for(int i=0; i<nargs; i++ )
					if(!!xargs1[i] && !!xargs1[i][0] && !isspace(xargs1[i][0]) &&
								!(xargs1[i].size()>1 && xargs1[i][0]=='-' && xargs1[i][1]=='e'))
						args.push_back(eatSpaces(xargs1[i].begin() ) );
			}

//			args.push_back("-bg");
//			args.push_back("yellow");


			//			std::cerr<<args<<"\n";
			args.push_back(0);

 			execvp(xterm, (char*const*)args.begin());
     	exit(1);
   	}
  	::close(client);

 		for(char c=0; c!='\n';)
			if(::read(serv, &c, 1)<1)
			{ setError("failed to read from xterm"); return; }
  	inM = dup(serv);
  	outM = dup(serv);
  	::close(serv);
		write("\033[1;1H           \033[1;1H",23);

// using pthread effects clients stdio
//		dup2(outM,1);
//		dup2(inM,0);

	}
	~Xterm()
	{ close(); }
	void close()
	{
		if(child_pid)
		{
 			kill(child_pid, SIGKILL);
 			wait(0);
		}
		if(!getError())
			::close(serv);
	}
	int write(const char *m, size_t count)
	{ 
		int s = ::write(outM, m, count); 
		if(s<0)
   	 setError("failed to write to serv");
		return s;

	} 
	int read(char *m, size_t count)
	{ 
		int s = ::read(inM, m, count); 
		if(s<0)
   	 setError("failed to read from serv");
		return s;
	}
};
#endif
	CTL_Inline void setPriority(int prior)
	{
		if(prior>0)
			setpriority(PRIO_PROCESS,0,10);
	}
		CTL_Inline void usleep(double usec)
		{
#ifdef CTL_SchedYield
			if(!usec)
			{
				::sched_yield();
				return;
			}
#endif
			timespec t;
 			t.tv_sec=int(1e-6*usec);;
			t.tv_nsec = long(1e3*int(usec-double(t.tv_sec)*1e6));
     	int ok =	::nanosleep(&t,0);
//if(ok<0)
//	fprintf(stderr,"waked up (%e) \n", usec); 
			(void)ok;
		}

	  CTL_Inline threadIdT getsTid()
		{ return (threadIdT)pthread_self(); }
	  CTL_Inline threadIdT getTid()
		{ return (threadIdT)getsTid()^(threadIdT)(getpid()<<7); }
		CTL_Inline void wakeUp(threadIdT thid)
		{	
//fprintf(stderr,"calling pthread_kill %LX %LX\n",thid,pthread_self() );
#ifdef CTL_CONT
			pthread_kill((pthread_t)thid, SIGCONT);
#endif
//fprintf(stderr,"called pthread_kill %LX %LX\n",thid,pthread_self() );
		}

		CTL_Inline int fcntl(int fd,int s, int flag)
		{ return ::fcntl(fd,s,flag); }
		CTL_Inline int close(int fd, bool /*isSock*/ =true)
		{ return ::close(fd); }
		CTL_Inline int write(int fd, const void *buf, size_t s, bool /*isSock*/ =true)
		{ return  ::write(fd,buf,s);; }
		inline iovec ioEntry(char *dat, uint8 b)
		{
			iovec io= { dat, b };
			return io;
		}
		CTL_Inline int writev(int fd,const  iovec *vec, int s, bool /*isSock*/ =true)
		{ return ::writev(fd,vec,s); }

		CTL_Inline int read(int fd, void *buf, size_t s, bool /*isSock*/ =true)
		{ return  ::read(fd,buf,s); }
		CTL_Inline ctl::string tmpnam(const ctl::string& prefix)
		{
			char buf[L_tmpnam];
			::tmpnam(buf);
			char* pos = strrchr(buf, '/');
			if(pos)
			{
				*pos = 0;
				return ctl::string(buf) + '/' + prefix + (pos+1);
			}
			return prefix + buf;

		}
		CTL_Inline ctl::string copyFile(const ctl::string& src, const ctl::string& dest_)
		{
			ctl::string dest=dest_;
			if(dest.empty())
				dest = tmpnam("ctltmp_");
			if(_copyFile(src, dest))
				return dest;
			return ctl::string();
		}
		CTL_Inline void *dlopen(const char *lib, bool local_binding=false)
#ifdef CTL_NO_RTLD_DEEPBIND
		{ return ::dlopen(lib,  RTLD_LAZY | (local_binding?(RTLD_LOCAL):RTLD_GLOBAL)); }
#else
		{ return ::dlopen(lib,  RTLD_LAZY | (local_binding?(RTLD_LOCAL|RTLD_DEEPBIND):RTLD_GLOBAL)); }
#endif
		CTL_Inline void *dlsym(void *handle, const char *sym)
		{ return ::dlsym(handle,sym); }
		CTL_Inline void dlclose(void *handle)
		{ ::dlclose(handle); }
		CTL_Inline const char* dlerror()
		{ return ::dlerror(); }
		CTL_Inline int popen2(const char *cmd, int &fdin, int &fdout)
		{
			int fdI[2], fdO[2];
			int pipe_err=::pipe(fdI);
			pipe_err = ::pipe(fdO);
			int pid = fork();
			if(pid==-1)
				return -1;
			if (pid == 0)
			{
				close(fdI[0]);
				close(fdO[1]);
				dup2(fdI[1],1);
				dup2(fdO[0],0);
				char cmd[128];
				read(fdO[0], cmd, 128);

				const char *const argv[4] = { "sh", "-c", (char*)cmd,  0};
				execve("/bin/sh", (char*const*)argv,  0);//environ);

				exit(127);
			}
			close(fdI[1]);
			close(fdO[0]);
			fdin = fdI[0];
			fdout = fdO[1];
			return 0;
		}
		CTL_Inline int system(const char *cmd, const ctl::string &path)
		{ 
			ctl::string cmdStr = ctl::string("cd ") + path + "; " + cmd;
			int res = ::system(cmdStr.c_str()); 
			std::cerr<<"system returned" << res<<"\n";
//			if(!WIFEXITED(res) )
			return res; 
		}

		CTL_Inline int spawn(const char *cmd, std::pair<FILE*, int> *pServOut=0, bool self = false )
		{
			if (cmd == 0 || (self && pServOut))
				return 1;
			if(!self && pServOut)
			{
				pServOut->first = pServOut->second < 0 ? popen(cmd,"w") : popen(cmd,"r");
//				std::cerr<<"popen"<<pServOut->first<<"\n";
				return pServOut->first!=0;
			}
			
			if(self)
			{
				Environment *penv = &ctl::env();
				void (*conn)() = penv->getConnector();
				int pid = fork();
				if(pid == -1)
					return -1;
				if (pid == 0)
				{
						void *venv=0;
						const char *arg[] = { "'forked'", cmd, 0 };
						int res = ctl::openMain(2, (char**)arg, &venv,  penv, sys::getLocatorName, conn);
//						ctl::closeLibrary(venv, true);
						exit(res);
				}
				else
					penv->ownedProcessIdM = pid;
				return 1;
			}
			ctl::string str = cmd;
			str += " &";
			return ::system(str.c_str())==0;
		}

#define CTL_runDclr		sys::threadResult
		CTL_Inline threadIdT startThread(void *( *run) (void *), void *arg)
		{
			pthread_t thread=0;
			int err=pthread_create(&thread, 0, run, arg);
			if(!err)
				pthread_detach(thread);
			return err ? 0 : threadIdT(thread);
		}
		CTL_Inline int mkfifo(const char *name, int mode =0)
		{ return ::mkfifo((char*)name, mode); }

		CTL_Inline int getCommandArgs(char **arg, int maxArg)
		{
			enum { maxS=1024 };
			FILE *fp = ::fopen("/proc/self/cmdline", "r");
			if(!fp)
				return 0;
			char buf[maxS], ch;
			int pos=0, cnt=0;	

			while ((ch = fgetc(fp)) != EOF && cnt<maxArg-1)
			{
				if ((ch == '\0') || (pos > maxS-1))
				{
					buf[pos] = '\0';
					int sz = strlen(buf)+1;
					arg[cnt] = new char[sz];
					strncpy(arg[cnt], buf, sz);
					arg[++cnt] = 0; 	pos = 0;
					continue;
				}
				buf[pos++] = ch;
			}
			::fclose(fp);
			return cnt;
		}

#endif
		device *createDev(char type, const location& loc, int &fdin, int &fdout)
		{
			fdin = 0; fdout = 1;
			switch(type)
			{
				case 'x': case 'X':
#ifdef CTL_XtermAvailable
				{ fileNoDevice *X = new Xterm(loc.exec.c_str(), loc.term.c_str()); fdin=X->getIn(), fdout = X->getOut(); return X;}
#else
			(void)loc;
#endif
				case 'n': case 'N':
				{ return new fileNoDevice(); }
			}
			return 0;	
		}
	} // sys
} // ctl

#endif

