/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_RESULT_H_
#define CTL_RESULT_H_

namespace ctl
{

class refCounter
{
	uint4 creditM;
	ctl::vector<uint4> valueM;
public:
	refCounter(): creditM(0) {}
	void get()
	{	creditM++;  }
	bool put(uint4 v);
	bool empty() const
	{ return creditM == 0; }
};

} // ctl

#include "link.h"

namespace ctl
{

/**/// base class for all remote callables
class Remote: public ctl::shared
{
friend class Environment;
protected:	

	mutable refCounter creditM;
public:
	bool guardedM;

	using shared::getSharedP;
	Environment *envM;
	mutable ctl::link linkM;
protected:
	Remote(bool guarded=true, const ctl::link &p=ctl::link(), bool getenv = true): guardedM(guarded),  envM(getenv ? &ctl::env() : 0), linkM(p)
	{}
public:
	virtual ~Remote() {}
// def. -> remote.inl

  void putCredit(uint4 v) const;
	void getCredit() const
	{ 
		if(creditM.empty() )
			ctl::addObject(envM, this);
		creditM.get();	
	}
//	bool emptyCredit() const
//	{ return creditM.empty(); }
	virtual typeId CTL_typeName() const { return "Remote"; }
	virtual const ctl::link &checkRP(const MID &, objectID &, PID &) const
	{ return linkM; }
	const CommunicatorSP &getCommunicator() const
	{ return linkM.getCommunicator(); }
	const ctl::link &getLink() const
	{ return linkM; }
	virtual status receive(ctl::iStream &) { CTL_Here; return success; }
	virtual const void *getCIP() const
	{ return 0; }
	virtual const void *extends(const typeId&)  const
	{ return 0; }
};

namespace internal
{
	struct creditT: public ctl::shared
	{
		mutable ctl::uint4 refVM;
		bool guardedM;
		objectID targetM;
		ctl::link linkM;
		ctl::typeId CInameM;
		PID rootIdM, threadIdM;
		creditT(ctl::uint4 refV, bool guarded, objectID obj, const link &p, const typeId &CIname,
					const PID &rootId, const PID &threadId):
				refVM(refV), guardedM(guarded),targetM(obj), linkM(p), CInameM(CIname), 
				rootIdM(rootId), threadIdM(threadId)
		{
			if(guardedM)
				ctl::log("#d -> credit for %s(%p) = %d\n", CInameM.c_str(), this, refVM);
		}
		~creditT()
		{
			if(guardedM)
			{
				ctl::log("#d <- credit for %s(%p) = %d\n", CInameM.c_str(), this, refVM);
				char buf[CTL_defaultBufferSize];
				if(!!linkM && targetM)
					linkM.writeMessage(ctl::serializeB(targetM, MID(0), refVM, buf, CTL_defaultBufferSize, linkM.getCommunicator() ) );
			}
		}

	};


template<class CI> struct rReference :public CI::CTL_RemoteBase
{
	typedef typename CI::CTL_RemoteBase baseT;
	using baseT::guardedM;
	using baseT::envM;
	typedef rReference<CI> thisT;
	cPointer<creditT>  creditM;
	objectID getObject() const
	{ return creditM->targetM; } 
	void incrCredit() const
	{ creditM->refVM++ ; }
	void decrCredit() const
	{ creditM->refVM++ ; }
	ctl::uint4 getCredit() const
	{ return creditM->refVM ; }
	const creditT *getCreditP() const
	{ return creditM.get(); }
	const ctl::link &checkRP(const MID &mid, objectID &trgt, PID &pid) const
	{
		pid = link().getPid();
		if(!pid)
			ctl::throwStr(" !! in class %s: probably method %d not implemented\n",CI::CTL_typeNameS().c_str(), int(mid));
		trgt = getObject();
		return link();
	}
	bool isLocal() const
	{ return false;}
	ctl::Remote *getRemoteP() const
	{ return (thisT*)this; }
	const void *getCIP() const
	{ return (const baseT*)this; } 

	bool residence(PID &rootId, PID &threadId) const
	{ rootId = creditM->rootIdM; threadId = creditM->threadIdM; return true; }

	rReference (const objectID &obj, const ctl::link &p, ctl::uint4 refV, const ctl::typeId &fullName, const PID &rootId, const PID &threadId):creditM(new creditT(refV, guardedM, obj, p, fullName, rootId, threadId))
	{
		link() = p;
		link().derelease();
		if(guardedM)
		{ holdReference(envM,this); }
	}
	rReference (const cPointer<creditT> &credit): creditM(credit)
	{
		link() = creditM->linkM;
		if(guardedM)
		{ holdReference(envM,this); }
	}
	~rReference()
	{	
		if(guardedM)
		{ releaseReference(envM,this); }
	}
// reimplement virtual Remote::methods
	const ctl::CommunicatorSP &getCommunicator() const
	{ return  ((const Remote*)this)->getCommunicator(); }
	ctl::link &link() const
	{ return static_cast<const Remote*>(this)->linkM; }
	void setName(const ctl::string &)
	{}
	ctl::typeId fullTypeName() const
	{ return  creditM->CInameM; }
	ctl::shared *getSharedP() const
	{ return static_cast<const Remote*>(this)->getSharedP(); }
	status receive(ctl::iStream &) { return failed; }
 };

} // internal

/**/// rPointer refers either to a rReference<CI> ( <==>  residence()!=Remote::localPID ) or to an CI
template<class CI> class rPointer: public ctl::cPointer<typename CI::CTL_Interface>
{
	typedef typename CI::CTL_Interface pointedT;
	typedef cPointer<pointedT> baseT;
	using baseT::release;
	using baseT::holdM;
	typedef internal::rReference<CI> refT;
public:
	using baseT::getRefCount;
	CTL_Template(ctl::rPointer, (CI), 1, isReference, (pointedT), 1);
	Remote* getRemoteP() const
	{ return get() ? get()->getRemoteP() : 0; }
	oStream &write (oStream &os) const
	{
		PID rootId, threadId;
		const pointedT *p=get();
// ctl::typeId() for later usage
		if(!p)
			return os<<rootId << threadId << ctl::objectID()<<ctl::typeId()<<ctl::uint4();
		bool remote=p->residence(rootId, threadId);
//		ctl::log("#m: host = (%d %s)\n", threadId.port,threadId.host.c_str() );
		typeId fullName = p ? p->fullTypeName() : ctl::typeId();
		if(!remote)   // <==> local implementation
		{
			ctl::uint4 refC =0;
			ctl::Remote *r=getRemoteP();
			if(os.writing() && r && r->guardedM)
			{
				r->getCredit();
//				ctl::addObject(r, p->fullTypeName().c_str());
			}
			return os << rootId << threadId << ctl::objectID(r)<<fullName<<refC;
		}
		const refT *ro=static_cast<const refT*>(p);
		if(os.counting())
			ro->incrCredit();	
//		if(os.commuM!=ro->getCommunicator())
//			pid = ro->getCommunicator()->getPhysId(pid);
//		ctl::log("W: host = %d %s\n", pid.port,pid.host.c_str() );
		return os << rootId << threadId << ro->getObject()<<fullName<<ro->getCredit();
	}
	ctl::iStream &read(ctl::iStream &is)
	{
		Environment *e = &env();
		PID rootId, threadId;
		ctl::objectID trgt;
		ctl::uint4 refV;
		ctl::typeId fullName;

		is >> rootId >> threadId >> trgt >> fullName >> refV;
		if(!trgt)
		{ *this = baseT(); return is; }	
		if(rootId==getRootId(e) ) // home again
		{
			ctl::Remote *target = trgt;
			if(target)
				target->putCredit(refV);
			*this = baseT(target ? static_cast<const pointedT*>(target->getCIP()) : 0 );
		}
		else
		{
			CommunicatorSP commu; commu = is.commuM;
			if(!threadId) // short for resultReceiver
			{
				*this = ctl::rPointer<CI>(trgt, ctl::link(commu, is.sourceM), refV, fullName, PID(), PID());
			}
			else
			{
				if(refV || rootId != threadId)	
					*this = ctl::rPointer<CI>(trgt, getRemoteLink(e, threadId), refV, fullName, rootId, threadId);
				else
					*this = ctl::rPointer<CI>(trgt, ctl::link(commu, is.sourceM), refV, fullName, rootId, threadId);
			}
		}
		return is;
	}
	rPointer(const baseT &p = baseT()): baseT(p) {}
	rPointer(const RemoteSP &p):baseT(static_cast<const pointedT*>(p->extends(CI::CTL_typeNameS()) ) ) {}
	rPointer(const ctl::rPointer<CI> &rp): baseT(rp) {}
	rPointer(const objectID &target, const ctl::link &p, ctl::uint4 refV, const ctl::typeId &fullName, const PID &rootId, const PID & threadId):
			baseT(new internal::rReference<CI> (target, p, refV, fullName, rootId, threadId)) {}
	~rPointer()
	{ 
		if(getRefCount()==1 && holdM) 
		{
			const ctl::link &p = get()->link();
			{
				ctl::link::envSet tmp(&p);
				get()->CTL_delete(p); 
			}
			release(); 
		} 
	}
	pointedT *get()
	{ return (pointedT*)((baseT*)this)->get(); }
	const pointedT *get() const
	{ return ((const baseT*)this)->get(); }
	bool operator!() const
	{	return !(const baseT&)*this || ( !get()->isLocal() && !get()->link() );
	}
};

namespace internal
{

struct resReceiverCI : public Remote
{
enum { extends = true };
	typedef resReceiverCI thisT;
	typedef thisT CTL_Interface;
	typedef thisT CTL_RemoteBase;
	using Remote::linkM;
	static ctl::typeId CTL_typeNameS()
	{ return "resReceiverCI<>"; }
	ctl::typeId fullTypeName () const
	{ return CTL_typeNameS(); }
	Remote*  getRemoteP () const
    { return (Remote*)this; }
	resReceiverCI(ctl::Flag):Remote(false/*= unguarded*/)
	{}
	resReceiverCI(const ctl::link &p=ctl::link() ): Remote(false/*= unguarded*/, p, false /*no env*/)
	{}// ((shared*)this)->ref();}
	template<class Args, class Y>  void send(const ctl::binaryObject &excpt, const Y &y, const Args &x, ioTag tag)
	{
		objectID trgt=0;
		PID pid=-1;
		ctl::link p = checkRP(MID(), trgt, pid);
		if(!p)
			return;
		p.setTag(tag);
		char buf[CTL_defaultBufferSize];
		CTL_Try
		{ p.writeMessage(ctl::serializeB(trgt,excpt,x,y,buf,CTL_defaultBufferSize,p.getCommunicator()) ); }
		CTL_Catch(const ctl::exception &e)
		{ p.writeMessage(ctl::serializeB(trgt, binaryObject(e), buf, CTL_defaultBufferSize, p.getCommunicator()) ); }
	}
	bool residence(PID &, PID &) const
	{ return false; }
	void CTL_delete(const ctl::link &)
	{ delete this;}
	ctl::link &link() const
	{ return static_cast<const Remote*>(this)->linkM; }
};

template<class Y> struct resultRecvBase: public  resReceiverCI
{
	typedef typename ctl::returnType<Y >::resultT valueT;
	using resReceiverCI::linkM;
	mutable  bool receivedM, emptyArgM, multiReadingM;
	resultRecvBase(const ctl::link &p = ctl::link()): resReceiverCI(p),receivedM(false), multiReadingM(false) {}
	ioTag getTag() const
	{ return linkM.getTag(); }
	bool emptyArg() const { return  emptyArgM; }
	virtual bool received(bool /*blocking*/, valueT &, cPointer<excptHandle>&) const
	{ return  receivedM; }
	~resultRecvBase()
	{ /*ctl::log("resultRecvBase<%s>\n", typeName<Y>().string().c_str() );*/ }
	virtual void setLinks() const=0;
};



template<class Z, class Y> struct resCaster: public resultRecvBase<Z>
{
	using resultRecvBase<Z>::emptyArgM;
	using resultRecvBase<Z>::receivedM;
	using resultRecvBase<Z>::linkM;
	typedef  typename resultRecvBase<Z>::valueT  Zvalue;
	typedef  typename resultRecvBase<Y>::valueT  Yvalue;
	typedef ctl::cPointer<internal::resultRecvBase<Y> > resYT;
	mutable resYT resM;
	mutable Yvalue tmpyM;
	resCaster(const resYT &res): resultRecvBase<Z>(res->getLink() ), resM(res)
	{	emptyArgM = res->emptyArg(); }
	bool received(bool blocking, Zvalue &z, cPointer<excptHandle>& excpt) const
	{
		if(!receivedM && resM->received(blocking, tmpyM, excpt) )
		{
			if(!excpt)
				z = binaryCast<Z>(tmpyM);
			receivedM = true;
		}
		return receivedM;
	}
	void setLinks() const
	{ if(!!resM) resM->setLinks(); }
};

template<class Y> inline void setLink(const Y &, const ctl::link &, ctl::falseT)
{}
template<class CI> inline void setLink(const CI &ci, const ctl::link &p, ctl::trueT)
{	((const typename CI::thisT&)ci).CTL_setLink(p); }

class setLinkApplier
{
	const ctl::link &p;
	setLinkApplier& operator=(const setLinkApplier&);
public:
	setLinkApplier(const ctl::link &p_):p(p_) {}
	template<class X> void operator()(X &x) const
	{ 
		enum { isCI=conversion<type2type<X>, X>::exist || sizeof(CTL_isCI((X*)0,(empty*)0) )>1 };
		setLink(x, p, ctl::int2type<isCI>() );
	} 
};

template<class Y, class Args =ctl::empty> struct resReceiverBase: public resultRecvBase<Y>
{
	using resultRecvBase<Y>::emptyArgM;
	using resultRecvBase<Y>::linkM;
	mutable Args argsM;
	resReceiverBase(const ctl::link &p, const Args &args): resultRecvBase<Y>(p), argsM((Args&)args)
	{ emptyArgM = Args::isEmpty; }
	void setLinks() const
	{ argsM.apply(setLinkApplier(linkM) ); }

};

template<class Y, class Args =ctl::empty> struct resReceiver: public resReceiverBase<Y, Args>
{
	using resReceiverBase<Y, Args>::emptyArgM;
	using resReceiverBase<Y, Args>::receivedM;
	using resReceiverBase<Y, Args>::argsM;
	using resReceiverBase<Y, Args>::linkM;
	using resReceiverBase<Y, Args>::multiReadingM;
	typedef typename resReceiverBase<Y, Args>::valueT valueT;
	mutable valueT *pyM;
	mutable cPointer<excptHandle> *pbinM;	
	resReceiver(const ctl::link &p, const Args &args):
							resReceiverBase<Y, Args>(p, args), pyM(0), pbinM(0)
	{}
	~resReceiver()
	{
		if(!receivedM)
			linkM.run(1,true);
	}
	bool received(bool blocking, valueT &y, cPointer<excptHandle> &excpt) const
	{
		if(!receivedM)
		{
//CTL_Profile("received");
			pyM = &y; pbinM = &excpt;
		  CTL_Try
			{
				if(multiReadingM)
				{
					Wait W;
					if(blocking)
						while(!receivedM)
						{
							linkM.run(1, false);
							W(true, true); // invkoke env::run
						}
						linkM.run(1, false);
				}
				else
					linkM.run(1,blocking);
			}

			CTL_Catch(const ctl::groupException &e)
			{ receivedM = true; excpt = new excptHandle(binaryObject(ctl::exception(e.whatM) ) ) ;	}
			CTL_Catch(const ctl::exception &e)
			{ receivedM = true; excpt = new excptHandle(binaryObject(e) );  }
		}
		return receivedM;
	}
	status receive(ctl::iStream &is)
	{
		CTL_Assert(pyM && pbinM);
		binaryObject bin;
		is>>bin;
		if(!!bin)
		{
  		*pbinM = new excptHandle(bin); 
			receivedM = true;
			is.clear();
			return success;
		}
		is>>argsM>>*pyM;
		is.clear();
		receivedM = true;
		return success;
	}
};

struct resCopy
{
	template<class X1, class X2> static void cp(const X1&, const X2&,trueT)  {}
	template<class X1, class X2> static void cp(X1 &x1, const X2 &x2, falseT) { x1 = x2; }
	template<class X1, class X2> void operator()(X1 &x1, const X2 &x2) const
	{ cp(x1, x2, int2type<ctl::constTraits<X1>::is>() ); }
};

template<class Y, class F,class Args, class refArgs> class resThreadReceiver: public resReceiverBase<Y, Args>
{
	typedef resThreadReceiver<Y, F, Args, refArgs> thisT;
	thisT& operator=(const thisT&);
public:
	using resReceiverBase<Y, Args>::receivedM;
	using resReceiverBase<Y, Args>::linkM;
	using resReceiverBase<Y, Args>::argsM;
	using resReceiverBase<Y, Args>::setLinks;
	using resReceiverBase<Y, Args>::multiReadingM;
	typedef  typename resultRecvBase<Y>::valueT  valueT;
	mutable valueT tmpyM;
	mutable binaryObject tmpbinM;
	mutable refArgs xRefM;

	F funcM;
	resThreadReceiver(const refArgs &xRef, const Args &args, F func, const ctl::link &p):resReceiverBase<Y,Args>(p, args),xRefM(xRef), funcM(func)
	{}

	~resThreadReceiver()
	{ 
		while(!receivedM)
			sys::usleep(0);
	}

	status receive(ctl::iStream &)
	{	return success;  	}
	bool received(bool blocking, valueT &y, cPointer<excptHandle> &bin) const
	{	
		if(!receivedM && !linkM.isValid())
		{
			tmpbinM = ctl::exception(" *** thread terminated unexpected ***");
			receivedM = true;
		}
		Wait W;
		if(multiReadingM)
		{
			while(blocking && !receivedM)
			{	W(true, true); }
		}
		else
		{
			while(blocking && !receivedM)
			{	W(true); }

		}
		bool done = receivedM;
		if(done)
		{
			if(!tmpbinM)
			{		
				y = tmpyM; 
				setLinkApplier appl(linkM);
				appl(y);
			  setLinks();
				xRefM.apply(resCopy(),argsM); 
			} 
			else
			{ bin = new excptHandle(tmpbinM); tmpbinM.clear(); }
		}
		return done;
	}
};

} // internal


template<class Y> class resultBase
{
public:
	enum { isCI=conversion<type2type<Y>, Y>::exist || sizeof(CTL_isCI((Y*)0,(empty*)0) )>1 };
	typedef resultBase<Y> thisT;
	typedef Y origValue;
	typedef typename internal::resultRecvBase<Y>::valueT  valueT;
	typedef ctl::cPointer<internal::resultRecvBase<Y> > SP;
	typedef rPointer<internal::resReceiverCI > RP;
	mutable valueT yM;
	mutable cPointer<excptHandle> excptM;
	mutable SP resM;
	mutable bool threadM, emptyM;
	mutable const char *prettyM;

	bool resReceived(bool blocking, bool thrw) const
	{
		bool recvd=!resM || resM->received(blocking, yM, excptM);
		if(recvd)
			resM = 0;
		if(thrw)
		{
			if(!!excptM && !!excptM->binM)
			{
				CTL_Log2(0, "#f => %s throws exception('%s')\n", prettyM, excptM->msg().c_str());
				excptM->throwIt();
			}
#ifndef CTL_NoLog
			if(!!prettyM)
			{	log("#f => %s\n", prettyM ); prettyM=0; }
#endif
		}
		return recvd;
	}
	void assign_(const resultBase<Y> &r)
	{		
		yM = r.yM;
		excptM = r.excptM;
		threadM = r.threadM;
		emptyM = r.emptyM;
		resM = r.resM;
		prettyM = r.prettyM;
		return;
	}
public:
	void clearExcpt() const
	{ 
		if(!!excptM)
			excptM->binM.clear();	
	}
	template<class Args>  resultBase(const ctl::link &p, const Args &args, const char *pretty): threadM(false),emptyM(false), prettyM(pretty)
	{ resM = SP(new internal::resReceiver<Y,Args>(p, args) ); }
	resultBase(internal::resultRecvBase<Y> *res, bool isEmpty): threadM(true),emptyM(false),prettyM(0)
	{
		resM = SP(res);
		if(!!resM)
			resM->emptyArgM = isEmpty;
	}
	resultBase(Flag = ctl::now): threadM(false),emptyM(true),prettyM(0)
	{}
	resultBase(const valueT &y) :yM(y), threadM(false),emptyM(false),prettyM(0)
	{}

	resultBase(const valueT &y, const cPointer<excptHandle> &excpt, const char *pretty) :yM(y), excptM(excpt), threadM(false), emptyM(false), prettyM(pretty)
	{}
//	resultBase(const valueT &y, const binaryObject &excpt, const char *pretty) :yM(y), threadM(false),emptyM(false), prettyM(pretty)
//	{ if(!!excpt) excptM = new excptHandle(excpt); }

	template<class Z> resultBase(const resultBase<Z> &r):
					excptM(r.excptM), threadM(r.threadM),emptyM(r.emptyM), prettyM(r.prettyM)
	{ 
		if(!emptyM)
		{
			if(!!r.resM)
				resM = SP(new internal::resCaster<Y,Z>(r.resM) );
			else // received lib
				yM = binaryCast<Y>(r.getValue());
		}
	}
	resultBase(const resultBase<Y> &r): yM(r.yM), excptM(r.excptM),resM(r.resM),threadM(r.threadM),emptyM(r.emptyM), prettyM(r.prettyM)
	{}
	~resultBase()
	{  	
		if(resM.getRefCount()==1) 
			while(!received(true, true)) ;
	}
	void setPretty(const char *pretty) const
	{ 
	//ctl::log("setPretty '%s'\n", pretty); 
		prettyM = pretty; 
	}
	void setLink(const ctl::link &p)
	{
		if(!!resM && !resM->linkM)
			resM->linkM = p;
	}
	ctl::link getLink() const
	{ return !resM ? ctl::link() : resM->linkM; }

	RP getRP()
	{
		ctl::cPointer<internal::resReceiverCI > lp(resM.get());
		return lp;
	}
   bool ready(bool blocking, bool thrw=true) const
   {
		if(resReceived(blocking, thrw) ||  !resM)
			return true;
		if(!resM->getCommunicator() ||  !resM->getCommunicator()->isValid() )
		{			
			resM.clear();
			if(thrw)
				ctl::throwStr(" *** result lost connection ***");
			return true;
		}
		return false;
	}
	bool received(bool blocking=false, bool thrw=true) const
	{
		bool received=false;
		if(!threadM)
		{
			if(ready(blocking, thrw))   // true for lib
				return true;
			else if(blocking)
				sys::usleep(0);
		}
		else
		{
			if(!(received=resReceived(blocking,thrw)) && blocking)
				sys::usleep(0);
		}
		return received;
	}
	bool operator!() const     // no throw
	{ return !received(false, false); }
	// listen to ALL incoming messages
	thisT& listen()
	{
		if(!!resM)
			resM->multiReadingM = true;
		return *this;
	}
	bool isEmpty() const
	{ return emptyM; }
	int getRefCount() const
	{ return isEmpty() ? 0 : resM.getRefCount(); }
	const valueT& getValue(bool logOff=false) const
	{		
(void)logOff;
		if(!resM)
		{
			if(!!excptM)
				excptM->throwIt();
#ifndef CTL_NoLog
			if(!!prettyM && !logOff)
			{	log("#f => %s\n", prettyM); prettyM=0; }
#endif
			return yM;
		}
		while(!received(true)) ;
/*
		if(!!resM)
		{
			resM->setLinks();
		}
*/
		return yM;
	}
	void assign(valueT& z) const
	{ if(!!resM) z = getValue(); }
};

template<typename Y, typename X> inline Y resCast(const X&, Y*);
template<class Y> class rBinResult: public resultBase<Y>
{
public:
	using  resultBase<Y>::getValue;
	using  resultBase<Y>::listen;
		typedef resultBase<Y> baseT;
		typedef Y origValue;
		typedef typename baseT::valueT  valueT;
//		typedef typename baseT::RP  RP;
		typedef typename baseT::SP SP;
		rBinResult<Y> &operator=(const rBinResult<Y> &r)
		{  	assign_(r);   return *this; }
		template<class Args>  rBinResult(const ctl::link &p, const Args &args, const char *pretty): baseT(p, args, pretty)
		{}
		rBinResult(internal::resultRecvBase<Y> *res, bool isEmpty): baseT(res, isEmpty)
		{}
		rBinResult(const valueT &y=valueT()): baseT(y)
		{}
//		rBinResult(const valueT &y=valueT(), const binaryObject &excpt =binaryObject()): baseT(y, excpt)
//		{}
		rBinResult(const valueT &y, const cPointer<excptHandle> &excpt, const char *pretty) : baseT(y, excpt, pretty)
		{}

		rBinResult(const rBinResult<Y> &r):baseT(r)
		{}
#ifdef CTL_TmplCastOper
		template<class Z> operator Z () const
		{	return resCast(*this, (Z*)0); }
//		{	return binaryCast<Z>(getValue()); }
#endif
		template<class Z> void assign(Z& z) const
		{ z = binaryCast<Z>(getValue()); }
	};

template<class Y> inline typename resultBase<Y>::valueT &operator<<(typename resultBase<Y>::valueT& z, const resultBase<Y> &r)
{ return z = r.getValue(); }

template<class Y, class Z> inline Z &operator<<(Z& z, const rBinResult<Y> &r)
{ return z = binaryCast<Z>(r.getValue()); }


//! accessor for results of methods or functions
template<class Y = ctl::empty> class result: public resultBase<Y>
{
	enum { isCI = conversion<type2type<Y>, Y>::exist || sizeof(CTL_isCI((Y*)0,(empty*)0) )>1 };
public:
	using  resultBase<Y>::getValue;
	using  resultBase<Y>::received;
	using  resultBase<Y>::resM;
	using  resultBase<Y>::listen;
		typedef resultBase<Y> baseT;
		typedef Y origValue;
		typedef  typename baseT::valueT  valueT;
		typedef typename baseT::SP SP;
		result<Y> &operator=(const result<Y> &r)
		{  	assign_(r);   return *this; }
		template<class Z> result<Y> &operator=(const rBinResult<Z> &r)
		{  return *this = result<Y>(r); }
		template<class Args>  result(ctl::link &p, const Args &args, const char *pretty): baseT(p, args, pretty)
		{}
		result(internal::resultRecvBase<Y> *res, bool isEmpty): baseT(res, isEmpty)
		{}
		result(Flag flag =ctl::now): baseT(flag)
		{}
		result(const valueT &y): baseT(y)
		{}
		result(const valueT &y, const cPointer<excptHandle> &excpt, const char *pretty) : baseT(y, excpt, pretty)
		{}
		template<class Z> result(const result<Z> &r): baseT((const resultBase<Z>&)r)
		{}
#ifndef CTL_TmplCastOper
		template<class Z> result(const rBinResult<Z> &r): baseT((const resultBase<Z>&)r)
		{}
#else
		template<class Z> explicit result(const rBinResult<Z> &r): baseT((const resultBase<Z>&)r)
		{}
#endif
//! non blocking: check whether either result is available or exception occured
		bool operator!() const
		{ return !received(false, false); }
		bool operator==(const  result<Y> &res) const
		{ return  resM == res.resM; }
		result(const result<Y> &r):baseT(r)
		{}
//!    blocking: return the returned valueT
		operator const valueT &() const
		{ return getValue(); }
	};

	template<class Y>  struct resultType
	{
		enum { isBin = sizeof(isBinary((const Y*)0) )>1 };
		typedef  typename ctl::select<isBin, rBinResult<Y>, result<Y> >::resultT  resultT;
	};

	template<typename Y, typename Res> inline Y resCast(const Res& r, Y*)
	{	return binaryCast<Y>(r.getValue()); }
	template<typename Y, typename Res> inline result<Y> resCast(const Res& r, result<Y>*)
	{	return result<Y>(r); }

} // ctl

#endif

