/******************************************************************************

 This file is part of the Component Template Library (CTL)

 Copyright (C) 2006  Rainer Niekamp
 email:              r.niekamp@tu-bs.de

 The CTL is free software; you can redistribute it and/or modify it under the
 terms of the GNU Lesser General Public License as published by the Free Software
 Foundation; either version 2.1 of the License, or (at your option) any later
 version. We only ask that proper credit be given to the author.

 This library is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 for more details.

 You should have received a copy of the GNU Lesser General Public License along
 with this library; if not, write to the Free Software Foundation,
 Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

 ******************************************************************************/

#ifndef CTL_SYSTEM_H_
#define CTL_SYSTEM_H_

#include "utl/timer.h"
#include <sstream>
#include <stdarg.h>
#include <math.h>

namespace ctl
{
	/**/// some forward declarations
	class Environment;
	class Communicator;
	class Remote;
	class objectID;
	class link;
	class group;
	class groupImpl;
	struct location;
	class TimeMarker;
	class socket;
	class any;
	struct binaryObject;

	template<class T> class result;
	typedef ctl::lPointer<Environment> EnvironmentSP;
	typedef ctl::lPointer<Communicator> CommunicatorSP;
	typedef ctl::cPointer<Remote> RemoteSP;
	typedef ctl::cPointer<groupImpl> groupSP;

	CTL_Inline Environment &env_(int n = 0, const char * const *args = 0,
	    const char* = 0);
	CTL_Inline const location &myLocation();
	CTL_Inline void addObject(Environment *, const RemoteSP &);
	CTL_Inline void addFunction(const char *fname, const RemoteSP &f);
	CTL_Inline void setTmpFunction(const char *fname, const RemoteSP &f);
	CTL_Inline RemoteSP getFunction(const char *pretty);
	CTL_Inline void holdReference(Environment*, const Remote *);
	CTL_Inline void releaseReference(Environment*, const Remote *);
	CTL_Inline link getLink(const typeId &scope);
	CTL_Inline link getLink(const typeId &scope, const ctl::any &prop,
	    const char *locator = 0);
	CTL_Inline void setLink(const typeId &scope, const link &p);
	CTL_Inline typeId getLocation(const typeId& scope);
	CTL_Inline void
	    setLocationModifier(const typeId &scope, const ctl::string &s);
	CTL_Inline typeId getLocationModifier(const typeId &scope);
	CTL_Inline void setClient(const ctl::link&);
	//! print link info to file
	CTL_Inline void messageInfo(FILE *file = stderr, const char* = 0);
	CTL_Inline link getClient();
	CTL_Inline group getGroup();
	CTL_Inline group getMainGroup();
	CTL_Inline group getMPIGroup();
	CTL_Inline group getDmnGroup();
	CTL_Inline void deleteGroup(const groupImpl *grp);

	CTL_Inline PID getRootId(Environment *e);
	CTL_Inline PID getThreadId(Environment *e);
	CTL_Inline const link &getRemoteLink(Environment *e, const PID &pid);
	//  CTL_Inline  bool sameThread(Environment *, Environment *);

	CTL_Export void openLibrary(void ** vmyenv, void *venv, void *proc,
	    const char *arg, const char* (*locator)(), void(*conn)());
	CTL_Export void closeLibrary(void *venv, bool);
	CTL_Inline void closeThread(void *venv);

	CTL_Inline int openMain(int n, char **arg, void **, void *,
	    const char* (*locator)(), void(*conn)());

	//  CTL_Inline void deleteEnv(Environment *myEnv,bool =false);

	CTL_Inline const char *getVariable(const char *name, bool global = true);
	CTL_Inline ctl::vector<std::pair<const char*, ctl::string> >
	    getAllVariables();
	CTL_Inline ctl::vector<ctl::location> readLocation(const char *file);
	CTL_Inline bool logStr(const char*, const Environment * = 0);
	//  CTL_Inline bool closeLog();
	CTL_Inline bool openLog(const char *logFile, int theme = -2);

	CTL_Inline void throwStr(const char*, ...);
	CTL_Inline void throwStr2bin(binaryObject &bin, const char* fmt, ...);
	CTL_Inline void rethrowStr(const std::exception &e, const char* fmt, ...);
	CTL_Inline void rethrowStr2bin(const std::exception &e, binaryObject &bin,
	    const char* fmt, ...);
	CTL_Inline TimeMarker &newTimeMark(const char*);
	//CTL_Inline group getMainGroup();

	CTL_Inline RemoteSP getNamedObject(
	    const std::pair<ctl::typeId, ctl::string>&key);
	CTL_Inline void setNamedObject(
	    const std::pair<ctl::typeId, ctl::string> &key, const RemoteSP &);

	CTL_Inline typeId prettyName(const char *y, const char *scope,
	    const char *f, const char *param, const char *args, const char *args_);
	CTL_Inline const char *prettyNameS(const char *y, const char *scope,
	    const char *f, const char *param, const char *args, const char *args_);

	CTL_Inline bool run(int n = 0, bool blocking = false);
	CTL_Inline const char *getcwd();
	CTL_Inline ctl::string getDOScwd();
	CTL_Inline int printf(const char *format, ...);
	CTL_Inline int scanf(const char *format, ...);
	CTL_Inline FILE *fopen(const char *file, const char *mode);
	inline int fclose(FILE *fp)
	{
		return ::fclose(fp);
	}
	CTL_Inline int system(const char *cmd);

#ifdef CTL_Monitor
	CTL_Inline bool monitoring(const ctl::string &, bool);
#endif 		

	namespace sys
	{
		typedef uint_ threadIdT;
#ifdef CTL_Unix
		CTL_Inline ctl::string copyFile(const ctl::string& src, const ctl::string& dest=ctl::string());
#else
		CTL_Inline ctl::string copyFile(const ctl::string& src, const ctl::string& dest);
#endif
		inline bool remove(const ctl::string& filename)
		{
			int err=::remove(filename.c_str());
			if(err)
				ctl::log("#d remove('%s') failed, err = %s\n", filename.c_str(), ::strerror(err) );
			return err==0;
		}

		CTL_Inline ctl::string getHostIP();
		CTL_Inline void setPriority(int prior);
		CTL_Inline ctl::string absolutePath(ctl::string mypath,
		    const ctl::string &path);
		CTL_Inline const char *getcwd(char *path, int len);
		CTL_Inline void usleep(double usec);
		CTL_Inline int getPid();
		CTL_Inline threadIdT getsTid();
		CTL_Inline threadIdT getTid();
		CTL_Inline void wakeUp(threadIdT pid);
		CTL_Inline const char *getEnv(const char *);
		CTL_Inline void setEnv(const char *, const char *);
		CTL_Inline void unsetEnv(const char *);
		CTL_Inline int system(const char *cmd, const ctl::string &path);
		CTL_Inline int getCommandArgs(char **arg, int maxArg);
		CTL_Inline const char *getLocatorName();

		CTL_Inline threadIdT startThread(void *(*run)(void *), void *arg);
	} // sys

	inline char *readLine(char *line, int n, FILE *fp)
	{
		char *c = 0;
		while((c = fgets(line, n, fp)))
		{
			while(isspace(*c))
				c++;
			if(*c == '#')
			{
				if(!c[1] || c[1] == '#')
					return 0;
				else
					continue;
			}
			char *end = strchr(c, '#');
			if(end)
				*end = 0;
			else
				end = c + strlen(c);
			while(isspace(*--end))
				*end = 0;
			if(*end == '\\')
			{
				char nLine[CTL_StringBufferSize];
				char *nc = readLine(nLine, CTL_StringBufferSize, fp);
				if(nc) memcpy(end, nc, std::min(strlen(nc) + 1, size_t(
				    CTL_StringBufferSize + (end - c))));
			}
			return c;
		}
		return 0;
	}

	CTL_Inline void addStatic(const std::pair<void*, void(*)(const void*)>, bool /*toParent*/=false);

// File descriptors and operator for io
	CTL_Inline int getFDin();
	CTL_Inline int getFDout();
	CTL_Inline int getFDerr();
	CTL_Inline int getFDlog();

	CTL_Inline char *gets(char *m, size_t max = 0);
	inline Environment &env(int n = 0, const char * const *arg = 0, bool ini =  true)
	{	return env_(n, arg, ini ? sys::getLocatorName() : 0); }


	typedef enum	{	cin = 0	 } cinType;
	typedef enum  {	cout = 1 } coutType;
	typedef enum	{	cerr = 2 } cerrType;
	typedef enum	{	clog = 3 } clogType;

	template<class T, class Cout> inline Cout leftShift(Cout os, const T& t, int fd)
	{
		std::stringstream S;
		S << t;
		::write(fd, (const void*) S.str().c_str(), (int) S.str().size());
		return os;
	}
	template<class T> inline coutType operator<<(coutType os, const T& t)
	{ return leftShift(os, t, ctl::getFDout());	}
	template<class T> inline cerrType operator<<(cerrType os, const T& t)
	{	return leftShift(os, t, ctl::getFDerr());	}
	template<class T> inline cinType operator>>(cinType is, T& t)
	{
		char buf[CTL_StringBufferSize];
		buf[CTL_StringBufferSize - 1] = 0;
		if(ctl::gets(buf, CTL_StringBufferSize))
		{
			std::stringstream S(buf);
			S >> t;
		}
		return is;
	}
#ifdef CTL_NoLog
	template<class T> inline clogType operator<<(clogType os, const T&)
	{	return os;}
	inline bool log(const Environment *,const char* ,...)
	{	return false;}
	inline bool log(const char* ,...)
	{	return false;}
#else
	template<class T> inline clogType operator<<(clogType os, const T& t)
	{
		return leftShift(os, t, ctl::getFDlog());
	}
	inline bool log(const Environment *e, const char* fmt, ...)
	{
		char msg[CTL_StringBufferSize];
		va_list ap;
		va_start(ap, fmt);
#ifndef CTL_VNstdio
		vsprintf(msg, fmt, ap);
#else
		vsnprintf(msg, CTL_StringBufferSize, fmt, ap);
#endif
		va_end(ap);
		return logStr(msg, e);
	}

	inline bool log(const char* fmt, ...)
	{
		char msg[CTL_StringBufferSize];
		va_list ap;
		va_start(ap, fmt);
#ifndef CTL_VNstdio
		vsprintf(msg, fmt, ap);
#else
		vsnprintf(msg, CTL_StringBufferSize, fmt, ap);
#endif
		va_end(ap);
		return logStr(msg);
	}
#endif

	class Wait
	{
		// variables in usec
		double t0M, qM, DM;
		mutable double dtM, tM;
		public:
		Wait(double t0, double q, double D) :
			t0M(t0), qM(q), DM(D), dtM(t0), tM(0)
		{
		}
		Wait(double D = 1e4, double M = 1e5, int N = 100) :
			DM(D), tM(0)
		{
			qM = M / (M - D);
			t0M = M / pow(qM, N);
			dtM = t0M;
		}
		void operator()(bool b, bool all = false) const
		{
			if(b)
			{
				dtM *= dtM < DM ? qM : 1;
				sys::usleep(dtM < 100. ? 0.0 : dtM);

				if(all && 1000 * dtM >= DM) ctl::run(1, false);
				tM += dtM;
			}
			else
			{
				dtM = t0M;
				tM = 0;
			};
		}
	};
} // ctl

#endif

