/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_FUNCTIONTRAITS_H_
#define CTL_FUNCTIONTRAITS_H_

#include "typefield.h"

namespace ctl
{

template<class oY, typename Func, class Args> class Functor
{
	typedef Functor<oY, Func, Args> thisT;
	thisT &operator=(const thisT&);
public:
	Func funcM;
	typedef Args args;
	typedef oY resultT;
	typedef typename noVoid<resultT>::resultT Y;
	typedef typename noVoid<resultT>::cast cast;
	Functor(Func func, const typeId& =typeId()): funcM(func) {}
	template<class arg> Y operator()(const arg &x) const
	{ return call((arg&)x, int2type<arg ::size>()); }
#define CTL_MethodCall(n)\
	template<class arg> Y call(arg &x, int2type<n>) const\
	{ (void)x; return cast((funcM)(CTL_ExprList(CTL_Unconst, x.t, Comma, n)) ), empty(); }
	CTL_MethodCall(0)
	CTL_MethodCall(1) CTL_MethodCall(2) CTL_MethodCall(3) CTL_MethodCall(4)
	CTL_MethodCall(5) CTL_MethodCall(6) CTL_MethodCall(7) CTL_MethodCall(8)
	CTL_MethodCall(9) CTL_MethodCall(10) CTL_MethodCall(11) CTL_MethodCall(12)
	CTL_MethodCall(13) CTL_MethodCall(14) CTL_MethodCall(15) CTL_MethodCall(16)

#undef CTL_MethodCall
};

template<class oY, class C, class Meth, class Args> class CFunctor
{
	typedef CFunctor<oY, C, Meth, Args> thisT;
	thisT &operator=(const thisT&);
	C &cM;
	Meth methM;
public:
	typedef Args args;
	typedef oY resultT;
	typedef typename noVoid<resultT>::resultT Y;
	typedef typename noVoid<resultT>::cast cast;
	CFunctor(C &c, Meth meth, const typeId& =typeId()): cM(c), methM(meth) {}
	template<class arg> Y operator()(const arg &x) const
	{ return call((arg&)x, int2type<arg ::size>()); }
#define CTL_MethodCall(n)\
	template<class arg> Y call(arg &x, int2type<n>) const\
	{ (void)x; return  cast((cM.*methM)(CTL_ExprList(CTL_Unconst, x.t, Comma, n)) ), empty() ; }
	CTL_MethodCall(0)
	CTL_MethodCall(1) CTL_MethodCall(2) CTL_MethodCall(3) CTL_MethodCall(4)
	CTL_MethodCall(5) CTL_MethodCall(6) CTL_MethodCall(7) CTL_MethodCall(8)
	CTL_MethodCall(9) CTL_MethodCall(10) CTL_MethodCall(11) CTL_MethodCall(12)
	CTL_MethodCall(13) CTL_MethodCall(14) CTL_MethodCall(15) CTL_MethodCall(16)

#undef CTL_MethodCall
};

template<class C, class Base, class Y>  CFunctor<Y,C,Y (Base::*)(), field0<> >
		functor(C &c, Y (Base::*meth)() )
{ return  CFunctor<Y,C,Y (Base::*)() , field0<>  >(c, meth); }
template<class C, class Base, class Y>  CFunctor<Y,const C,Y (Base::*)() const, field0<> >
		functor(const C &c, Y (Base::*meth)() const)
{ return  CFunctor<Y,const C,Y (Base::*)() const, field0<> >(c, meth); }
template<class Y>  Functor<Y,Y (*)(), field0<>  >
		functor(Y (*func)() )
{ return  Functor<Y,Y (*)() , field0<> >(func); }

#define CTL_Functors(n)\
template<class C, class Base, class Y, CTL_ExprList(CTL_ClassList, X, Comma,n) >  CFunctor<Y,C,Y (Base::*)(CTL_List(X,n) ), CTL_Field_Type((CTL_List(X,n)),n ) >\
		functor(C &c, Y (Base::*meth)(CTL_List(X,n)) ){ return  CFunctor<Y,C,Y (Base::*)(CTL_List(X,n) ), CTL_Field_Type((CTL_List(X,n)),n ) >(c, meth); }\
template<class C, class Base, class Y, CTL_ExprList(CTL_ClassList, X, Comma, n) >  CFunctor<Y,const C,Y (Base::*)(CTL_List(X,n) ) const, CTL_Field_Type((CTL_List(X,n)),n ) >\
		functor(const C &c, Y (Base::*meth)(CTL_List(X,n)) const )\
{ return  CFunctor<Y,const C,Y (Base::*)(CTL_List(X,n) ) const, CTL_Field_Type((CTL_List(X,n)),n ) >(c, meth); }\
template<class Y, CTL_ExprList(CTL_ClassList, X, Comma, n) >  Functor<Y,Y (*)(CTL_List(X,n) ), CTL_Field_Type((CTL_List(X,n)),n ) >\
functor(Y (*func)(CTL_List(X,n)) ) { return  Functor<Y,Y (*)(CTL_List(X,n) ), CTL_Field_Type((CTL_List(X,n)),n ) >(func); }

CTL_Functors(1)
CTL_Functors(2)
CTL_Functors(3)
CTL_Functors(4)
CTL_Functors(5)
CTL_Functors(6)
CTL_Functors(7)
CTL_Functors(8)
CTL_Functors(9)
CTL_Functors(10)
CTL_Functors(11)
CTL_Functors(12)
CTL_Functors(13)
CTL_Functors(14)
CTL_Functors(15)
CTL_Functors(16)
CTL_Functors(17)

#undef CTL_Functors
#ifdef CTL_NoLog
struct logger
{
	logger(char =1, const char * =0) {}
	logger &operator<<(const char *)
	{ return *this; }
};
#else
struct logger
{
	int thema;
	ctl::string msg;
	logger(char th=1, const ctl::string &m =ctl::string()) :thema(th), msg(m) {}
	~logger()
	{ if(thema) ctl::log("#%c %s \n", thema, msg.c_str() ); }
	logger &operator<<(const ctl::string &m)
	{ msg += m; return *this; }
};
#endif

template<class E0=empty,class E1=empty,class E2=empty, class E3=empty,class E4=empty,class E5=empty> struct tryCallCatchList
{
	static void connect()
	{ ctl::connectType((E0*)0); tryCallCatchList<E1, E2, E3, E4, E5>::connect(); }
	template<class Y, class F, class Arg> static Y call(const F &func, const Arg &x, Y *y, ctl::cPointer<ctl::excptHandle> &excpt)
	{		
		CTL_Try { return tryCallCatchList<E1, E2, E3, E4, E5>::call(func, x, y, excpt); }
		CTL_Catch(const E0 &excption)
		{
			ctl::binaryObject a(excption);
			ctl::log("#x !! user exception %s caught\n", ctl::typeName<E0>().string().c_str());
			excpt = new ctl::excptHandle(a);
 		}
		return Y();
	}
};
template<> struct tryCallCatchList<>
{
	static void connect() {}
	template<class Y, class F, class Arg> static Y call(const F &func, const Arg &x, Y *, ctl::cPointer<ctl::excptHandle> &)
	{ return func(x); }
};

template<class E0=empty,class E1=empty,class E2=empty, class E3=empty,class E4=empty,class E5=empty> struct tryCallCatch
{
	enum { isEmpty = true }; 
	static void connect()
	{ tryCallCatchList<E0, E1, E2, E3, E4, E5>::connect(); }
	template<class Y, class F, class Arg> Y call(const F &func, const Arg &x, Y *y, ctl::cPointer<ctl::excptHandle>& excpt, const char *pretty)
	{
		if(!pretty)
			return Y();
		CTL_Log1(0, "#f -> %s\n", pretty)
		logger lg('f', "<= ");
		lg << pretty;		
//		CTL_Try { return func(x); }
		CTL_Try { return tryCallCatchList<E0, E1, E2, E3, E4, E5>::call(func, x, y, excpt); }
		CTL_Catch(const ctl::exception &e)
		{ excpt = new ctl::excptHandle; ctl::rethrowStr2bin(e, excpt->binM, " *** in %s: ***", pretty); 	} 
		CTL_Catch(const std::exception &e)
    { excpt = new ctl::excptHandle; ctl::rethrowStr2bin(e, excpt->binM, " *** in %s: ***", pretty);       }
		CTL_CatchAll
		{ excpt = new ctl::excptHandle; ctl::throwStr2bin(excpt->binM, " *** unknown exception in %s: ***", pretty); 	}
		lg<<" throws "<< excpt->binM.type.c_str()<<"\n";		
		return Y();
	}
};


} // ctl

#endif

