/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_SERIALIZE_INL_
#define CTL_SERIALIZE_INL_

#include <map>
#include <set>
#include <list>
#include <vector>
#include <string>
#include <complex>
#include <memory>
#include <queue>

#include "vector.h"
#include "factory.h"
#include "types.h"

namespace ctl 
{

template<class T> inline oStream &operator <<(oStream &os, const T &x);
template<class T> inline iStream &operator >>(iStream &is, T &x);

inline oStream &operator <<(oStream &os,const real8 x) { return os.write(&x,1); }
inline iStream &operator >>(iStream &is,real8 &x) { return is.read(&x,1); }

inline oStream &operator <<(oStream &os,const real4 &x) { return os.write(&x,1); }
inline iStream &operator >>(iStream &is,real4 &x) { return is.read(&x,1); }

inline oStream &operator <<(oStream &os,const int8 &x) { return os.write(&x,1); }
inline iStream &operator >>(iStream &is,int8 &x) { return is.read(&x,1); }

inline oStream &operator <<(oStream &os,const uint8 &x) { return os.write(&x,1); }
inline iStream &operator >>(iStream &is,uint8 &x) { return is.read(&x,1); }

inline oStream &operator <<(oStream &os,const long &x) { return os.write(&x,1); }
inline iStream &operator >>(iStream &is,long &x) { return is.read(&x,1); }

inline oStream &operator <<(oStream &os,const unsigned long &x) { return os.write(&x,1); }
inline iStream &operator >>(iStream &is,unsigned long &x) { return is.read(&x,1); }

inline oStream &operator <<(oStream &os,const int4 &x) { return os.write(&x,1); }
inline iStream &operator >>(iStream &is,int4 &x) { return is.read(&x,1); }

inline oStream &operator <<(oStream &os,const uint4 &x) { return os.write(&x,1); }
inline iStream &operator >>(iStream &is,uint4 &x) { return is.read(&x,1); }

inline oStream &operator <<(oStream &os,const int2 &x) { return os.write(&x,1); }
inline iStream &operator >>(iStream &is,int2 &x) { return is.read(&x,1); }

inline oStream &operator <<(oStream &os,const uint2 &x) { return os.write(&x,1); }
inline iStream &operator >>(iStream &is,uint2 &x) { return is.read(&x,1); }

inline oStream &operator <<(oStream &os,const char &x) { return os.write(&x,1); }
inline iStream &operator >>(iStream &is,char &x) { return is.read(&x,1); }

inline oStream &operator <<(oStream &os,const unsigned char &x) { return os.write(&x,1); }
inline iStream &operator >>(iStream &is,unsigned char &x) { return is.read(&x,1); }

inline oStream &operator <<(oStream &os,const bool &x) { return os.write(&x,1); }
inline iStream &operator >>(iStream &is,bool &x) { return is.read(&x,1); }

inline oStream &operator <<(oStream &os,const empty &) { return os; }
inline iStream &operator >>(iStream &is,empty &) { return is; }

inline oStream &operator <<(oStream &os,const PID::IPvX &IP)
{ os<<IP.port; return os.write(IP.ip,8); }	
inline iStream &operator >>(iStream &is,PID::IPvX &IP)
{ is>>IP.port; return is.read(IP.ip,8); }	

inline oStream &operator <<(oStream &os, const PID  &p)
{ return os<<PID::IPvX(p); }
inline iStream &operator >>(iStream &is, PID &p)
{ PID::IPvX ip; is>>ip; p = ip; return is; }

template<class T> inline oStream &write(oStream &os, const T &t)
{ return t.write(os); }
template<class T> inline iStream &read(iStream &is, T &t)
{ return t.read(is); }


template<class S, class T> inline oStream &write(oStream &os, const std::pair<S,T> &);
template<class S, class T> inline iStream &read(iStream &is, std::pair<S,T>&);

template<class T> 	inline oStream &lPointer<T>::write(oStream &os) const
{  return os<<soM; }

template<class T> 	inline iStream &lPointer<T>::read(iStream &is)
{
	clear();
	T *t=0;  		is>>t;
	const lPointer<T> rd(is.getPhysAddrSP(t));
	if(!rd)
	{
		soM = t;
		if(t)
			is.appendPhysAddrSP(t, *this);
	}
	else
		*this = rd;				
	return 	is;
}


inline oStream &operator<<(oStream &os, const char *str)
{
	if(!str)
		return os<<char(0);
	return os.write(str,strlen(str)+1);
};
                                  
inline iStream &operator>>(iStream &is, char *&str)
{	
	size_t size;
	bool end = is.strTerm(size);
	str = new char[size+1];
	is.read(str, end ? size : size+1);
	if(end)
		str[size] = '\0';
	return is;
};

inline iStream &operator>>(iStream &is, const char *&str)
{ return is >> (char*&)str; }

inline oStream &operator<<(oStream &os, const char * const *str)
{
	while(str && *str)
		os.write(*str,strlen(*str)+1), /*ctl::log("<<'%s[%p]'", *str, str) ,*/str++;
	return os<<char(0);
};
inline oStream &operator<<(oStream &os, const char **str)
{ return os<<(const char * const *)str; }
inline oStream &operator<<(oStream &os, char *const *str)
{ return os<<(const char * const *)str; }
inline oStream &operator<<(oStream &os, char **str)
{ return os<<(const char * const *)str; }

inline iStream &operator>>(iStream &is, char ** &str)
{	
	size_t size, n=0;
	bool end = is.strTerm(size,n);
	char c;
	if(!n)
	{
		if(!end)
			is>>c;
		str = 0;
		return is;
	}
	char *buf = new char[size+n];
	is.read(buf, end ? size+n-1 : size+n);
	if(!end)
		is>>c;
	else
		buf[size+n] = '\0';
	str = new char*[n+1];
	for(size_t i=0; i<n; i++)
	{
		str[i] = buf;
		buf +=strlen(buf) + 1;
	}
	str[n] = 0;
	return is;
};

inline iStream &operator>>(iStream &is, const char ** &str)
{ return is>>(char**&)str; }

inline oStream &operator<<(oStream &os, const std::string &str)
{
	if(str.empty())
		return os<<char(0);
	return os.write(str.c_str(),str.size()+1);
};
                                  
inline iStream &operator>>(iStream &is, std::string &str)
{	
	size_t size;
	bool end = is.strTerm(size);
	ctl::vector<char> buf(size+1);
	is.read(buf.begin(), end ? size : size+1);
	if(end)
		buf[size] = '\0';
	str = buf.begin();
	return is;
};

#ifdef CTL_wchar
inline oStream &operator<<(oStream &os, const std::wstring &str)
{
	if(str.empty())
		return os<<wchar(0);
	return os.write(str.c_str(),sizeof(wchar)*str.size()+1);
};
                                  
inline iStream &operator>>(iStream &is, std::wstring &str)
{	
	size_t size;
	bool end = is.wstrTerm(size);
	ctl::vector<wchar> buf(size+1);
	is.read(buf.begin(), end ? size : size+1);
	if(end)
		buf[size] = wchar(0);
	str = std::wstring((const wchar_t*)buf.begin() );
	return is;
};
#endif

inline oStream &operator<<(oStream &os, const ctl::string &str)
{
	if(str.empty())
		return os<<char(0);
	return os.write(str.c_str(),str.size());
};
                                  
inline iStream &operator>>(iStream &is, ctl::string &str)
{
	size_t size;
	bool end = is.strTerm(size);
	if(!size)
	{
		char c=1;
		is.read(&c,1);
		CTL_Assert(c==0);
		str.clear();
	}
	else
	{	
		str.resize(size+1);
		is.read(str.begin(), end ? size : size+1);
		if(end)
			str[size] = '\0';
	}
	return is;
};


inline oStream &operator<<(oStream &os, const ctl::exception &ex)
{  return os<<ex.whatM; }

inline iStream &operator>>(iStream &is, ctl::exception &ex)
{ return is>>ex.whatM; }


namespace internal
{
	template<class T> inline oStream &sendStdContainer(oStream &os, const T &t)
	{
		uint8 size=t.size();
		os<<size;
		for(typename T::const_iterator i=t.begin(); i!=t.end(); i++)
			os<<*i;
		return os;
	}
	template<class T> inline iStream &readStdContainer(iStream &is, T &t)
	{
		uint8 size;
		is>>size;
		t.resize(size);
		for(typename T::iterator i=t.begin(); i!=t.end(); i++)
			is>>*i;
		return is;
	}
} // internal


template<class T, class cmp, class alloc>  inline  oStream &operator<<(oStream &os, const std::set<T, cmp, alloc> &y)
{ return internal::sendStdContainer(os, y); }

template<class T, class cmp, class alloc> inline  iStream &operator>>(iStream &is, std::set<T, cmp, alloc> &y)
{
		uint8 size;
		is>>size;

		for(uint8 i=0; i<size; i++)
		{
			T x;
			is>>x;
			y.insert(x);
		}
		return is;
}

template<class T, class seq>  inline  oStream &operator<<(oStream &os, const std::queue<T, seq> &y)
{
	std::queue<T, seq> cy = y;
	os<<int8(cy.size());
	while(!cy.empty())
	{
		os << cy.front();
		cy.pop();
	}	
	return os;
}

template<class T, class seq> inline  iStream &operator>>(iStream &is, std::queue<T, seq> &y)
{
		uint8 size;
		is>>size;
       std::queue<T, seq> tmp;
		for(uint8 i=0; i<size; i++)
		{
			T x;
			is>>x;
			tmp.push(x);
		}
		y = tmp;
		return is;
}

template<class S, class T, class cmp, class alloc>  inline  oStream &operator<<(oStream &os, const std::map<S,T,cmp,alloc> &y)
{ return internal::sendStdContainer(os, y); }

template<class S, class T, class cmp, class alloc> inline  iStream &operator>>(iStream &is, std::map<S,T,cmp,alloc> &y)
{
 		uint8 size;
		is>>size;

		for(uint8 i=0; i<size; i++)
		{
			S s;
			T t;			
			is>>s>>t;
			y[s] = t;
		}
		return is;
}

namespace internal
{
	template<class T> inline  oStream &write(oStream &os,const vector<T> &y, falseT)
	{	return internal::sendStdContainer(os, y); }
	template<class T> inline  iStream &read(iStream &is, vector<T> &y,  falseT)
	{	return internal::readStdContainer(is, y); }

	template<class T> inline  oStream &write(oStream &os,const vector<T> &y, trueT)
	{
		uint8 size = y.size();
		os<<size;
		return os.write(y.begin(), size);
	}
	template<class T> inline  iStream &read(iStream &is, vector<T> &y,  trueT)
	{
		uint8 size;
		is>>size;
		y.resizeRaw(size);
		return is.read(y.begin(), size);
	}

}

template<class T> inline  oStream &operator<<(oStream &os,const vector<T> &y)
{
	enum { pod=fundamental<T>::is && !pointerTraits<T>::is };
	return internal::write(os, y, int2type<pod>() );
}

template<class T> inline  iStream &operator>>(iStream &is, vector<T> &y)
{
	enum { pod=fundamental<T>::is && !pointerTraits<T>::is };
	return internal::read(is, y, int2type<pod>() );
}

template<class T, class Alloc> inline  oStream &operator<<(oStream &os,const std::vector<T, Alloc> &y)
{
	enum { pod=fundamental<T>::is && !pointerTraits<T>::is };
	size_t size=y.size();
	if(pod && size>0 && size_t(&y[size-1] -&y[0] +1) == size)
	{
		os<<uint8(size);
		return os.write(&y[0], size);	
	}
	return internal::sendStdContainer(os, y);
}

template<class T, class Alloc> inline  iStream &operator>>(iStream &is, std::vector<T,Alloc> &y)
{ return internal::readStdContainer(is, y); }

template<class Alloc> inline  iStream &operator>>(iStream &is, std::vector<bool,Alloc> &y)
{
	uint8 size;
	is>>size;
	y.resize(size);
	for(typename std::vector<bool,Alloc>::iterator i=y.begin(); i!=y.end(); i++)
	{	bool b; is>>b; *i = b; }
	return is;
}

template<class Alloc> inline  oStream &operator<<(oStream &os, const  std::vector<bool,Alloc> &y)
{
	uint8 size=y.size();
	os<<size;
	for(typename std::vector<bool,Alloc>::const_iterator i=y.begin(); i!=y.end(); i++)
	{	bool b = *i; os<<b; }
	return os;
}

template<class X, typename T0> inline iStream& readByCtor(iStream& is, X& x, const T0&)
{
	T0 t0;
	is>>t0;
	x = X(t0);
	return is;
}

template<class X, typename T0, typename T1> inline iStream& readByCtor(iStream& is, X& x, const T0&, const T1&)
{
	T0 t0;	T1 t1;
	is>>t0>>t1;
	x = X(t0, t1);
	return is;
}

template<class X, typename T0, typename T1, typename T2> inline iStream& readByCtor(iStream& is, X& x, const T0&, const T1&, const T2&)
{
	T0 t0;	T1 t1;	T2 t2;
	is>>t0>>t1>>t2;
	x = X(t0, t1, t2);
	return is;
}

template<class X, typename T0, typename T1, typename T2, typename T3> inline iStream& readByCtor(iStream& is, X& x, const T0&, const T1&, const T3&)
{
	T0 t0;	T1 t1;	T2 t2;	T3 t3;
	is>>t0>>t1>>t2>>t3;
	x = X(t0, t1, t2, t3);
	return is;
}

inline bool oStream::write(const void *t, const typeId &type)
{
	if(!t)
	{	*this<<true<<int4(-1); return false; }
	int4 log = getLogAddr(t);
	if(log>0)
	{	*this<<true<<log; return false; }
	appendLogAddr(t);
	*this<<false << type;
	return !type.empty();
}

template<class T> inline oStream &operator<<(oStream &os, T *const &t)
{
//ctl::log(" oStream &operator<<(oStream &is, T* ** &str %p)\n", t);
	typeId type;
	if(t)
	{
		typeId typeI = factoryS().find(&typeid(*t));
		type = typeI.empty() ? typeName<T>(*t).string() : typeI;
//ctl::log(" oStream &operator<<(oStream &is, T* ** &str %p) type = '%s' '%s'\n", t, type.c_str(), getTypeId<T>().c_str());
	}
	if(os.write(t,  type ) )
	{
		os << blockBegin();	
		if(!factoryS().write(t,  type, os) )
			ctl::throwStr(" *** fatal io error in serializing pointer of %s ***", type.c_str());		
		os << blockEnd();
	}
	return os;
}

inline bool iStream::read(void *&t, typeId &type)
{
	bool b;
	*this>>b;
	t = 0;
	if(b)
	{
		int4 log;
		*this>>log;
		if(log>0)
			t = (void*)getPhysAddr(log);
		return false;
	}
	*this>>type;
//ctl::log(" iStream read(iStream &is, T* type=%s &str %p, %d)\n",type.c_str(), t, int(b));
	return !type.empty();
}

template<class T> inline iStream &operator>>(iStream &is, T *&t)
{
//void *MARKER=0;
	typeId type;
//ctl::log(" iStream &operator>> |T|=%d\n", sizeof(T) );
	if(is.read((void*&)t, type) )
	{
		is>>blockBegin();	
		factoryS().read((const T*&)t, type.c_str(), is);
		is>>blockEnd();
//ctl::log(" iStream operator(iStream &is, T* type=%s &str %p)\n",type.c_str());
	}
	return is;
}

inline oStream &operator<<(oStream &os, const any &obj)
{ 		
	typeId type=obj.getTypeName();
	const binaryObject *gobj = pointer_cast<const binaryObject>(obj);
	if(gobj)
		return os<<gobj->type<<gobj->buffer;
/*
	if(type.empty() || type[0] == '?' || !!gobj)
	{
		if(gobj)
			return os<<gobj->type<<gobj->buffer;
		return	os<<typeId();
	}
*/		
	os<<typeId(type);	
	os<<blockBegin();	
	obj.write(os);
/*
	if(!ctl::factory_write(obj, type, os) )
	{
		ctl::string err = ctl::string(" *** fatal io error in serializing any of ") + type + " ***";
		ctl::log("#w %s\n",err.c_str());
		CTL_Throw(exception(err) );
	}
*/
	os<<blockEnd();

	return os;
}

inline iStream &operator>>(iStream &is, any &obj)
{
	typeId type;
 	is>>type;   	
	if(!type.empty())
	{
		is>>blockBegin();
		if(!ctl::factory_read(obj,  type.c_str(), is) )
			obj = binaryObject(type, int4(is.currentBlock()), is.currentBuffer());	
		is>>blockEnd();
	}
	return is;
}


template<class T> inline oStream &operator<<(oStream &os, const std::auto_ptr<T> &sp)
{  return os<<sp.get(); }

template<class T> inline iStream &operator>>(iStream &is, std::auto_ptr<T> &sp)
{	
	T *t;
	is>>t; 
	std::auto_ptr<T> p(t);
	sp = p; 
	return is; 
}

template<class T> inline oStream &operator<<(oStream &os, const type2type<T>)
{ return os; }

template<class T> inline iStream &operator>>(iStream &is, const type2type<T>)
{ return is; }

inline oStream &binaryObject::write(oStream &os) const
{
	os<<type<<msg;
	if(!type.empty())
		os<<buffer;
	return os;
}

inline iStream &binaryObject::read(iStream &is)
{
	is>>type>>msg;
	if(!type.empty())
		is>>buffer;
	else
		buffer.clear();
	return is;
}


template<class T> inline oStream &operator <<(oStream &os, const T &t)
{ return ctl::write(os, t); }

template<class T> inline iStream &operator >>(iStream &is, T &t)
{ return ctl::read(is, t); }

} // ctl
#endif
