/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_SORT_H_
#define CTL_SORT_H_

# include<algorithm>
#	include<stdlib.h>
#	include<stdio.h>

namespace ctl
{
template<class T>	struct stdLess
{
	bool operator()(const T&a,const T&b) const
	{ return a<b; }
};

namespace internal
{

template<class less,class X> struct pointerLess
{
	bool operator()(const X *r0,const X *r1) const
	{ return less()(*r0,*r1); }
};

template<class less,class X> struct pointerLessTotal
{
	bool operator()(const X *r0, const X *r1) const
	{
		if(less()(*r0,*r1))
			return true;
		if(less()(*r1,*r0))
			return false;	
		return r0<r1;
	}
};

template<class X> void perm(X *a, const X **q0, int  n)
{
	const X **q=q0,**e=q0+n;
	while(q<e)
	{
		if(q-q0!=*q-a)
		{
			X d = a[q-q0];
			const X **p=q;
			for(p=q; *(q=q0+(*p-a)); p=q)
			{  a[p-q0] = **p; *p = 0;  }
			a[p-q0] = d; *p = 0;
		}
		while(++q<e && *q==0) ;
	}
}

} // internal

template <class less,class X> void qSort(X *const a, X *const b, less lt =less())
{
	if(b-a>1)
	{
		if(b-a==2)
		{
			if(lt(*(a+1),*a))
		   {
		   		X c=*a;
		   		*a = *(a+1);
	   			*(a+1) = c;
			}
			return;
		}
		X *i=a, *j=b, *m=i+(j-i)/2, v=*m;
		*m = *a;
		for(;;)
		{
			while(!lt(*--j,v))
				if(i==j) goto L;
			*i = *j;
			while(!lt(v,*++i))
				if(i==j) goto L;
			*j = *i;
		}
L:	*j++ = v;
		qSort(a,i,lt); qSort(j,b,lt);
	}
}

template <class less,class X> void pqSort(X *a, X *b, less)
{
	const int n=b-a;
	const X **p= new const X*[n];
	for(int i=0; i<n; i++)
		p[i] = a+i;
//	std::sort(p,p+n, internal::pointerLess<less,X>());
//	qSort(p,p+n, internal::pointerLess<less,X>());
	qSort(p,p+n, internal::pointerLessTotal<less,X>());
	internal::perm(a,p,n);
	delete[] p;
}
template <class X> void pqSort(X *a,X *b)
{ pqSort(a,b,stdLess<X>()); }

template<typename less,class T>
	int lowerBound(const T *a0,const T *a1,const T &t, less lt =less())
	{
		if(a1<=a0)
			return -1;
		const T *a=a0;

		while(a1-a0>1)
		{
			const T* m=a0+((a1-a0)>>1);
			if(lt(t,*m))
				a1=m;
			else
				a0=m;
		}
		if(a0==a && lt(t, *a) )
			return -1;
		return a0-a;
	}
} // ctl

#endif
