/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_TYPEFIELD_H_
#define CTL_TYPEFIELD_H_

#include "serialize.h"


namespace ctl
{


#ifdef CTL_MSC

template<class P> struct typeMap
{
	template<class X> struct map
	{ typedef X resultT; };
	template<class X, class Y> struct map2
	{ typedef X resultT; };
};

template<class P, class Y> struct mapper
{
	typedef typeMap<P> typemap;
	typedef typename typemap::map<Y> M;
	typedef typename M::resultT resultT;
};

template<class P, class Y0, class Y1> struct mapper2
{
	typedef typeMap<P> typemap;
	typedef typename typemap::map2<Y0, Y1> M;
	typedef typename M::resultT resultT;
};

#define CTL_Forward1(tmpl)\
template<class F> struct CTL_Get##_##tmpl##H\
{\
	template<typename T> struct dum\
	{\
		typedef dum<T> resultT;\
		template<typename P> struct tmpl {};\
	};\
	template<>	struct dum<F>\
	{	typedef F resultT;	};\
	template<class P> struct tmpl\
	{\
		typedef typename dum<typename F::thisT>::resultT M;\
		typedef typename M::tmpl<P> tmpl##_;\
	};\
};\
template<class F, class P> struct CTL_Get##_##tmpl\
{\
	typedef typename CTL_Get##_##tmpl##H <F>::tmpl<P> T;\
	typedef typename T::tmpl##_ resultT;\
};

#define CTL_Forward2(tmpl)\
template<class F> struct CTL_Get##_##tmpl##H\
{\
	template<typename T> struct dum\
	{\
		typedef dum<T> resultT;\
		template<typename P, typename Q> struct tmpl {};\
	};\
	template<>	struct dum<F>\
	{	typedef F resultT;	};\
	template<class P, class Q> struct tmpl\
	{\
		typedef typename dum<typename F::thisT>::resultT M;\
		typedef typename M::tmpl<P,Q> tmpl##_;\
	};\
};\
template<class F, class P, class Q> struct CTL_Get##_##tmpl\
{\
	typedef typename CTL_Get##_##tmpl##H <F>::tmpl<P,Q> T;\
	typedef typename T::tmpl##_ resultT;\
};

CTL_Forward1(map)
CTL_Forward2(map2)

template<class F, class P> struct fieldMap
{
	typedef typename CTL_Get_map<F,P>::resultT map;
	typedef typename map::resultT resultT;
};
template<class F0, class F1, class P> struct fieldMap2
{
	typedef typename CTL_Get_map2<F0,F1,P>::resultT map;
	typedef typename map::resultT resultT;
};
#else
template<class P> struct typeMap;

template<class F, class P> struct fieldMap
{  typedef typename F::CTL_TemplateDclr map<P>::resultT resultT;  };

template<class F0,class F1, class P> struct fieldMap2
{ 	typedef typename F0::CTL_TemplateDclr map2<F1,P>::resultT resultT; };
#endif

template<class T> inline bool sizeChanged(const T&)
{ return false; }


#define CTL_FieldTypeDef(T,n) typedef T##n Y##n
#define CTL_FieldMemb(t,n) Y##n t##n
#define CTL_FieldEmpty(T,n) same<empty,T##n>::value
#	define CTL_EmptyConstr(n)\
	field##n(): CTL_ExprList(CTL_EmptyInit, T, Comma,n){}
#ifdef CTL_MSC
#	define CTL_EmptyInit(Z,n)  t##n(T##n())
#	define CTL_FieldMapApply(Z,n)  typename mapper<P,T##n>::resultT
#	define CTL_FieldMapApply2(Z,n)  typename mapper2<P,T##n,typename F::Y##n>::resultT
#	define CTL_FieldMap(n)\
	template<class P>  struct map \
	{ typedef field##n< CTL_ExprList(CTL_FieldMapApply, T, Comma, n) > resultT; }; \
	template<class F, class P>  struct map2 \
	{ typedef field##n< CTL_ExprList(CTL_FieldMapApply2, Y, Comma, n) > resultT; };

#else
#	define CTL_EmptyInit(Z,n)  t##n()
#	define CTL_sizeChanged(Z,n)  ctl::sizeChanged(t##n)
#	define CTL_FieldMapApply(Z,n)  typename typemap::CTL_TemplateDclr map<T##n>::resultT
#	define CTL_FieldMapApply2(Z,n)  typename typemap::CTL_TemplateDclr map2<T##n,typename F::Y##n>::resultT
#	define CTL_FieldMap(n)\
	template<class P>  struct map \
	{ typedef typeMap<P> typemap; typedef field##n< CTL_ExprList(CTL_FieldMapApply, T, Comma, n) > resultT; }; \
	template<class F, class P>  struct map2 \
	{ typedef typeMap<P> typemap; typedef field##n< CTL_ExprList(CTL_FieldMapApply2, Y, Comma, n) > resultT; };

#endif
#define CTL_Unconst(t,n)  unconst(t##n)
#define CTL_Apply_f(t,n)  f(t##n)
#define CTL_Apply_f2(t,n)  f(t##n, fld.t##n)
#define CTL_FieldInit(f,n)  t##n(f##n)

#define CTL_Field(n)\
CTL_Tmpl_Dclr2((CTL_List(T,n)),n) struct fieldH##n {};\
CTL_Tmpl_Dclr2((CTL_List(T,n)),n) struct field##n \
{\
	typedef CTL_Field_Type((CTL_List(T,n)),n ) thisT;\
	thisT &operator=(const thisT &);\
	CTL_ExprList(CTL_FieldTypeDef, T, Semicolon ,n); \
	CTL_ExprList(CTL_FieldMemb, t, Semicolon, n); \
	enum { size = n, isEmpty = CTL_ExprList(CTL_FieldEmpty, Y, And, n) }; \
	CTL_FieldMap(n)\
	CTL_EmptyConstr(n) \
	template<class F> field##n(const F &f): CTL_ExprList(CTL_FieldInit, ((F&)f).t, Comma, n) {} \
	field##n(CTL_ExprList(CTL_FieldMemb, tt, Comma, n) ): CTL_ExprList(CTL_FieldInit,tt,Comma,n) {}\
	bool sizeChanged() const\
	{ return  CTL_ExprList(CTL_sizeChanged, t, Or, n); }\
	ctl::oStream& write(ctl::oStream &os) const\
	{ return os << CTL_ListSep(t, LShift, n); } \
	ctl::iStream& read(ctl::iStream &is) \
	{ return is >> CTL_ExprList(CTL_Unconst,t, RShift, n); }\
	template<class F> void apply(const F &f)\
	{ CTL_ExprList(CTL_Apply_f,t, Semicolon, n); }\
	template<class F, class Field> void apply(const F &f, const Field &fld)\
	{ CTL_ExprList(CTL_Apply_f2,t, Semicolon, n); }\
};

template<class T0 = empty> struct field0
{
//void *t0;
	typedef field0<T0> thisT;
	thisT &operator=(const thisT &);
	enum { isEmpty = true, size = 0 };

	template<class P>  struct map
	{ typedef field0<> resultT; };
	template<class P, class F>  struct map2
	{ typedef field0<> resultT; };
	bool sizeChanged() const
	{ return false; }
	ctl::oStream& write(ctl::oStream &os) const
	{ return os; }
	ctl::iStream& read(ctl::iStream &is) const
	{ return is; }
	template<class F> void apply(const F&) const
	{}
	template<class F, class Field> void apply(const F&, const Field&)
	{}
};

template<class T0> struct field1
{
	typedef field1<T0> thisT;
	thisT &operator=(const thisT &);
	typedef T0 Y0;
	Y0 t0;
	enum { isEmpty = same<empty,Y0>::value , size = 1 };
	CTL_FieldMap(1)
	template<class T> field1(const field1<T> &f): t0(((field1<T>&)f).t0) {}
	CTL_EmptyConstr(1)
	field1(Y0 tt0): t0(tt0) {}
	bool sizeChanged() const
	{ return ctl::sizeChanged(t0); }
	ctl::oStream& write(ctl::oStream &os) const
	{ return os << t0; }
	ctl::iStream& read(ctl::iStream &is)
	{ return is >> unconst(t0); }
	template<class F> void apply(const F &f)
	{ f(t0); }
	template<class F, class Field> void apply(const F &f, const Field &fld)
	{ f(t0, fld.t0); }
};

struct fieldH0 {};
template<class T0> struct fieldH1 {};

template<class T0> inline ctl::oStream& write(ctl::oStream &os,const ctl::field1<T0> &f)
{ return os<<f.t0; }

template<class T0> inline ctl::iStream& read(ctl::iStream &is, ctl::field1<T0> &f)
{ return is>>unconst(f.t0); }

CTL_Field(2)
CTL_Field(3)
CTL_Field(4)
CTL_Field(5)
CTL_Field(6)
CTL_Field(7)
CTL_Field(8)
CTL_Field(9)
CTL_Field(10)
CTL_Field(11)
CTL_Field(12)
CTL_Field(13)
CTL_Field(14)
CTL_Field(15)
CTL_Field(16)
CTL_Field(17)

#undef CTL_FieldTypeDef
#undef CTL_FieldMapApply
#undef CTL_FieldMapApply2
#undef CTL_FieldMemb
#undef CTL_sizeChanged
#undef CTL_FieldEmpty
#undef CTL_FieldInit
#undef CTL_Field

} // ctl


#endif

