/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_TYPENAME_H_
#define CTL_TYPENAME_H_

#include "ctl/gen/utl.mcr"

#define CTL_TypeNameDef(name, CONN)\
	ctl::typeId CTL_typeName() const\
	{ return CONN; }\
	static ctl::typeId CTL_typeNameS()\
	{ ctl::typeId s(name); simplifyTypeName(s); return s;	}\
	operator ctl::property<ctl::hasTypeName>() const

#define CTL_TypeName(T) CTL_TypeName2(T,T)
#define CTL_TypeNameS(T) CTL_TypeName2(T,T)\
	CTL_TypeNameDef(#T, CTL_typeNameS());)
	
#define CTL_TypeName2(T, type)\
	CTL_TypeNameDef(#T, ctl::connectType2(( type*)0, CTL_typeNameS());)
	
#define CTL_TemplateName(T, TList, nT)\
	 CTL_TypeNameDef(CTL_TmplNameStr3(T,TList, nT) , ctl::connectType2(( CTL_Tmpl_Name(T,TList,nT)*)0, CTL_typeNameS() );)

#define CTL_TemplateNameS(T, TList, nT)\
	 CTL_TypeNameDef(CTL_TmplNameStr3(T,TList, nT) , CTL_typeNameS() )

#define CTL_SetTypeName(T)    CTL_SetTypeName2(T,T)
#define CTL_SetTypeName2(T, type)\
namespace ctl	 {\
	inline ctl::typeId type_name(const type *)\
{ ctl::connectType2(( type*)0, CTL_Str(T) ); return #T; }\
 int setProperty(ctl::property<ctl::hasTypeName>*,const type*);   }

#define CTL_SetTemplateName(T, TList, nT)\
namespace ctl	 {\
	CTL_Tmpl_Dclr2(TList,nT) inline ctl::typeId type_name(const CTL_Tmpl_Name(T,TList,nT) *)\
{ const ctl::typeId s = CTL_TmplNameStr3(T,TList, nT); ctl::connectType2(( CTL_Tmpl_Name(T,TList,nT)*)0, s); return s;	}\
CTL_Tmpl_Dclr2(TList,nT) int setProperty(ctl::property<ctl::hasTypeName>*,CTL_Tmpl_Name(T,TList,nT)*);   }

#include "vector.h"
#include <string>
#include <vector>
#include <list>
#include <set>
#include <map>

namespace ctl
{
typedef ctl::string typeId;

inline void simplifyTypeName(ctl::typeId& name)
{
	char* s=name.begin(), * s0(s);
	while(s!=name.end())
	{
		while(s[0]==',' && (s[1]==',' || s[1]=='>') )
			s++;
		*s0++ = *s++;
	}
	name.resize(s0-name.begin());
}
template<class T> inline typeId connectType2(T* t, const typeId &type);

template<class T> struct typeName;

template<class T> inline typeId typeNameA(falseT, const T *t)
{ return type_name(t); }

template<class T> inline typeId typeNameA(trueT, const T *t)
{ return t ? t->CTL_typeName() : T::CTL_typeNameS(); }

template<class T> inline typeId typeNameP(trueT,T *const * t)
{
	typedef int2type<ctl::property<ctl::hasTypeName, T>::intrinsic> has;
	typeId cname;
	const T * const ct=  t ? *t : 0;
	if(!has::value)
		cname = pointerTraits<T>::is ? typeNameP(trueT(), ct) : type_name(ct);
	else
		cname=typeNameA(has(),ct);
	const typeId name= !cname.empty() ?
		typeId(cname) + (constTraits<T>::is ? "*const" : "*") : typeId();
	return name;
}

template<class T> inline typeId typeNameP(falseT, const T *t)
{
	typedef int2type<ctl::property<ctl::hasTypeName, T>::intrinsic> has;
	return typeNameA(has(),t);
}

inline typeId typeNameP(trueT, const void*  =0)
{  return typeId(); }

#ifdef CTL_MSC
template<class T> inline typeId typeNameR(T **t =0)
{ return typeNameP(int2type<pointerTraits<T>::is>(),t ? *t : (T*)0);
}
template<class T> inline typeId typeNameR(T *t =0)
{ return typeNameP(int2type<pointerTraits<T>::is>(),&t ? t : (T*)0);
}

template<class T, class Int> inline T* refToPointer(type2type<T>,Int)
{ return 0; }
#else

template<class T> inline typeId typeNameR(const T *t =0)
{ return typeNameP(int2type<pointerTraits<T>::is>(), t );
}
template<class T> inline T* refToPointer(type2type<T>, falseT)
{ return 0; }

template<class T> inline T* refToPointer(type2type<T&>, trueT)
{ return 0; }

#endif

template<class T> struct typeName
{
	typedef typename referenceTraits<T>::reference rT;
	typedef typename referenceTraits<T>::constReference crT;

	typeId nameM;
	static typeId qualifiedName(const typeId &name)
	{	return !name.empty() ? typeId(constTraits<T>::is ? "const " : "") + name + (referenceTraits<T>::is ? "&" : ""): typeId(); }
	typeName()
	{
		typeId pure = typeNameR(refToPointer(type2type<T>(), int2type<referenceTraits<T>::is>()));
		nameM = qualifiedName(pure);
	}
	typeName(crT t)
	{ nameM = typeNameP(int2type<pointerTraits<T>::is>(),&t); }
	operator typeId() const
	{ return string(); }
	typeId string() const
	{ return !nameM.empty()  ? nameM : typeId("?"); }
};

template<> struct typeName<void>
{
	operator const char *() const
	{ return "void"; }
	typeId string() const
	{ return "void"; }
};

template<> struct typeName<empty>
{
	operator const char *() const
	{ return ""; }
	typeId string() const
	{ return ""; }
};

template<class X> inline typeId getTypeId(const X* =0)
{ return typeName<X>().string(); }


template<int n> inline typeId getTypeId(int2type<n>* =0)
{ return string::int2string(n); }


inline int compare(const std::type_info &t0, const std::type_info &t1)
{ return strcmp(t0.name(), t1.name() ); }

inline bool equal(const std::type_info &t0, const std::type_info &t1)
{
	if(&t0==&t1)
		return true;
	return compare(t0, t1) == 0;
}

}  // ctl

template<class T>	inline std::ostream &operator<<(std::ostream &os, const ctl::typeName<T> &a)
{ return os<<a.string(); }


#endif
