/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_TYPES_H_
#define CTL_TYPES_H_

#include "ctl/config.h"
#include "ctl/gen/ctl.mcr"
#if defined(CTL_Alpha) && !defined(CTL_GNUC)
#include <iostream>

#endif

namespace ctl
{	
	typedef int		int4;
	typedef unsigned int 	uint4;
	typedef short		int2;
	typedef unsigned short 	uint2;
	typedef char	int1;
	typedef unsigned char 	uint1;
	typedef float		real4;
	typedef double real8;
//	typedef unsigned char uchar;	
	typedef short			wchar;
	
#ifdef CTL_MSC
	typedef __int64			int8;
	typedef unsigned __int64 	uint8;

#else
	typedef long long		int8;
	typedef unsigned long long 	uint8;
#endif
	
#ifdef CTL_PartialSpec
	
	template<bool b,class X,class Y>  struct select
	{ typedef Y resultT; };
	
	template<class X,class Y>  struct select<true,X,Y>
	{ typedef X resultT; };
	
#else
	
	namespace internal
	{
		
		template<bool b> struct selectorH
		{
			template<class X,class Y>  struct which
			{ typedef Y resultT; };
		};
		
		template<> struct selectorH<true>
		{
			template<class X,class Y>  struct which
			{ typedef X resultT; };
		};
		
	} // e.o. internal
	
	template<bool b,class X,class Y>  struct select
	{ typedef typename internal::selectorH<b>::which<X,Y>::resultT resultT; };
	
#endif
	struct empty
	{
		enum { isEmpty = true };
		empty() {}
		static const char* CTL_typeNameS() { return ""; }
//		static typeId CTL_typeNameS() { return ""; }
		template<class T> empty(const T &x) { (void)x; }
		empty(const empty &x) {  (void)x; }
	};
	template<class T> class typeTraits;
	template<class Y> struct referenceTraits;


#ifdef CTL_SGI
} // ctl 
char CTL_isCI(const void*, empty*);

char  setProperty(void*,  const void *);
namespace ctl
#endif


#ifndef CTL_PartialSpec

namespace internal
{

	template<class X> struct conversionH
	{
		template<class Y> struct converts
		{
		struct H
		{
			typedef char Small;
			typedef typename referenceTraits<Y>::constReference YY;
 
			static char f(X x, int);
			static int   f(ctl::empty, float);
			static YY Y_(int);
		};
			enum { exist = sizeof(H::f(H::Y_(0), 0))==sizeof(char), same = false };
		};
		template<> struct converts<X>
		{ enum { exist = true, same = true }; };
		template<> struct converts<void>
		{ enum { exist = false, same = false }; };
	};

	template<class X> struct sameH
	{
		template<class Y> struct is
		{ enum { value = false }; };
		template<> struct is<X>
		{ enum { value = true }; };
	};
	template<> struct conversionH<void>
	{
		template<class Y> struct converts
		{ enum { exist = true, same = false };	};
		template<> struct converts<void>
		{ enum { exist = true, same = true }; };
	};

} //.internal
#endif

typedef select<sizeof(void*)==4, int4, int8>::resultT int_;
typedef select<sizeof(void*)==4, uint4, uint8>::resultT uint_;

#ifdef CTL_PartialSpec
	//! Conversion::exist = Y converts to X
	template<class X,class Y> struct conversion
	{
		typedef typename referenceTraits<Y>::constReference YY;
		struct H
		{
			static char f(X x, int);
			static int  f(ctl::empty, float);
			static YY Y_(int);
		};
		enum { exist = sizeof((H::f)((H::Y_)(0), 0))==sizeof(char) };
	};
	
	template<class X> struct conversion<X,X>
	{ enum { exist = true }; };
	template<class X> struct conversion<void, X>
	{ enum { exist = true }; };
	template<class X> struct conversion<X, void>
	{ enum { exist = false }; };
	template<> struct conversion<void, void>
	{ enum { exist = true }; };
	
	template<class X, class Y> struct same
	{ enum { value=false }; };
	template<class X> struct same<X,X>
	{ enum { value = true }; };
#else
	
	template<class X,class Y> struct conversion
	{  enum { exist = internal::conversionH<X>::converts<Y>::exist }; };
	template<class X, class Y> struct same
	{ enum { value = internal::sameH<X>::is<Y>::value }; };
#endif
	
	
	template<int n0, int n1=0, int n2=0, int n3=0, int n4=0, int n5=0, int n6=0, int n7=0> struct int2type
	{ enum { value = n0, v0=n0, v1=n1, v2=n2, v3=n3, v4=n4, v5=n5, v6=n6, v7=n7 }; };
	
	typedef int2type<true> trueT;
	typedef int2type<false> falseT;

	template<class X=empty,class X1=empty,class X2=empty,class X3=empty, class X4=empty,class X5=empty,class X6=empty,class X7=empty> struct type2type
	{ typedef X resultT; };
	
	template<class T> inline const T *nullPointer(const T&)
	{ return (const T*)0; }

	template<class X> inline X& unconst(const X &x)
	{ return (X&)x; }

	enum {hasBinaryCheck=1, hasTypeName, isAbstract};
	template<int P, class T=ctl::empty> struct property
	{		
		enum { extrinsic = (sizeof(setProperty((property<P>*)0,(const T*)0 ) ) > sizeof(char) )  };
		enum { intrinsic = ctl::conversion<property<P>, T>::exist };
		enum { value = extrinsic |  intrinsic };
		typedef int2type<value> resultT;
	};
	template<int P> struct property<P,empty>
	{
		enum { value = true };
		typedef int2type<value> resultT;
	};

	
	char  setProperty(void*,  const void *);

	template<class T, bool Const> struct constSelect
	{ typedef typename ctl::select<Const, const T, T>::resultT resultT; };


	
#ifdef CTL_PartialSpec
	
	template<class U> struct pointerTraits
	{ 	
		typedef U resultT;		enum { is = false };
		typedef U* pointer;
		typedef const U * constPointer;
	};
	template<class U> struct pointerTraits<U&>
	{ 	
		typedef U resultT;		enum { is = false };
		typedef U* pointer;
		typedef const U * constPointer;
	};
	template<class U> struct pointerTraits<const U&>
	{ 	
		typedef U resultT;		enum { is = false };
		typedef U* pointer;
		typedef const U * constPointer;
	};
	template<class U> struct pointerTraits<U *>
	{
		typedef U resultT;			enum { is = true };
		typedef U* pointer;
		typedef const U * constPointer;
	};
	template<class U, int n> struct pointerTraits<U [n]>
	{
		typedef U resultT;			enum { is = true };
		typedef U* pointer;
		typedef const U * constPointer;
	};
	template<class U,int n> struct pointerTraits<const U [n]>
	{
		typedef U resultT;			enum { is = true };
		typedef U* pointer;
		typedef const U * constPointer;
	};
	template<class U> struct pointerTraits<U *const>
	{ 	
		typedef U resultT;			enum { is = true };
		typedef U* pointer;
		typedef const U * constPointer;
	};
	
	template<class X> struct referenceTraits
	{ 	
		typedef X resultT; enum { is = false };	
		typedef X& reference;
		typedef const X& constReference;
	};
	template<class X> struct referenceTraits<X &>
	{ 	
		typedef X resultT;	enum { is = true };
		typedef X& reference;
		typedef const X& constReference;
	};
	template<class X> struct referenceTraits<const X &>
	{ 	
		typedef const X resultT;	enum { is = true };
		typedef X& reference;
		typedef const X& constReference;
	};
	template<> struct referenceTraits<void>
	{ 	
		typedef void resultT; enum { is = false };	
		typedef void reference;
		typedef void constReference;
	};
	
	template<class U> struct constTraits
	{ 	typedef U resultT;	enum { is = false };};
	template<class U> struct constTraits<U&>
	{ 	typedef U resultT;	enum { is = false };};
	template<class U> struct constTraits<const U>
	{ 	typedef U resultT;	enum { is = true }; };
	template<class U> struct constTraits<const U&>
	{ 	typedef U resultT;	enum { is = true }; };
	
	template<class U> struct unRefConst
	{ 	typedef U resultT;	};
	template<class U> struct unRefConst<const U&>
	{ 	typedef U resultT;	};
	template<class U> struct unRefConst<U&>
	{ 	typedef U resultT;	};
	template<class U> struct unRefConst<const U>
	{ 	typedef U resultT;	};
	
#else
	namespace internal
	{
		template<bool> struct refTypeH
		{ template<class T> struct type { typedef T& value; }; };
		template<> struct refTypeH<true>
		{ template<class T> struct type { typedef T value; }; };
		
		template<bool> struct pointerTypeH
		{ template<class T> struct type { typedef T* value; }; };
		template<> struct pointerTypeH<true>
		{ template<class T> struct type { typedef T value; }; };
		
		template<class T> struct traitsH
		{
			static ctl::int4 h(type2type<T>);
			static char h(...);
		};
		template<class T> type2type<const T> checkRef(const T&);
		template<class T> char checkPnt(const T*);
		ctl::int4 checkPnt(...);
		template<class T> type2type<T&> checkConstRef(const T&);
		
	} // e.o. internal

	template<class T> inline T refToRef(type2type<T>);	
	
	template<class T> struct pointerTraits
	{
		enum { is = sizeof(internal::checkPnt(refToRef(type2type<T>())))==sizeof(char) };
	
	};
	template<class T> struct referenceTraits
	{
		typedef typename internal::traitsH<const T> ch;
		enum { is = sizeof(ch::h(internal::checkRef(refToRef(type2type<T>()))))==sizeof(char) };
		typedef typename internal::refTypeH<is>::type<T>::value reference;
		typedef typename internal::refTypeH<is>::type<const T>::value constReference;
	};
	template<> struct referenceTraits<void>
	{ enum { is=false }; typedef void reference; typedef void constReference;};
	template<class T> struct constTraits
	{
		typedef typename referenceTraits<T>::reference rT;
		typedef typename internal::traitsH<rT> ch;
		enum { is = sizeof(ch::h(internal::checkConstRef(refToRef(type2type<T>()))))==sizeof(char) };
	};
	template<> struct constTraits<void>
	{ enum { is = false }; };
#endif

		template<class T> struct fundamental { enum { is = pointerTraits<T>::is }; };
		template<> struct fundamental<bool> { enum { is = true }; };
		template<> struct fundamental<long> { enum { is = true }; };
		template<> struct fundamental<unsigned long> { enum { is = true }; };
		template<> struct fundamental<char> { enum { is = true }; };
		template<> struct fundamental<unsigned char> { enum { is = true }; };
		template<> struct fundamental<ctl::int2> { enum { is = true }; };
		template<> struct fundamental<ctl::uint2> { enum { is = true }; };
		template<> struct fundamental<ctl::int4> { enum { is = true }; };
		template<> struct fundamental<ctl::uint4> { enum { is = true }; };
		template<> struct fundamental<ctl::int8> { enum { is = true }; };
		template<> struct fundamental<ctl::uint8> { enum { is = true }; };
		template<> struct fundamental<ctl::real4> { enum { is = true }; };
		template<> struct fundamental<ctl::real8> { enum { is = true }; };
		template<> struct fundamental<const bool> { enum { is = true }; };
		template<> struct fundamental<const char> { enum { is = true }; };
		template<> struct fundamental<const unsigned char> { enum { is = true }; };
		template<> struct fundamental<const ctl::int2> { enum { is = true }; };
		template<> struct fundamental<const ctl::uint2> { enum { is = true }; };
		template<> struct fundamental<const ctl::int4> { enum { is = true }; };
		template<> struct fundamental<const ctl::uint4> { enum { is = true }; };
		template<> struct fundamental<const ctl::int8> { enum { is = true }; };
		template<> struct fundamental<const ctl::uint8> { enum { is = true }; };
		template<> struct fundamental<const ctl::real4> { enum { is = true }; };
		template<> struct fundamental<const ctl::real8> { enum { is = true }; };

namespace internal
{

	template<int N> struct sizeType
	{ char D[N]; };

	template<int N> struct whichProp;

#define CTL_Registrate(i, type)\
	template<> struct whichProp<i>\
	{ typedef type resultT; };\
	sizeType<i> slctProp(type)

	CTL_Registrate(1, int1);
	CTL_Registrate(2, uint1);
	CTL_Registrate(3, int2);
	CTL_Registrate(4, uint2);
	CTL_Registrate(5, int4);
	CTL_Registrate(6, uint4);
	CTL_Registrate(7, int8);
	CTL_Registrate(8, uint8);
	CTL_Registrate(9, real4);
	CTL_Registrate(10, real8);

	template<typename S, typename T, char op> struct resultType;

	template<typename S, typename T> struct resultType<S, T, '*'>
	{
		enum { p=sizeof(slctProp( S() * T() ) ) };
		typedef typename whichProp<p>::resultT resultT;
	};
	template<typename S, typename T> struct resultType<S, T, '+'>
	{
		enum { p=sizeof(slctProp( S() + T() ) ) };
		typedef typename whichProp<p>::resultT resultT;
	};

}
	template<class T> struct optArgType
	{
		enum { isVar = !constTraits<T>::is & referenceTraits<T>::is, isFund = fundamental<T>::is,
						self = isVar | isFund | constTraits<T>::is};
		typedef typename ctl::select<self, T, typename referenceTraits<T>::constReference>::resultT resultT;
	};
	
	template<class T> class typeTraitsBase
	{
public:
#ifdef CTL_PartialSpec
	typedef  typename unRefConst<T>::resultT unrefconst;
#endif

	};
	
	template<> class typeTraitsBase<void>
	{
public:
	typedef void unrefconst;
	};
	
	template<class T> class typeTraits: public typeTraitsBase<T>
	{
		typedef typeTraitsBase<T> base;		
public:

	enum { 	isPointer = pointerTraits<T>::is,
		isReference = referenceTraits<T>::is,
		isConst = constTraits<T>::is,
		isPure = !isConst & !isReference & !isPointer,
		isEmpty = false };
	};
	
	template<>  class typeTraits<void>
	{
public:
	enum { isPointer = false, isReference = false,  isConst = false, isEmpty = true, isFundamental = false };
	};
	
	template<>  	class typeTraits<ctl::empty>: public typeTraitsBase<ctl::empty>
	{
public:
	typedef empty unrefconst;
	enum { isPointer = false, isReference = false,  isConst = false,isEmpty = true, isFundamental = false };
	};

	template<class X> struct stringType
	{ X x; };
	typedef stringType<char> charT;
	typedef stringType<wchar> wcharT;

	template<class X> class noVoid
	{
	public:
		noVoid<X>& operator=(const noVoid&);
		enum { isVoid = false };
		typedef X resultT;
		class cast
		{	
			cast operator=(const cast&);
		public:
			typedef typename referenceTraits<X>::constReference XRef;
			XRef xM;
			cast(XRef x): xM(x) {}
			XRef operator, (ctl::empty)
			{ return xM; }
		};
	};
	
	template<> class noVoid<void>
	{
	public:
		enum { isVoid = true };
		typedef ctl::empty resultT;
		typedef void cast;
	};

  	bool log(const char* fmt,...);
  	void throwStr(const char* fmt,...);

template<class F, int a, int n, int c=0> class binTreeCall
{
private:
	enum { value=c+n/2 };
public:
	static bool call(F,falseT)
	{ return true; }
	static bool call(F f,trueT)
	{
		return  binTreeCall<F,a,n/2,c>::call(f, int2type<(n>1)>()) &&
		f(int2type<a+value>()) &&
		binTreeCall<F,a,(n-1)/2, value+1>::call(f, int2type<(n>2)>());
	}
	static bool call(F f)
	{ return call(f,int2type< (n>0) >()); }

};

template<int a, int o, class F> inline bool forLoop(F &f)
{ return binTreeCall<F&,a,o-a>::call(f); }

template<int n> struct Assert;
template<> struct Assert<1> {};

template<class T> struct abstractT
{};

template<class T> inline void createObject(T *&,void *, trueT /*isAbstract*/)
{ ctl::throwStr("Can't create object of abstract type '%s'", T::CTL_typeNameS().c_str());  }

template<class T> inline void createObject(const T *&pt,void *p, falseT /*isAbstract*/)
{ pt = p ? new(p) T() :  new T(); }

template<class T> inline void cloneObject(T *&, void *, const T &, trueT /*isAbstract*/)
{ ctl::throwStr("Can't clone object of abstract type '%s'", T::CTL_typeNameS().c_str()); }

template<class T> inline void cloneObject(T *&pt, void *p, const T& t,falseT /*isAbstract*/)
{ pt = p ? new(p) T(t) :  new T(t); }


char typeCheck(empty*);
template<class T> int typeCheck(empty(*)(T) );

#define CTL_typeCheck(X, id)\
	ctl::empty CTL_checkExpr##id(X);\
	enum { id = sizeof(typeCheck(&CTL_checkExpr##id))>sizeof(char) }

} // ctl
#endif
