/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_VECTOR_H_
#define CTL_VECTOR_H_

#include <iostream>
#include <string.h>
#include <new>
#include <assert.h>
#include "types.h"
#include "sort.h"
//#include "timer.h"


namespace ctl
{

namespace internal
{

template<class T, class X> inline void uFill(T *a,  T *o, const X &x, falseT)
{
	while(a<o)
		new(a++) T(x);
}

template<class T, class X> inline void uFill(T *a,  T *o, const X &x, trueT)
{
	if(x==0)
		memset((void*)a, 0, (o-a)*sizeof(T) );
	else
		while(a<o)
			*a++ = x;
}

template<class T, class X> inline void uFillP(T *a,  T *o, const X *x, falseT)
{
	while(a<o)
	{
		new(a) T(*x);
		a++; x++;
	}
}

template<class T, class X> inline void uFillP(T *a,  T *o, const X *x, trueT)
{
  if(same<T,X>::value)
		memcpy((void*)a, (const void*)x,  (o-a)*sizeof(T) );
	else
		while(a<o)
			*a++ = T(*x++);
	
}
template<class T, class X> inline void pFill(T *a,  T *o, const X &x, trueT)
{  uFill(a,o, x, trueT() ); }

template<class T, class X> inline void pFill(T *a,  T *o, const X &x, falseT)
{
	while(a<o)
		*a++ = x;
}

template<class T> struct less
{
	bool operator()(const T &a, const T &b)  const
	{ return a<b; }
};

} // e.o. internal

template<typename T> class vector;

template<class T, class H> class vectorRep : public H
{
friend class vector<T>;
	enum { fundamental = fundamental<T>::is };
	typedef int2type<fundamental> fillType;
	size_t sizeM;
#ifdef CTL_SelfDelete
#define CTL_SetDel  delM(del),
	void (*delM)(vectorRep<T,H>*);
	void operator delete(void *, size_t) {}
	void operator delete(void *p)
	{ 
		vectorRep<T,H> *rep=static_cast<vectorRep<T,H>*>(p);
		(*rep->delM)(rep);
	}
	static void del(vectorRep<T,H> *p)
	{ delete[] (char*)p;}

#else
#define CTL_SetDel
	void operator delete(void *p)
	{ delete[] (char*)p; }
	void operator delete(void *, size_t) {}
#endif
	T tM;
	void *operator  new(size_t mysize,size_t size)
	{ return new char[mysize + (size-1) * sizeof(T) ]; }

	template<class X>
		vectorRep(size_t size, const X &x, bool init=true): sizeM(size), CTL_SetDel tM(x)
	{ if(init || !fundamental)
					internal::uFill(&tM+1, &tM+sizeM, x, fillType()); }
	template<class X>
		vectorRep(size_t size, const X *x, int): sizeM(size), CTL_SetDel tM(*x)
	{ internal::uFillP(&tM+1, &tM+sizeM, x+1, fillType()); }
	vectorRep(size_t size, const vectorRep<T,H> &hv,const T &t, bool init=true): H(hv),sizeM(size), CTL_SetDel tM(hv.tM)
	{
		size_t m=sizeM<hv.sizeM ? sizeM : hv.sizeM ;
		internal::uFillP(&tM+1, &tM+m, &hv.tM+1, fillType());

		if(init || !fundamental)
			internal::uFill(&tM+m, &tM+sizeM, t, fillType());
	}
	void clear(trueT)
	{}
	void clear(falseT)
	{
		for(size_t i=1; i<sizeM; i++)
			(&tM+i)->~T();			 	
	}
	~vectorRep()
	{	clear(fillType()); }
#undef CTL_SetDel
	size_t size() const
	{ return sizeM; }
	const T *p() const
	{ return &tM; }
	T *p()
	{ return &tM; }
	const T&operator[](size_t i) const
	{ return *(&tM+i); }
	T&operator[](size_t i)
	{ return *(&tM+i); }
};

/**/// late copy implementation of a vector: assumes default and copy constr on T
template<class T> class vector
{
public:
	typedef size_t sizeT;
private:
	struct Header
	{
		int4 refCM;
		Header(): refCM(1) {}
		Header(const Header &): refCM(1) {}
	};
	typedef vectorRep<T, Header> repType;

	sizeT sizeM;
	repType *dataM;
	void unique()
	{
// check dataM!=0 before call to unique
		if(dataM->refCM>1)
		{
			  dataM->refCM--;
			  repType *newRep = new(sizeM) repType(sizeM, *dataM, T());
			  dataM = newRep;
		}
	}
	void unref()
	{
		if(dataM && --dataM->refCM==0)
			delete dataM; 
	}
	void init(const T* t)
	{
		if(sizeM)
			dataM = t ? new(sizeM) repType(sizeM,t,0) :  new(sizeM) repType(sizeM, T());
		else
			dataM = 0;		
	}
public:
	int getRefCount() const
	{ return dataM ? dataM->refCM : -1; }
//	void getBuffer(const void *&buf, uint8 &bytes) const
//	{ buf = &dataM->sizeM; bytes = sizeof(repType) + (sizeM-1) * sizeof(T) - ( ((char*)buf) - ((char*)dataM) ); }
	vector() : sizeM(0),dataM(0) 
	{}
	explicit vector(sizeT size, const T&t=T(), bool init = true): sizeM(size)
	{
		if(sizeM)
			dataM = new(sizeM) repType(sizeM,t, init);
		else
			dataM = 0;
	}
	vector(const T *a, const T *o): sizeM(o-a)
	{ init(a); }
	vector(sizeT size, const T *t): sizeM(size)
	{ init(t); }

	template<class X>
		explicit vector(const vector<X> &v) : sizeM(v.size())
	{ 	dataM = new(sizeM) repType(sizeM, v.p(), 0 );	 }
	vector(const vector<T> &v) : sizeM(v.sizeM)
	{
		if(v.dataM)
			v.dataM->refCM++;
		dataM = v.dataM;
	 }

	template<class X>
		vector<T> &operator=(const vector<X> &v)
	{
		unref();
		sizeM = v.size();
		dataM = sizeM ? new(sizeM) repType(sizeM, v.p(), 0 ) : 0;
		return *this;
	}
	vector<T> &operator=(const vector<T> &v)
	{
		if(v.dataM)
			v.dataM->refCM++;
		unref();	
		sizeM = v.sizeM;
	 	dataM = v.dataM;
	 	return *this;
	}
	void fill(const T &t)
	{
		typedef typename repType::fillType fillType;
		if(dataM)
		{
			unique();
			T *a=p();
			internal::pFill(a,a+sizeM, t, fillType());
		}
	}
	~vector()
	{ unref(); }
// invariant: sizeM <=  dataM->sizeM
	void resizeRaw(sizeT newSize)
	{ resize(newSize, T(), false); }
	void resize(sizeT newSize, const T &t=T(), bool init=true)
	{
		sizeM = newSize;		
		if(!dataM)
		{
			if(newSize>0)
				dataM = new(newSize) repType(newSize,t, init);			
			return;
		}
		if(newSize<=dataM->size())
			return;
		
		repType *newRep = new(newSize) repType(newSize,*dataM,t, init);
		unref();
		dataM = newRep;
	}
	void clear()
	{
		unref();
		dataM = 0;
		sizeM = 0;
	}
	sizeT size() const
	{ return sizeM; }
	sizeT max_size() const
  { return sizeT(-1) / sizeof(T); }
	sizeT capacity() const
	{ return dataM ? dataM->sizeM : sizeT(0); }
	const T *p() const
	{ return dataM ? dataM->p() : 0; }
	T *p()
	{
		if(!dataM)
			return 0;
		unique();
		return dataM->p();
	}
	const T& operator[](sizeT i) const
	{ CTL_Assert(dataM && i<sizeM); return (*dataM)[i]; }
	T& operator[](sizeT i)
	{
		CTL_Assert(dataM && i<sizeM);
		unique();
		return (*dataM)[i];
	}
	bool operator !() const
	{ return sizeM==0; }
	T& append(const T &s)
	{
		sizeT oldSize = sizeM;
		if(!dataM || dataM->size()<sizeM+1)
			resize(sizeM<8 ? 9 : sizeM + (sizeM>>2));
		sizeM = oldSize + 1;	
		unique();
		return (*dataM)[oldSize] = s;		
	}
	template<class Int>
	vector<T> subVector(const vector<Int> &index) const
	{
		vector<T> s(index.size());
		T *a=s.p(), *o=a+s.size();
		const T *A=p();
		const Int *i=index.p();
		while(a<o)	  	
		{
			CTL_Assert(size_t(*i) < size() );
		  	*a++ =   A[*i++];
		}
		return s;
	}
// assumes f: T->Y

	template<class Y, class F> vector<Y> apply(const F &f) const
	{
		vector<Y> y(size() );
		const T *a=begin(), *o = a+size();
		Y *p = y.begin();
		while(a<o)
			*p++ = f(*a++);
		return y;
	}
	template<class Y, class X, class F> vector<Y> apply(const F &f, const vector<X> &x) const
	{
		CTL_Assert(size()==x.size());
		vector<Y> y(size() );
		const T *a=begin(), *o = a+size();
		Y *p = y.begin();
		const X *px=x.begin();
		while(a<o)
			*p++ = f(*a++, *px++);
		return y;
	}
//    assumes T has algebraic operator
	vector<T> &operator +=(const vector<T> &);
	vector<T> &operator -=(const vector<T> &);
	vector<T> &operator *=(const T &);
	vector<T> &operator /=(const T &);
	void add(const vector<T> &, const T&);

//    assumes T is total ordered by <  or less
	template< bool (*less)(const T&,const T&) > struct Less
	{
		bool operator()(const T&a,const T&b) const
		{ return (*less)(a,b); }
	};
	template<typename less>	int lowerBound(const T &t, less lt = less())  const
	{ return ctl::lowerBound(p(),p()+size(),t,lt);	}	
	template<bool (*less)(const T&,const T&)>  int lowerBoundF(const T&t) const
	{ return lowerBound(t,Less<less>()); }
	int lowerBound(const T &t) const
	{ return lowerBound(t, stdLess<T>()); }
// assuming sorted by less
	template<typename less> int position(const T &t, less lt = less())  const
	{
		int pos=lowerBound(t, lt);
		return pos<0 || lt((*dataM)[pos], t) ? -1 : pos;
	}	
   template<bool (*less)(const T&,const T&)> int positionF(const T&t) const
	{ return position(t, Less<less>()); }
	int position(const T &t) const
	{ return position(t, stdLess<T>()); }
	
	template<typename less>	void sort(bool large=false, less lt=less())
	{
		if(size()<2)
			return;
		if(large || sizeof(T)>64)
			ctl::pqSort(p(), p()+size(), lt);
		else
			ctl::qSort(p(), p()+size(), lt);
	}
	template<bool (*less)(const T&,const T&)> void sortF(bool large=false)
	{ return sort(large, Less<less>()); }
	void sort(bool large=false)
	{ sort(large, stdLess<T>()); }

	vector<int> sortIndex() const
	{
		vector<const T*> P(sizeM);
		sizeT i;
		for(i=0; i<sizeM; i++)
			P[i] = p()+i;
		typedef internal::pointerLessTotal<stdLess<T>,T> cmp;

		qSort(P.p(),P.p()+sizeM, cmp());
		vector<int> index(sizeM);
		for(i=0; i<sizeM; i++)
			index[i] = P[i]-p();
		return index;
	}

	bool operator==(const vector<T> &v) const
	{
		if(size()!=v.size())
			return false;
		const T *a=begin(), *b= v.begin(), *o = a+ size();
		while(a<o)
			if(*a++!=*b++)
				return false;
		return true;
	}
	bool operator!=(const vector<T> &v) const
	{ return !operator==(v); }
	int find(const T &key) const
	{
		const T *a=begin(), *o = a+ size();
		while(a<o)
			if(*a++==key)
				return a-begin() - 1;
		return -1;
	}
	int rfind(const T &key) const
	{
		const T *a=begin(), *o = a+ size();
		while(o>a)
			if(*--o==key)
				return o-begin();
		return -1;
	}
	// stl interface
	typedef T* iterator;
	typedef const T* const_iterator;
	typedef T* reverse_iterator;
	typedef const T* const_reverse_iterator;
 
	typedef T value_type;

  reverse_iterator rbegin() { return end(); }
  const_reverse_iterator rbegin() const { return end(); }
  reverse_iterator rend() { return begin(); }
  const_reverse_iterator rend() const { return begin(); }

	iterator begin() { return p(); }
	const_iterator begin() const { return p(); }
	iterator end() { return p()+size(); }
	const_iterator end() const { return p() + size(); }
	void push_back(const T&t)
	{ append(t); }
	void push_back(const vector<T> &v)
	{ 
		if(!v)
			return;
		typedef typename repType::fillType fillType;
		sizeT s=size();
		resize(s+v.size());
		internal::uFillP(begin()+s, end(), v.begin(), fillType());
	}
	T &insert(const T &s, size_t pos)
	{
		CTL_Assert(pos<=sizeM);
		push_back(s);
		T *t=p();
		for(sizeT i=sizeM-1; i>pos; i--)
			t[i] = t[i-1];
		return t[pos] = s;
	}
	T& push_front(const T &s)
	{ return insert(s,0); }

	const T& front() const
	{ return (*this)[0]; }
	T& front()
	{ return (*this)[0]; }
	const T& back() const
	{ return (*this)[size()-1]; }
	T& back()
	{ return (*this)[size()-1]; }
	void erase(size_t a)
	{ erase(a,a+1); }
	void erase(size_t a, size_t b)
	{
		if(b>sizeM)
			b = sizeM;
		if(a>=b)
			return;
		sizeT d= b-a;
		T *t = p();
		for(sizeT i=b; i<sizeM ; i++)
			t[i-d] = t[i];
		resize(sizeM - d);
	}
	bool empty() const
	{ return !sizeM; }
};

} // ctl

#include <string>

namespace ctl {

// zero padded vector
	template<class T> class sequence : public vector<T>
	{
	public:
		using vector<T>::resize;
		using vector<T>::size;
		using vector<T>::begin;
		using vector<T>::end;
		typedef typename vector<T>::sizeT sizeT;
		typedef typename vector<T>::iterator iterator;
		typedef typename vector<T>::const_iterator const_iterator;
		sequence() : vector<T>()
		{}
		explicit sequence(sizeT size, const T&t=T(), bool init = true): vector<T>(size, t, init)
		{}
		sequence(const T *a, const T *o): vector<T>(a, o)
		{}
		sequence(sizeT size, const T *t): vector<T>(size, t)
		{}
		template<class X>
			explicit sequence(const sequence<X> &v) : vector<T>(v)
		{}
		sequence(const vector<T> &v) : vector<T>(v)
		{}
		T operator[](sizeT i) const
		{ return i<size() ? begin()[i] : T(); }
		T& operator[](sizeT i)
		{
			if(i>=size())
				resize(i+1);
			return begin()[i];
		}
	//    assumes T has algebraic operator
		sequence<T> &operator +=(const sequence<T> &);
		sequence<T> &operator -=(const sequence<T> &);
		sequence<T> &operator *=(const T &);
		sequence<T> &operator /=(const T &);
		void add(const sequence<T> &, const T&);
	};

	class string : public vector<char>
	{
	typedef vector<char> baseT;
	string(size_t size): baseT(size)
	{}

public: 
	static string int2string(int m)
	{
		if(!m)
			return "0";
		char num[32], *c=num+31;
		bool neg=m<0;
		if(neg)
			m = - m;
		*c = 0;
		while(m)
			*--c = '0' + (m%10), m /=10;
		if(neg)
			*--c = '-';
		return c;
	}

	string(const string &str) : baseT(str)
	{}
	string(const std::string &str) : baseT(str.size()+1, str.c_str())
	{}
	operator std::string() const
	{ return begin(); }
	string &operator=(const string &str)
	{ baseT::operator=(str);  return *this; }
	string() {}
	string(const char *s) : baseT((s && *s) ? strlen(s) + 1 : 0, s)
	{}
	string(const char *a, const char *o) : baseT((a && a<o && *a) ? (std::min)(size_t(o-a), strlen(a)) + 1 : 0, a)
	{ begin()[size()-1] = 0; }
	void resize(size_t s)
	{ 
		if(s)
		{
			baseT::resize(s);
			begin()[s-1] = 0;
		}
		else
			clear();
	}
	string &operator +=(char c)
	{ 
		if(!c)
			return* this;
		if(size()==0)
		{ push_back(c); push_back(0); }
		else
		{	begin()[size()-1] = c; push_back(0); }
		return *this; 
	}
	string &operator +=(const string &str)
	{ return *this = *this + str; }
	string &operator +=(int num)
	{ return *this += int2string(num); }
	inline friend string operator+(const string& str, int num)
	{ string cpy(str); return cpy += num; }
	string operator +(const string &str) const
	{
		if(size()==0)
			return str;
		if(str.size()==0)
			return *this;
		string cat(size() + str.size()-1);
		memcpy(cat.begin(), begin(), size()-1 );
		memcpy(cat.begin()+size()-1, str.begin(), str.size()-1 );
		return cat;
	}
	string operator +(char c) const
	{
		string cpy(*this);
		return cpy += c;
	}
	inline friend string operator +(char c, const string& str)
	{ 
		string cpy(str);
		if(c)
			cpy.push_front(c);
		else
			cpy.clear();
		return cpy;
	}
	inline friend string operator +(const char* s, const string& str)
	{
		if(s==0 || *s==0)
			return str;
		if(str.size()==0)
			return s;
		size_t l=strlen(s);
		string cat(l + str.size());
		memcpy(cat.begin(), s, l );
		memcpy(cat.begin()+l, str.begin(), str.size()-1 );
		return cat;
	} 
	string substr(size_t a) const
	{ CTL_Assert(a<size()); return substr(a, size()-1); }
	string substr(size_t a, size_t b) const
	{
		if(!size() )
			return string();
		if(b>=size())
			b = size()-1;
		if(b<=a)
			return string();

		CTL_Assert(a<b && b<size());
		string sub(b-a+1);
		memcpy(sub.begin(), begin()+a, b-a);
		return sub;
	}
	bool operator <(const string &v) const
	{
		if(!v)
			return false;
		if(!size())
			return true;
		return strcmp(begin(), v.begin() ) <0;
	}
	bool operator==(const string &v) const
	{ return baseT::operator==(v); }
	bool operator!=(const string &v) const
	{ return !operator==(v); }
	const char *c_str() const
	{ return begin() ? begin() : ""; }
	inline friend std::ostream &operator<<(std::ostream &os, const ctl::string &str)
	{
		if(!str)
			return os<<"(null)";
		return os<<str.begin();
	}

};

template<class T> inline std::ostream &operator<<(std::ostream &os, const ctl::vector<T> &v)
{
	os<<"[ ";
	for(size_t i=0; i<v.size(); i++)
		os<<v[i]<<' ';
	return os<<"]";
}

} //ctl


#endif
